/*
   OpenGUI - Drawing & Windowing library

   Copyright (C) 1996,2005  Marian Krivos

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   nezmar@atlas.sk

   base.h - base graphics support headers

  $Id: base.h,v 1.29 2004/09/28 08:20:41 majo Exp $

  $Log: base.h,v $
  Revision 1.29  2004/09/28 08:20:41  majo
  Calibration stuff

  Revision 1.28  2004/09/27 13:26:06  majo
  cal4

  Revision 1.27  2004/09/27 10:53:42  majo
  calibrating part III.

  Revision 1.26  2004/09/27 06:59:57  majo
  SetParameter(void *)
  Calibration part I.

  Revision 1.25  2004/08/26 05:45:35  majo
  added FGDrawBuffer::arc2()

  Revision 1.24  2004/08/25 21:16:10  majo
  priebezne zmeny

  Revision 1.23  2004/07/13 18:07:04  majo
  FGSIze docs

  Revision 1.22  2004/07/13 17:42:22  majo
  Added FGSize

  Revision 1.21  2004/07/07 16:51:32  majo
  FGSize
  FGWindowMenu::AddteText()
  FGMatrix::scale_with_center()

  Revision 1.20  2004/05/30 11:52:55  majo
  dox

  Revision 1.19  2004/05/24 17:40:41  majo
  bugs for FGRect
  added FGCircle

  Revision 1.18  2004/04/27 17:37:10  majo
  fixed bug in FGRect and FGPointArray

  Revision 1.17  2004/04/16 08:45:25  majo
  fixed clipping error

  Revision 1.16  2004/03/24 17:13:58  majo
  added FGXColor
  added FGColorPicker

  Revision 1.15  2004/03/18 16:05:27  majo
  readded changes to branch

  Revision 1.13  2004/03/09 20:23:58  majo
  added FGImage
  improved FGColor & and its docs

  Revision 1.12  2004/03/07 20:24:23  majo

  kompilacia FG*

  Revision 1.11  2004/03/03 20:12:38  majo
  added class FGColor
  fixed Solaris compiling

  Revision 1.10  2004/03/01 15:19:15  majo
  _GTRANSP -> _GCOLORKEY
  added _GTRANSPARENT

  Revision 1.9  2004/02/29 19:58:19  majo
  devel state

  Revision 1.8  2004/02/26 21:41:28  majo
  class FGInternalState
  char* -> const char*
  int ink = -1  -> unsigned ink = UNDEFINED_COLOR

  Revision 1.7  2004/02/24 17:34:59  majo
  FGTwostateButton

  Revision 1.6  2004/02/23 20:08:46  majo
  *

  Revision 1.5  2004/02/23 20:03:50  majo
  added FGPoint, FGPointArray (inspired by QT)
  FGDrawBuffer::save_state(), restore_state().
  all polygon functions are using FGPointArray instead (int, int [][2]).
  All classes are with prexix FG* on now (DrawBuffer -> FGDrawBuffer).
    but there are some macros to hide this behaviour.
  all FGDrawBuffer by parameters overloaded methods like box(x,y,w.h,ink)
    are now all-in-one.

  Revision 1.4  2004/02/10 16:18:59  majo
  fixed compile warnings

  Revision 1.3  2004/02/02 15:12:39  majo
  *** empty log message ***

  Revision 1.2  2004/01/22 17:49:11  majo
  Increased version number

  Revision 1.1.1.1  2004/01/22 16:25:08  majo
  initial release

*/

#ifndef __BASE_H
#define __BASE_H

//#define FG_NAMESPACE		// uncoment this line if you want namespace support

#ifdef __BORLANDC__
#pragma option push -b
#pragma option push -a4
#endif

#define FG_VERSION "4.9.7"
#define FG_MAJOR 	4
#define FG_MINOR 	9
#define FG_PATCH 	7

#include "internal.h"

/*
  todo 4.9.x:
   + extend & document FGRect, FGPointArray
   + finalize & document FGMatrix
   + translate polygons to use FGPointArray
   + restructure Color management & document one
   + SSE FGmemcpy
*/
/*
  todo 5.x:
   - new Drawable for Windows
   + new Transformation layer for FGDrawBuffer
   - Save Images Bitmap
   - FGTheme
*/

#ifdef FG_NAMESPACE
/**
    @brief Global library namespace.
    On now namespace support in the library is fully implemented, but disabled.
	The switch to 'fgl::' namespace will be in near future.
	You can modify your code to support OpenGUI namespace right now by
	uncommenting #define FG_NAMESPACE line in 'base.h' headers;
*/
namespace fgl {
#else
#define fgl
#endif

#ifdef INDEX_COLORS
/**
	The unsigned integer that is equal in size with the one graphic pixel
	representation in VRAM, i.e. char for indexed colors, int for true colors (24 or 32 bit modes).
	@ingroup Types
*/
typedef unsigned char FGPixel;	// type of color (must correspond with bpp)
#endif

#if (DIRECT_COLORS==15) | (DIRECT_COLORS==16)
typedef unsigned short FGPixel;		// type of color (must correspond with bpp)
#endif

#ifdef TRUE_COLORS
typedef unsigned int FGPixel;	// type of color (must correspond with bpp)
#endif

typedef void (*blitter_t) (FGPixel *, FGPixel *, int, int, int, int);
typedef void (*blitter2_t) (FGPixel *, FGPixel *, int, int, int, int, unsigned);
typedef void (*blitter_a_t) (FGPixel *, FGPixel *, int, int, int, int, int, unsigned ppop);
typedef int  (*shapper_t) (int, int, int, int, FGPixel ink, unsigned ppop);
typedef void (*clipper_t) (int, int, int, int);
typedef void (*plotter_t) (FGPixel *ptr, int x, int y, int ww, FGPixel ink, unsigned ppop);
typedef void (*paletter_t) (unsigned, unsigned);
typedef int  (*mouser_t) (int, int, int, int&);

#define	CR		13
#define	ESC		27
#define	TAB		9
#define	LF		10

#define	MIN(a,b)	( (a) < (b) ? (a) : (b) )
#define MAX(a,b)	( (a) > (b) ? (a) : (b) )

/**
	The ROPs. Intended for a drawing functions.
	@ingroup Enums
*/
enum ENUM_PPOP
{
	_GSET,				// 0
	_GXOR,
	_GAND,
	_GOR,
	_GPLUS,
	_GMINUS,
	_GNOT,
	_GREPLACE_GE16,     // 7
	_GTRANSPARENT,      // 8
	_GCOLORKEY,         // 9
	_GREPLACE_LE32,      // 10
	_GREPLACE_LE48,      // 10
	_GREPLACE_LE64,      // 10
	_GREPLACE_LE80,      // 10
	_GREPLACE_LESS      // 11
};

/**
	Constants for apropriate videomodes.
	@ingroup Enums
*/
enum ENUM_VIDEOMODE {
	GTEXT,
	G320x200,
	G640x480,
	G800x600,
	G1024x768,
	G1280x1024,
	G1600x1200,
	GCUSTOM
};

/**
	Constants for bitmap rotations.
	@ingroup Enums
	@see FGDrawBuffer::rotate()
*/
enum ENUM_ROTATE_DIRECTION
{
	ROTATE_90_CW,
	ROTATE_90_CCW,
	ROTATE_180,
	ROTATE_HORIZONTAL,
	ROTATE_VERTICAL
};

#define	FG_BUTTON_NONE			0
#define	FG_BUTTON_LEFT			1
#define	FG_BUTTON_RIGHT			2
#define	FG_BUTTON_MIDDLE		4
#define	FG_BUTTON_WHEEL_UP		8
#define	FG_BUTTON_WHEEL_DOWN	16

/**
	Flags for use with FGDrawBuffer::bitblit()
	@addtogroup Defines
	@{
*/
//! do only raw copy (no OR, AND etc.)
const unsigned BITBLIT_COPY	=			0x0000;
//! do copy with current dest. operator (_GSET, _GOR ...), see set_ppop()
const unsigned BITBLIT_WITH_OPERATOR =	0x0200;
//! do copy and assume the source alpha value (transparency)
const unsigned BITBLIT_WITH_ALPHA =		0x0400;
//! clips the coordinates before drawing (use if unsure)
const unsigned BITBLIT_WITH_CLIPPING =	0x1000;
//! to copy the overlapped blocks
const unsigned BITBLIT_WITH_DECREMENT =	0x2000;

#ifdef INDEX_COLORS
//! Constant for the BLACK color
const FGPixel CBLACK =  0;
//! Constant for the DARK_GRAY color
const FGPixel CDARK	= 1;
//! Constant for the another GRAY color
const FGPixel CGRAYED =  2;
//! Constant for the GRAY color #1
const FGPixel CGRAY1 =  3;
//! Constant for the GRAY color #2
const FGPixel CGRAY2 =  4;
//! Constant for the GRAY color #3
const FGPixel CGRAY3 =  5;
//! Constant for the BLUE color
const FGPixel CBLUE	=  6;
//! Constant for the LIGHT BLUE color
const FGPixel CBLUELIGHT=  7;
//! Constant for the GREEN color
const FGPixel CGREEN =  8;
//! Constant for the LIGHT GREEN color
const FGPixel CGREENLIGHT = 9;
//! Constant for the RED color
const FGPixel CRED	=  10;
//! Constant for the LIGHT RED color
const FGPixel CREDLIGHT =  11;
//! Constant for the BROWN color
const FGPixel CBROWN =  12;
//! Constant for the YELLOW color
const FGPixel CYELLOW = 13;
//! Constant for the DUN WHITE color
const FGPixel CWHITED = 14;
//! Constant for the 100% WHITE color
const FGPixel CWHITE =  15;

#else
const FGPixel CBLACK = 0;
const FGPixel CDARK	= FGDirectColor(0x444444U);
const FGPixel CGRAYED = FGDirectColor(0x585858U);
const FGPixel CGRAY1 = FGDirectColor(0x7c8074U);
const FGPixel CGRAY2 = FGDirectColor(0x687884U);
const FGPixel CGRAY3 =	FGDirectColor(0xa8b8acU);
const FGPixel CBLUE	= FGDirectColor(0x000078U);
const FGPixel CBLUELIGHT = FGDirectColor(0x0000f8U);
const FGPixel CGREEN = FGDirectColor(0x007800U);
const FGPixel CGREENLIGHT = FGDirectColor(0x00f800U);
const FGPixel CRED = FGDirectColor(0xd44448U);
const FGPixel CREDLIGHT	= FGDirectColor(0xf80000U);
const FGPixel CBROWN = FGDirectColor(0x684028U);
const FGPixel CYELLOW = FGDirectColor(0xfce000U);
const FGPixel CWHITED = FGDirectColor(0xd8d0b8U);
const FGPixel CWHITE = FGDirectColor(0xffffffU);
#endif // TRUE_COLORS

const unsigned UNDEFINED_COLOR = unsigned(0xffffffffU);
const unsigned DEFAULT_COLOR_KEY = unsigned(0x12345678U);
const unsigned DEFAULT_ALPHA_VALUE = unsigned(0);

#ifdef INDEX_COLORS
/**
	Remap the indexed colors to the truecolor directly by palette color map.
	Useful macro when you port your 256 color modes app to the truecolors.
*/
#define FGCOLOR(i) (i)
#else
#define FGCOLOR(i) ((fgl::FGPixel) FGDirectColor(fgl::_fgl_palette[i]))
#endif

/**
	@}
*/

extern unsigned int long _fgl_palette[256];

/**
	The all event types.
	@ingroup Enums
*/
enum ENUM_EVENTS  {
/**
a dummy event.
*/
NOEVENT,
/**
This events indicates key pressing (the keyboard input is going to the active window only).
System has some reserved key combination (ALT+X, CTRL+TAB etc), and these aren't possible
to testing. These reserved keys are converted to appropriate events (QUITEVENT, LOSTFOCUSEVENT & GETFOCUSEVENT).
<b>Used with:</b> Application and Window handlers.
*/
KEYEVENT,
/**
Internal event.
*/
MOUSEEVENT,
/**
The last event, that is sent only to the application procedure. This event can generated
in one of the two ways: by user pressing <ALT+X> or by calling AppDone().
<b>Used with:</b> Application handler only.
*/
QUITEVENT,
/**
This is a very important event that is generated by system when you activate
some Controls (Button, EditBox, ..). The member key contains the id
of current control. The callback handler of the object [you can assign
to each Control your callback that will be called at the right time] is
called after this event parsing into your procedure. The test of some controls:
<b>Used with:</b> Application and Window handlers.

	void WindowHandler(GuiEvent *p)
	{
		switch(p->GetType())
		{
			case ACCELEVENT:
			if (p->GetKey() == button1->GetId())
			{
				// some code
			}
		}
	}
*/
ACCELEVENT,
/**
Last event in the app life. You can delete your stuff after this, the window is deleted now.
<b>Used with:</b> Window handlers only.
*/
TERMINATEEVENT,
/**
CLose  the most inner application loop only.
*/
CLOSEEVENT,
/**
Indicates any movement of the mouse (including button clicks). The members are
assigned as follows: If cursor is out of any window, member key is 0,
and items x & y contains the position of mouse cursor in screen co-ordination
([0,0] is top left corner). When the cursor is over some window, then event
is sent only to the active window. The key member is set to 0 if the window
is active, -1 when window with cursor is not active, and if the member key
is greater than 0, then value is an id of controls in window. The co-ordinations
for movement in case of window are relative to the window workspace (if cursor is
at the title of window, then y co-ordination will be negative!). For a good overview
of this, look at the events example.
<b>Used with:</b> Application and Window handlers.
*/
MOVEEVENT,
/**
This event is sent when a simple click on mouse's right button is detected. You must use window
flag WCLICKABLE to enable this feature. Position reported in the event
is in the screen co-ordinates, or relative to windows, according to placement.
<b>Used with:</b> Application and Window handlers.
*/
CLICKRIGHTEVENT,
/**
This event is sent when a simple click on mouse's left button is detected. You must use window
flag WCLICKABLE to enable this feature. Position reported in the event
is in the screen co-ordinates, or relative to windows, according to placement.
<b>Used with:</b> Application and Window handlers.
*/
CLICKLEFTEVENT,
/**
This event is sent when a simple click on mouse's middle button is detected. You must use window
flag WCLICKABLE to enable this feature. Position reported in the event
is in the screen co-ordinates, or relative to windows, according to placement.
<b>Used with:</b> Application and Window handlers.
*/
CLICKMIDDLEEVENT,
/**
<b>Used with:</b> Application and Window handlers.
*/
DRAGLEFTEVENT,
/**
<b>Used with:</b> Application and Window handlers.
*/
DRAGRIGHTEVENT,
/**
	The window has got the input focus.
	<b>Used with:</b> Window handlers only.
*/
GETFOCUSEVENT,
/**
	The window lost the input focus.
	<b>Used with:</b> Window handlers only.
*/
LOSTFOCUSEVENT,
/**
The first event that is sent to the window when a window is created and it is drawn
at the screen. It is a right place for drawing controls and other stuff.
<b>Used with:</b> Window handlers only.
*/
INITEVENT,
/**
If you hold the mouse button at the Controls PUSHBUTTON a while,
system will repeatedly emit this event. See also the SetRepeatDelay() function.
<b>Used with:</b> Application and Window handlers.
*/
BUTTONHOLDEVENT,
/**
If the window is resize-able (switch WSIZEABLE), then you can by dragging
bottom right corner of windows, resize one. After each size change,
this event is generated. You may redraw window at this point (Controls
are draws automatically). The current size (of window workspace)
is available by calling GetWW() a GetHW().
<b>Used with:</b> Window handlers only.
*/
WINDOWRESIZEEVENT,
/**
The event is generated when window moves at new position.
<b>Used with:</b> Window handlers only.
*/
WINDOWMOVEEVENT,
/**
If you use switch APP_MAGNIFIER, you can select a rectangle area by
the mouse dragging. The members are assigned with [x,y] position
of top left corner and [w,h] the size of one. If you want exactly
first and end point position of dragging, you must use GetDragVector().
<b>Used with:</b> Application and Window handlers.
*/
STARTDRAGLEFTEVENT,
/**
If you use switch APP_MAGNIFIER, you can select a rectangle area by
the mouse dragging. The members are assigned with [x,y] position
of top left corner and [w,h] the size of one. If you want exactly
first and end point position of dragging, you must use GetDragVector().
<b>Used with:</b> Application and Window handlers.
*/
STARTDRAGRIGHTEVENT,
/**
The window is repainted. Place your drawing code here.
<b>Used with:</b> Window handlers only.
*/
REPAINTEVENT,
/**
<b>Used with:</b> Window handlers only.
*/
NOTIFYEVENT,
/**
<b>Used with:</b> Window handlers only.
*/
ICONIZEEVENT,
/**
<b>Used with:</b> Application handler only.
*/
CURSOROUTEVENT,
/**
This event is sent when a double click on mouse's left button is detected. You must use window
flag WCLICKABLE to enable this feature. Position reported in the event
is in the screen co-ordinates (app handler), or relative to windows, according to placement.
<b>Used with:</b> Application and Window handlers.
*/
DBLCLICKRIGHTEVENT,
/**
This event is sent when a double click on mouse's right button is detected. You must use window
flag WCLICKABLE to enable this feature. Position reported in the event
is in the screen co-ordinates (app handler), or relative to windows, according to placement.
<b>Used with:</b> Application and Window handlers.
*/
DBLCLICKLEFTEVENT,
/**
This event is sent when a double click on mouse's middle button is detected. You must use window
flag WCLICKABLE to enable this feature. Position reported in the event
is in the screen co-ordinates (app handler), or relative to windows, according to placement.
<b>Used with:</b> Application and Window handlers.
*/
DBLCLICKMIDDLEEVENT,
/**
This event is sent when mouse's wheel spin is detected. You must use window
flag WCLICKABLE to enable this feature. GetButton() returns -1 or 1 for direcions. Position reported in the event
is in the screen co-ordinates (app handler), or relative to windows, according to placement.
<b>Used with:</b> Application and Window handlers.
*/
MOUSEWHEELEVENT,
/**
    Sent on the TabPage switch. The key member is pointer the name of one.
*/
TABSWITCHEVENT,
//! NOTE> this must be at the end - used as terminator.
LASTEVENT
};

struct variable_record;
struct fonthdr;

/**
Internal structure to describe built-in bitmap fonts;
*/
struct FGFont
{
		char	name[24];
		int type, width, height;
		variable_record *var; // null or ptr
		FGPixel *fontimg;
	public:
		FGFont()
		{
			memset(this, 0, sizeof(*this));
		}
};

/**
* @ingroup Types
* Basic type for coordinates storage.
*/
typedef int FGCOORD;

/**
	@brief The FGPoint class defines a point in the plane.

	@ingroup images
	@ingroup graphics

	A point is specified by an x coordinate and a y coordinate.

	The coordinate type is @c FGCOORD (a 32-bit integer). The minimum
	value of @c FGCOORD is @c FGCOORD_MIN (-2147483648) and the maximum
	value is  @c FGCOORD_MAX (2147483647).

	The coordinates are accessed by the functions x() and y(); they
	can be set by setX() and setY() or by the reference functions rx()
	and ry().

	Given a point @e p, the following statements are all equivalent:
	@code
	p.setX( p.x() + 1 );
	p += FGPoint( 1, 0 );
	p.rx()++;
	@endcode

	A FGPoint can also be used as a vector. Addition and subtraction
	of FGPoints are defined as for vectors (each component is added
	separately). You can divide or multiply a FGPoint by an @c int or a
	@c double. The function manhattanLength() gives an inexpensive
	approximation of the length of the FGPoint interpreted as a vector.

	Example:
	@code
	//FGPoint oldPos is defined somewhere else
	MyWidget::mouseMoveEvent( QMouseEvent *e )
	{
		FGPoint vector = e->pos() - oldPos;
		if ( vector.manhattanLength() > 3 )
		... //mouse has moved more than 3 pixels since oldPos
	}
	@endcode

	FGPoints can be compared for equality or inequality, and they can
	be written to and read from a QStream.

	@sa FGPointArray FGRect
*/
struct FGPoint
{
	FGCOORD x, y;
	/*!
		Constructs a point with coordinates (0, 0) (isNull() returns TRUE).
	*/
	FGPoint()
	{
		x=y=0;
	}
	/*!
		Constructs a point with x value @a xpos and y value @a ypos.
	*/
	FGPoint(int xpos, int ypos)
	{
		x = xpos;
		y = ypos;
	}

	/*!
		Constructs a point with x value @a xpos and y value @a ypos.
	*/
	FGPoint(double xpos, double ypos)
	{
		x = (int)xpos;
		y = (int)ypos;
	}

	/*!
		Returns TRUE if both the x value and the y value are 0; otherwise
		returns FALSE.
	*/
	inline bool isNull() const
	{ return x == 0 && y == 0; }

	/*!
		Adds point @a p to this point and returns a reference to this
		point.

		Example:
		@code
		FGPoint p(  3, 7 );
		FGPoint q( -1, 4 );
		p += q;            // p becomes (2,11)
		@endcode
	*/
	inline FGPoint &operator+=( const FGPoint &p )
	{ x+=p.x; y+=p.y; return *this; }

	/*!
		Subtracts point @a p from this point and returns a reference to
		this point.

		Example:
		@code
		FGPoint p(  3, 7 );
		FGPoint q( -1, 4 );
		p -= q;            // p becomes (4,3)
		@endcode
	*/
	inline FGPoint &operator-=( const FGPoint &p )
	{ x-=p.x; y-=p.y; return *this; }

	/*!
		Multiplies this point's x and y by @a c, and returns a reference
		to this point.

		Example:
		@code
		FGPoint p( -1, 4 );
		p *= 2;            // p becomes (-2,8)
		@endcode
	*/
	inline FGPoint &operator*=( int c )
	{ x*=(FGCOORD)c; y*=(FGCOORD)c; return *this; }

	/*!
		Multiplies this point's x and y by @a c, and returns a reference
		to this point.

		Example:
		@code
		FGPoint p( -1, 4 );
		p *= 2.5;          // p becomes (-3,10)
		@endcode

		Note that the result is truncated because points are held as
		integers.
	*/
	inline FGPoint &operator*=( double c )
	{ x=(FGCOORD)(x*c); y=(FGCOORD)(y*c); return *this; }
	/*!
		Divides both x and y by @a c, and returns a reference to this
		point.

		Example:
		@code
		FGPoint p( -2, 8 );
		p /= 2;            // p becomes (-1,4)
		@endcode
	*/
	inline FGPoint &operator/=( int c )
	{
		x/=(FGCOORD)c;
		y/=(FGCOORD)c;
		return *this;
	}

	/*!
		Divides both x and y by @a c, and returns a reference to this
		point.

		Example:
		@code
		FGPoint p( -3, 10 );
		p /= 2.5;           // p becomes (-1,4)
		@endcode

		Note that the result is truncated because points are held as
		integers.
	*/
	inline FGPoint &operator/=( double c )
	{
		x=(FGCOORD)(x/c);
		y=(FGCOORD)(y/c);
		return *this;
	}

	/**
		Returns the sum of the absolute values of x() and y(),
		traditionally known as the "Manhattan length" of the vector from
		the origin to the point. The tradition arises because such
		distances apply to travelers who can only travel on a rectangular
		grid, like the streets of Manhattan.

		This is a useful, and quick to calculate, approximation to the
		true length: sqrt(pow(x(),2)+pow(y(),2)).
	*/
	int ManhattanLength(void) const
	{
		return labs(x)+labs(y);
	}

	friend inline bool	 operator==( const FGPoint &, const FGPoint & );
	friend inline bool	 operator!=( const FGPoint &, const FGPoint & );
	friend inline const FGPoint operator+( const FGPoint &, const FGPoint & );
	friend inline const FGPoint operator-( const FGPoint &, const FGPoint & );
	friend inline const FGPoint operator*( const FGPoint &, int );
	friend inline const FGPoint operator*( int, const FGPoint & );
	friend inline const FGPoint operator*( const FGPoint &, double );
	friend inline const FGPoint operator*( double, const FGPoint & );
	friend inline const FGPoint operator-( const FGPoint & );
	friend inline const FGPoint operator/( const FGPoint &, int );
	friend inline const FGPoint operator/( const FGPoint &, double );
};

/*!
	@relates FGPoint
	Returns TRUE if @a p1 and @a p2 are equal; otherwise returns FALSE.
*/
inline bool operator==( const FGPoint &p1, const FGPoint &p2 )
{ return p1.x == p2.x && p1.y == p2.y; }

/*!
	@relates FGPoint

	Returns TRUE if @a p1 and @a p2 are not equal; otherwise returns FALSE.
*/
inline bool operator!=( const FGPoint &p1, const FGPoint &p2 )
{ return p1.x != p2.x || p1.y != p2.y; }


/*!
	@relates FGPoint

	Returns the sum of @a p1 and @a p2; each component is added separately.
*/
inline const FGPoint operator+( const FGPoint &p1, const FGPoint &p2 )
{ return FGPoint(p1.x+p2.x, p1.y+p2.y); }


/*!
	@relates FGPoint

	Returns @a p2 subtracted from @a p1; each component is subtracted
	separately.
*/

inline const FGPoint operator-( const FGPoint &p1, const FGPoint &p2 )
{ return FGPoint(p1.x-p2.x, p1.y-p2.y); }

/*!
	@relates FGPoint

	Returns the FGPoint formed by multiplying both components of @a p
	by @a c.
*/
inline const FGPoint operator*( const FGPoint &p, int c )
{ return FGPoint(p.x*c, p.y*c); }

/*!
	@relates FGPoint

	Returns the FGPoint formed by multiplying both components of @a p
	by @a c.
*/
inline const FGPoint operator*( int c, const FGPoint &p )
{ return FGPoint(p.x*c, p.y*c); }

/*!
	@relates FGPoint

	Returns the FGPoint formed by multiplying both components of @a p
	by @a c.

	Note that the result is truncated because points are held as
	integers.
*/
inline const FGPoint operator*( const FGPoint &p, double c )
{ return FGPoint(p.x*c, p.y*c); }

/*!
	@relates FGPoint

	Returns the FGPoint formed by multiplying both components of @a p
	by @a c.

	Note that the result is truncated because points are held as
	integers.
*/
inline const FGPoint operator*( double c, const FGPoint &p )
{ return FGPoint((FGCOORD)(p.x*c), (FGCOORD)(p.y*c)); }

/*!
	@relates FGPoint

	Returns the FGPoint formed by changing the sign of both components
	of @a p, equivalent to @c {FGPoint(0,0) - p}.
*/
inline const FGPoint operator-( const FGPoint &p )
{ return FGPoint(-p.x, -p.y); }

/*!
	@relates FGPoint

	Returns the FGPoint formed by dividing both components of @a p by
	@a c.
*/
inline const FGPoint operator/( const FGPoint &p, int c )
{
	return FGPoint(p.x/c, p.y/c);
}
/*!
	@relates FGPoint

	Returns the FGPoint formed by dividing both components of @a p
	by @a c.

	Note that the result is truncated because points are held as
	integers.
*/
inline const FGPoint operator/( const FGPoint &p, double c )
{
	return FGPoint((FGCOORD)(p.x/c), (FGCOORD)(p.y/c));
}

/**
  \brief The FGSize class defines the size of a two-dimensional object.

  \ingroup images
  \ingroup graphics

  A size is specified by a width and a height.

  The coordinate type is FGCOORD (defined as \c int).
  The minimum value of FGCOORD is FGCOORD_MIN (-2147483648) and the maximum
  value is FGCOORD_MAX (2147483647).

  The size can be set in the constructor and changed with setWidth()
  and setHeight(), or using operator+=(), operator-=(), operator*=()
  and operator/=(), etc. You can swap the width and height with
  transpose(). You can get a size which holds the maximum height and
  width of two sizes using expandedTo(), and the minimum height and
  width of two sizes using boundedTo().

  \sa FGPoint, FGRect
*/
class FGSize
{
	enum ScaleMode
	{
		ScaleFree,
		ScaleMin,
		ScaleMax
	};
public:
	FGCOORD wd;
	FGCOORD ht;

	FGSize();
	FGSize( int w, int h );

	bool   isNull()	const;
	bool   isEmpty()	const;
	bool   isValid()	const;

	int	   width()	const;
	int	   height()	const;
	void   setWidth( int w );
	void   setHeight( int h );
	void   transpose();

	FGSize expandedTo( const FGSize & ) const;
	FGSize boundedTo( const FGSize & ) const;

	void scale( int w, int h, ScaleMode mode );
	void scale( const FGSize &s, ScaleMode mode );

	FGCOORD &rwidth();
	FGCOORD &rheight();

	FGSize &operator+=( const FGSize & );
	FGSize &operator-=( const FGSize & );
	FGSize &operator*=( int c );
	FGSize &operator*=( double c );
	FGSize &operator/=( int c );
	FGSize &operator/=( double c );

	friend inline bool	operator==( const FGSize &, const FGSize & );
	friend inline bool	operator!=( const FGSize &, const FGSize & );
	friend inline const FGSize operator+( const FGSize &, const FGSize & );
	friend inline const FGSize operator-( const FGSize &, const FGSize & );
	friend inline const FGSize operator*( const FGSize &, int );
	friend inline const FGSize operator*( int, const FGSize & );
	friend inline const FGSize operator*( const FGSize &, double );
    friend inline const FGSize operator*( double, const FGSize & );
	friend inline const FGSize operator/( const FGSize &, int );
    friend inline const FGSize operator/( const FGSize &, double );
};

/*****************************************************************************
  FGSize inline functions
 *****************************************************************************/

inline FGSize::FGSize()
{ wd = ht = -1; }

inline FGSize::FGSize( int w, int h )
{ wd=(FGCOORD)w; ht=(FGCOORD)h; }

inline bool FGSize::isNull() const
{ return wd==0 && ht==0; }

inline bool FGSize::isEmpty() const
{ return wd<1 || ht<1; }

inline bool FGSize::isValid() const
{ return wd>=0 && ht>=0; }

inline int FGSize::width() const
{ return wd; }

inline int FGSize::height() const
{ return ht; }

inline void FGSize::setWidth( int w )
{ wd=(FGCOORD)w; }

inline void FGSize::setHeight( int h )
{ ht=(FGCOORD)h; }

inline FGCOORD &FGSize::rwidth()
{ return wd; }

inline FGCOORD &FGSize::rheight()
{ return ht; }

inline FGSize &FGSize::operator+=( const FGSize &s )
{ wd+=s.wd; ht+=s.ht; return *this; }

inline FGSize &FGSize::operator-=( const FGSize &s )
{ wd-=s.wd; ht-=s.ht; return *this; }

inline FGSize &FGSize::operator*=( int c )
{ wd*=(FGCOORD)c; ht*=(FGCOORD)c; return *this; }

inline FGSize &FGSize::operator*=( double c )
{ wd=(FGCOORD)(wd*c); ht=(FGCOORD)(ht*c); return *this; }

inline bool operator==( const FGSize &s1, const FGSize &s2 )
{ return s1.wd == s2.wd && s1.ht == s2.ht; }

inline bool operator!=( const FGSize &s1, const FGSize &s2 )
{ return s1.wd != s2.wd || s1.ht != s2.ht; }

inline const FGSize operator+( const FGSize & s1, const FGSize & s2 )
{ return FGSize(s1.wd+s2.wd, s1.ht+s2.ht); }

inline const FGSize operator-( const FGSize &s1, const FGSize &s2 )
{ return FGSize(s1.wd-s2.wd, s1.ht-s2.ht); }

inline const FGSize operator*( const FGSize &s, int c )
{ return FGSize(s.wd*c, s.ht*c); }

inline const FGSize operator*( int c, const FGSize &s )
{  return FGSize(s.wd*c, s.ht*c); }

inline const FGSize operator*( const FGSize &s, double c )
{ return FGSize((FGCOORD)(s.wd*c), (FGCOORD)(s.ht*c)); }

inline const FGSize operator*( double c, const FGSize &s )
{ return FGSize((FGCOORD)(s.wd*c), (FGCOORD)(s.ht*c)); }

inline FGSize &FGSize::operator/=( int c )
{
	wd/=(FGCOORD)c; ht/=(FGCOORD)c;
	return *this;
}

inline FGSize &FGSize::operator/=( double c )
{
	wd=(FGCOORD)(wd/c); ht=(FGCOORD)(ht/c);
	return *this;
}

inline const FGSize operator/( const FGSize &s, int c )
{
	return FGSize(s.wd/c, s.ht/c);
}

inline const FGSize operator/( const FGSize &s, double c )
{
	return FGSize((FGCOORD)(s.wd/c), (FGCOORD)(s.ht/c));
}

inline FGSize FGSize::expandedTo( const FGSize & otherSize ) const
{
	return FGSize( MAX(wd,otherSize.wd), MAX(ht,otherSize.ht) );
}

inline FGSize FGSize::boundedTo( const FGSize & otherSize ) const
{
	return FGSize( MIN(wd,otherSize.wd), MIN(ht,otherSize.ht) );
}

/**
	Circle object.
*/
struct FGCircle : public FGPoint
{
	FGCOORD     r;

	/*!
		Constructs an invalid rectangle.
	*/
	FGCircle()
	 : FGPoint()
	{
		r=0;
	}

	/*!
		Constructs a circle with the \a top, \a left corner and \a
		width and \a height.
	*/
	FGCircle(int xpos, int ypos, int radius)
	: FGPoint(xpos,ypos)
	{
		r = radius;
	}

	/*!
	*/
	FGCircle(const FGPoint point, int radius)
	: FGPoint(point)
	{
		r = radius;
	}
	/*!
		Returns TRUE if the circle is valid; otherwise returns FALSE.

		A valid circle has a left() \<= right() and top() \<= bottom().

		Note that non-trivial operations like intersections are not defined
		for invalid circles.

		\c {isValid() == !isEmpty()}

		\sa isNull(), isEmpty(), normalize()
	*/
	inline bool isValid(void) const
	{
		return r >= 0;
	}

	/*!
		Returns TRUE if the circle is a null circle; otherwise
		returns FALSE.

		A null rectangle has both the width and the height set to 0, that
		is right() == left() - 1 and bottom() == top() - 1.

		Note that if right() == left() and bottom() == top(), then the
		circle has width 1 and height 1.

		A null circle is also empty.

		A null circle is not valid.

		\sa isEmpty(), isValid()
	*/
	inline bool isNull(void) const
	{ return r == 0; }

	/*!
		Returns TRUE if the circle is empty; otherwise returns FALSE.

		An empty circle has a left() \> right() or top() \> bottom().

		An empty circle is not valid. \c {isEmpty() == !isValid()}

		\sa isNull(), isValid(), normalize()
	*/
	inline bool isEmpty(void) const
	{
		return r < 0;
	}
	/*!
		Returns TRUE if the point \a x, \a y is inside this circle,
		otherwise returns FALSE.

		If \a proper is TRUE, this function returns TRUE only if the point
		is entirely inside (not on the edge).
	*/
	bool   contains( double xc, double yc) const
	{
		if ( ((x-xc)*(x-xc) + (y-yc)*(y-yc)) <= ((double)r*r) )
			return true;
		return 0;
	}
};


/*!
	@brief The FGRect class defines a rectangle in the plane.

	@ingroup images
	@ingroup graphics

	A rectangle is internally represented as an upper-left corner and
	a bottom-right corner, but it is normally expressed as an
	upper-left corner and a size.

	The coordinate type is FGCOORD (defined in @c qwindowdefs.h as @c
	int). The minimum value of FGCOORD is FGCOORD_MIN (-2147483648) and
	the maximum value is  FGCOORD_MAX (2147483647).

	Note that the size (width and height) of a rectangle might be
	different from what you are used to. If the top-left corner and
	the bottom-right corner are the same, the height and the width of
	the rectangle will both be 1.

	Generally, @ e{width = right - left + 1} and @e {height = bottom -
	top + 1}. We designed it this way to make it correspond to
	rectangular spaces used by drawing functions in which the width
	and height denote a number of pixels. For example, drawing a
	rectangle with width and height 1 draws a single pixel.

	The default coordinate system has origin (0, 0) in the top-left
	corner. The positive direction of the y axis is down, and the
	positive x axis is from left to right.

	A FGRect can be constructed with a set of left, top, width and
	height integers, from two FGPoints or from a FGPoint and a FGSize.
	After creation the dimensions can be changed, e.g. with setLeft(),
	setRight(), setTop() and setBottom(), or by setting sizes, e.g.
	setWidth(), setHeight() and setSize(). The dimensions can also be
	changed with the move functions, e.g. moveBy(), moveCenter(),
	moveBottomRight(), etc. You can also add coordinates to a
	rectangle with addCoords().

	You can test to see if a FGRect contains a specific point with
	contains(). You can also test to see if two FGRects intersect with
	intersects() (see also intersect()). To get the bounding rectangle
	of two FGRects use unite().

	@sa FGPoint
*/
struct FGRect : public FGPoint
{
	FGCOORD     w;
	FGCOORD     h;
	/*!
		Constructs an invalid rectangle.
	*/
	FGRect()
	 : FGPoint()
	{
		w=h=0;
	}

	//! copy
	FGRect(const FGRect& old) { memcpy(this, &old, sizeof(*this)); }

	/*!
		Constructs a rectangle with the \a top, \a left corner and \a
		width and \a height.

		Example (creates two identical rectangles):
		\code
		FGRect r1( FGPoint(100,200), FGPoint(110,215) );
		FGRect r2( 100, 200, 11, 16 );
		\endcode
	*/
	FGRect(int xpos, int ypos, int width, int height)
	: FGPoint(xpos,ypos)
	{
		w = width;
		h = height;
	}

	/*!
		Constructs a rectangle with the top left corner at (0,0) and \a
		width and \a height.
	*/
	FGRect(int width, int height)
	: FGPoint(0, 0), w(width), h(height)
	{}

	/*!
		Constructs a rectangle with \a topLeft as the top-left corner and
		\a bottomRight as the bottom-right corner.
	*/
	FGRect(const FGPoint r0, const FGPoint r1)
	: FGPoint(r0)
	{
		w = r1.x - x;
		h = r1.y - y;
	}

	/*!
		Constructs a rectangle with \a topLeft as the top-left corner and
		\a bottomRight as the bottom-right corner.
	*/
	FGRect(const FGSize& sz)
	: FGPoint()
	{
		w = sz.width();
		h = sz.height();
	}

	/*!
		Constructs a rectangle with \a topLeft as the top-left corner and
		\a bottomRight as the bottom-right corner.
	*/
	FGRect(const FGCircle c)
	: FGPoint(c.x-c.r, c.y-c.r)
	{
		w = h = c.r*2;
	}

	/*!
		Returns TRUE if the rectangle is valid; otherwise returns FALSE.

		A valid rectangle has a left() \<= right() and top() \<= bottom().

		Note that non-trivial operations like intersections are not defined
		for invalid rectangles.

		\c {isValid() == !isEmpty()}

		\sa isNull(), isEmpty(), normalize()
	*/
	inline bool isValid(void) const
	{
		return w >= 0 && h >= 0;
	}

	/*!
		Returns TRUE if the rectangle is a null rectangle; otherwise
		returns FALSE.

		A null rectangle has both the width and the height set to 0, that
		is right() == left() - 1 and bottom() == top() - 1.

		Note that if right() == left() and bottom() == top(), then the
		rectangle has width 1 and height 1.

		A null rectangle is also empty.

		A null rectangle is not valid.

		\sa isEmpty(), isValid()
	*/
	inline bool isNull(void) const
	{ return w == 0 && h == 0; }

	/*!
		Returns TRUE if the rectangle is empty; otherwise returns FALSE.

		An empty rectangle has a left() \> right() or top() \> bottom().

		An empty rectangle is not valid. \c {isEmpty() == !isValid()}

		\sa isNull(), isValid(), normalize()
	*/
	inline bool isEmpty(void) const
	{
		return w < 0 || h > 0;
	}

	/*!
		Returns a normalized rectangle, i.e. a rectangle that has a
		non-negative width and height.

		It swaps left and right if left() \> right(), and swaps top and
		bottom if top() \> bottom().

		\sa isValid()
	*/
	FGRect  normalize(void)	const
	{
		return *this;
	}

	/*!
		Returns the left coordinate of the rectangle. Identical to x().

		\sa setLeft(), right(), topLeft(), bottomLeft()
	*/
	FGCOORD left(void) const { return x; }

	/*!
		Returns the right coordinate of the rectangle.

		\sa setRight(), left(), topRight(), bottomRight()
	*/
	FGCOORD right(void) const { return x+w; }
	/*!
		Returns the top coordinate of the rectangle. Identical to y().

		\sa setTop(), bottom(), topLeft(), topRight()
	*/
	FGCOORD top(void) const { return y; }
	/*!
		Returns the bottom coordinate of the rectangle.

		\sa setBottom(), top(), bottomLeft(), bottomRight()
	*/
	FGCOORD bottom(void) const { return y+h; }
	/*!
		Returns the top-left position of the rectangle.

		\sa setTopLeft(), moveTopLeft(), bottomRight(), left(), top()
	*/
	FGPoint topLeft(void) const { return FGPoint(x,y); }
	/*!
		Returns the bottom-right position of the rectangle.

		\sa setBottomRight(), moveBottomRight(), topLeft(), right(), bottom()
	*/
	FGPoint bottomRight(void) const { return FGPoint(int(x+w), int(y+h)); }
	/*!
		\sa setTopRight(), moveTopRight(), bottomLeft(), top(), right()
	*/
	FGPoint topRight(void) const { return FGPoint(x+w,y); }
	/*!
		Returns the bottom-left position of the rectangle.

		\sa setBottomLeft(), moveBottomLeft(), topRight(), bottom(), left()
	*/
	FGPoint bottomLeft(void) const { return FGPoint(x,y+h); }
	/*!
		Returns the center point of the rectangle.

		\sa moveCenter(), topLeft(), bottomRight(), topRight(), bottomLeft()
	*/
	FGPoint center(void) const { return FGPoint( int(x+w/2), int(y+h/2)); }
	/*!
		Returns the bounding rectangle of this rectangle and rectangle \a
		r.

		The bounding rectangle of a nonempty rectangle and an empty or
		invalid rectangle is defined to be the nonempty rectangle.

		\sa operator|=(), operator&(), intersects(), contains()
	*/
	FGRect  operator|(const FGRect &r) const;
	/*!
		Returns the intersection of this rectangle and rectangle \a r.

		Returns an empty rectangle if there is no intersection.

		\sa operator&=(), operator|(), isEmpty(), intersects(), contains()
	*/
	FGRect  operator&(const FGRect &r) const;
	/*!
		Unites this rectangle with rectangle \a r.
	*/
	FGRect&  operator|=(const FGRect &r);
	/*!
		Intersects this rectangle with rectangle \a r.
	*/
	FGRect&  operator&=(const FGRect &r);
	/*!
		Returns TRUE if the point \a p is inside or on the edge of the
		rectangle; otherwise returns FALSE.

		If \a proper is TRUE, this function returns TRUE only if \a p is
		inside (not on the edge).
	*/
	bool   contains( const FGPoint &p, bool proper=false ) const;
	/*!
		Returns TRUE if the point \a x, \a y is inside this rectangle;
		otherwise returns FALSE.

		If \a proper is TRUE, this function returns TRUE only if the point
		is entirely inside (not on the edge).
	*/
	bool   contains( int x, int y, bool proper=false ) const;
	/*!
		\overload

		Returns TRUE if the rectangle \a r is inside this rectangle;
		otherwise returns FALSE.

		If \a proper is TRUE, this function returns TRUE only if \a r is
		entirely inside (not on the edge).

		\sa unite(), intersect(), intersects()
	*/
	bool   contains( const FGRect &r, bool proper=false ) const;
	/*!
		Returns the bounding rectangle of this rectangle and rectangle \a
		r. \c {r.unite(s)} is equivalent to \c {r|s}.
	*/
	FGRect unite( const FGRect &r ) const;
	/*!
		Returns the intersection of this rectangle and rectangle \a r.
		\c {r.intersect(s)} is equivalent to \c {r&s}.
	*/
	FGRect intersect( const FGRect &r ) const;
	/*!
		Returns TRUE if this rectangle intersects with rectangle \a r
		(there is at least one pixel that is within both rectangles);
		otherwise returns FALSE.

		\sa intersect(), contains()
	*/
	bool   intersects( const FGRect &r ) const
	{
		return ( MAX( x, r.x ) <= MIN( w, r.w ) &&
			 MAX( y, r.y ) <= MIN( h, r.h ) );
	}
	friend bool operator==( const FGRect &, const FGRect & );
	friend bool operator!=( const FGRect &, const FGRect & );
};

/*!
	@relates FGRect
	@returns TRUE if \a r1 and \a r2 are equal; otherwise returns FALSE.
*/
bool operator==( const FGRect &r1, const FGRect &r2 );
/*!
	@relates FGRect
	@returns TRUE if \a r1 and \a r2 are different; otherwise returns FALSE.
*/
bool operator!=( const FGRect &r1, const FGRect &r2 );

/*!
	@brief The FGPointArray class provides an array of points.

	@ingroup images
	@ingroup graphics

	A FGPointArray is an array of FGPoint objects. In addition to the
	functions provided by QMemArray, FGPointArray provides some
	point-specific functions.

	For convenient reading and writing of the point data use
	setPoints(), putPoints(), point(), and setPoint().

	For geometry operations use boundingRect() and translate(). There
	is also the FGMatrix::map() function for more general
	transformations of FGPointArrays. You can also create arcs and
	ellipses with makeArc() and makeEllipse().

	Among others, FGPointArray is used by QPainter::drawLineSegments(),
	QPainter::drawPolyline(), QPainter::drawPolygon() and
	QPainter::drawCubicBezier().

	Note that because this class is a QMemArray, copying an array and
	modifying the copy modifies the original as well, i.e. a shallow
	copy. If you need a deep copy use copy() or detach(), for example:

	@code
	void drawGiraffe( const FGPointArray & r, QPainter * p )
	{
		FGPointArray tmp = r;
		tmp.detach();
		// some code that modifies tmp
		p->drawPoints( tmp );
	}
	@endcode

	If you forget the tmp.detach(), the const array will be modified.

	@sa FGMatrix
*/
struct FGPointArray
{
	unsigned int    vertices;
	FGPoint*        array;
	bool			allocated;

	/**
	*  Resizes the array of the FGPoints.
	*/
	void resize(unsigned new_size)
	{
		if (new_size != vertices)
		{
			if (allocated)
				delete [] array;

			array = new FGPoint[vertices = new_size];
			allocated = true;
		}
	}
	/*!
		Returns TRUE if the array is empty; otherwise returns FALSE.

		An empty rectangle has a zero number of vertices.

	*/
	inline bool isEmpty(void) const
	{
		return vertices <= 0;
	}
	/*!
		Constructs a null point array.
		\sa isNull()
	*/
	FGPointArray()
	{
		vertices=0;
		array = 0;
		allocated = false;
	}
	/*!
		Destroys the point array.
	*/
	~FGPointArray()
	{
		if (allocated)
			if (array) delete [] array;
		array = 0;
	}
	/*!
		Constructs a point array with room for \a size points. Makes a
		null array if \a size == 0.

		\sa resize(), isNull()
	*/
	FGPointArray(unsigned int nPoints)
	: vertices(nPoints)
	{
		if (vertices > 0)
		{
			array = new FGPoint[vertices];
			allocated = true;
		}
	}
	/*!
		Constructs a deep copy of the point array \a a.
	*/
	FGPointArray(const FGPointArray& old)
	{
		vertices = old.vertices;
		array = new FGPoint[vertices];;
		allocated = true;

		for(unsigned int i=0; i<vertices; i++)
			array[i] = old.array[i];
	}
	/*!
	  Constructs a point array with \a nPoints points, taken from the
	  \a FGPoint array.

	  Equivalent to setPoints(nPoints, points).
	*/
	FGPointArray(unsigned int nPoints, FGPoint* points)
	: vertices(nPoints)
	{
		array = new FGPoint[vertices];
		allocated = true;
		for(unsigned int i=0; i<vertices; i++)
			array[i] = points[i];
	}
	/*!
	  Constructs a point array with \a nPoints points, taken from the
	  \a points array.

	  Equivalent to SetPoints(nPoints, points).
	*/
	FGPointArray(unsigned int nPoints, FGCOORD points[][2])
	: vertices(nPoints)
	{
		array = new FGPoint[vertices];
		allocated = true;

		SetPoints(vertices, (FGCOORD *)points);
	}
	/*!
		Constructs a point array from the rectangle \a r.

		If \a closed is FALSE, then the point array just contains the
		following four points in the listed order: r.topLeft(),
		r.topRight(), r.bottomRight() and r.bottomLeft().

		If \a closed is TRUE, then a fifth point is set to r.topLeft().
	*/
	FGPointArray( const FGRect &r, bool closed=false )
	{
		vertices = closed ? 5 : 4;
		array = new FGPoint[vertices];
		allocated = true;

		array[0] = r.topLeft();
		array[1] = r.bottomRight();
		array[2] = r.topRight();
		array[3] = r.bottomLeft();
		if (vertices == 5)
			array[4] = r.topLeft();
	}
	/*!
	  Constructs a point array with \a nPoints points, taken from the
	  \a points array.

	  Equivalent to SetPoints(nPoints, points).
	*/
	FGPointArray( unsigned int nPoints, const FGCOORD *points )
	 : vertices(nPoints)
	{
		array = new FGPoint[vertices];
		allocated = true;

		SetPoints(vertices, points);
	}
	/*!
	  \internal
	  Resizes the array to \a nPoints and sets the points in the array to
	  the values taken from \a points.

	  Returns TRUE if successful, or FALSE if the array could not be
	  resized (normally due to lack of memory).

	  The example code creates an array with two points (1,2) and (3,4):
	  \code
		static FGCOORD points[] = { 1,2, 3,4 };
		FGPointArray a;
		a.SetPoints( 2, points );
	  \endcode

	  \sa resize(), putPoints()
	*/
	void SetPoints( unsigned int nPoints, const FGCOORD *points )
	{
		resize(nPoints);
		for(unsigned int i=0; i<vertices; i++)
		{
			array[i].x = *points++;
			array[i].y = *points++;
		}
	}
	/*!
		\overload

		Resizes the array to \a nPoints and sets the points in the array
		to the values taken from the variable argument list.

		Returns TRUE if successful, or FALSE if the array could not be
		resized (typically due to lack of memory).

		The example code creates an array with two points (1,2) and (3,4):

		\code
		FGPointArray a;
		a.SetPoints( 2, 1,2, 3,4 );
		\endcode

		The points are given as a sequence of integers, starting with \a
		firstx then \a firsty, and so on.

		\sa resize(), putPoints()
	*/
	void SetPoints( unsigned nPoints, int firstx, int firsty, ... )
	{
		SetPoints(vertices, (FGCOORD *) &firstx);
	}
	/*!
	  \internal
	  Resizes the array to \a nPoints and sets the points in the array to
	  the values taken from \a points.

	  Returns TRUE if successful, or FALSE if the array could not be
	  resized (normally due to lack of memory).

	  The example code creates an array with two points (1,2) and (3,4):
	  \code
		static FGCOORD points[] = { 1,2, 3,4 };
		FGPointArray a;
		a.SetPoints( 2, points );
	  \endcode

	  \sa resize(), putPoints()
	*/
	void SetPoints( unsigned int nPoints, const double *points )
	{
		resize(nPoints);
		for(unsigned int i=0; i<vertices; i++)
		{
			array[i].x = FGCOORD(*points++);
			array[i].y = FGCOORD(*points++);
		}
	}
	/*!
		Translates all points in the array by \a (dx, dy).
	*/
	void translate( int dx, int dy )
	{
		register FGPoint *p = array;
		register int i = vertices;
		FGPoint pt( dx, dy );
		while ( i-- )
		{
			*p += pt;
			p++;
		}
	}
	/**
		Compute the bounding rect of polygon vertices.
	*/
	FGRect BoundingRect() const;

	/*!
		@return TRUE if the point \a p is inside this object;
		otherwise returns FALSE.
	*/
	bool   contains(const FGPoint &p) const;
	/*!
		Returns TRUE if the point \a x, \a y is inside this rectangle;
		otherwise returns FALSE.
	*/
	bool   contains(int x, int y) const;
};


#define MAX_FONTS		64

/**
	Predefined const for a built-in fixed and proportional fonts.
	@ingroup Enums
*/
enum ENUM_FONTS
{
	FONT0406, FONT0808, FONT0816, FONT1222, FONT1625, FONT2034, FONTSYS, FONTSYSMEDIUM, FONTSYSLIGHT, FONTLAST=(MAX_FONTS-1)
};

/**
 Describes the TrueType font provided by freetype library.
*/
class FGFontProperty
{
	public:
		/** for rendering */
		enum FONT_STYLE { ttfSolid, ttfShaded, ttfBlended };
	private:
		void	 	*_font;
		FONT_STYLE 	algorithm;
	public:
		/**
			Creates a TTF font face object for a Window text output.
			You can use a function Valid() to test for a valid font face.
			@param n a filename ("arial.ttf" by example).
			@param sz a size of font face in pixels.
		*/
		FGFontProperty(char *n, int sz);
		/**
		Unload the font face from memory.
		*/
		~FGFontProperty();
		/**
		Sets a 'bold' font style.
		*/
		void Bold(int value);
		/**
		Sets an 'underline' font style.
		*/
		void Underline(int value);
		/**
		Sets an 'italic' font style.
		*/
		void Italic(int value);
		/**
		Sets a solid rendering, fast but not very nice. Default.
		*/
		void Solid(void) { algorithm = ttfSolid; }
		/**
		Sets a shaded rendering, slow but very nice.
		*/
		void Shaded(void) { algorithm = ttfShaded; }
		/**
		Sets a blended rendering, slow but very nice.
		@bug still not implemented.
		*/
		void Blended(void) { algorithm = ttfBlended; } // not implemented yet
		/**
		Test the object for a valid data.
		@return true if object contains a valid font face.
		*/
		bool Valid(void) { return !!_font; }
		FONT_STYLE GetAlgorithm(void) { return algorithm; }
		void * GetFace(void) { return _font; }
		char * PrepareFont(variable_record *var, int& size, int count, int offset, char *name);
};

/**
	The OpenGUI interface to the font manager.
	It registers and deregisters variable in-memory or in-file fonts.
	These fonts maybe with fixed width or proportional width.
	Object provides these fonts to others.
*/
class FGFontManager
{
		static FGFont 	FGFonts[MAX_FONTS];
		static int 		font_counter;
		// register variable font from memory
		// img and data must be allocated with malloc
		static int FGAPI register_font_var(variable_record *data, char *img, int size, char *desc);
		// register variable font from compiled font
		static int FGAPI register_font_image(fonthdr * hdr, char *img, char *desc);
		static void FGAPI deregister_fonts(void);
		static int FGAPI _register_font(unsigned char *source, int width, int height, int count, int offset, variable_record * type, int index, char *desc);
	public:
		FGFontManager();
		~FGFontManager();

		static int FGAPI register_font_ttf(char *filename, int points, int count=128-' ', int offset=' ', char *desc=0);
		static int FGAPI register_font_fix(unsigned char *source, int width, int height, int count, int offset, char *desc);
		static int FGAPI register_font_file(const char *source, char *desc);

		static int FGAPI textwidth(int ID, const char *txt);
		static int FGAPI textwidth(int ID, const char *txt, int cnt);
		static int FGAPI charwidth(int ID, int c);

		/**
		Tests font ID and returns pointer to.
		@return a valid font record if exists one, or 0 if doesn't exist.
		@param ID font ID for prepared font
		*/
		static FGFont * FGAPI get_font_ptr(int ID)
		{
			if (ID<0 || ID>font_counter)
				return 0;
			// test for defined fonts
			if (!FGFonts[ID].fontimg)
				return 0;
			return &FGFonts[ID];
		}
		/**
		@return returns the width in pixels for font ID or 0 if doesn't exists.
		for proportional fonts there is the width of bigger glyph in the font.
		@param ID font ID
		*/
		static inline int FGAPI GetW(int ID)
		{
			return FGFonts[ID].width;
		}
		/**
		@return returns the height in pixels for font ID or 0 if doesn't exists.
		@param ID font ID
		*/
		static inline int FGAPI GetH(int ID)
		{
			return FGFonts[ID].height;
		}
		/**
		@return returns the pointer to a structure for proportional
		font description or 0 if font is fixed size.
		@param ID font ID
		*/
		static inline variable_record * FGAPI IsVariableFont(int ID)
		{
			return FGFonts[ID].var;
		}
		/**
		@return returns the font pixelmap or 0 if desn't exists.
		@param ID font ID
		*/
		static inline FGPixel * FGAPI GetFontImage(int ID)
		{
			return FGFonts[ID].fontimg;
		}
		/**
		@return the font name (max. 23 chars in length+'\\0').
		@param ID font ID
		*/
		static inline char * FGAPI GetFontName(int ID)
		{
			return FGFonts[ID].name;
		}
		/**
		@return the number of registered fonts in system.
		*/
		static inline int FGAPI GetCounter(void)
		{
			return font_counter;
		}
};

extern FGFontManager fontman;

/**
	The structure describes a pattern for the patterned line drawing.
*/
struct FGPattern
{
	FGPixel *data;
	int len, masked, reset, pos;
	FGPixel next(void);

	FGPattern(FGPixel * pp, int l, int m, int r = 1):data(pp), len(l),
	  masked(m), reset(r), pos(0)
	{
	}
};

void delay(unsigned int);

#ifdef _WIN32
void sound(unsigned);
void nosound(void);
#endif

void outpb(unsigned, unsigned);
void outpw(unsigned, unsigned);
unsigned char inpb(unsigned);
unsigned short inpw(unsigned);
unsigned int inpl(unsigned);
void outpl(unsigned, unsigned);

extern FGPattern PatternDot , PatternSlashDot , PatternSlashDotSlash , PatternSlash ;

//unsigned long long _rdtsc(void);
#ifndef __BORLANDC__
char *strupr(register char *s);
int stricmp(const char *s1, const char *s2);
int strnicmp(const char *s1, const char *s2, size_t cnt);
char *strlwr(char *_s);
#endif
/**
	@addtogroup Graphics
	@{
*/
int  get_point(int x, int y);
void draw_hline(int x, int y, int w, FGPixel ink, unsigned ppop);
void draw_box(int x, int y, int w, int h, FGPixel ink, unsigned ppop);
void fill_box(int x, int y, int w, int h, FGPixel ink, unsigned ppop);
void draw_point(int x, int y, FGPixel color, unsigned ppop);
int  draw_line(int x, int y, int x1, int y1, FGPixel color, unsigned ppop);
#define draw_line(a,b,c,d,e,f) vector_draw_line(a,b,c,d,e,f)
int  FGAPI drawto_line(int x, int y, FGPixel ink, unsigned ppop);
void FGAPI draw_pattern_line(int x, int y, int x1, int y1, FGPattern *, unsigned ppop);
void FGAPI drawto_pattern_line(int x, int y, FGPattern *, unsigned ppop);
void FGAPI draw_pattern_box(int x, int y, int a, int b, FGPattern *pat, unsigned ppop);
void FGAPI draw_spline(int points[8], FGPixel ink, unsigned ppop);
void FGAPI calc_spline(int points[8], int npts, int *x, int *y);
void FGAPI moveto(int x, int y);
unsigned int FGAPI areasize(int x, int y);

void FGAPI _set_default_palette(void);
void FGAPI _set_fgl_palette(void);
unsigned int FGAPI get_palette(unsigned int i);
void _palette(unsigned, unsigned int);
void FGAPI CreatePaletteEntry(int rc, int gc, int bc, int idx);

int FGAPI text_out(int x, int y, char *txt, FGPixel ink, FGPixel paper, unsigned ppop, int f=FONT0816);
void FGAPI fill_convex(int c, int a[][2], FGPixel ink, unsigned ppop);
void FGAPI draw_convex(int c, int a[][2], FGPixel ink, unsigned ppop);
void FGAPI get_block(int x, int y, int a, int b, FGPixel * p);
void FGAPI put_block(int x, int y, int a, int b, FGPixel * p, unsigned ppop);
void FGAPI set_clip_rect(int w, int h, int x, int y);

void FGAPI fill_ellipse(int x, int y, int rx, int ry, FGPixel ink, unsigned ppop);
void FGAPI draw_ellipse(int x, int y, int rx, int ry, FGPixel ink, unsigned ppop);
void FGAPI draw_arc(int x, int y, double ang1, double ang2, int r, FGPixel ink, unsigned ppop);
void FGAPI draw_circle(int x, int y, int r, FGPixel ink, unsigned ppop);
void FGAPI fill_circle(int xs, int ys, int r, FGPixel ink, unsigned ppop);

void RamToVideo(FGPixel * Image, int x, int y, int xm, int ym, int w, int h, int cx_max, int cy_max, int opcia, unsigned ppop);
void VideoToRam(FGPixel * Image, int x, int y, int xm, int ym, int w, int h, int cx_max, int cy_max);
void RamToRam(int x, int y, int xmax, int ymax, int xdst, int ydst, int xmaxdst, int ymaxdst, int w, int h, FGPixel * from, FGPixel * to, unsigned alpha, unsigned ppop);
int gprintf(FGPixel ink, FGPixel paper, int x, int y, const char *format,...);
/** @} Graphics */

/**
	@addtogroup Colors
	@{
*/
int FGAPI GetFreeColors(void);
void FGAPI SetColorFuzzy(int a);
FGPixel FGAPI CreateColor(int, int, int, int);
void FGAPI DeleteColor(int);
void FGAPI set_transpcolor(FGPixel);
/** @} Colors */

void FGmemcpy(FGPixel * to, FGPixel * from, unsigned c);
void FGmemset(FGPixel * to, FGPixel data, unsigned c);
void FGmemcpy2(FGPixel * to, FGPixel * from, unsigned c, unsigned ppop);
void FGmemset2(FGPixel * to, FGPixel data, unsigned c, unsigned ppop);

/**
	Clears the whole screen.
	@ingroup Misc
	@param c background color
*/
void clear_frame_buffer(FGPixel c);
int graph_set_mode(int mode);
/**
	Change from one to other graphics mode with its resolution.
	@ingroup Misc
	@param ww the new X axis resolution
	@param hh the new Y axis resolution
	@return TRUE if OK
*/
int graph_change_mode(int ww, int hh);
void cleanup(void);
int FGAPI get_colordepth(void);
unsigned FGAPI GetXRes(void);
unsigned FGAPI GetYRes(void);
FGPixel * FGAPI GetFrameBuffer(void);

/**
	For Buffered animation.
	@ingroup Enums
*/
enum ENUM_BUFFERING
{
	FG_DOUBLEBUFFER=2,
	FG_TRIPLEBUFFER,
	FG_QUADBUFFER
};

/**
	Returns absolute time in msec from the start of application. Useful for timing.
	@ingroup Misc
*/
unsigned long FGAPI FGClock(void);
void FGAPI Snd(int a, int b);
void FGAPI Puk(void);

/**
	Returns the version string, by example: "4.1.0"
	@ingroup Misc
*/
const char* GetVer(void);
void UpdateRect(int x, int y, int xm, int ym, int w, int h);
int  GetPriv(void);

class FGDriver;

extern FGDriver *__fg_driver;

// draw clipped to screen line
extern shapper_t   vector_draw_line;

// draw clipped to screen rect
extern clipper_t   vector_clip_rect;

// draw clipped to screen plot
extern plotter_t   vector_draw_point;

// draw clipped to screen box
extern shapper_t   vector_fill_box;

// set 8 bit palette (shift needed for 6-bit palette)
// palette_8 is accepted
extern paletter_t  vector_palette;

// bitblit copy
extern blitter_t   vector_blit_copy;
extern blitter2_t   vector_blit_op;
extern blitter_a_t vector_blit_a;

/**
	This object is provided for common communication between
	any two OpenGUI graphics objects. By default is this unconnected.
	When you want connect the two objects to sending signal from one
	object to other, simply call the Connect() method. In called object
	you must overload virtual methods OnSignal(). When signal
	from one to other object is emited, there is one optional parameter
	that is transmited. You can set this parameter with Connect() or
	late by calling SetParam().
	@see FGCONNECT
*/
class FGConnector
{
		void* 	parameter1;
		class 	FGConnector *sendTo;
	public:
		/**
		This is an trivial way to solve problem with unavailable related user defined
		parameter. You can use method SetParam() to any Window to assign some parameter
		of int type. Since you can read this parameter at other place by GetParam().
		If returned value is -1, value is not assigned.
		*/
		void	FGAPI SetParameter(void* a) { parameter1 = a; }
		/**
		This is an trivial way to solve problem with unavailable related user defined
		parameter. You can use method SetParam() to any Window to assign some parameter
		of int type. Since you can read this parameter at other place by GetParam().
		If returned value is -1, value is not assigned.
		*/
		void*	FGAPI GetParameter(void)  const { return parameter1; }

		FGConnector()
		{
			SetParameter((void*)-1L);
			sendTo = 0;
		}
		/**
		Connects other object pointed by 'from' and set an optional value.
		@note There is allowed only one bindings between two objects at one time,
		'The last, is the winner'!
		@param from other object of this kind to connect with us
		@param par user value/event code/magic number
		*/
		void 	FGAPI Connect(FGConnector *from, void* par)
		{
			sendTo = from;
			SetParameter(par);
		}
		//! do nothing - override in your code
		virtual void OnSignal(FGConnector *sender, void* value)
		{
		}
		//! Try emit defined signal if there is any.
		void FGAPI RunSignal(void)
		{
			if (sendTo)
				sendTo->OnSignal(this, parameter1);
		}
};

/**
 MACRO to connect one object with other. See class FGConnector for more.
 @ingroup Defines
 @see FGConnector::Connect()
*/
#define		FGCONNECT(sender, THIS, PARAM1) { (sender)->Connect((THIS), int(PARAM1)); }

/**
	The types of mutexes.
	@ingroup Enums
*/
enum ENUM_MUTEX_TYPE
{
	fastMutEx=1,
	recursiveMutEx
};

/**
	Synchronisation object - MUTEX. Used when multithrading
	support is enabled.
*/
class FGMutex
{
#ifdef _PTHREAD_H
		int				type;
		pthread_mutex_t m_mutex;
#endif
#ifdef _WIN32
		CRITICAL_SECTION m_mutex;
#endif
#ifdef FG_THREADED
		void init(ENUM_MUTEX_TYPE typ);
#endif
	public:
		//! Create a MUTEX. You can choose from fast or recursive one.
		FGMutex(ENUM_MUTEX_TYPE typ = fastMutEx)
#ifdef FG_THREADED
		;
#else
		{}
#endif
#ifdef FG_THREADED
		FGMutex(const FGMutex& old);
#endif
		~FGMutex()
#ifdef FG_THREADED
		;
#else
		{}
#endif
		//! Lock the object for others.
		void Lock()
#ifdef FG_THREADED
		;
#else
		{}
#endif
		//! Unlock the object for others.
		void Unlock()
#ifdef FG_THREADED
		;
#else
		{}
#endif
		//! Try lock the object for others. Returns false if one is already locked.
		bool TryLock()
#ifdef FG_THREADED
		;
#else
		{ return true; }
#endif
};

/**
	A base drawing object.
	This object is used as a parent for the all graphics stuff.
	You can use this one for in-memory drawing and after drawing
	complete update the screen with contents of this object.
*/
class FGDrawBuffer : public FGConnector
{
		struct internal_state
		{
			unsigned       _ink;
			unsigned	   _paper;
			int		 	   _font;	// font code
			unsigned       _ppop;
			unsigned	   colorkey;
			unsigned 	   alpha;
		};

	public:
		enum ObjectType {	WINDOW=0x4000, CHECKBUTTON, PUSHBUTTON, POINTBUTTON, EDITBOX, PROGRESSBAR,
			BASEMENU, BASEMENUITEM, MENUWINDOW, SLIDEBAR, MULTILINEEDIT, LISTBOX, LABEL, WINDOWPANEL,
			BMP_FILE, BMP_IMAGE, BMP_MEM, ROOTWINDOW, FGFRAMEBUFFER, COMBOBOX, WINDOWTEXT, WINDOWIMAGE,
			BMP_NONE=0 };
		friend class GLSurface;
		friend class listbox;
	private:
		static	int ID_counter;
		int		id;
		FGMutex	atomic;

		inline void	lock(void)
		{
			atomic.Lock();
		}
		inline void	unlock(void)
		{
			atomic.Unlock();
		}
		void	FGAPI _symetry(int xs, int ys, int x,	int	y, FGPixel color);
		void	FGAPI _symetry2(int xs, int ys, int x, int y, FGPixel color);
		void	FGAPI _init(int ww, int hh, ObjectType t, FGPixel ii=CBLACK, FGPixel pp=CWHITE, const char *nm=0);
		void	FGAPI turn_bitmap(int cnt, FGPixel *from, FGPixel *to);
	protected:
		int		w;
		int		h;
		int		xoff,yoff;	// coor
		int		wwrk,hwrk;	// sizes!!!
		unsigned long	status;
		FGMutex	clip_lock;
		FGPixel	*image;
		char 	*name;
		union 	{ ObjectType	type; int _asd; };

		internal_state state;
		internal_state state_backup;
		bool        state_saved;

		static	const char *empty_string;
		static	int oldx, oldy;

		FGPixel *	FGAPI SetArray(FGPixel *array)
		{
			if (image) free(image);
			return image = array;
		}

		FGDrawBuffer(int ww, int hh, const char * nm, ObjectType t, FGPixel i, FGPixel p);

		void 	FGAPI SetStatus(long s) {	status = s;	}
		static int FGAPI GetCounter(void) { return ID_counter; }
		//! Clip coordinates and size to the client space.
		inline int FGAPI clip(int& xr, int& yr, int& wr, int& hr)
		{
			if (xr<xoff)
			{
				wr -= (xoff-xr);
				xr = xoff;
			}
			if (yr<yoff)
			{
				hr -= (yoff-yr);
				yr = yoff;
			}
			if (xr+wr > xoff+wwrk)
			{
// FIXME
				wr = wwrk-(xr-xoff);
//				wr = wwrk;
			}
			int yoff2 = h - (hwrk+yoff);
			if (yr+hr > h-yoff2)
			{
				hr = h-yr-yoff2;
			}
			return (wr<1 || hr<1) ? 0 : 1; // return if valid
		}
		inline int FGAPI clip_shape(int& xr, int& yr, int& wr, int& hr)
		{
			if (xr<0)
			{
				wr += xr;
				xr = 0;
			}
			if (yr<0)
			{
				hr += yr;
				yr = 0;
			}
			if (xr+wr>w)
			{
				wr = w-xr;
			}
			if (yr+hr>h)
			{
				hr = h-yr;
			}
			return (wr<1 || hr<1) ? 0 : 1; // return if valid
		}
		void 	FGAPI modify_point(FGPixel *ptr, FGPixel ink);
		void 	__line(int,int,int,int, unsigned color);
	public:
        //! Save the internal state of object described by struct internal_state (ink, paper, font, alpha, colorkey anf ppop).
        //! @note the state is not stacked, i.e. you can call this methotd only once before restore_state().
	    void save_state(void)
		{
			lock();
			if (state_saved)
			{
			    ::printf("FGDrawBuffer::save_state() multiple call!\n");
			}
			else
			{
				state_backup = state;
			}
			state_saved = true;
			unlock();
		}
        //! Restore the state saved by save_state().
		void restore_state(void)
		{
			lock();
			if (state_saved == false)
			{
				::printf("FGDrawBuffer::restore_state() called without save_state()!\n");
			}
			else
			{
				state = state_backup;
			}
			state_saved = false;
			unlock();
		}
		/**
		Calculate the absolute address of [x:y] pixel. There are no relative offsets used.
		*/
		inline FGPixel* FGAPI CalcAddr(int x, int y)
		{
			return image + (w * y + x);
		}
		/**
		Resizes the buffer by relative sizes.
		*/
		void	FGAPI Resize(int dx, int dy);
		/**
		Returns the status of object.
		*/
		long	FGAPI GetStatus(void)  const	{ return status; }
		//! Try get the font image and set requested font ID
		//! if success for the object as current.
		inline 	void FGAPI set_font(int font_id)
		{
			if (FGFontManager::get_font_ptr(font_id))
				state._font = font_id;
		}
		//! Returns a current object's font ID.
		inline 	int FGAPI get_font(void)
		{
			return state._font;
		}
		//! Returns a width of the 'text' in pixels for current font.
		inline int	FGAPI textwidth(const char *text)
		{
			return FGFontManager::textwidth(state._font, text);
		}
        //! Returns a width for the first 'cnt' characters
        //! of the 'text' in pixels for current font.
		inline int	FGAPI textwidth(char *txt, int cnt)
		{
			return FGFontManager::textwidth(state._font, txt, cnt);
		}
        //! Returns a width of the character 'c' in pixels for current font.
		inline int	FGAPI charwidth(int c)
		{
			return FGFontManager::charwidth(state._font, c);
		}
		//! Returns a width of one character for current font if one is fixed-size.
		inline int	FGAPI GetFontW(void)
		{
			return FGFontManager::GetW(state._font);
		}
        //! Returns a height of one character for current font if one is fixed-size.
		inline int	FGAPI GetFontH(void)
		{
			return FGFontManager::GetH(state._font);
		}
		//! Set the ROP for all drawing operations used with the object.
		//! See enum ENUM_PPOP for more details.
		inline unsigned FGAPI set_ppop(unsigned op)
		{
			unsigned oldp = state._ppop;
			state._ppop = op;
			return oldp;
		}
        //! Set the foreground color for drawing operations.
		//! This color will be used when you choose function
		//! that don't uses color as parameter immediatelly.
		inline 	void FGAPI set_fcolor(FGPixel val) { state._ink = val; }
		//! Set the background color for text drawing operations.
		//! This color will be used when you choose function
		//! that don't uses color as parameter immediatelly.
		inline 	void FGAPI set_bcolor(FGPixel val) { state._paper = val; }
		//! Set the foreground (ink) and the background (paper)
		//! color for drawing operations.
		inline 	void FGAPI set_colors(FGPixel i, FGPixel p) { state._ink = i; state._paper = p; }
		//! Returns the ID of the object.
        /**
        The all controls have an exclusive ID. This value goes from 0 to ..,
		for each window and its control items. For example: when window
		contains 5 buttons, then these buttons have local id values
        from 0 to 4,according to the order of its creating. This is
		useful for fast and easy testing which button (when you use
        one call-back procedure for more buttons i.e.) especially
        with using GetLastActive();
		*/
		int		FGAPI GetId(void)  const	{ return id; }
		//! Set position and size of the client space for the object (ww & hh are size).
		inline bool FGAPI set_clip_rect(int xx, int yy, int ww, int hh)
		{
			if (hh<1 || ww<1 || xx<0 || yy<0) return false;
			xoff = xx;
			yoff = yy;
			if ((xx + ww)	> w) ww -= ((xx + ww) - w);
			if ((yy + hh)	> h) hh -= ((yy + hh) - h);
			wwrk = ww;
			hwrk = hh;
			return true;
		}
        //! Returns current values of the client space of the object.
		inline void	FGAPI get_clip_rect(int& x,int& y,int& w,int& h)
		{
			x = xoff;
			y = yoff;
			w = wwrk;
			h = hwrk;
		}
		FGDrawBuffer(int ww=0, int hh=0, ObjectType=BMP_MEM, int color=CBLACK, FGPixel *buf=0);
		FGDrawBuffer(int ww, int hh, FGPixel *buf);
        //! Makes the deep copy of object.
		FGDrawBuffer(FGDrawBuffer &);

		virtual ~FGDrawBuffer();

		// copy block from one buffer to other (with ppop or alpha)
		// this is software blitter only.
		int			FGAPI bitblit(int xdst, int ydst, int xsrc, int ysrc, int ww, int hh, class FGDrawBuffer *src, int opcia=BITBLIT_COPY);
		int			FGAPI hline(int x, int y, int size, FGPixel color);
		int			FGAPI vline(int x, int y, int size, FGPixel color);

		//! Sets the start position for first time used lineto().
		void    	FGAPI moveto(int x, int y) { oldx = x; oldy = y; }
		int			FGAPI line(int x, int y, int x2, int y2, unsigned color = UNDEFINED_COLOR);
		int			FGAPI lineto(int x, int y, unsigned color = UNDEFINED_COLOR);
		int			FGAPI pline(int x1, int y1, int x2, int y2, FGPattern *p);

		void		FGAPI spline(int points[8], unsigned color = UNDEFINED_COLOR);

		int			FGAPI rect(int xdst, int ydst, int ww, int hh, unsigned color = UNDEFINED_COLOR);
		int			FGAPI box(int xdst, int ydst, int ww, int hh, unsigned color = UNDEFINED_COLOR);

		int 		FGAPI text(int x, int y, const char *txt, unsigned i = UNDEFINED_COLOR, unsigned p = UNDEFINED_COLOR);
#undef printf
		int			printf(int , int , const char * , ... );
		int			FGAPI putpixel(int x, int y, unsigned color = UNDEFINED_COLOR);
		/**
		 Returns the value of color at the position [x:y]
		 of the object or (-1) if coordinates are out of range.
		*/
		int			FGAPI getpixel(int x, int y)
		{
			if (x<0	|| y<0 || x>=wwrk || y>=hwrk)
				return 0xFFFFFFFF;
			return *(image+(x+xoff)+(y+yoff)*w);
		}
		void		FGAPI stretch(int _w, int _h);
		void		FGAPI stretch(double ratiox, double ratioy);

		int			FGAPI arc(int	x,	int	y,	double ang1, double ang2, int r, unsigned color = UNDEFINED_COLOR);
		int			FGAPI arc2(int	x,	int	y,	int ang1, int ang2, int r, unsigned color = UNDEFINED_COLOR);
		int			FGAPI ellipse(int x, int y, int rx, int ry, unsigned color = UNDEFINED_COLOR);
		int			FGAPI fellipse(int x, int y, int rx, int ry, unsigned color = UNDEFINED_COLOR);
		int			FGAPI circle(int	xs,	int	ys,	int	r, unsigned color = UNDEFINED_COLOR);
		int			FGAPI fcircle(int	xs,	int	ys,	int	r, unsigned color = UNDEFINED_COLOR);

		void 		FGAPI polygon(const FGPointArray& vertices, unsigned color = UNDEFINED_COLOR);
		FGRect		FGAPI fpolygon(const FGPointArray& vertices, unsigned color = UNDEFINED_COLOR);

		FGRect		FGAPI ftriangle(int x1, int y1, int x2, int y2, int x3, int y3, unsigned color = UNDEFINED_COLOR);

		//! Returns the address of pixel data storage. There are pixels, from right to left, line by line.
		FGPixel*    FGAPI GetArray(void) const { return image; }
		//! Returns an ALPHA (transparency) value of the object.
		unsigned	FGAPI GetAlpha(void) const { return state.alpha; }
		//! Returns a foreground color of the object.
		FGPixel	  	FGAPI GetInk(void)  const { return (FGPixel)state._ink; }
		//! Returns a background color of the object.
		FGPixel	  	FGAPI GetPaper(void)  const { return	(FGPixel)state._paper; }
		//! Returns a width of the object.
		int		  	FGAPI GetW(void)  const { return	w; }
		//! Returns a height of the object.
		int		  	FGAPI GetH(void)  const { return	h; }
		//! Returns the X coordinate of the client space of the object.
		int	   		FGAPI GetXW(void) const { return xoff; }
		//! Returns the Y coordinate of the client space of the object.
		int	   		FGAPI GetYW(void) const { return yoff; }
		//! Returns a client space width of the object.
		int			FGAPI GetWW(void) const {	return wwrk; }
		//! Returns a client space height of the object.
		int			FGAPI GetHW(void) const {	return hwrk; }
		//! Returns the type of the object (WINDOW, PUSHBUTTON ..)
		enum ObjectType 	FGAPI GetType(void)  const {	return type; }
		//! Returns the COLOR_KEY of the object (i.e. color that is used as transparent).
		unsigned	FGAPI GetColorKey(void) const { return state.colorkey; }
		//! Returns the name of the object.
		char*       FGAPI GetName(void)  const {	return name; }
        //! Set the COLOR_KEY of the object (i.e. color that will be transparent when bitblit
		//! operation will be performed with this object).
		void		FGAPI SetColorKey(unsigned ck)
		{
			state.colorkey = ck;
		}
		void	  	FGAPI SetName(const char *s);
        //! Set the level of transparency for the object when bitblit
        //! operation will be performed with this object. The level
        //! is number from 0 to 255, where 0 is completely transparent
        //! and 255 completely opaque.
		void		FGAPI SetAlpha(unsigned a) { state.alpha = a; }
		//! Clear the entire object image with this color.
		void		FGAPI clear(int color) { FGmemset(image, color, w*h); }
		/**
			Rotates image to apropriate direction.
		*/
		void		FGAPI rotate(ENUM_ROTATE_DIRECTION direction);
};

#define	SAVE_CLIPPING(THIS) int _a,_b,_c,_d; clip_lock.Lock(); _a = (THIS)->xoff; _b = (THIS)->yoff; _c = (THIS)->wwrk; _d = (THIS)->hwrk; (THIS)->xoff = 0; (THIS)->yoff = 0; (THIS)->wwrk = (THIS)->w; (THIS)->hwrk = (THIS)->h;
#define	SAVE_CLIPPING2(THIS) clip_lock.Lock(); _a = (THIS)->xoff; _b = (THIS)->yoff; _c = (THIS)->wwrk; _d = (THIS)->hwrk; (THIS)->xoff = 0; (THIS)->yoff = 0; (THIS)->wwrk = (THIS)->w; (THIS)->hwrk = (THIS)->h;
#define	RESTORE_CLIPPING(THIS)	(THIS)->xoff = _a; (THIS)->yoff = _b; (THIS)->wwrk = _c; (THIS)->hwrk = _d; clip_lock.Unlock();

/**
	Structure to provide user defined mouse cursor.
*/
struct FGMouseCursor
{
	FGPixel *bitmap;
	FGPixel *mask;
	int	xoff, yoff;
	unsigned w,h;

	FGMouseCursor(FGPixel *a, FGPixel *b, int c, int d, int e, int f)
		: bitmap(a), mask(b), xoff(c), yoff(d), w(e), h(f)
	{
	}
};

/**
	Internal - one palette record
*/
struct FGPalette
{
		unsigned char b;
		unsigned char g;
		unsigned char r;
		unsigned char alfa;
	  public:
		FGPalette(void)
		{
			r = g = b = alfa = 0;
		}
		FGPalette(unsigned p)
		{
			*((unsigned *) this) = p;
		}
		FGPalette(unsigned char _r, unsigned char _g, unsigned char _b)
		{
			r = _r;
			g = _g;
			b = _b;
			alfa = 0;
		}
		unsigned int GetValue(void)
		{
			return *((unsigned *) this);
		}
};

extern const FGMouseCursor IDC_NORMAL_SMALL;
extern const FGMouseCursor IDC_NORMAL_LARGE;
extern const FGMouseCursor *idc_normal;

/**
	Interface for user touchscreen + calibration framework.
*/
class FGInputDevice
{
	protected:
		bool working;

		struct POINT
		{
			int x, y;
		};

		struct MATRIX
		{
			/* This arrangement of values facilitates
			 *  calculations within getDisplayPoint()
			 */
			double An,					/* A = An/Divider */
			  Bn,						/* B = Bn/Divider */
			  Cn,						/* C = Cn/Divider */
			  Dn,						/* D = Dn/Divider */
			  En,						/* E = En/Divider */
			  Fn,						/* F = Fn/Divider */
			  Divider;
		};

		/* ( 15, 15)  */
		/* ( 50, 85)  */
		/* ( 85, 50)  */
		POINT imaginary[3];
		POINT real[3];
		MATRIX matrix;

		void init(int xmax, int ymax);
		void save_values(void);
		void load_values();
		int setCalibrationMatrix(POINT * display, POINT * screen, MATRIX * matrix);
		int getDisplayPoint(POINT * display, POINT * screen, MATRIX * matrix);
		void DoPoint(POINT * display, POINT * screen);
	public:
		FGInputDevice();
		virtual ~FGInputDevice() {}

		virtual bool Attach(void) = 0;
		virtual int GetInputEvent(int& type, int& key, int& x, int& y, int& button) = 0;
		virtual void Scale(int& x, int& y);

		void Calibration(void);
};


#ifdef FG_NAMESPACE
}
#endif

#ifdef __BORLANDC__
#pragma option pop /* pop -a switch */
#pragma option pop /* pop -b */
#endif

#ifndef NO_ABSOLETE
#define Pattern     FGPattern
#define Panel       FGPanel
#define Label       FGLabel
#define App         FGApp
#define Window      FGWindow
#define Control     FGControl
#define FontProperty FGFontProperty
#define DrawBuffer  FGDrawBuffer
#define MenuWindow  FGMenuWindow
#define TextEditor  FGTextEditor
#define PushButton  FGPushButton
#define CheckButton FGCheckButton
#define PointButton FGPointButton
#define BaseGui     FGBaseGui
#define GuiEvent    FGEvent
#define FontDialog  FGFontDialog
#define FileDialog  FGFileDialog
#define ColorDialog FGColorDialog
#define Bitmap      FGBitmap
#define EditBox     FGEditBox
#define ButtonGroup FGButtonGroup
#define ProgressBar FGProgressBar
#define Slider      FGSlider
#define SlideBarV   FGSlideBarV
#define SlideBarH   FGSlideBarH
#define RadioGroupV FGRadioGroupVertical
#define RadioGroupH FGRadioGroupHorizontal
#define TwoStateButton FGTwoStateButton
#define ColorScheme FGColorScheme
#endif

#endif // base_h
