/*
   Copyright (C) 1996,2004  Marian Krivos

   nezmar@atlas.sk

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   bitmap.cc - bitmap engine

  $Id: bitmap.cpp,v 1.12 2004/08/30 06:38:37 majo Exp $

  $Log: bitmap.cpp,v $
  Revision 1.12  2004/08/30 06:38:37  majo
  warningy

  Revision 1.11  2004/08/26 05:45:35  majo
  added FGDrawBuffer::arc2()

  Revision 1.10  2004/05/30 11:52:55  majo
  dox

  Revision 1.9  2004/05/24 17:42:24  majo
  FGCircle
  bugfixes

  Revision 1.8  2004/04/27 17:39:00  majo
  fixed BitmapSave()
  added _Gxxx PPOP

  Revision 1.7  2004/04/16 08:50:19  majo
  slightly reimplemented class FGBitmap
  added JpegSave(char *name)
  removed member FGBitmap::original_name and its methods

  Revision 1.6  2004/03/03 20:12:39  majo
  added class FGColor
  fixed Solaris compiling

  Revision 1.5  2004/03/01 15:19:15  majo
  _GTRANSP -> _GCOLORKEY
  added _GTRANSPARENT

  Revision 1.4  2004/02/29 19:58:19  majo
  devel state

  Revision 1.3  2004/02/26 21:41:28  majo
  class FGInternalState
  char* -> const char*
  int ink = -1  -> unsigned ink = UNDEFINED_COLOR

  Revision 1.2  2004/02/23 20:08:01  majo
  all classes are with prefix FG* on now
  polygon functions uses FGPointArray from now
  class GuiEvent is renamed to FGEvent
  some by parameters overloaded methods was removed (class FGWindow)
  many other small changes

  Revision 1.1.1.1  2004/01/22 16:25:08  majo
  initial release

 */

#include <stdarg.h>

#include "config.h"
#include "fastgl.h"
#include "_fastgl.h"

#ifdef FG_JPEG
extern "C" {		// jpeglib.h forgot this... (Oron Peled)
#include <jpeglib.h>
#include <jerror.h>
#ifdef __BORLANDC__
#ifdef __linux__
#pragma link "libjpeg.so"
#else
#pragma link "libjpeg.lib"
#endif
#endif
}
#endif

#ifdef FG_TIFF
#include <tiffio.h>
#ifdef __BORLANDC__
#ifdef __linux__
#pragma link "libtiff.so"
#else
#pragma link "libtiff.lib"
#endif
#endif
#endif

#ifdef FG_PNG
#include <png.h>
#ifdef __BORLANDC__
#ifdef __linux__
#pragma link "libpng.so"
#pragma link "libz.so"
#else
#pragma link "libpng.lib"
#pragma link "zlib.lib"
#endif
#endif
#else
#include <setjmp.h>
#endif

#include "gif.h"

#define 	TRANSP_KEY			0xFE00FE
#define     MAX_LWZ_BITS        12
#define		INTERLACE			0x40
#define		LOCALCOLORMAP		0x80

#define BitSet(byte, bit)	(((byte) & (bit)) == (bit))
#define	ReadOK(file,buffer,len) (fread(buffer, len, 1, file) != 0)
#define MKINT(a,b)		(((b)<<8)|(a))
#define NEW(x)			((x *)malloc(sizeof(x)))

/***************************************************************************
*
*  ERROR()    --  should not return
*  INFO_MSG() --  info message, can be ignored
*
***************************************************************************/

#define INFO_MSG(fmt)
#define _ERROR(str) 	IError(str, 0), longjmp(setjmp_buffer, 1)

/***************************************************************************/

#ifdef FG_NAMESPACE
namespace fgl {
#endif

#ifdef FG_JPEG
struct ima_error_mgr {
	struct jpeg_error_mgr pub;	/* "public" fields */
	jmp_buf setjmp_buffer;		/* for return to caller */
	char* buffer;				/* error message <CSC>*/
};

typedef ima_error_mgr *ima_error_ptr;

static void fgl_jpeg_error_exit(j_common_ptr cinfo)
{
	/* cinfo->err really points to a my_error_mgr struct, so coerce pointer */
	ima_error_ptr myerr = (ima_error_ptr) cinfo->err;
	/* Create the message */
	myerr->pub.format_message (cinfo, myerr->buffer);
	/* Send it to stderr, adding a newline */
	/* Return control to the setjmp point */
	longjmp(myerr->setjmp_buffer, 1);
}

#endif

static int readColorMap(FILE *, int, unsigned char[GIF_MAXCOLORS][3]);
static int GetDataBlock(FILE *, unsigned char *);
static void readImage(FILE *, int, int, int, unsigned char *);
static void read_RLE8_compressed_image(FILE * f, unsigned char *bmp, int w, int h);

static jmp_buf setjmp_buffer;
int FGDrawBuffer::ID_counter = 1;		// incremental ID of all graph. objects
int FGDrawBuffer::oldx, FGDrawBuffer::oldy;

/* TRUEVISION-XFILE magic signature string */
static unsigned char magic[18] =
{
  0x54, 0x52, 0x55, 0x45, 0x56, 0x49, 0x53, 0x49, 0x4f,
  0x4e, 0x2d, 0x58, 0x46, 0x49, 0x4c, 0x45, 0x2e, 0x0
};

/***************************************************************************/

#ifdef FG_JPEG
// thanks to Chris Shearer Cooper <cscooper(at)frii(dot)com>
class CxFileJpg : public jpeg_destination_mgr, public jpeg_source_mgr
	{
public:
	enum { eBufSize = 4096 };

	CxFileJpg(FILE* pFile)
	{
		m_pFile = pFile;

		init_destination = InitDestination;
		empty_output_buffer = EmptyOutputBuffer;
		term_destination = TermDestination;

		init_source = InitSource;
		fill_input_buffer = FillInputBuffer;
		skip_input_data = SkipInputData;
		resync_to_restart = jpeg_resync_to_restart; // use default method
		term_source = TermSource;
		next_input_byte = NULL; // => next byte to read from buffer
		bytes_in_buffer = 0;	// # of bytes remaining in buffer

		m_pBuffer = new unsigned char[eBufSize];
	}
	~CxFileJpg()
	{
		delete [] m_pBuffer;
	}

	static void InitDestination(j_compress_ptr cinfo)
	{
		CxFileJpg* pDest = (CxFileJpg*)cinfo->dest;
		pDest->next_output_byte = pDest->m_pBuffer;
		pDest->free_in_buffer = eBufSize;
	}

	static boolean EmptyOutputBuffer(j_compress_ptr cinfo)
	{
		CxFileJpg* pDest = (CxFileJpg*)cinfo->dest;
		if (fwrite(pDest->m_pBuffer,1,eBufSize, pDest->m_pFile)!=(size_t)eBufSize)
			ERREXIT(cinfo, JERR_FILE_WRITE);
		pDest->next_output_byte = pDest->m_pBuffer;
		pDest->free_in_buffer = eBufSize;
		return TRUE;
	}

	static void TermDestination(j_compress_ptr cinfo)
	{
		CxFileJpg* pDest = (CxFileJpg*)cinfo->dest;
		size_t datacount = eBufSize - pDest->free_in_buffer;
		/* Write any data remaining in the buffer */
		if (datacount > 0) {
			if (!fwrite(pDest->m_pBuffer,1,datacount,pDest->m_pFile))
				ERREXIT(cinfo, JERR_FILE_WRITE);
		}
		fflush(pDest->m_pFile);
		/* Make sure we wrote the output file OK */
		if (errno) ERREXIT(cinfo, JERR_FILE_WRITE);
		return;
	}

	static void InitSource(j_decompress_ptr cinfo)
	{
		CxFileJpg* pSource = (CxFileJpg*)cinfo->src;
		pSource->m_bStartOfFile = TRUE;
	}

	static boolean FillInputBuffer(j_decompress_ptr cinfo)
	{
		size_t nbytes;
		CxFileJpg* pSource = (CxFileJpg*)cinfo->src;
		nbytes = fread(pSource->m_pBuffer,1,eBufSize,pSource->m_pFile);
		if (nbytes <= 0){
			if (pSource->m_bStartOfFile)	// Treat empty input file as fatal error
				ERREXIT(cinfo, JERR_INPUT_EMPTY);
			WARNMS(cinfo, JWRN_JPEG_EOF);
			// Insert a fake EOI marker
			pSource->m_pBuffer[0] = (JOCTET) 0xFF;
			pSource->m_pBuffer[1] = (JOCTET) JPEG_EOI;
			nbytes = 2;
		}
		pSource->next_input_byte = pSource->m_pBuffer;
		pSource->bytes_in_buffer = nbytes;
		pSource->m_bStartOfFile = FALSE;
		return TRUE;
	}

	static void SkipInputData(j_decompress_ptr cinfo, long num_bytes)
	{
		CxFileJpg* pSource = (CxFileJpg*)cinfo->src;
		if (num_bytes > 0){
			while (num_bytes > (long)pSource->bytes_in_buffer){
				num_bytes -= (long)pSource->bytes_in_buffer;
				FillInputBuffer(cinfo);
				// note we assume that fill_input_buffer will never return FALSE,
				// so suspension need not be handled.
			}
			pSource->next_input_byte += (size_t) num_bytes;
			pSource->bytes_in_buffer -= (size_t) num_bytes;
		}
	}

	static void TermSource(j_decompress_ptr cinfo)
	{
		return;
	}
protected:
	FILE  *m_pFile;
	unsigned char *m_pBuffer;
	bool m_bStartOfFile;
public:
	enum CODEC_OPTION
	{
		ENCODE_BASELINE = 0x1,
		ENCODE_ARITHMETIC = 0x2,
		ENCODE_GRAYSCALE = 0x4,
		ENCODE_OPTIMIZE = 0x8,
		ENCODE_PROGRESSIVE = 0x10,
		ENCODE_LOSSLESS = 0x20,
		ENCODE_SMOOTHING = 0x40,
		DECODE_GRAYSCALE = 0x80,
		DECODE_QUANTIZE = 0x100,
		DECODE_DITHER = 0x200,
		DECODE_ONEPASS = 0x400,
		DECODE_NOSMOOTH = 0x800
	};

	int m_nPredictor;
	int m_nPointTransform;
	int m_nSmoothing;
	int m_nQuantize;
	J_DITHER_MODE m_nDither;

};
#endif // jpeg

/***************************************************************************/

FGBitmap::ReturnValues FGBitmap::LoadImage(const char* nm, int imagefmt)
{
	int ys, one;
	ReturnValues retval;
	unsigned int c, bits = 0;
	register unsigned char *src;
#ifndef INDEX_COLORS
	register FGPixel *bp;
	int sz, d, i;
	FGPixel *pom;
#endif

	type = BMP_NONE;
	init();

    SetName(nm);

	if (imagefmt == IMAGE_AUTO)
		format = detect(name);
	else
	    format = imagefmt;

	switch(format)
	{
		case IMAGE_GIF:
            retval = gif(name,one,bits,c);
			if (retval) return retval;
			break;
		case IMAGE_PCX:
			retval = pcx(name,one,bits,c);
			if (retval) return retval;
			break;
#ifdef FG_PNG
		case IMAGE_PNG:
			retval = png(name,one,bits,c);
			if (retval) return retval;
			break;
#endif
#ifdef FG_JPEG
		case IMAGE_JPG:
			retval = jpg(name,one,bits,c);
			if (retval) return retval;
			break;
#endif
		case IMAGE_TGA:
			retval = tga(name,one,bits,c);
			if (retval) return retval;
			break;
#ifdef FG_TIFF
		case IMAGE_TIF:
			retval = tiff(name,one,bits,c);
			if (retval) return retval;
			break;
#endif
		case IMAGE_BMP:
			retval = bmp(name,one,bits,c);
			if (retval) return retval;
			break;
		default:
			IError("Unknown image format!", 0);
			return IMAGE_UNKNOWN_FORMAT;
	}

#ifndef INDEX_COLORS
	if (bits<=8)
	{
		bp = (FGPixel *) calloc(areasize(w, h), 1);
		pom = bp;
		src = (unsigned char *) GetArray();
		sz = w * h;

		if (state.colorkey == 0xffffffffU)
			for (i = 0; i < sz; i++)
		{
			d = *((long *) ((char *) &rgb + *src * one));
			*pom = FGDirectColor(d);
			pom++;
			src++;
		}
		else
		{
			for (i = 0; i < sz; i++)
			{
				if (*src == state.colorkey)
					d = TRANSP_KEY;
				else
					d = *((long *) ((char *) &rgb + *src * one));
				*pom++ = FGDirectColor(d);
				src++;
			}
			state.colorkey = FGDirectColor(TRANSP_KEY);
		}
		free(image);
		image = bp;
	}
#endif
	type = BMP_FILE;
	wwrk = w;
	hwrk = h;
#ifdef INDEX_COLORS
	DoPalette(one);
#endif
}

FGBitmap::ReturnValues FGBitmap::SaveImage(const char* name, int imagefmt)
{
	switch(imagefmt)
	{
		case IMAGE_BMP:
		    return BitmapSave(name, false);
		case IMAGE_PNG:
		    return PngSave(name);
		case IMAGE_JPG:
		    return JpegSave(name);
	}
	return IMAGE_UNKNOWN_FORMAT;
}

int FGBitmap::detect(const char *nm)
{
	if ( strstr(nm, ".gif") || strstr(nm, ".GIF") )
	{							// decode *.gif file
		return IMAGE_GIF;
	}
	else if ( strstr(nm, ".pcx") || strstr(nm, ".PCX") )
	{							// decode *.gif file
		return IMAGE_PCX;
	}
	else if ((strstr(nm, ".tga")) || (strstr(nm, ".TGA")))
	{							// decode *.gif file
		return IMAGE_TGA;
	}
	else if ((strstr(nm, ".png")) || (strstr(nm, ".PNG")))
	{							// decode *.gif file
		return IMAGE_PNG;
	}
	else if ((strstr(nm, ".jpg")) || (strstr(nm, ".JPG")) || (strstr(nm, ".jpeg")) || (strstr(nm, ".JPEG")))
	{
		return IMAGE_JPG;
	}
	else if ((strstr(nm, ".tiff")) || (strstr(nm, ".TIFF")) || (strstr(nm, ".tif")) || (strstr(nm, ".TIF")))
	{
		return IMAGE_TIF;
	}
	else if ((strstr(nm, ".bmp")) || (strstr(nm, ".BMP")))
	{							// decode *.gif file
		return IMAGE_BMP;
	}
	return IMAGE_AUTO;
}

FGBitmap::FGBitmap()
 : FGDrawBuffer()
{
	type = BMP_NONE;
	init();
}

/**
	Creates a BITMAP object from one of supported image formats.
	Parameter 'n' is pointer to filename. You can test to fail
	after return. The member 'type' must not be BMP_NONE, else
	object is no valid. Supported file formats are BMP, TGA, GIF
	and PCX. You can enable JPG and PNG by editing 'config.mak'
	optionaly too. These two are supported by third party libs
	(libjpeg & libpng).
*/
FGBitmap::FGBitmap(const char *n)
: FGDrawBuffer()
{
    LoadImage(n, IMAGE_AUTO);
}

FGBitmap::ReturnValues FGBitmap::bmp(char *n, int& one, unsigned int& bits, unsigned int& c)
{
	int ys;
	int	compressed = 0;
	int rest;
	FILE *f=0;
	long pixel;
	register unsigned char *src;
#ifndef INDEX_COLORS
	register FGPixel *bp;
#endif

	f = fopen(n, "rb");
	if (!f)
	{
		char nm[128];
		sprintf(nm, "FGBitmap '%s' not found!", n);
		IError(nm, 0);
		return IMAGE_FILE_NOT_FOUND;
	}
	name = new char[strlen(n) + 1];
	strcpy(name, n);
	if (sizeof(bmfh) != 10)
	{
		::printf("Ooops, your data structure is not good aligned!!!\a\n");
		assert(0);
		abort();			// 100% sure to stop
	}
	fread(&bmfh, sizeof(bmfh), 1, f);	// read first 10 bytes
	// return if no BMP
	if (bmfh.usType != (unsigned short) ('M' * 256 + 'B'))
	{
		char jj[128];
		sprintf(jj,"%s isn't supported image file!", n);
		fclose(f);
		IError(jj, 0);
		return IMAGE_UNKNOWN_FORMAT;
	}

	fread(&bmi, sizeof(bmi), 1, f);		// read next 4 bytes
	memset(&bmih, 0, sizeof(bmih));
	fread(&bmih, 4, 1, f);	// read size of bitmap info header
	fread(&bmih.bmih1.cx, bmih.bmih1.size - 4, 1, f);	// if 0x0c then OS/2 bitmap
	if (bmih.bmih1.size == 0x0C)
		one = 3;
	else
		one = 4;

	c = (bmi.imageOffset - (bmih.bmih1.size + 14)) / one;	// compute numbers of used colors

	if (c == 0)
	{
#ifdef INDEX_COLORS
		fclose(f);
		IError("True colors BMP not supported!", 0);
		return IMAGE_UNKNOWN_FORMAT;
#else
		w = bmih.bmih2.cx;
		h = bmih.bmih2.cy;
		bits = bmih.bmih2.cBitCount;
		if (bmih.bmih2.cPlanes != 1)
		{
			fclose(f);
			IError("This color depth BMP is not supported!", 0);
			return IMAGE_UNKNOWN_FORMAT;
		}
#endif
	}
	else if (one == 3)
	{
		if ((compressed = bmih.bmih1.ulCompression) > 1)
		{
			IError("No this compressed BMP supported!", 0);
			return IMAGE_UNKNOWN_FORMAT;
		}
		fread(&rgb, sizeof(RGB3), c, f);
		w = bmih.bmih1.cx;
		h = bmih.bmih1.cy;
		bits = bmih.bmih1.cBitCount;
		if (bmih.bmih1.cPlanes != 1)
		{
			fclose(f);
			IError("Multi-planes BMP not supported!", 0);
			return IMAGE_UNKNOWN_FORMAT;
		}
	}
	else if (one == 4)
	{
		if ((compressed = bmih.bmih2.ulCompression) > 1)
		{
			IError("Unknown compress format!", 0);
			return IMAGE_UNKNOWN_FORMAT;
		}
		fread(&rgb, sizeof(RGB4), c, f);
		w = bmih.bmih2.cx;
		h = bmih.bmih2.cy;
		bits = bmih.bmih2.cBitCount;
		if (bmih.bmih2.cPlanes != 1)
		{
			fclose(f);
			IError("Multi-planes BMP not supported!", 0);
			return IMAGE_UNKNOWN_FORMAT;
		}
	}

	assert(SetArray((FGPixel *) calloc(areasize(w, h), 1)));

	//
	// common image decoder
	//
	if (f) switch (bits)
	{
		case 1:
			for (ys = h - 1; ys >= 0; ys--)
			{
				src = (unsigned char *) GetArray() + ys * w;
				for (int xs = w / 8; xs > 0; xs--)
				{
					fread((unsigned char *) &pixel, 1, 1, f);
					*src++ = (pixel >> 7) & 1;
					*src++ = (pixel >> 6) & 1;
					*src++ = (pixel >> 5) & 1;
					*src++ = (pixel >> 4) & 1;
					*src++ = (pixel >> 3) & 1;
					*src++ = (pixel >> 2) & 1;
					*src++ = (pixel >> 1) & 1;
					*src++ = (pixel) & 1;
				}
				rest = w % 8;
				if (rest)
				{
					fread((unsigned char *) &pixel, 1, 1, f);
					*src++ = (pixel >> 7) & 1;
					if (--rest == 0)
						goto end_expand;
					*src++ = (pixel >> 6) & 1;
					if (--rest == 0)
						goto end_expand;
					*src++ = (pixel >> 5) & 1;
					if (--rest == 0)
						goto end_expand;
					*src++ = (pixel >> 4) & 1;
					if (--rest == 0)
						goto end_expand;
					*src++ = (pixel >> 3) & 1;
					if (--rest == 0)
						goto end_expand;
					*src++ = (pixel >> 2) & 1;
					if (--rest == 0)
						goto end_expand;
					*src++ = (pixel >> 1) & 1;
				}
			  end_expand:
				rest = (w * bits / 8) % 4;
				if (w % 8)
					rest++;
				while (rest & 3)
				{
					rest++;
					fgetc(f);
				}
			}
			break;
		case 4:
			for (ys = h - 1; ys >= 0; ys--)
			{
				src = (unsigned char *) GetArray() + ys * w;
				for (int xs = w / 2; xs > 0; xs--)
				{
					fread((unsigned char *) &pixel, 1, 1, f);
					*src++ = (pixel >> 4) & 15;
					*src++ = pixel & 15;
				}
				rest = w % 2;
				if (rest)
				{
					fread((unsigned char *) &pixel, 1, 1, f);
					*(src++) = (pixel >> 4) & 15;
				}
				rest = (w * bits / 8) % 4;
				if (w % 2)
					rest++;
				while (rest & 3)
				{
					rest++;
					fgetc(f);
				}
			}
			break;
		case 8:
			if (!compressed)
				for (ys = h - 1; ys >= 0; ys--)
				{
					fread((unsigned char *) GetArray() + ys * w, w, 1, f);
					rest = (w * bits / 8) % 4;
					while (rest & 3)
					{
						rest++;
						fgetc(f);
					}
				}
			else read_RLE8_compressed_image(f, (unsigned char *) GetArray(), w, h);
			break;
#ifndef INDEX_COLORS
		case 24:
			for (ys = h - 1; ys >= 0; ys--)
			{
				bp = (image + ys * w);
				for (int xs = 0; xs < w; xs++)
				{
					fread(&pixel, 3, 1, f);
					*bp = FGDirectColor(pixel);
					bp++;
				}
			}
			rest = (w * bits / 8) % 4;
			while (rest & 3)
			{
				rest++;
				fgetc(f);
			}
			break;
#endif
#ifdef DIRECT_COLORS
		case 16:
		case 15:
			for (ys = h - 1; ys >= 0; ys--)
			{
				bp = (image + ys * w);
				for (int xs = 0; xs < w; xs++)
				{
					fread(bp, 2, 1, f);
					bp++;
				}
			}
			rest = (w * bits / 8) % 4;
			while (rest & 3)
			{
				rest++;
				fgetc(f);
			}
			break;
#endif
		default:
			IError("Bad FGBitmap format!", 0);
			fclose(f);
			free(GetArray());
			SetArray(0);
			return IMAGE_UNKNOWN_FORMAT;
	}
	if (f) fclose(f);

	return IMAGE_OK;
}

#ifdef INDEX_COLORS
void FGBitmap::DoPalette(int one)
{
	cUsedTable = new int[256];
	int sz = w * h, d, i;
	FGPixel *bm = GetArray();

	memset(cUsedTable, 0, sizeof(long) * 256);	// reset color table
	// calculate which colors is used in image
	for (i = 0; i < sz; i++)
	{
		cUsedTable[*bm++]++;
	}

	bm = GetArray();
	// allocating colors
	for (i = 0; i < 256; i++)
	{
		if (cUsedTable[i])
		{
			d = *((long *) ((char *) &rgb + i * one));
			if (i!=state.colorkey)
				cUsedTable[i] = (char) CreateColor((d >> 16)&0xff, (d >> 8)&0xff, d&0xff, -1); // CORRECTED
			else cUsedTable[i] = i;
		}
		else cUsedTable[i] = -1;
	}

	// remapping colors
	for (i = 0; i < sz; i++)
	{
		bm[i] = (char) cUsedTable[bm[i]];
	}
}
#endif

/* read_RLE8_compressed_image:
 *  For reading the 8 bit RLE compressed BMP image format.
 */
static void read_RLE8_compressed_image(FILE * f, unsigned char *bmp, int w, int h)
{
	unsigned char count, val, val0;
	int j, pos, line;
	int eolflag, eopicflag;

	eopicflag = 0;
	line = h - 1;

	while (eopicflag == 0)
	{
		pos = 0;				/* x position in bitmap */
		eolflag = 0;			/* end of line flag */

		while ((eolflag == 0) && (eopicflag == 0))
		{
			count = (char) getc(f);
			val = (char) getc(f);

			if (count > 0)
			{					/* repeat pixel count times */
				for (j = 0; j < count; j++)
				{
					bmp[line * w + pos] = val;
					pos++;
				}
			}
			else
			{
				switch (val)
				{
					case 0:		/* end of line flag */
						eolflag = 1;
						break;
					case 1:		/* end of picture flag */
						eopicflag = 1;
						break;
					case 2:		/* displace picture */
						count = (char) getc(f);
						val = (char) getc(f);
						pos += count;
						line += val;
						break;
					default:	/* read in absolute mode */
						for (j = 0; j < val; j++)
						{
							val0 = (char) getc(f);
							bmp[line * w + pos] = val0;
							pos++;
						}

						if (j % 2 == 1)
							/*val0 = (char) */getc(f);	/* align on word boundary */
						break;
				}
			}
			if (pos > w)
				eolflag = 1;
		}
		line--;
		if (line < 0)
			eopicflag = 1;
	}
}

static void copy_gif_palette(unsigned char *to, unsigned char *data, int cnt)
{
	int r,g,b;
	for(;cnt>0;cnt--)
	{
		r = *data++;
		g = *data++;
		b = *data++;
		*to++ = b;
		*to++ = g;
		*to++ = r;
	}
}

#ifdef FG_JPEG
struct r_jpeg_error_mgr
{
	struct jpeg_error_mgr pub;
    jmp_buf envbuffer;
};

int fh_jpeg_id(char *name)
{
    FILE * fd;
	unsigned char id[10];
	fd=fopen(name,"rb");
    if(fd==0)
        return 0;
    fread(id,1,10,fd);
	fclose(fd);

    if(id[6]=='J' && id[7]=='F' && id[8]=='I' && id[9]=='F') return(1);
    if(id[0]==0xff && id[1]==0xd8 && id[2]==0xff) return(1);

	return 0;
}


void jpeg_cb_error_exit(j_common_ptr cinfo)
{
    struct r_jpeg_error_mgr *mptr;
    mptr=(struct r_jpeg_error_mgr*) cinfo->err;
    (*cinfo->err->output_message) (cinfo);
    longjmp(mptr->envbuffer,1);
}

#endif

FGBitmap::ReturnValues FGBitmap::jpg(char *n, int& one, unsigned int& bits, unsigned int& c)
{
#ifdef FG_JPEG
	struct jpeg_decompress_struct cinfo;
	struct jpeg_decompress_struct *ciptr;
	struct r_jpeg_error_mgr emgr;
	FGPixel *bp;
	FILE *fh;
	JSAMPLE *lb;

	ciptr=&cinfo;
	if(fh_jpeg_id(n)==0) return IMAGE_FILE_NOT_FOUND;
    if(!(fh=fopen(n,"rb"))) return IMAGE_FILE_NOT_FOUND;
	ciptr->err=jpeg_std_error(&emgr.pub);
    emgr.pub.error_exit=jpeg_cb_error_exit;
    if(setjmp(emgr.envbuffer)==1)
    {
		// FATAL ERROR - Free the object and return...
		jpeg_destroy_decompress(ciptr);
		fclose(fh);
		return IMAGE_UNKNOWN_FORMAT;
    }

    jpeg_create_decompress(ciptr);
    jpeg_stdio_src(ciptr,fh);
    jpeg_read_header(ciptr,TRUE);
    ciptr->out_color_space=JCS_RGB;
    jpeg_start_decompress(ciptr);

    w = ciptr->output_width;
	h = ciptr->output_height;
	int cc = ciptr->output_components;
	c = 0;
	bits = 24;
	one = 0;

	assert(SetArray((FGPixel *) calloc(areasize(w, h), 1)));
	bp = GetArray();

    if(cc==3)
    {
		unsigned char pixel[4];
		lb = (JSAMPLE *)(*ciptr->mem->alloc_small)(
			(j_common_ptr) ciptr,
			JPOOL_PERMANENT,
			cc*w);
		while (ciptr->output_scanline < ciptr->output_height)
		{
		    jpeg_read_scanlines(ciptr, &lb, 1);
			for (int xs = 0; xs < w; xs++)
			{
				pixel[2] = *(lb+3*xs);
				pixel[1] = *(lb+3*xs+1);
				pixel[0] = *(lb+3*xs+2);
				*bp = FGDirectColor(*(unsigned int *)(&pixel));
				bp++;
			}
		}
    }
    jpeg_finish_decompress(ciptr);
    jpeg_destroy_decompress(ciptr);
	fclose(fh);
	return IMAGE_OK;
#else
	return IMAGE_UNSUPPORTED_FORMAT;
#endif
}

FGBitmap::ReturnValues FGBitmap::tiff(char *n, int& one, unsigned int& bits, unsigned int& c)
{
#if defined(FG_TIFF) && !defined(INDEX_COLORS)
	TIFF* tif = TIFFOpen(n, "r");
	if (tif)
	{
		TIFFRGBAImage img;
		char emsg[1024];

		if (TIFFRGBAImageBegin(&img, tif, 0, emsg)) {
			size_t npixels;
			uint32* raster;

			npixels = img.width * img.height;
			raster = (uint32*) _TIFFmalloc(npixels * sizeof (uint32));
			if (raster != NULL)
			{
				if (TIFFRGBAImageGet(&img, raster, img.width, img.height))
				{
					w = img.width;
					h = img.height;
					c = 0;
					bits = 24;
					one = 0;

					assert(SetArray((FGPixel *) calloc(areasize(w, h), 1)));
					unsigned char pixel[4]={0,0,0,0}, *lb = (unsigned char *)raster;
					FGPixel *bp = GetArray();

					for (int ys = img.height-1; ys>=0; ys--)
					{
						lb = (unsigned char *)(raster+img.width*ys);
						for (int xs = 0; xs < w; xs++)
						{
							pixel[2] = *(lb+4*xs);
							pixel[1] = *(lb+4*xs+1);
							pixel[0] = *(lb+4*xs+2);
							*bp = FGDirectColor(*(unsigned int *)(&pixel));
							bp++;
						}
					}
				}
				_TIFFfree(raster);
			}
			TIFFRGBAImageEnd(&img);
		}
		else
		{
			TIFFError(n, emsg);
			TIFFClose(tif);
			return IMAGE_FILE_NOT_FOUND;
		}
		TIFFClose(tif);
	}
	return IMAGE_OK;
#else
	return IMAGE_UNSUPPORTED_FORMAT;
#endif
}

#ifdef FG_PNG

#define PNG_BYTES_TO_CHECK 4
#define min(x,y) ((x) < (y) ? (x) : (y))

int fh_png_id(char *name)
{
	FILE* fd;
	char id[4];
	fd=fopen(name,"rb");
	if(fd==0) return(0);
	fread(id,1,4,fd);
	fclose(fd);
	if(id[1]=='P' && id[2]=='N' && id[3]=='G') return(1);
	return(0);
}
#endif // PNG

FGBitmap::ReturnValues FGBitmap::png(char *n, int& one, unsigned int& bits, unsigned int& c)
{
#ifdef FG_PNG
	png_structp png_ptr;
	png_infop info_ptr;
	png_uint_32 width, height;
	int i;
	int bit_depth, color_type, interlace_type;
	int number_passes,pass;
	char *rp;
	png_bytep rptr[2];
	FGPixel *fbptr;
	FILE *fh;

	if(fh_png_id(n)==0) return IMAGE_FILE_NOT_FOUND;
	if(!(fh=fopen(n,"rb"))) return IMAGE_FILE_NOT_FOUND;

	png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING,NULL,NULL,NULL);
	if (png_ptr == NULL) return IMAGE_UNKNOWN_FORMAT;
	info_ptr = png_create_info_struct(png_ptr);
	if (info_ptr == NULL)
	{
		png_destroy_read_struct(&png_ptr, (png_infopp)NULL, (png_infopp)NULL);
		fclose(fh);
        return IMAGE_UNKNOWN_FORMAT;
	}
	rp=0;
	if (setjmp(png_ptr->jmpbuf))
	{
		png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
		if(rp) free(rp);
		fclose(fh);
        return IMAGE_UNKNOWN_FORMAT;
	}

	png_init_io(png_ptr,fh);

	png_read_info(png_ptr, info_ptr);
	png_get_IHDR(png_ptr, info_ptr, &width, &height, &bit_depth, &color_type,&interlace_type, NULL, NULL);

	if (color_type == PNG_COLOR_TYPE_PALETTE)
		png_set_expand(png_ptr);
	if (bit_depth < 8)
		png_set_packing(png_ptr);
	if (color_type == PNG_COLOR_TYPE_GRAY || color_type== PNG_COLOR_TYPE_GRAY_ALPHA)
		png_set_gray_to_rgb(png_ptr);
    if( bit_depth==16)
		png_set_strip_16(png_ptr);
	png_set_filter(png_ptr, 0, PNG_NO_FILTERS);
/*
	If, for some reason, you don't need the alpha channel on an image,
	and you want to remove it rather than combining it with the background
	(but the image author certainly had in mind that you *would* combine
	it with the background, so that's what you should probably do):
*/
    if (color_type & PNG_COLOR_MASK_ALPHA)
		png_set_strip_alpha(png_ptr);
/*
	In PNG files, the alpha channel in an image
	is the level of opacity.  If you need the alpha channel in an image to
	be the level of transparency instead of opacity, you can invert the
	alpha channel (or the tRNS chunk data) after it's read, so that 0 is
	fully opaque and 255 (in 8-bit or paletted images) or 65535 (in 16-bit
	images) is fully transparent, with
*/
//    png_set_invert_alpha(png_ptr);

	number_passes = png_set_interlace_handling(png_ptr);
	png_read_update_info(png_ptr,info_ptr);

    rp=(char*) malloc(width*3);
    rptr[0]=(png_bytep) rp;
	w = width;
	h = height;
	c = 0;
	bits = 24;
	one = 0;
	assert(SetArray((FGPixel *) calloc(areasize(w, h), 1)));
	unsigned char pixel[4];

	for (pass = 0; pass < number_passes; pass++)
	{
		fbptr=GetArray();
		for(i=0;i<(int)height;i++)
		{
			png_read_rows(png_ptr, rptr, NULL, 1);
			for (int xs = 0; xs < w; xs++)
			{
				pixel[2] = *(rp+3*xs);
				pixel[1] = *(rp+3*xs+1);
				pixel[0] = *(rp+3*xs+2);
				*fbptr = FGDirectColor(*(unsigned int *)(&pixel));
				fbptr++;
			}
		}
	}
	free(rp);
	png_read_end(png_ptr, info_ptr);
	png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
	fclose(fh);
	return IMAGE_OK;
#else
	return IMAGE_UNSUPPORTED_FORMAT;
#endif
}

FGBitmap::ReturnValues FGBitmap::pcx(char *n, int& one, unsigned int& bits, unsigned int& c)
{
	char hdr[128];
	FILE *f=0;
	char nm[128];
	int counter, byte, data;
	register unsigned char *src;

		f = fopen(n, "rb");
		if (!f)
		{
			sprintf(nm, "FGBitmap '%s' not found!", n);
			IError(nm, 0);
			return IMAGE_FILE_NOT_FOUND;
		}
		fread(hdr, 128, 1, f);
		if (hdr[0] != 0xA || hdr[1] != 5)
		{
error:
			sprintf(nm, "FGBitmap '%s' isn't PCX 5.0 file!", n);
			IError(nm, 0);
			fclose(f);
			return IMAGE_UNKNOWN_FORMAT;
		}
		w = *(short *)(hdr+66);
		h = *(short *)(hdr+10)+1;
		bits = hdr[3];
		assert(SetArray((FGPixel *) calloc(areasize(w, h), 1)));
		int sz = w * h;
		src = (unsigned char *)GetArray();
		unsigned char *end = src+sz;
		while (src < end)
		{
			byte = fgetc(f);
			if (0xC0 == (0xC0 & byte))
			{
				counter = 0x3F & byte;
				data = fgetc(f);
				while(counter--)
				{
					*src++ = data;
				}
			}
			else
			{
				*src++ = byte;
			}
		}
		byte = fgetc(f);
		if (byte != 0xC) goto error;
		fread(&rgb, 768, 1, f);
		copy_gif_palette((unsigned char *)&rgb, (unsigned char *)&rgb, 256);
		fclose(f);
//		f = 0;
		c = 256;
		one = 3;
		return IMAGE_OK;
}

static int rle_read(FILE * fp, unsigned char * buffer, tga_info * info)
{
	static int repeat = 0;
	static int direct = 0;
	static unsigned char sample[4];
	int head;
	int x, k;

	for (x = 0; x < info->width; x++)
	{
		if (repeat == 0 && direct == 0)
		{
			head = getc(fp);

			if (head == EOF)
			{
				return EOF;
			}
			else if (head >= 128)
			{
				repeat = head - 127;

				if (fread(sample, info->bytes, 1, fp) < 1)
					return EOF;
			}
			else
			{
				direct = head + 1;
			}
		}

		if (repeat > 0)
		{
			for (k = 0; k < info->bytes; ++k)
			{
				buffer[k] = sample[k];
			}

			repeat--;
		}
		else					/* direct > 0 */
		{
			if (fread(buffer, info->bytes, 1, fp) < 1)
				return EOF;

			direct--;
		}

		buffer += info->bytes;
	}

	return 0;
}

static void flip_line(unsigned char * buffer, tga_info * info)
{
	unsigned char temp;
	unsigned char *alt;
	int x, s;

	alt = buffer + (info->bytes * (info->width - 1));

	for (x = 0; x * 2 <= info->width; x++)
	{
		for (s = 0; s < info->bytes; ++s)
		{
			temp = buffer[s];
			buffer[s] = alt[s];
			alt[s] = temp;
		}

		buffer += info->bytes;
		alt -= info->bytes;
	}
}

static void read_line(FILE * fp,
		  FGPixel * row, unsigned char * buffer, tga_info * info, unsigned *palette)
{
 	int width=info->width;
#ifndef INDEX_COLORS
	int x;
	unsigned char *dest = (unsigned char *)row;
#endif
	// read from the file
	if (info->imageCompression == TGA_COMP_RLE)
	{
		rle_read(fp, buffer, info);
	}
	else
	{
		fread(buffer, info->bytes, info->width, fp);
	}
	// flip
	if (info->flipHoriz)
	{
		flip_line(buffer, info);
	}

	switch(info->imageType)
	{
	case TGA_TYPE_COLOR:
#ifdef INDEX_COLORS
		IError("Not supported BPP", 1);
#endif
#ifdef DIRECT_COLORS
		if (info->bpp == 16)
		{
			for (x = 0; x < width; x++)
			{
				dest[0] = buffer[0];
				dest[1] = buffer[1];
#ifdef BPP16
				dest[1] <<= 1;
				dest[1] |= (dest[0]&0x80? 1:0);
				dest[0] = (dest[0]&0x1f)+((dest[0]&0x60)<<1);
#endif
				buffer+=2;
				dest+=2;
			}
		}
		else
		{
			int pad=info->bytes;
			for (x = 0; x < width; x++)
			{
				*((short *)dest) = FGDirectColor(*(unsigned *)buffer);
				buffer+=pad;
				dest+=2;
			}
		}
#endif
#ifdef TRUE_COLORS
		switch (info->bpp)
		{
		case 16:	// 16 > 32bit
			for (x = 0; x < width; x++)
			{
				dest[0] = ((buffer[0] << 3) & 0xf8);
				dest[0] += (dest[2] >> 5);

				dest[1] = ((buffer[0] & 0xe0) >> 2) + ((buffer[1] & 0x03) << 6);
				dest[1] += (dest[1] >> 5);

				dest[2] = ((buffer[1] << 1) & 0xf8);
				dest[2] += (dest[0] >> 5);

				buffer+=2;
				dest+=4;
			}
			break;
		case 24:	// 24 > 32bit
			for (x = 0; x < width; x++)
			{
				dest[0] = buffer[0];
				dest[1] = buffer[1];
				dest[2] = buffer[2];
				dest[3] = 0;
				buffer+=3;
				dest+=4;
			}
			break;
		case 32:  // ONE to ONE
			memcpy (row, buffer, info->width * info->bytes);
			break;
		}
#endif
		break;

	case TGA_TYPE_GRAY:
#ifdef INDEX_COLORS
		memcpy (row, buffer, info->width * info->bytes);
#endif
#ifdef DIRECT_COLORS
		for (x = 0; x < width; x++)
		{
			*((short *)dest) = FGDirectColor(palette[*buffer]);
			buffer++;
			dest+=2;
		}
#endif
#ifdef TRUE_COLORS
		for (x = 0; x < width; x++)
		{
			*(unsigned *)dest = palette[*buffer];
			buffer++;
			dest+=4;
		}
#endif
		break;
	case TGA_TYPE_MAPPED:
#ifdef TRUE_COLORS
			for (x = 0; x < width; x++)
			{
				*(unsigned *)dest = palette[*buffer];
				buffer++;
				dest+=4;
			}
#endif
#ifdef DIRECT_COLORS
		for (x = 0; x < width; x++)
		{
			*((short *)dest) = FGDirectColor(palette[*buffer]);
			buffer++;
			dest+=2;
		}
#endif
#ifdef INDEX_COLORS
		// this works only for BPP8, either this point will be never reached
		memcpy (row, buffer, width * info->bytes);
#endif
		break;
	}
}

//
// tga helper
//
static int ReadImage(FILE * fp, tga_info * info, char * filename, unsigned char *tga_cmap, FGPixel *data)
{
	unsigned char *buffer;
 	FGPixel *row;
	int i, y;

	unsigned int cmap_bytes;
	/* Handle colormap */

	if (info->colorMapType == 1)
	{
		memset(tga_cmap, 0, sizeof(tga_cmap));
		cmap_bytes = (info->colorMapSize + 7) / 8;
		if (cmap_bytes <= 4)
		{
			for (i=0; i<info->colorMapLength; i++)
				fread(tga_cmap+(i*4), cmap_bytes, 1, fp);
		}
		else
		{
			::printf("TGA: File is truncated or corrupted \"%s\"\n", filename);
			return -1;
		}
	}

	/* Allocate the data buffer, with file BPP */
	buffer = (unsigned char *) malloc (info->width * info->bytes);

	if (info->flipVert)
		row = data + (info->width * info->height);
	else
		row = data;

	for (y = 0; y < info->height; ++y)
	{
		if (info->flipVert)
			row -= info->width;

		// read the one line
		read_line(fp, row, buffer, info, (unsigned *)tga_cmap);
//::printf("row = %x\n",row);
		if (!info->flipVert)
			row += info->width;
	}
	free(buffer);
	return 0;
}

FGBitmap::ReturnValues FGBitmap::tga(char *filename, int& one, unsigned int& bits, unsigned int& c)
{
	FILE *fp;
	char nm[128];
//	int counter, byte, data;
	register unsigned char *src;

	tga_info info;
	unsigned char header[18],
		footer[26],
		extension[495];
	long offset;

	memset(&info, 0, sizeof(info));
	fp = fopen(filename, "rb");
	if (!fp)
	{
		sprintf(nm, "FGBitmap '%s' not found!", filename);
		IError(nm, 0);
		return IMAGE_FILE_NOT_FOUND;
	}

	if (!fseek(fp, -26L, SEEK_END))
	{							/* Is file big enough for a footer? */
		if (fread(footer, sizeof(footer), 1, fp) != 1)
		{
			::printf("TGA: Cannot read footer from %s\n", filename);
			return IMAGE_UNKNOWN_FORMAT;
		}
		else if (memcmp(footer + 8, magic, sizeof(magic)) == 0)
		{
			/* Check the signature. */
			offset = footer[0] + (footer[1] * 256) + (footer[2] * 65536)
				+ (footer[3] * 16777216);

			if (fseek(fp, offset, SEEK_SET) ||
				fread(extension, sizeof(extension), 1, fp) != 1)
			{
				::printf("TGA: Cannot read extension from \"%s\"\n", filename);
				return IMAGE_UNKNOWN_FORMAT;
			}

			/* Eventually actually handle version 2 TGA here */
		}
	}

	if (fseek(fp, 0, SEEK_SET) || fread(header, sizeof(header), 1, fp) != 1)
	{
		::printf("TGA: Cannot read header from \"%s\"\n", filename);
		return IMAGE_UNKNOWN_FORMAT;
	}

	switch (header[2])
	{
		case 1:
			info.imageType = TGA_TYPE_MAPPED;
			info.imageCompression = TGA_COMP_NONE;
			break;
		case 2:
			info.imageType = TGA_TYPE_COLOR;
			info.imageCompression = TGA_COMP_NONE;
			break;
		case 3:
			info.imageType = TGA_TYPE_GRAY;
			info.imageCompression = TGA_COMP_NONE;
			break;

		case 9:
			info.imageType = TGA_TYPE_MAPPED;
			info.imageCompression = TGA_COMP_RLE;
			break;
		case 10:
			info.imageType = TGA_TYPE_COLOR;
			info.imageCompression = TGA_COMP_RLE;
			break;
		case 11:
			info.imageType = TGA_TYPE_GRAY;
			info.imageCompression = TGA_COMP_RLE;
			break;

		default:
			info.imageType = 0;
	}

	info.idLength = header[0];
	info.colorMapType = header[1];

	info.colorMapIndex = header[3] + header[4] * 256;
	info.colorMapLength = header[5] + header[6] * 256;
	info.colorMapSize = header[7];

	info.xOrigin = header[8] + header[9] * 256;
	info.yOrigin = header[10] + header[11] * 256;
	info.width = header[12] + header[13] * 256;
	info.height = header[14] + header[15] * 256;

	bits = info.bpp = header[16];
	info.bytes = (info.bpp + 7) / 8;
	info.alphaBits = header[17] & 0x0f;	/* Just the low 4 bits */
	info.flipHoriz = (header[17] & 0x10) ? 1 : 0;
	info.flipVert = (header[17] & 0x20) ? 0 : 1;

	switch (info.imageType)
	{
		case TGA_TYPE_MAPPED:
			if (info.bpp != 8)
			{
				::printf("TGA: Unhandled sub-format in \"%s\"\n", filename);
				return IMAGE_UNSUPPORTED_FORMAT;
			}
			break;
		case TGA_TYPE_COLOR:
			if (info.alphaBits == 0 && info.bpp == 32)
			{
				::printf("TGA: Possibly incorrect alpha in \"%s\"\n",
						  filename);
				info.alphaBits = 8;
			}
			if (info.bpp != 16 && info.bpp != 24
				&& (info.alphaBits != 8 || info.bpp != 32))
			{
				::printf("TGA: Unhandled sub-format in \"%s\"\n", filename);
				return IMAGE_UNSUPPORTED_FORMAT;
			}
			break;
		case TGA_TYPE_GRAY:
			if (info.alphaBits == 0 && info.bpp == 16)
			{
				::printf("TGA: Possibly incorrect alpha in \"%s\"\n",
						  filename);
				info.alphaBits = 8;
			}
			if (info.bpp != 8 && (info.alphaBits != 8 || info.bpp != 16))
			{
				::printf("TGA: Unhandled sub-format in \"%s\"\n", filename);
				return IMAGE_UNSUPPORTED_FORMAT;
			}
			break;

		default:
			::printf("TGA: Unknown image type for \"%s\"\n", filename);
			return IMAGE_UNSUPPORTED_FORMAT;
	}

	/* Plausible but unhandled formats */
	if (info.bytes * 8 != info.bpp)
	{
		::printf("TGA: No support yet for TGA with these parameters\n");
		return IMAGE_UNSUPPORTED_FORMAT;
	}

	/* Check that we have a color map only when we need it. */
	if (info.imageType == TGA_TYPE_MAPPED && info.colorMapType != 1)
	{
		::printf("TGA: indexed image has invalid color map type %d\n",
				  info.colorMapType);
		return IMAGE_UNSUPPORTED_FORMAT;
	}
	else if (info.imageType != TGA_TYPE_MAPPED && info.colorMapType != 0)
	{
		::printf("TGA: non-indexed image has invalid color map type %d\n",
				  info.colorMapType);
		return IMAGE_UNSUPPORTED_FORMAT;
	}
#ifdef INDEX_COLORS
	else if (info.imageType == TGA_TYPE_COLOR)
	{
		::printf("TGA: non-indexed image not supported with BPP==8\n");
		return IMAGE_UNSUPPORTED_FORMAT;
	}
#endif
	/* Skip the image ID field. */
	if (info.idLength && fseek(fp, info.idLength, SEEK_CUR))
	{
		::printf("TGA: File is truncated or corrupted \"%s\"\n", filename);
		return IMAGE_UNKNOWN_FORMAT;
	}

	w = info.width;
	h = info.height;
	bits = info.bpp;

	if (info.bpp>8)
	{
		c = 0;
		one = 0;
	}
	else
	{
		c = 256;
		one = 4;
	}

	assert(SetArray((FGPixel *) calloc(areasize(w, h), 1)));
	src = (unsigned char *)GetArray();

	// generate gray ramp
	if (info.imageType == TGA_TYPE_GRAY)
		for(int i=0; i<(int)c; i++)
			*(unsigned *)(&rgb.rgb4[i]) = i+(i<<8)+(i<<16);

	ReadImage(fp, &info, filename, (unsigned char *)&rgb, GetArray());

	// mark as already converted
	bits = 24;
	fclose(fp);
	return IMAGE_OK;
}

FGBitmap::ReturnValues FGBitmap::gif(char *n, int& one, unsigned int& bits, unsigned int& c)
{
	GIFStream *gif=0;
	char nm[128];

		gif = GIFRead(n);
		if (gif)
		{
			w = gif->width;
			h = gif->height;
			one = 3;
			c = gif->cmapSize;
			bits = 8;
			if (c) copy_gif_palette((unsigned char *)&rgb, gif->cmapData[0], c);
			assert(SetArray((FGPixel *) malloc(areasize(w, h))));
			memset(GetArray(), gif->background, areasize(w, h));
			GIFData *d = gif->data;
			while(d)
			{
				if (d->type == gif_image)
				{
					if (d->data.image.cmapSize)
					{
						c = d->data.image.cmapSize;
						if (c) copy_gif_palette((unsigned char *)&rgb, d->data.image.cmapData[0], c);
					}
					state.colorkey = d->info.transparent;
					memcpy(GetArray(), d->data.image.data, d->width*d->height);
					free(d->data.image.data);
				}
				else if(d->type == gif_text || d->type == gif_comment)
				{
					if (d->data.comment.text) free(d->data.comment.text);
				}
				GIFData *pom = d;
				d = d->next;
				free(pom);
			}
			free(gif);
		}
		else
		{
			sprintf(nm, "GIF file '%s' loading error!", n);
			IError(nm, 0);
			return IMAGE_FILE_NOT_FOUND;
		}
		return IMAGE_OK;
}

/**
	Performs image rectangle copy from one FGDrawBuffer object to other with specified
	ROP and clipping.
	@return false if no drawing was performed, or true if all ran OK.
	@param xdst x coord into the destination block
	@param ydst y coord into the destination block
	@param xsrc x coord into the source block
	@param ysrc y coord into the source block
	@param ww a width of copied block (maybe bigger than src or dest block)
	@param hh a height of copied block (maybe bigger than src or dest block)
	@param src a source object that will be copied
	@param opcia a bit oriented value that describes the operational mode of procedure:
-	BITBLIT_COPY - copy only
-	BITBLIT_WITH_OPERATOR - copy with operator
-	BITBLIT_WITH_ALPHA - use the source alpha value
-	BITBLIT_WITH_CLIPPING - perform clipping before
-	BITBLIT_WITH_DECREMENT - for use with overlapping blocks
	@see Defines for more about one.
*/
int FGAPI FGDrawBuffer::bitblit(int xdst, int ydst, int xsrc, int ysrc, int ww, int hh, FGDrawBuffer *src, int opcia)
{
	int ppp = state._ppop, oldt = fgstate.color_key;
	int _ww=src->w, _hh=src->h;

	if (src->image == 0 || src->type == BMP_NONE || image==0)
		return 0;

	xsrc += src->xoff;
	ysrc += src->yoff;
	xdst += xoff;
	ydst += yoff;

	// compute the tile size
	if (src->clip(xsrc, ysrc, _ww, _hh)==0) return 0;
	if (ww<1 || hh<1) return 0;

	// test for transparency ... and force one
	if (src->GetAlpha())
		opcia = (src->GetAlpha() | BITBLIT_WITH_OPERATOR);

	if (src->GetColorKey() != DEFAULT_COLOR_KEY)
	{
		ppp = _GCOLORKEY;
		fgstate.color_key = src->GetColorKey();
		opcia |= BITBLIT_WITH_OPERATOR;
	}
	for (int i = 0; i < hh;)
	{
		for (int j = 0; j < ww;)
		{
			int _x=xdst+j,
				_y=ydst+i,
				_w=j+_ww>ww?ww-j:_ww,
				_h=i+_hh>hh?hh-i:_hh;

			if (clip(_x, _y, _w, _h))
				RamToRam(xsrc+(_x-(xdst+j)), ysrc+(_y-(ydst+i)), src->w, src->h,
					_x, _y, w, h,
					_w, _h, src->image, image, opcia, ppp);
			j += _ww;
		}
		i += _hh;
	}
	fgstate.color_key = oldt;
	return 1;
}

FGBitmap::~FGBitmap()
{
#ifdef INDEX_COLORS
	if (cUsedTable)
	{
		for (int i = 0; i < 256; i++)
			if (cUsedTable[i] != -1)
				DeleteColor(cUsedTable[i]);
		delete cUsedTable;
	}
#endif
}

/**
	Create a BITMAP object in memory with specified size and color layout.
*/
FGBitmap::FGBitmap(int ww, int hh, FGPixel color)
: FGDrawBuffer(ww, hh, BMP_MEM, color)
{
	init();
}

/**
	Create a BITMAP object in memory as exact copy of the FGDrawBuffer object.
	This one has dimensions and contents the same as source buffer.
	The image has been copied from, that you can't expect that any next changes
	are appears in it. The application of one are when you want save
	the contents of a window to the BMP file by example.
*/
FGBitmap::FGBitmap(FGDrawBuffer * wPtr)
:FGDrawBuffer(wPtr->GetW(), wPtr->GetH())
{
	init();
	memcpy(image, wPtr->GetArray(), areasize(wPtr->GetW(), wPtr->GetH()));
}

void FGBitmap::init(void)
{
	cUsedTable = 0;
	format = IMAGE_UNKNOWN;
	name = 0;
	memset(&bmih, 0, sizeof(bmih));
	state._ink = CWHITE;
	state._paper = CBLACK;
	state._font = 2;
	state._ppop = _GSET;
	state.colorkey = DEFAULT_COLOR_KEY;
	state.alpha = DEFAULT_ALPHA_VALUE;
}

/**
	Saves a BITMAP object into the current directory.
	The file will be standard Windows BMP file. Extension
	is added automagically.
	@param filename
	@param convert convert to PNG image too.
	@return value is false if operation fail. On linux platform
	is file saved to PNG format also, if imagemagic is installed.
*/
FGBitmap::ReturnValues FGBitmap::BitmapSave(const char *filename, bool convert)
{
	FILE *fp=0;

	if (GetType() == BMP_NONE)
		return IMAGE_UNKNOWN_FORMAT;

#ifndef INDEX_COLORS
	int j, rest;
	char *bmp2;
#endif

	bmfh.usType = 'M' * 256 + 'B';
	bmfh.reserve1 = 0;
	bmfh.reserve2 = 0;
	bmi.reserve = 0;
	bmih.bmih2.size = 0x28;
	bmih.bmih2.cx = w;
	bmih.bmih2.cy = h;
	bmih.bmih2.cPlanes = 1;
	bmih.bmih2.ulCompression = 0;
	bmih.bmih2.cbImage = w * h;
	bmih.bmih2.cxResolution = 0;
	bmih.bmih2.cyResolution = 0;
#ifdef INDEX_COLORS
	bmi.imageOffset = 0x36 + 0x400;
	bmfh.cbSizeLow = w * h + 1024 + 0x36;
	bmfh.cbSizeHigh = (w * h + 1024 + 0x36) >> 16;
	bmih.bmih2.cclrUsed = 256;
	bmih.bmih2.cclrImportant = 256;
	bmih.bmih2.cBitCount = 8;
	memcpy(&rgb, &_fgl_palette, sizeof(rgb));
#else
	bmi.imageOffset = 0x36;
	bmfh.cbSizeLow = w * h + 0x36;
	bmfh.cbSizeHigh = (w * h + 0x36) >> 16;
	bmih.bmih2.cclrUsed = 0;
	bmih.bmih2.cclrImportant = 0;
	bmih.bmih2.cBitCount = 24;
#endif // index

	char* ext;
	if ((ext = strstr(filename,".bmp")) == 0)
	{
		char new_filename[128];
		strncpy(new_filename, filename, sizeof(new_filename)-5);
		strcat(new_filename, ".bmp");
		fp = fopen(new_filename, "wb");
	}
	else
		fp = fopen(filename, "wb");

	if (fp == 0)
	{
		IError("Bitmap not saved because of disk write error!", 0);
		return IMAGE_FILE_NOT_FOUND;
	}
	assert(sizeof(bmfh) == 10);
	fwrite(&bmfh, sizeof(bmfh), 1, fp);
	fwrite(&bmi, sizeof(bmi), 1, fp);
	fwrite(&bmih, sizeof(bmih.bmih2), 1, fp);

#ifdef INDEX_COLORS
	fwrite(&rgb, sizeof(rgb.rgb4), 1, fp);
	rotate(ROTATE_VERTICAL);
	fwrite(image, w * h, 1, fp);
	rotate(ROTATE_VERTICAL);
#endif // index

#ifndef INDEX_COLORS
#if (FASTGL_BPP == 15 || FASTGL_BPP == 16)
	unsigned long cc, aa;
#endif // BPP16
	for (int i = h - 1; i >= 0; i--)
	{
		bmp2 = (char *) (image + w * i);
		for (j = 0; j < w; j++)
		{
#ifdef TRUE_COLORS
			fwrite(bmp2, 3, 1, fp);
			bmp2 += 4;
#else // true
			cc = *(FGPixel *) bmp2;
#if (FASTGL_BPP == 15)
			aa = ((cc << 9) & 0xF80000) | ((cc << 6) & 0xF800) | ((cc << 3) & 0xF8);
#else
			aa = ((cc << 8) & 0xF80000) | ((cc << 5) & 0xFc00) | ((cc << 3) & 0xF8);
#endif // direct
			fwrite(&aa, 3, 1, fp);
			bmp2 += 2;
#endif // true
		}
#ifdef TRUE_COLORS
		rest = (w * 3) % 4;
#else
		rest = (w * 2) % 4;
#endif // true
		while (rest & 3)
		{
			rest++;
			fputc(0, fp);
		}
	}
#endif // not index_color
	fclose(fp);
#ifdef __linux__
	if (convert)
	{
		char ss[64];
		sprintf(ss, "/usr/X11R6/bin/convert %s.bmp %s.png", filename, filename);
		system(ss);
	}
#endif
	return IMAGE_OK;
}

// swaps the blue and red components (for RGB images)
// > buffer: pointer to the pixels
// > lenght: number of bytes to swap. lenght may not exceed the scan line.
void FGBitmap::RGBtoBGR(unsigned char *buffer, unsigned char *input, int length)
{
	if (buffer)
	{
		length *= 3;
		for (int i=0;i<length;i+=3)
		{
			buffer[i+2] = *input++;
			buffer[i+1] = *input++;
			buffer[i] = *input++;
			input++;
		}
	}
}

FGBitmap::ReturnValues FGBitmap::JpegSave(const char *filename)
{
#ifdef FG_JPEG

#ifdef INDEX_COLORS
	IError("JPEG can save only RGB or GreyScale images", 0);
	return IMAGE_UNSUPPORTED_FORMAT;
#endif

	FILE *outfile = fopen(filename, "wb");

	if (outfile == 0)
	{
		IError("Bitmap not saved because of disk write error!", 0);
		return IMAGE_FILE_WRITE_ERROR;
	}

	// This struct contains the JPEG compression parameters and pointers to
	// working space (which is allocated as needed by the JPEG library).
	// It is possible to have several such structures, representing multiple
	// compression/decompression processes, in existence at once.  We refer
	// to any one struct (and its associated working data) as a "JPEG object".
	struct jpeg_compress_struct cinfo;

	// This struct represents a JPEG error handler.  It is declared separately
	// because applications often want to supply a specialized error handler
	// (see the second half of this file for an example).  But here we just
	// take the easy way out and use the standard error handler, which will
	// print a message on stderr and call exit() if compression fails.
	// Note that this struct must live as long as the main JPEG parameter
	// struct, to avoid dangling-pointer problems.
	//
	struct ima_error_mgr jerr;
	jerr.buffer = szLastError;
	// More stuff
	int row_stride;
	JSAMPARRAY buffer;

	// Step 1: allocate and initialize JPEG compression object
	// We have to set up the error handler first, in case the initialization
	// step fails.  (Unlikely, but it could happen if you are out of memory.)
	// This routine fills in the contents of struct jerr, and returns jerr's
	// address which we place into the link field in cinfo.
	//
	// We set up the normal JPEG error routines, then override error_exit.
	cinfo.err = jpeg_std_error(&jerr.pub);
	jerr.pub.error_exit = fgl_jpeg_error_exit;

	// Establish the setjmp return context for my_error_exit to use.
	if (setjmp(jerr.setjmp_buffer))
	{
		// If we get here, the JPEG code has signaled an error.
		// We need to clean up the JPEG object, close the input file, and return.
		strcpy(szLastError, jerr.buffer);
		jpeg_destroy_compress(&cinfo);
		return IMAGE_FILE_WRITE_ERROR;
	}

	// Now we can initialize the JPEG compression object.
	jpeg_create_compress(&cinfo);
	// Step 2: specify data destination (eg, a file)
	// Note: steps 2 and 3 can be done in either order.
	// Here we use the library-supplied code to send compressed data to a
	// stdio stream.  You can also write your own code to do something else.
	// VERY IMPORTANT: use "b" option to fopen() if you are on a machine that
	// requires it in order to write binary files.

//	jpeg_stdio_dest(&cinfo, outfile);
	CxFileJpg dest(outfile);
	cinfo.dest = &dest;

	// Step 3: set parameters for compression
	// First we supply a description of the input image.
	// Four fields of the cinfo struct must be filled in:
	cinfo.image_width = w; 	// image width and height, in pixels
	cinfo.image_height = h;
/*
	if (IsGrayScale())
	{
		cinfo.input_components = 1;			// # of color components per pixel
		cinfo.in_color_space = JCS_GRAYSCALE; // colorspace of input image
	}
	else
	{
*/		cinfo.input_components = 3; 	// # of color components per pixel
		cinfo.in_color_space = JCS_RGB; // colorspace of input image
//	}

	/* Now use the library's routine to set default compression parameters.
	* (You must set at least cinfo.in_color_space before calling this,
	* since the defaults depend on the source color space.)
	*/
	jpeg_set_defaults(&cinfo);
	/* Now you can set any non-default parameters you wish to.
	* Here we just illustrate the use of quality (quantization table) scaling:
	*/
	//jpeg_set_quality(&cinfo, info.nQuality, TRUE /* limit to baseline-JPEG values */);

#ifdef C_ARITH_CODING_SUPPORTED
	if ((GetCodecOption() & ENCODE_ARITHMETIC) != 0)
		cinfo.arith_code = TRUE;
#endif
#ifdef ENTRPY_OPT_SUPPORTED
	if ((GetCodecOption() & ENCODE_OPTIMIZE) != 0)
		cinfo.optimize_coding = TRUE;
#endif
//	if ((GetCodecOption() & ENCODE_GRAYSCALE) != 0)
//		jpeg_set_colorspace(&cinfo, JCS_GRAYSCALE);
//	if ((GetCodecOption() & ENCODE_SMOOTHING) != 0)
//		cinfo.smoothing_factor = m_nSmoothing;
	jpeg_set_quality(&cinfo, 60, true);
	cinfo.optimize_coding = TRUE;
#ifdef C_PROGRESSIVE_SUPPORTED
	if ((GetCodecOption() & ENCODE_PROGRESSIVE) != 0)
		jpeg_simple_progression(&cinfo);
#endif
#ifdef C_LOSSLES_SUPPORTED
	if ((GetCodecOption() & ENCODE_LOSSLESS) != 0)
		jpeg_simple_lossless(&cinfo, m_nPredictor, m_nPointTransform);
#endif

	cinfo.density_unit=1;
	cinfo.X_density=(unsigned short)75;
	cinfo.Y_density=(unsigned short)75;

	/* Step 4: Start compressor */
	/* TRUE ensures that we will write a complete interchange-JPEG file.
	* Pass TRUE unless you are very sure of what you're doing.
	*/
	jpeg_start_compress(&cinfo, TRUE);

	/* Step 5: while (scan lines remain to be written) */
	/*           jpeg_write_scanlines(...); */
	/* Here we use the library's state variable cinfo.next_scanline as the
	* loop counter, so that we don't have to keep track ourselves.
	* To keep things simple, we pass one scanline per call; you can pass
	* more if you wish, though.
	*/
	row_stride = w;	/* JSAMPLEs per row in image_buffer */

	//<DP> "8+row_stride" fix heap deallocation problem during debug???
	buffer = (*cinfo.mem->alloc_sarray)
		((j_common_ptr) &cinfo, JPOOL_IMAGE, 8+row_stride*3, 1);

	while (cinfo.next_scanline < cinfo.image_height)
	{
		// not necessary if swapped red and blue definition in jmorecfg.h;ln322 <W. Morrison>
		// swap R & B for RGB images
		RGBtoBGR(buffer[0], (unsigned char *)CalcAddr(0, cinfo.next_scanline), row_stride); // Lance : 1998/09/01 : Bug ID: EXP-2.1.1-9
		jpeg_write_scanlines(&cinfo, buffer, 1);
	}

	jpeg_finish_compress(&cinfo);

	/* And we're done! */
	return IMAGE_OK;
#else
	return IMAGE_UNSUPPORTED_FORMAT;
#endif
}


FGBitmap::ReturnValues FGBitmap::PngSave(const char *filename)
{
	return IMAGE_UNSUPPORTED_FORMAT;
}


/* +-------------------------------------------------------------------+ */
/* | Copyright 1990 - 1994, David Koblas. (koblas@netcom.com)          | */
/* |   Permission to use, copy, modify, and distribute this software   | */
/* |   and its documentation for any purpose and without fee is hereby | */
/* |   granted, provided that the above copyright notice appear in all | */
/* |   copies and that both that copyright notice and this permission  | */
/* |   notice appear in supporting documentation.  This software is    | */
/* |   provided "as is" without express or implied warranty.           | */
/* +-------------------------------------------------------------------+ */

GIFStream *GIFReadFP(FILE * fd)
{
	unsigned char buf[256];
	unsigned char c;
	GIFStream *stream=0;
	GIFData *cur, **end;
	GIF89info info;
	int resetInfo = TRUE;
	int n;

	if (fd == NULL)
		return NULL;

	if (setjmp(setjmp_buffer))
		goto out;

	if (!ReadOK(fd, buf, 6))
		_ERROR("error reading magic number");

	if (strncmp((char *) buf, "GIF", 3) != 0)
		_ERROR("not a GIF file");

	if ((strncmp((const char *) buf + 3, "87a", 3) != 0) &&
		(strncmp((const char *) buf + 3, "89a", 3) != 0))
		_ERROR("bad version number, not '87a' or '89a'");

	if (!ReadOK(fd, buf, 7))
		_ERROR("failed to read screen descriptor");

	stream = NEW(GIFStream);

	stream->width = MKINT(buf[0], buf[1]);
	stream->height = MKINT(buf[2], buf[3]);

	stream->cmapSize = 2 << (buf[4] & 0x07);
	stream->colorMapSize = stream->cmapSize;
	stream->colorResolution = ((int) (buf[4] & 0x70) >> 3) + 1;
	stream->background = buf[5];
	stream->aspectRatio = buf[6];

	stream->data = NULL;

	end = &stream->data;

	/*
	   **  Global colormap is present.
	 */
	if (BitSet(buf[4], LOCALCOLORMAP))
	{
		if (readColorMap(fd, stream->cmapSize, stream->cmapData))
			_ERROR("unable to get global colormap");
	}
	else
	{
		stream->cmapSize = 0;
		stream->background = -1;
	}

	if (stream->aspectRatio != 0 && stream->aspectRatio != 49)
	{
		float r;

		r = ((float) stream->aspectRatio + 15.0) / 64.0;
		INFO_MSG(("warning - non-square pixels; to fix do a 'pnmscale -%cscale %g'",
				  r < 1.0 ? 'x' : 'y',
				  r < 1.0 ? 1.0 / r : r));
	}

	while (ReadOK(fd, &c, 1) && c != ';')
	{
		if (resetInfo)
		{
			info.disposal = (GIFDisposalType) 0;
			info.inputFlag = 0;
			info.delayTime = 0;
			info.transparent = -1;
			resetInfo = FALSE;
		}
		cur = NULL;

		if (c == '!')
		{						/* Extension */
			if (!ReadOK(fd, &c, 1))
				_ERROR("EOF / read error on extention function code");
			if (c == 0xf9)
			{					/* graphic control */
				(void) GetDataBlock(fd, buf);
				info.disposal = (GIFDisposalType) ((buf[0] >> 2) & 0x7);
				info.inputFlag = (buf[0] >> 1) & 0x1;
				info.delayTime = MKINT(buf[1], buf[2]);
				if (BitSet(buf[0], 0x1))
					info.transparent = buf[3];

				while (GetDataBlock(fd, buf) != 0)
					;
			}
			else if (c == 0xfe || c == 0x01)
			{
				int len = 0;
				int size = 256;
				char *text;

				/*
				   **  Comment or Plain Text
				 */

				cur = NEW(GIFData);

				if (c == 0x01)
				{
					(void) GetDataBlock(fd, buf);

					cur->type = gif_text;
					cur->info = info;
					cur->x = MKINT(buf[0], buf[1]);
					cur->y = MKINT(buf[2], buf[3]);
					cur->width = MKINT(buf[4], buf[5]);
					cur->height = MKINT(buf[6], buf[7]);

					cur->data.text.cellWidth = buf[8];
					cur->data.text.cellHeight = buf[9];
					cur->data.text.fg = buf[10];
					cur->data.text.bg = buf[11];

					resetInfo = TRUE;
				}
				else
				{
					cur->type = gif_comment;
				}

				text = (char *) malloc(size);

				while ((n = GetDataBlock(fd, buf)) != 0)
				{
					if (n + len >= size)
						text = (char *) realloc(text, size += 256);
					memcpy(text + len, buf, n);
					len += n;
				}

				if (c == 0x01)
				{
					cur->data.text.len = len;
					cur->data.text.text = text;
				}
				else
				{
					cur->data.comment.len = len;
					cur->data.comment.text = text;
				}
			}
			else
			{
				/*
				   **  Unrecogonized extension, consume it.
				 */
				while (GetDataBlock(fd, buf) > 0)
					;
			}
		}
		else if (c == ',')
		{
			if (!ReadOK(fd, buf, 9))
				_ERROR("couldn't read left/top/width/height");

			cur = NEW(GIFData);

			cur->type = gif_image;
			cur->info = info;
			cur->x = MKINT(buf[0], buf[1]);
			cur->y = MKINT(buf[2], buf[3]);
			cur->width = MKINT(buf[4], buf[5]);
			cur->height = MKINT(buf[6], buf[7]);
			cur->data.image.cmapSize = 1 << ((buf[8] & 0x07) + 1);
			if (BitSet(buf[8], LOCALCOLORMAP))
			{
				if (readColorMap(fd, cur->data.image.cmapSize,
								 cur->data.image.cmapData))
					_ERROR("unable to get local colormap");
			}
			else
			{
				cur->data.image.cmapSize = 0;

			}
			cur->data.image.data = (unsigned char *) malloc(cur->width * cur->height);
			cur->data.image.interlaced = BitSet(buf[8], INTERLACE);
			readImage(fd, BitSet(buf[8], INTERLACE),
					  cur->width, cur->height, cur->data.image.data);

			resetInfo = TRUE;
		}
		else
		{
			INFO_MSG(("bogus character 0x%02x, ignoring", (int) c));
		}

		if (cur != NULL)
		{
			*end = cur;
			end = &cur->next;
			cur->next = NULL;
		}
	}

	if (c != ';')
		_ERROR("EOF / data stream");

  out:
	return stream;
}

GIFStream *GIFRead(char *file)
{
	FILE *fp = fopen(file, "rb");
	GIFStream *stream = NULL;

	if (fp != NULL)
	{
		stream = GIFReadFP(fp);
		fclose(fp);
	}
	return stream;
}

static int readColorMap(FILE * fd, int size,
						unsigned char data[GIF_MAXCOLORS][3])
{
	int i;
	unsigned char rgb[3 * GIF_MAXCOLORS];
	unsigned char *cp = rgb;

	if (!ReadOK(fd, rgb, size * 3))
		return TRUE;

	for (i = 0; i < size; i++)
	{
		data[i][0] = *cp++;
		data[i][1] = *cp++;
		data[i][2] = *cp++;
	}

	return FALSE;
}

/*
   **
 */

static int ZeroDataBlock = FALSE;

static int GetDataBlock(FILE * fd, unsigned char *buf)
{
	unsigned char count;

	if (!ReadOK(fd, &count, 1))
	{
		INFO_MSG(("error in getting DataBlock size"));
		return -1;
	}

	ZeroDataBlock = count == 0;
	if ((count != 0) && (!ReadOK(fd, buf, count)))
	{
		INFO_MSG(("error in reading DataBlock"));
		return -1;
	}
	return count;
}

/*
   **
   **
 */

/*
   **  Pulled out of nextCode
 */
static int curbit, lastbit, get_done, last_byte;
static int return_clear;

/*
   **  Out of nextLWZ
 */
static int stack[(1 << (MAX_LWZ_BITS)) * 2], *sp;
static int code_size, set_code_size;
static int max_code, max_code_size;
static int clear_code, end_code;

static void initLWZ(int input_code_size)
{
	set_code_size = input_code_size;
	code_size = set_code_size + 1;
	clear_code = 1 << set_code_size;
	end_code = clear_code + 1;
	max_code_size = 2 * clear_code;
	max_code = clear_code + 2;

	curbit = lastbit = 0;
	last_byte = 2;
	get_done = FALSE;

	return_clear = TRUE;

	sp = stack;
}

static int nextCode(FILE * fd, int code_size)
{
	static unsigned char buf[280];
	static int maskTbl[16] =
	{
		0x0000, 0x0001, 0x0003, 0x0007,
		0x000f, 0x001f, 0x003f, 0x007f,
		0x00ff, 0x01ff, 0x03ff, 0x07ff,
		0x0fff, 0x1fff, 0x3fff, 0x7fff,
	};
	int i, j, ret, end;

	if (return_clear)
	{
		return_clear = FALSE;
		return clear_code;
	}

	end = curbit + code_size;

	if (end >= lastbit)
	{
		int count;

		if (get_done)
		{
			if (curbit >= lastbit)
				_ERROR("ran off the end of my bits");
			return -1;
		}
		buf[0] = buf[last_byte - 2];
		buf[1] = buf[last_byte - 1];

		if ((count = GetDataBlock(fd, &buf[2])) == 0)
			get_done = TRUE;

		last_byte = 2 + count;
		curbit = (curbit - lastbit) + 16;
		lastbit = (2 + count) * 8;

		end = curbit + code_size;
	}

	j = end / 8;
	i = curbit / 8;

	if (i == j)
		ret = buf[i];
	else if (i + 1 == j)
		ret = buf[i] | (buf[i + 1] << 8);
	else
		ret = buf[i] | (buf[i + 1] << 8) | (buf[i + 2] << 16);

	ret = (ret >> (curbit % 8)) & maskTbl[code_size];

	curbit += code_size;

	return ret;
}

#define readLWZ(fd) ((sp > stack) ? *--sp : nextLWZ(fd))

static int nextLWZ(FILE * fd)
{
	static int table[2][(1 << MAX_LWZ_BITS)];
	static int firstcode, oldcode;
	int code, incode;
	register int i;

	while ((code = nextCode(fd, code_size)) >= 0)
	{
		if (code == clear_code)
		{
			for (i = 0; i < clear_code; ++i)
			{
				table[0][i] = 0;
				table[1][i] = i;
			}
			for (; i < (1 << MAX_LWZ_BITS); ++i)
				table[0][i] = table[1][i] = 0;
			code_size = set_code_size + 1;
			max_code_size = 2 * clear_code;
			max_code = clear_code + 2;
			sp = stack;
			do
			{
				firstcode = oldcode = nextCode(fd, code_size);
			}
			while (firstcode == clear_code);

			return firstcode;
		}
		if (code == end_code)
		{
			int count;
			unsigned char buf[260];

			if (ZeroDataBlock)
				return -2;

			while ((count = GetDataBlock(fd, buf)) > 0)
				;

//			if (count != 0)
//				INFO_MSG(("missing EOD in data stream"));
			return -2;
		}

		incode = code;

		if (code >= max_code)
		{
			*sp++ = firstcode;
			code = oldcode;
		}

		while (code >= clear_code)
		{
			*sp++ = table[1][code];
			if (code == table[0][code])
				_ERROR("circular table entry BIG ERROR");
			code = table[0][code];
		}

		*sp++ = firstcode = table[1][code];

		if ((code = max_code) < (1 << MAX_LWZ_BITS))
		{
			table[0][code] = oldcode;
			table[1][code] = firstcode;
			++max_code;
			if ((max_code >= max_code_size) &&
				(max_code_size < (1 << MAX_LWZ_BITS)))
			{
				max_code_size *= 2;
				++code_size;
			}
		}

		oldcode = incode;

		if (sp > stack)
			return *--sp;
	}
	return code;
}

static void readImage(FILE * fd, int interlace, int width, int height,
					  unsigned char *data)
{
	unsigned char *dp, c;
	int v, xpos = 0, ypos = 0;

	/*
	   **  Initialize the Compression routines
	 */
	if (!ReadOK(fd, &c, 1))
		_ERROR("EOF / read error on image data");

	initLWZ(c);

	if (fgstate.verbose)
		INFO_MSG(("reading %d by %d%s GIF image",
				  width, height, interlace ? " interlaced" : ""));

	if (interlace)
	{
		int i;
		int pass = 0, step = 8;

		for (i = 0; i < height; i++)
		{
			dp = &data[width * ypos];
			for (xpos = 0; xpos < width; xpos++)
			{
				if ((v = readLWZ(fd)) < 0)
					goto fini;

				*dp++ = v;
			}
			if ((ypos += step) >= height)
			{
				do
				{
					if (pass++ > 0)
						step /= 2;
					ypos = step / 2;
				}
				while (ypos > height);
			}
		}
	}
	else
	{
		dp = data;
		for (ypos = 0; ypos < height; ypos++)
		{
			for (xpos = 0; xpos < width; xpos++)
			{
				if ((v = readLWZ(fd)) < 0)
					goto fini;

				*dp++ = v;
			}
		}
	}

  fini:
	if (readLWZ(fd) >= 0)
		INFO_MSG(("too much input data, ignoring extra..."));

	return;
}

FGDrawBuffer::FGDrawBuffer(int ws, int hs, const char *nm, ObjectType typ,  FGPixel i, FGPixel p)
{
	_init(ws, hs, typ, i, p, nm);
}

/**
	Create a DRAWBUFFER object.
	Parameter 'ww' and 'hh' are width and height of rectangle, 'buf'
	is pointed to the image data taht will be copied into created object.
*/
FGDrawBuffer::FGDrawBuffer(int ww, int hh, FGPixel * buf)
{
	_init(ww, hh, BMP_MEM, 0);
	Resize(0, 0);
	memcpy(image, buf, areasize(w,h));
}

/**
	Creates the object with specified size, type, color layout or already existing pixel data.
*/
FGDrawBuffer::FGDrawBuffer(int ww, int hh, ObjectType t, int color, FGPixel * buf)
{
	_init(ww, hh, t, color);
	if (ww > 0 && hh > 0)
	{
		if (!buf)
		{
			Resize(0, 0);
			clear(color);
		}
		else image = buf;
	}
	else image = 0;
}

/**
	Release object's members.
*/
FGDrawBuffer::~FGDrawBuffer()
{
	if (name && name != empty_string) delete []name;
	if (type && image && type!=FGFRAMEBUFFER && type != BMP_IMAGE)
	{
		free(image);
	}
}

FGDrawBuffer::FGDrawBuffer(FGDrawBuffer& old)
{
	*this = old;
	id = ID_counter++;
	atomic = FGMutex();
	if (old.name && old.name != empty_string)
	{
		assert(name	 = new char[strlen(old.name)+1]);
		strcpy(name, old.name);
	}
	else
	{
		name = (char *)empty_string;
	}
	if (old.image && old.type!=FGFRAMEBUFFER && old.type != BMP_IMAGE)
	{
		image = (FGPixel *) malloc(areasize(old.w, old.h));
		memcpy(image, old.image, areasize(old.w, old.h));
	}
}

void FGAPI FGDrawBuffer::_init(int ww, int hh, ObjectType t, FGPixel ii,  FGPixel pp, const char *nm)
{
	lock();
	id = ID_counter++;
	if (nm)
	{
		assert(name	 = new char[strlen(nm)+1]);
		strcpy(name,nm);
	}
	else
	{
		name = (char *)empty_string;
	}
	state_saved = false;
	image = 0;
	state.alpha = DEFAULT_ALPHA_VALUE;
	state.colorkey = DEFAULT_COLOR_KEY;
	w = ww;
	h = hh;
	xoff=yoff=0;
	wwrk=ww;
	hwrk=hh;
	type = t;
	status = WHIDEN|WLOCKED;
	state._ink = ii;
	state._paper = pp;
	state._ppop = _GSET;
	set_font(FONT0816);
	unlock();
}

void FGAPI FGDrawBuffer::modify_point(FGPixel *ptr, FGPixel ink)
{
	switch (state._ppop)
	{
		default:
		case 0:
			*ptr = ink;
			break;
		case 1:					// xor
			*ptr^= ink;
			break;
		case 2:					// and
			*ptr &= ink;
			break;
		case 3:					// or
			*ptr |= ink;
			break;
		case 4:					// plus
			*ptr += ink;
			break;
		case 5:					// minus
			*ptr -= ink;
			break;
		case 6:					// not
			*ptr ^= 0xffffff;
			break;
		case _GREPLACE_GE16:			// replace
			if (*ptr>=16) *ptr = ink;
			break;
		case _GTRANSPARENT:					// alpha transparency
			if (*ptr != 0) *ptr = ink;
			break;
		case _GCOLORKEY:					// color key transparency
			if (*ptr != fgstate.color_key) *ptr = ink;
			break;
		case _GREPLACE_LE32:
			if (*ptr < 32) *ptr = ink;
			break;
		case _GREPLACE_LE48:
			if (*ptr < 48) *ptr = ink;
			break;
		case _GREPLACE_LE64:
			if (*ptr < 64) *ptr = ink;
			break;
		case _GREPLACE_LE80:
			if (*ptr < 80) *ptr = ink;
			break;
		case _GREPLACE_LESS:
			if (*ptr < ink) *ptr = ink;
			break;
	}
}

/**
	Draw a horizontal line from [x:y] with size 'c' and color 'ink'.

	Coordinate [0:0] is top-left corner of the screen. INK is color
	index in 8-bit videomodes, or absolute color in 16/24 bit modes.
	Raster Operaration is the current value of ENUM_PPOP. The default
	is _GSET (i.e. simply copy) See set_ppop() for more details.
	Coordinate are clipped if needed before drawing.
	Function return false if no drawing (i.e. object has been
	clipped-out completelly). The object XY-offset is added
	before drawing.
*/
int FGAPI FGDrawBuffer::hline(int x, int y, int c, FGPixel ink)
{
	FGPixel *ptr;
	int	j,stuff=1;

	x += xoff;
	y += yoff;

	if (clip(x,y,c,stuff) == 0)
		return 0;

	ptr	= image	+ w * y + x;

	switch(state._ppop)
	{
		default:
		case 0:	 // set
			FGmemset(ptr, ink, c);
			break;
		case 1:	 // xor
			for	(j = c;	j; j--)
			{
				*ptr++ ^= ink;
			}
			break;
		case 2:	 // and
			for	(j = c;	j; j--)
			{
				*ptr++ &= ink;
			}
			break;
		case 3:	 // or
			for	(j = c;	j; j--)
			{
				*ptr++ |= ink;
			}
			break;
		case 4:	 // plus
			for	(j = c;	j; j--)
			{
				*ptr++ += ink;
			}
			break;
		case 5:	 // minus
			for	(j = c;	j; j--)
			{
				*ptr++ -= ink;
			}
			break;
		case 6:	 // not
			for	(j = c;	j; j--)
			{
				*ptr++ ^= 0x00ffffff;
			}
			break;
		case _GREPLACE_GE16:	 // replace
			for	(j = c;	j; j--)
			{
				if (*(ptr) >= 16)
					*ptr = ink;
				ptr++;
			}
			break;
		case _GTRANSPARENT:
			for	(j = c;	j; j--)
			{
				if (*ptr != 0)
					*ptr = ink;
				ptr++;
			}
			break;
		case _GCOLORKEY:	 // replace
			for	(j = c;	j; j--)
			{
				if (*ptr != fgstate.color_key)
					*ptr = ink;
				ptr++;
			}
			break;
		case _GREPLACE_LE32:	 // replace2
			for	(j = c;	j; j--)
			{
				if (*(ptr) < 32) *ptr = ink;
				ptr++;
			}
			break;
		case _GREPLACE_LE48:	 // replace2
			for	(j = c;	j; j--)
			{
				if (*(ptr) < 48) *ptr = ink;
				ptr++;
			}
			break;
		case _GREPLACE_LE64:	 // replace2
			for	(j = c;	j; j--)
			{
				if (*(ptr) < 64) *ptr = ink;
				ptr++;
			}
			break;
		case _GREPLACE_LE80:	 // replace2
			for	(j = c;	j; j--)
			{
				if (*(ptr) < 80) *ptr = ink;
				ptr++;
			}
			break;
		case _GREPLACE_LESS:	 // replace4
			for	(j = c;	j; j--)
			{
				if (*(ptr) < ink) *ptr = ink;
				ptr++;
			}
			break;
	}
	return 1;
}

/**
	Draw a vertical line from [x:y] with size 'c' and color 'ink'.

	Coordinate [0:0] is top-left corner of the screen. INK is color
	index in 8-bit videomodes, or absolute color in 16/24 bit modes.
	Raster Operaration is the current value of ENUM_PPOP. The default
	is _GSET (i.e. simply copy) See set_ppop() for more details.
	Coordinate are clipped if needed before drawing.
	Function return false if no drawing (i.e. object has been
	clipped-out completelly). The object XY-offset is added
	before drawing.
*/
int FGAPI FGDrawBuffer::vline(int x, int y, int c, FGPixel ink)
{
	FGPixel *ptr;
	int	i, stuff=1;

	x += xoff;
	y += yoff;

	if (clip(x,y,stuff,c) == 0)
		return 0;

	ptr	= image	+ w * y + x;

	for	(i = c;	i; i--)
	{
		modify_point(ptr, ink);
		ptr	+= w;
	}
	return 1;
}

/**
	Draw a pixel at position [x:y] with 'color'.

	Coordinate [0:0] is top-left corner of the screen. INK is color
	index in 8-bit videomodes, or absolute color in 16/24 bit modes.
	Raster Operaration is the current value of ENUM_PPOP. The default
	is _GSET (i.e. simply copy) See set_ppop() for more details.
	Coordinates are clipped if needed before drawing.
	Function return false if no drawing (i.e. object has been
	clipped-out completelly). The object XY-offset is added
	before drawing.
*/
int FGAPI FGDrawBuffer::putpixel(int x, int y, unsigned color)
{
	x += xoff;
	y += yoff;

	if (y <	yoff || x < xoff || y >= (hwrk+yoff) || x >= (wwrk+xoff))
		return 0;
	modify_point(image + w * y + x, color == UNDEFINED_COLOR ? state._ink : color);
	return 1;
}

// internal for circle draw to RAM
void   FGAPI FGDrawBuffer::_symetry(int xs, int ys, int x,	int	y, FGPixel color)
{
	putpixel(xs + x, ys + y, color);
	putpixel(xs - x, ys + y, color);
	putpixel(xs + x, ys - y, color);
	putpixel(xs - x, ys - y, color);
	putpixel(xs + y, ys + x, color);
	putpixel(xs - y, ys + x, color);
	putpixel(xs + y, ys - x, color);
	putpixel(xs - y, ys - x, color);
}

// internal for circle fill to RAM
void	FGAPI FGAPI FGDrawBuffer::_symetry2(int xs, int ys, int x, int y, FGPixel color)
{
	hline(xs-x,ys+y,x+x+1, color);
	hline(xs-x,ys-y,x+x+1, color);
	hline(xs-y,ys+x,y+y+1, color);
	hline(xs-y,ys-x,y+y+1, color);
}

/**
	Draw a circle shape at position [x:y] with radius and with 'color'.

	Coordinate [0:0] is top-left corner of the screen. INK is color
	index in 8-bit videomodes, or absolute color in 16/24 bit modes.
	Raster Operaration is the current value of ENUM_PPOP. The default
	is _GSET (i.e. simply copy) See set_ppop() for more details.
	Coordinate are clipped if needed before drawing.
	Function return false if no drawing (i.e. object has been
	clipped-out completelly). The object XY-offset is added
	before drawing.
*/
int FGAPI FGDrawBuffer::circle(int	xs,	int	ys,	int	r, unsigned color)
{
	int	x =	0;
	int	y =	r;
	int	p =	3 -	2 *	r;

	if ((ys	+ r) < 0 || (xs +	r) < 0
		||	(ys	- r) >=	h || (xs -	r) >= w
		||	r <	1) return 0;

	if (color == UNDEFINED_COLOR)
	    color = state._ink;

	while (x < y)
	{
		_symetry(xs, ys, x,	y, color);
		if (p <	0)
			p += 4 * (x++) + 6;
		else
			p += 4 * ((x++)	- (y--)) + 10;
	}
	if (x == y)
		_symetry(xs, ys, x,	y, color);
	return 1;
}

/**
	Fill a circle shape at position [x:y] with radius and with 'color'.

	Coordinate [0:0] is top-left corner of the screen. INK is color
	index in 8-bit videomodes, or absolute color in 16/24 bit modes.
	Raster Operaration is the current value of ENUM_PPOP. The default
	is _GSET (i.e. simply copy) See set_ppop() for more details.
	Coordinate are clipped if needed before drawing.
	Function return false if no drawing (i.e. object has been
	clipped-out completelly). The object XY-offset is added
	before drawing.
*/
int FGAPI FGDrawBuffer::fcircle(int	xs,	int	ys,	int	r, unsigned color)
{
	int	x =	0;
	int	y =	r;
	int	p =	3 -	2 *	r;

	if ((ys	+ r) < 0 || (xs +	r) < 0
		||	(ys	- r) >=	h || (xs -	r) >= w
		||	r <	1) return 0;

	if (color == UNDEFINED_COLOR)
	    color = state._ink;

	while (x < y)
	{
		_symetry2(xs, ys, x, y, color);
		if (p <	0)
			p += 4 * (x++) + 6;
		else
			p += 4 * ((x++)	- (y--)) + 10;
	}
	if (x == y)
		_symetry2(xs, ys, x, y, color);
	return 1;
}

/**
	Draw an ellipse shape at position [x:y] with radius 'rx' & 'ry'and with 'color'.

	Coordinate [0:0] is top-left corner of the screen. INK is color
	index in 8-bit videomodes, or absolute color in 16/24 bit modes.
	Raster Operaration is the current value of ENUM_PPOP. The default
	is _GSET (i.e. simply copy) See set_ppop() for more details.
	Coordinate are clipped if needed before drawing.
	Function return false if no drawing (i.e. object has been
	clipped-out completelly). The object XY-offset is added
	before drawing.
*/
int FGAPI FGDrawBuffer::ellipse(int	x,	int	y,	int	rx, int ry, unsigned color)
{
	if ((y	+ ry) < 0 || (x + rx) < 0 || (y - ry) >= h || (x	- rx) >= w || rx<1 || ry<1)
		return 0;
	int ix, iy;
	int h, i, j, k;
	int oh, oi, oj, ok;

	if (color == UNDEFINED_COLOR)
	    color = state._ink;

	if (rx < 1)
		rx = 1;

	if (ry < 1)
		ry = 1;

	h = i = j = k = 0xFFFF;

	if (rx > ry)
	{
		ix = 0;
		iy = rx * 64;

		do
		{
			oh = h;
			oi = i;
			oj = j;
			ok = k;

			h = (ix + 32) >> 6;
			i = (iy + 32) >> 6;
			j = (h * ry) / rx;
			k = (i * ry) / rx;

			if (((h != oh) || (k != ok)) && (h < oi))
			{
				putpixel(x + h, y + k, color);
				if (h)
					putpixel(x - h, y + k, color);
				if (k)
				{
					putpixel(x + h, y - k, color);
					if (h)
						putpixel(x - h, y - k, color);
				}
			}

			if (((i != oi) || (j != oj)) && (h < i))
			{
				putpixel(x + i, y + j, color);
				if (i)
					putpixel(x - i, y + j, color);
				if (j)
				{
					putpixel(x + i, y - j, color);
					if (i)
						putpixel(x - i, y - j, color);
				}
			}

			ix = ix + iy / rx;
			iy = iy - ix / rx;
		}
		while (i > h);
	}
	else
	{
		ix = 0;
		iy = ry * 64;

		do
		{
			oh = h;
			oi = i;
			oj = j;
			ok = k;

			h = (ix + 32) >> 6;
			i = (iy + 32) >> 6;
			j = (h * rx) / ry;
			k = (i * rx) / ry;

			if (((j != oj) || (i != oi)) && (h < i))
			{
				putpixel(x + j, y + i, color);
				if (j)
					putpixel(x - j, y + i, color);
				if (i)
				{
					putpixel(x + j, y - i, color);
					if (j)
						putpixel(x - j, y - i, color);
				}
			}

			if (((k != ok) || (h != oh)) && (h < oi))
			{
				putpixel(x + k, y + h, color);
				if (k)
					putpixel(x - k, y + h, color);
				if (h)
				{
					putpixel(x + k, y - h, color);
					if (k)
						putpixel(x - k, y - h, color);
				}
			}

			ix = ix + iy / ry;
			iy = iy - ix / ry;

		}
		while (i > h);
	}
	return 1;
}

/**
	Fill an ellipse shape at position [x:y] with radius 'rx' & 'ry'and with 'color'.

	Coordinate [0:0] is top-left corner of the screen. INK is color
	index in 8-bit videomodes, or absolute color in 16/24 bit modes.
	Raster Operaration is the current value of ENUM_PPOP. The default
	is _GSET (i.e. simply copy) See set_ppop() for more details.
	Coordinate are clipped if needed before drawing.
	Function return false if no drawing (i.e. object has been
	clipped-out completelly). The object XY-offset is added
	before drawing.
*/
int FGAPI FGDrawBuffer::fellipse(int	x,	int	y,	int	rx, int ry, unsigned color)
{
	if ((y	+ ry) < 0 || (x + rx) < 0 || (y - ry) >= h || (x	- rx) >= w || rx<1 || ry<1)
		return 0;
	int ix, iy;
	int a, b, c, d;
	int da, db, dc, dd;
	int na, nb, nc, nd;

	if (color == UNDEFINED_COLOR)
	    color = state._ink;

	if (rx > ry)
	{
		dc = -1;
		dd = 0xFFFF;
		ix = 0;
		iy = rx * 64;
		na = 0;
		nb = (iy + 32) >> 6;
		nc = 0;
		nd = (nb * ry) / rx;

		do
		{
			a = na;
			b = nb;
			c = nc;
			d = nd;

			ix = ix + (iy / rx);
			iy = iy - (ix / rx);
			na = (ix + 32) >> 6;
			nb = (iy + 32) >> 6;
			nc = (na * ry) / rx;
			nd = (nb * ry) / rx;

			if ((c > dc) && (c < dd))
			{
				hline(x - b, y + c, b * 2, color);
				if (c)
					hline(x - b, y - c, b * 2, color);
				dc = c;
			}

			if ((d < dd) && (d > dc))
			{
				hline(x - a, y + d, a * 2, color);
				hline(x - a, y - d, a * 2, color);
				dd = d;
			}

		}
		while (b > a);
	}
	else
	{
		da = -1;
		db = 0xFFFF;
		ix = 0;
		iy = ry * 64;
		na = 0;
		nb = (iy + 32) >> 6;
		nc = 0;
		nd = (nb * rx) / ry;

		do
		{
			a = na;
			b = nb;
			c = nc;
			d = nd;

			ix = ix + (iy / ry);
			iy = iy - (ix / ry);
			na = (ix + 32) >> 6;
			nb = (iy + 32) >> 6;
			nc = (na * rx) / ry;
			nd = (nb * rx) / ry;

			if ((a > da) && (a < db))
			{
				hline(x - d, y + a, d * 2, color);
				if (a)
					hline(x - d, y - a, d * 2, color);
				da = a;
			}

			if ((b < db) && (b > da))
			{
				hline(x - c, y + b, c * 2, color);
				hline(x - c, y - b, c * 2, color);
				db = b;
			}

		}
		while (b > a);
	}
	return 1;
}

/**
Draws a circular arc with centre x, y and radius r, in an anticlockwise direction
starting from the angle a and ending when it reaches b. These values are specified
in radian format, with M_PI*2 equal to a full circle.
Zero is to the right of the centre point, and larger values rotate anticlockwise from there.
@param x x coordinate of the arc centre
@param y y coordinate of the arc centre
@param ang1 starting angle
@param ang2 ending angle
@param r radius in pixels
@param color color
*/
int FGAPI FGDrawBuffer::arc(int x, int y, double ang1, double ang2, int r, unsigned color)
{
	int px, py;
	int ex, ey;
	int px1, px2, px3;
	int py1, py2, py3;
	int d1, d2, d3;
	int ax, ay;
	int q, qe;
	double tg_cur, tg_end;
	int done = 0;
	double rr1, rr2, rr3;
	int rr = (r * r);
	int carry_qv=0;

	rr1 = r;
	rr2 = x;
	rr3 = y;

	/* evaluate the start point and the end point */
	px = (int) (rr2 + rr1 * cos(ang1));
	py = (int) (rr3 - rr1 * sin(ang1));
	ex = (int) (rr2 + rr1 * cos(ang2));
	ey = (int) (rr3 - rr1 * sin(ang2));

	if (color == UNDEFINED_COLOR)
	    color = state._ink;

	/* start quadrant */
	if (px >= x)
	{
		if (py <= y)
			q = 1;				/* quadrant 1 */
		else
			q = 4;				/* quadrant 4 */
	}
	else
	{
		if (py < y)
			q = 2;				/* quadrant 2 */
		else
			q = 3;				/* quadrant 3 */
	}

	/* end quadrant */
	if (ex >= x)
	{
		if (ey <= y)
			qe = 1;				/* quadrant 1 */
		else
			qe = 4;				/* quadrant 4 */
	}
	else
	{
		if (ey < y)
			qe = 2;				/* quadrant 2 */
		else
			qe = 3;				/* quadrant 3 */
	}

#define loc_tg(_y, _x)  (_x-x) ? (double)(_y-y)/(_x-x) : (double)(_y-y)

	tg_end = loc_tg(ey, ex);
    if (q==qe)  // su v rovnakom kvadrante
    { if (ang1 > ang2) // kreslenie cez 4 kv
      {
         carry_qv=1;
         qe+=4;
    } }

	while (!done)
	{
		putpixel(px, py, color);

		/* from here, we have only 3 possible direction of movement, eg.
		 * for the first quadrant:
		 *
		 *    OOOOOOOOO
		 *    OOOOOOOOO
		 *    OOOOOO21O
		 *    OOOOOO3*O
		 */

		/* evaluate the 3 possible points */
		switch (q)
		{

			case 1:
				px1 = px;
				py1 = py - 1;
				px2 = px - 1;
				py2 = py - 1;
				px3 = px - 1;
				py3 = py;

				/* 2nd quadrant check */
				if (px != x)
				{
					break;
				}
				else
				{
					/* we were in the end quadrant, changing is illegal. Exit. */
					if (qe == q)
						done = 1;
					q++;
                    if (carry_qv) {carry_qv=0;qe-=4;}
				}
				/* fall through */

			case 2:
				px1 = px - 1;
				py1 = py;
				px2 = px - 1;
				py2 = py + 1;
				px3 = px;
				py3 = py + 1;

				/* 3rd quadrant check */
				if (py != y)
				{
					break;
				}
				else
				{
					/* we were in the end quadrant, changing is illegal. Exit. */
					if (qe == q)
						done = 1;
					q++;
                    if (carry_qv) {carry_qv=0;qe-=4;}
				}
				/* fall through */

			case 3:
				px1 = px;
				py1 = py + 1;
				px2 = px + 1;
				py2 = py + 1;
				px3 = px + 1;
				py3 = py;

				/* 4th quadrant check */
				if (px != x)
				{
					break;
				}
				else
				{
					/* we were in the end quadrant, changing is illegal. Exit. */
					if (qe == q)
						done = 1;
					q++;
                    if (carry_qv) {carry_qv=0;qe-=4;}
				}
				/* fall through */

			case 4:
				px1 = px + 1;
				py1 = py;
				px2 = px + 1;
				py2 = py - 1;
				px3 = px;
				py3 = py - 1;

				/* 1st quadrant check */
				if (py == y)
				{
					/* we were in the end quadrant, changing is illegal. Exit. */
					if (qe == q)
						done = 1;

					q = 1;
                    if (carry_qv) {carry_qv=0;qe-=4;}
					px1 = px;
					py1 = py - 1;
					px2 = px - 1;
					py2 = py - 1;
					px3 = px - 1;
					py3 = py;
				}
				break;

			default:
				return 1;
		}

		/* now, we must decide which of the 3 points is the right point.
		 * We evaluate the distance from center and, then, choose the
		 * nearest point.
		 */
		ax = x - px1;
		ay = y - py1;
		rr1 = ax * ax + ay * ay;

		ax = x - px2;
		ay = y - py2;
		rr2 = ax * ax + ay * ay;

		ax = x - px3;
		ay = y - py3;
		rr3 = ax * ax + ay * ay;

		/* difference from the main radius */
		if (rr1 > rr)
			d1 = (int) (rr1 - rr);
		else
			d1 = (int) (rr - rr1);
		if (rr2 > rr)
			d2 = (int) (rr2 - rr);
		else
			d2 = (int) (rr - rr2);
		if (rr3 > rr)
			d3 = (int) (rr3 - rr);
		else
			d3 = (int) (rr - rr3);

		/* what is the minimum? */
		if (d1 <= d2)
		{
			px = px1;
			py = py1;
		}
		else if (d2 <= d3)
		{
			px = px2;
			py = py2;
		}
		else
		{
			px = px3;
			py = py3;
		}

		/* are we in the final quadrant? */
		if (qe == q)
		{
			tg_cur = loc_tg(py, px);

			/* is the arc finished? */
			switch (q)
			{

				case 1:
					/* end quadrant = 1? */
					if (tg_cur <= tg_end)
						done = 1;
					break;

				case 2:
					/* end quadrant = 2? */
					if (tg_cur <= tg_end)
						done = 1;
					break;

				case 3:
					/* end quadrant = 3? */
					if (tg_cur <= tg_end)
						done = 1;
					break;

				case 4:
					/* end quadrant = 4? */
					if (tg_cur <= tg_end)
						done = 1;
					break;
			}
		}
	}

	/* draw the last evaluated point */
	putpixel(px, py, color);
	return 1;
}


static void CartesFromPolar(float azimut, float range, int& coord_x, int& coord_y)
{
		double _x,_y, pi=M_PI;

		double dAz=azimut*pi/180.0;        /* azimut v radianoch     */

		_x=(double)range*sin(dAz);
		_y=(double)range*cos(dAz);

		if(dAz > (pi/2.0) &&  dAz < (1.5*pi))
		{
				if(_y>0)  coord_y = int(-_y+0.5);
				else coord_y = int(_y+0.5);
		}
		else
		{
				if(_y>0)  coord_y = int(_y+0.5);
				else coord_y = int(-_y+0.5);
		}

		if(dAz > pi && dAz <= (2.0*pi))
		{
				if(_x>0)  coord_x = int(-_x+0.5);
				else coord_x = int(_x+0.5);
		}
		else
		{
				if(_x>0)  coord_x = int(_x+0.5);
				else coord_x = int(-_x+0.5);
		}
}


/**
Draws a circular arc with centre x, y and radius r, in an clockwise direction
starting from the angle a and ending when it reaches b. These values are specified
in degrees, with 360 equal to a full circle. The zero angle is at the 12 o'clock.
Zero is to the right of the centre point, and larger values rotate clockwise from there.
@param x x coordinate of the arc centre
@param y y coordinate of the arc centre
@param ang1 starting angle
@param ang2 ending angle
@param r radius in pixels
@param color color
*/
int FGAPI FGDrawBuffer::arc2(int x, int y, int ang1, int ang2, int r, unsigned color)
{
#define NORMALIZE_ANGLE(a) a = a%360
	int _x,_y;

	NORMALIZE_ANGLE(ang1);
	NORMALIZE_ANGLE(ang2);

	CartesFromPolar(ang1, r, _x, _y);

	moveto(x+_x, y+ -_y);
	do
	{
		ang1++;
		NORMALIZE_ANGLE(ang1);
		CartesFromPolar(ang1, r, _x, _y);
		lineto(x+_x, y+ -_y, color);
	} while (ang1 != ang2);

	return true;
}

#define swap(a,b)           {a^=b; b^=a; a^=b;}
#define	MIN(a,b)	( (a) < (b) ? (a) : (b) )
#define MAX(a,b)	( (a) > (b) ? (a) : (b) )

void FGDrawBuffer::__line(int a1, int b1, int a2, int b2, unsigned color)
{
	int xend, yend, dx, dy, c1, c2, step;
	register int p, x, y;

	dx = abs(a2 - a1);
	dy = abs(b2 - b1);

	if (dx > 32000)
		return;					// bulgar const

	if (dy > 32000)
		return;

	if (color == UNDEFINED_COLOR)
		color = state._ink;

	if (dx > dy)
	{							/* slope < 1 => step in x direction */
		x = MIN(a1, a2);
		xend = MAX(a1, a2);
		if (x == a1)
		{
			y = b1;
			step = ((b2 - y) < 0 ? -1 : 1);
		}
		else
		{
			y = b2;
			step = ((b1 - y) < 0 ? -1 : 1);
		}

		p = 2 * dy - dx;
		c1 = 2 * dy;
		c2 = 2 * (dy - dx);

		putpixel(x, y, color);
		while (x < xend)
		{
			x++;
			if (p < 0)
			{
				p += c1;
			}
			else
			{
				y += step;
				p += c2;
			}
			putpixel(x, y, color);
		}
	}
	else
	{							/* slope > 1 => step in y direction */
		y = MIN(b1, b2);
		yend = MAX(b1, b2);
		if (y == b1)
		{
			x = a1;
			step = ((a2 - x) < 0 ? -1 : 1);
		}
		else
		{
			x = a2;
			step = ((a1 - x) < 0 ? -1 : 1);
		}

		p = 2 * dx - dy;
		c1 = 2 * dx;
		c2 = 2 * (dx - dy);

		putpixel(x, y, color);

		while (y < yend)
		{
			y++;
			if (p < 0)
			{
				p += c1;
			}
			else
			{
				x += step;
				p += c2;
			}
			putpixel(x, y, color);
		}
	}
}

/**
	Draw a line from position [x:y] to the position [a:b] with 'color'.

	Coordinate [0:0] is top-left corner of the screen. INK is color
	index in 8-bit videomodes, or absolute color in 16/24 bit modes.
	Raster Operaration is the current value of ENUM_PPOP. The default
	is _GSET (i.e. simply copy) See set_ppop() for more details.
	Coordinate are clipped if needed before drawing.
	Function return false if no drawing (i.e. object has been
	clipped-out completelly). The object XY-offset is added
	before drawing.
*/
int FGAPI FGDrawBuffer::line(int x, int y, int a, int b, unsigned color)
{
	int cl=1;
	oldx = a;
	oldy = b;
	cl = ClipLine(x,y,a,b,0,0,w,h);
	if (cl)	__line(x,y,a,b,color); // null if not clipped
	return cl;
}

static FGMutex TextMutex;

/**
	Draw a text with the current object font, background 'p' and foreground 'i' from the position [x:y].

	Coordinate [0:0] is top-left corner of the screen. INK is color
	index in 8-bit videomodes, or absolute color in 16/24 bit modes.
	Raster Operaration is the current value of ENUM_PPOP. The default
	is _GSET (i.e. simply copy) See set_ppop() for more details.
	Coordinate are clipped if needed before drawing.
	Function return false if no drawing (i.e. object has been
	clipped-out completelly). The object XY-offset is added
	before drawing.
*/
int FGAPI FGDrawBuffer::text(int x, int y, const char *txt, unsigned i, unsigned p)
{
	FGPixel *dst, *dst2;
	int alloc=0;

	TextMutex.Lock();

	x+=xoff;
	y+=yoff;

	int	width =	textwidth((char *)txt);
	int _x=x, _y=y,ww=GetFontW(), bufw=width+3;
	int height = GetFontH();

	if (clip(x,y,width,height)==0 || width<2)
	{
		TextMutex.Unlock();
		return 0;
	}

	if (i == UNDEFINED_COLOR)
	    i = state._ink;
	if (p == UNDEFINED_COLOR)
		p = state._paper;

	_x = x-_x;
	_y = y-_y;

	unsigned  bytes = (bufw)*height;

	if (bytes < FGTEXTBUFSZ)
		dst	= fgstate.fgimagebuffer;
	else
	{
		alloc = 1;
		dst = (FGPixel *) malloc(bytes*sizeof(FGPixel));
	}

	dst2 = dst;

	if (state._ppop == _GCOLORKEY)
	{
		p = fgstate.color_key;
	}
//	set_mmx();
	while (*txt)
	{
		variable_record *fontvar = FGFontManager::IsVariableFont(state._font);
		FGPixel *font = FGFontManager::GetFontImage(state._font);

		if (fontvar)
		{
			ww = fontvar->w[*(unsigned char *)txt];
		}
		CharOutClip(dst,
			fontvar ? fontvar->off[*(unsigned char *)txt++] + font:
				*(unsigned char *)txt++ * (GetFontW()*GetFontH()) + font,
			bufw,
			(ww+3)&~3,
			GetFontH(), i, p);
		dst	+= ww;
	}
//	reset_mmx();
	RamToRam(_x, _y, bufw, GetFontH(), x, y, w, h, width, height, dst2, image, BITBLIT_WITH_OPERATOR, state._ppop);
	if (alloc==1)
		free(dst2);

	TextMutex.Unlock();


	return 1;
}

/**
	Draw a patterned line from position [a1:b1] to the position [a2:b2].

	Coordinate [0:0] is top-left corner of the screen. INK is color
	index in 8-bit videomodes, or absolute color in 16/24 bit modes.
	Raster Operaration is the current value of ENUM_PPOP. The default
	is _GSET (i.e. simply copy) See set_ppop() for more details.
	Coordinate are clipped if needed before drawing.
	Function return false if no drawing (i.e. object has been
	clipped-out completelly). The object XY-offset is added
	before drawing.
*/
int FGAPI FGDrawBuffer::pline(int a1, int b1, int a2, int b2, FGPattern *pat)
{
	int xend, yend, dx, dy, c1, c2, step;
	register int p, x, y;
	int cl;
	FGPixel iscolor;

	oldx = a2;
	oldy = a2;

	cl = ClipLine(a1,b1,a2,b2,0,0,w,h);

	if (cl==0) return 0;

	dx = abs(a2 - a1);
	dy = abs(b2 - b1);

	if (dx > 32000)
		return 0;			// bulgar const

	if (dy > 32000)
		return 0;

	if (pat->reset)
		pat->pos = 0;

	if (dx > dy)
	{							/* slope < 1 => step in x direction */
		x = MIN(a1, a2);
		xend = MAX(a1, a2);
		if (x == a1)
		{
			y = b1;
			step = ((b2 - y) < 0 ? -1 : 1);
		}
		else
		{
			y = b2;
			step = ((b1 - y) < 0 ? -1 : 1);
		}

		p = 2 * dy - dx;
		c1 = 2 * dy;
		c2 = 2 * (dy - dx);

		iscolor=pat->next();
		if (iscolor)
			putpixel(x, y, iscolor);
		while (x < xend)
		{
			x++;
			if (p < 0)
			{
				p += c1;
			}
			else
			{
				y += step;
				p += c2;
			}
			iscolor=pat->next();
			if (iscolor)
				putpixel(x, y, iscolor);
		}
	}
	else
	{							/* slope > 1 => step in y direction */
		y = MIN(b1, b2);
		yend = MAX(b1, b2);
		if (y == b1)
		{
			x = a1;
			step = ((a2 - x) < 0 ? -1 : 1);
		}
		else
		{
			x = a2;
			step = ((a1 - x) < 0 ? -1 : 1);
		}

		p = 2 * dx - dy;
		c1 = 2 * dx;
		c2 = 2 * (dx - dy);

		iscolor=pat->next();
		if (iscolor)
			putpixel(x, y, iscolor);

		while (y < yend)
		{
			y++;
			if (p < 0)
			{
				p += c1;
			}
			else
			{
				x += step;
				p += c2;
			}
			iscolor=pat->next();
			if (iscolor)
				putpixel(x, y, iscolor);
		}
	}
	return cl;
}

/**
	Draw a rectangle from position [x:y] with size [w:h] with 'color'.

	Coordinate [0:0] is top-left corner of the screen. INK is color
	index in 8-bit videomodes, or absolute color in 16/24 bit modes.
	Raster Operaration is the current value of ENUM_PPOP. The default
	is _GSET (i.e. simply copy) See set_ppop() for more details.
	Coordinate are clipped if needed before drawing.
	Function return false if no drawing (i.e. object has been
	clipped-out completelly). The object XY-offset is added
	before drawing.
*/
int FGAPI FGDrawBuffer::rect(int x, int y, int w, int h, unsigned color)
{
	if (color == UNDEFINED_COLOR)
		color = state._ink;

	hline(x, y,	w,color);
	hline(x, y + h - 1,	w,color);
	vline(x, y,	h,color);
	vline(x	+ w	- 1, y,	h,color);
	return 1;
}

/**
	Fill a rectangle from position [x:y] with size [w:h] with 'color'.

	Coordinate [0:0] is top-left corner of the screen. INK is color
	index in 8-bit videomodes, or absolute color in 16/24 bit modes.
	Raster Operaration is the current value of ENUM_PPOP. The default
	is _GSET (i.e. simply copy) See set_ppop() for more details.
	Coordinate are clipped if needed before drawing.
	Function return false if no drawing (i.e. object has been
	clipped-out completelly). The object XY-offset is added
	before drawing.
*/
int FGAPI FGDrawBuffer::box(int x, int y, int ww, int hh, unsigned color)
{
	x += xoff;
	y += yoff;

	if (clip(x, y, ww, hh) == 0) return 0;

	if (color == UNDEFINED_COLOR)
	    color = state._ink;
	L1Box(CalcAddr(x,y), ww, hh, color, w, state._ppop);
	return 1;
}

/**
	Fill convex polygon with 'n' vertices. The each vertice is
	a pair of two integers X and Y. Parameter 'pt' is an array
	of these pairs. FUnction return the Rectangle of the polygon
	boundary.

	Coordinate [0:0] is top-left corner of the screen. INK is color
	index in 8-bit videomodes, or absolute color in 16/24 bit modes.
	Raster Operaration is the current value of ENUM_PPOP. The default
	is _GSET (i.e. simply copy) See set_ppop() for more details.
	Coordinate are clipped if needed before drawing.
	Function return false if no drawing (i.e. object has been
	clipped-out completelly). The object XY-offset is added
	before drawing.
*/
FGRect FGAPI FGDrawBuffer::fpolygon(const FGPointArray& vertices, unsigned color)
{
	FGRect rect;
	edge *edges=edge_array, *ep;
	scan *scans=scan_array, *sp, *points, *segments;
	int	xmin, xmax,	ymin, ymax;
	int	ypos, nedges, mem = 0;

	int n = vertices.vertices;
	FGPoint* pt = vertices.array;
	if (color == UNDEFINED_COLOR)
	    color = state._ink;

	if ((n > 1)	&&
	  (pt[0].x == pt[n - 1].x) && (pt[0].y == pt[n - 1].y))
	{
		n--;
	}
	if (n <	1)
	{
		return rect;
	}
	if (n>128)
	{
		edges =	(edge *) malloc(sizeof(edge) * (n +	2));
		scans =	(scan *) malloc(sizeof(scan) * (n +	8));
		mem = 1;
	}
	if (edges && scans)
	{
		int	prevx =	xmin = xmax	= pt[0].x;
		int	prevy =	ymin = ymax	= pt[0].y;

		nedges = 0;
		ep = edges;
		while (--n >= 0)
		{
			if (pt[n].y >=	prevy)
			{
				ep->e.x	= prevx;
				ep->e.y	= prevy;
				ep->e.xlast	= prevx	= pt[n].x;
				ep->e.ylast	= prevy	= pt[n].y;
			}
			else
			{
				ep->e.xlast	= prevx;
				ep->e.ylast	= prevy;
				ep->e.x	= prevx	= pt[n].x;
				ep->e.y	= prevy	= pt[n].y;
			}
			if ((ep->e.y > (hwrk+yoff)) || (ep->e.ylast <	(0)))
				continue;
			{
				if (ep->e.y	< 0)
				{
					ep->e.x	+= (((int) (((long)	(((int)	((ep->e.xlast -	ep->e.x)) << 1)) * (long) ((0	- ep->e.y))) / (long) ((ep->e.ylast	- ep->e.y))) + (((int) ((ep->e.xlast - ep->e.x)) ^ (int) ((0 - ep->e.y)) ^ (int) ((ep->e.ylast - ep->e.y))) >> ((sizeof(int) * 8)	- 1)) +	1) >> 1);
					ep->e.y	= 0;;
				}
			};
			if (ymin > ep->e.y)
				ymin = ep->e.y;
			if (ymax < ep->e.ylast)
				ymax = ep->e.ylast;
			if (xmin > ep->e.x)
				xmin = ep->e.x;
			if (xmax < ep->e.xlast)
				xmax = ep->e.xlast;
			{
				(&ep->e)->dy = (&ep->e)->ylast - (&ep->e)->y;
				(&ep->e)->dx = (&ep->e)->xlast - (&ep->e)->x;
				if ((&ep->e)->dx < 0)
				{
					(&ep->e)->xstep	= (-1);
					(&ep->e)->dx = (-(&ep->e)->dx);
				}
				else
				{
					(&ep->e)->xstep	= 1;
				}
				if ((&ep->e)->dx > (&ep->e)->dy)
				{
					(&ep->e)->xmajor = 1;
					(&ep->e)->error	= (&ep->e)->dx >> 1;
				}
				else
				{
					(&ep->e)->xmajor = 0;
					(&ep->e)->error	= ((&ep->e)->dy	- ((1 -	(&ep->e)->xstep) >>	1))	>> 1;
				}
			};
			ep->status = inactive;
			nedges++;
			ep++;
		}
		rect.x = xmin;
		rect.y = ymin;
		rect.w = xmax-xmin;
		rect.h = ymax-ymin;
		if ((nedges	> 0) &&	(xmin <= wwrk+xoff)	&& (xmax >=	0))
		{
			if (xmin < 0)
				xmin = 0;
			if (ymin < 0)
				ymin = 0;
			if (xmax > (wwrk+xoff))
				xmax = (wwrk+xoff);
			if (ymax > (hwrk+yoff))
				ymax = (hwrk+yoff);

			for	(ypos =	ymin; ypos <= ymax;	ypos++)
			{
				sp = scans;
				points = 0;
				segments = 0;
				for	(n = nedges, ep	= edges; --n >=	0; ep++)
				{
					switch (ep->status)
					{
						case inactive:
							if (ep->e.y	!= ypos)
								break;
							if (ep->e.dy ==	0)
							{
								ep->status = passed;
								xmin = ep->e.x;
								xmax = ep->e.xlast;
								{
									if ((int) (xmin) > (int) (xmax))
									{
										int	_swap_tmpval_ =	(xmin);

										(xmin) = (xmax);
										(xmax) = _swap_tmpval_;
									}
								};
								{
									scan *prev = 0;
									scan *work = segments;
									int	overlap	= 0;

									while (work	!= 0)
									{
										if ((work->x1 <= xmax) && (xmin	<= work->x2))
										{
											overlap	= 1;
											if (xmin < work->x1)
												work->x1 = xmin;
											if (xmax > work->x2)
											{
												prev = work;
												while ((work = work->next) != 0)
												{
													if (work->x1 > xmax)
														break;
													if (work->x2 > xmax)
														xmax = work->x2;
												}
												prev->x2 = xmax;
												prev->next = work;
											}
											break;
										}
										if (work->x1 > xmax)
											break;
										prev = work;
										work = work->next;
									}
									if (!overlap)
									{
										sp->x1 = xmin;
										sp->x2 = xmax;
										sp->next = work;
										if (prev)
											prev->next = sp;
										else
											segments = sp;
									}
								};
								sp++;
								break;
							}
							ep->status = active;
						case active:
							xmin = xmax	= ep->e.x;
							if (ep->e.ylast	== ypos)
							{
								ep->status = passed;
								xmax = ep->e.xlast;
								{
									if ((int) (xmin) > (int) (xmax))
									{
										int	_swap_tmpval_ =	(xmin);

										(xmin) = (xmax);
										(xmax) = _swap_tmpval_;
									}
								};
								{
									scan *prev = 0;
									scan *work = points;

									while (work	!= 0)
									{
										if (work->x1 > xmin)
											break;
										prev = work;
										work = work->next;
									}
									sp->x1 = xmin;
									sp->x2 = xmax;
									sp->next = work;
									if (prev)
										prev->next = sp;
									else
										points = sp;
								};
								sp++;
							}
							else if	(ep->e.xmajor)
							{
								for	(;;)
								{
									(&ep->e)->x	+= (&ep->e)->xstep;
									if (((&ep->e)->error -=	(&ep->e)->dy) <	0)
									{
										(&ep->e)->error	+= (&ep->e)->dx;
										break;
									}
								};
								xmax = ep->e.x - ep->e.xstep;
								{
									if ((int) (xmin) > (int) (xmax))
									{
										int	_swap_tmpval_ =	(xmin);

										(xmin) = (xmax);
										(xmax) = _swap_tmpval_;
									}
								};
							}
							else
							{
								{
									if (((&ep->e)->error -=	(&ep->e)->dx) <	0)
									{
										(&ep->e)->x	+= (&ep->e)->xstep;
										(&ep->e)->error	+= (&ep->e)->dy;
									}
								};
							}
							{
								scan *prev = 0;
								scan *work = points;

								while (work	!= 0)
								{
									if (work->x1 > xmin)
										break;
									prev = work;
									work = work->next;
								}
								sp->x1 = xmin;
								sp->x2 = xmax;
								sp->next = work;
								if (prev)
									prev->next = sp;
								else
									points = sp;
							};
							sp++;
							break;
						default:
							break;
					}
				}
				while (points != 0)
				{
					scan *nextpt = points->next;

					if (!nextpt)
						break;
					xmin = points->x1;
					xmax = nextpt->x2;
					points = nextpt->next;
					{
						scan *prev = 0;
						scan *work = segments;
						int	overlap	= 0;

						while (work	!= 0)
						{
							if ((work->x1 <= xmax) && (xmin	<= work->x2))
							{
								overlap	= 1;
								if (xmin < work->x1)
									work->x1 = xmin;
								if (xmax > work->x2)
								{
									prev = work;
									while ((work = work->next) != 0)
									{
										if (work->x1 > xmax)
											break;
										if (work->x2 > xmax)
											xmax = work->x2;
									}
									prev->x2 = xmax;
									prev->next = work;
								}
								break;
							}
							if (work->x1 > xmax)
								break;
							prev = work;
							work = work->next;
						}
						if (!overlap)
						{
							nextpt->x1 = xmin;
							nextpt->x2 = xmax;
							nextpt->next = work;
							if (prev)
								prev->next = nextpt;
							else
								segments = nextpt;
						}
					};
				}
				while (segments	!= 0)
				{
					xmin = segments->x1;
					xmax = segments->x2;
					segments = segments->next;
					{
						if (xmin > wwrk+xoff)
						{
							continue;
						}
						if (xmax < 0)
						{
							continue;
						}
						if (xmin < 0)
						{
							xmin = 0;;
						}
						if (xmax > wwrk+xoff)
						{
							xmax = wwrk+xoff;
						}
					}
					hline(xmin, ypos, xmax - xmin +	1, color);
				}
			}
		}
	}
	if (mem)
	{
		free(edges);
		free(scans);
	}
	return rect;
}

/**
	Draw convex polygon with 'n' vertices. The each vertice is
	a pair of two integers X and Y. Parameter 'pt' is an array
	of these pairs. The used color is current object's foreground color.

	Coordinate [0:0] is top-left corner of the screen. INK is color
	index in 8-bit videomodes, or absolute color in 16/24 bit modes.
	Raster Operaration is the current value of ENUM_PPOP. The default
	is _GSET (i.e. simply copy) See set_ppop() for more details.
	Coordinate are clipped if needed before drawing.
	Function return false if no drawing (i.e. object has been
	clipped-out completelly). The object XY-offset is added
	before drawing.
*/

/**
	Draw convex polygon with 'n' vertices. The each vertice is
	a pair of two integers X and Y. Parameter 'pt' is an array
	of these pairs.

	Coordinate [0:0] is top-left corner of the screen. INK is color
	index in 8-bit videomodes, or absolute color in 16/24 bit modes.
	Raster Operaration is the current value of ENUM_PPOP. The default
	is _GSET (i.e. simply copy) See set_ppop() for more details.
	Coordinate are clipped if needed before drawing.
	Function return false if no drawing (i.e. object has been
	clipped-out completelly). The object XY-offset is added
	before drawing.
*/
void FGAPI FGDrawBuffer::polygon(const FGPointArray& vertices, unsigned color)
{
	int doClose=0;
	int n = vertices.vertices;
	FGPoint* pt = vertices.array;

	if (color == UNDEFINED_COLOR)
		color = state._ink;

	if (n <= 1)	return;
	if (n == 2) doClose	= 1;

	moveto(pt[0].x, pt[0].y);

	for	(int i = 1;	i <	n; i++)
	{
		lineto(pt[i].x, pt[i].y, color);
	}

	if (doClose || (pt[0].x!=pt[n-1].x && pt[0].y != pt[n-1].y))
		lineto(pt[0].x, pt[0].y, color);
}

/**
	Draws a bezier spline using the four control points specified in the points array.

	Coordinate [0:0] is top-left corner of the screen. INK is color
	index in 8-bit videomodes, or absolute color in 16/24 bit modes.
	Raster Operaration is the current value of ENUM_PPOP. The default
	is _GSET (i.e. simply copy) See set_ppop() for more details.
	Coordinate are clipped if needed before drawing.
	Function return false if no drawing (i.e. object has been
	clipped-out completelly). The object XY-offset is added
	before drawing.
*/
void FGAPI FGDrawBuffer::spline(int points[8], unsigned color)
{
	#define NPTS   64

	int xpts[NPTS], ypts[NPTS];
	int i;

	if (color == UNDEFINED_COLOR)
	    color = state._ink;

	calc_spline(points, NPTS, xpts, ypts);
	moveto(xpts[0], ypts[0]);
	for (i=1; i<NPTS; i++)
	{
		lineto(xpts[i], ypts[i], color);
	}
}

/**
	Draw a line from the latest PEN position to the position [x:y]

	Coordinate [0:0] is top-left corner of the screen. INK is color
	index in 8-bit videomodes, or absolute color in 16/24 bit modes.
	Raster Operaration is the current value of ENUM_PPOP. The default
	is _GSET (i.e. simply copy) See set_ppop() for more details.
	Coordinate are clipped if needed before drawing.
	Function return false if no drawing (i.e. object has been
	clipped-out completelly). The object XY-offset is added
	before drawing.
*/
int FGAPI FGDrawBuffer::lineto(int x, int y, unsigned color)
{
	int rc = line(oldx, oldy, x, y, color);
	if (color == UNDEFINED_COLOR)
		color = state._ink;
	if (state._ppop == _GXOR)
		putpixel(x, y, color);
	return rc;
}

/**
	Draw a text via standard "C" behaviour with the current object font & colors from the position [x:y].

	Coordinate [0:0] is top-left corner of the screen. INK is color
	index in 8-bit videomodes, or absolute color in 16/24 bit modes.
	Raster Operaration is the current value of ENUM_PPOP. The default
	is _GSET (i.e. simply copy) See set_ppop() for more details.
	Coordinate are clipped if needed before drawing.
	Function return false if no drawing (i.e. object has been
	clipped-out completelly). The object XY-offset is added
	before drawing.
*/
int FGDrawBuffer::printf(int xx, int yy, const char *format,...)
{
	char s[256];
	va_list arglist;

	va_start(arglist, format);

#ifndef _MSC_VER
	vsnprintf(s, 255, format, arglist);
#else
	vsprintf(s, format, arglist);
#endif

	va_end(arglist);
	int len = strlen(s);
	if (s[len-1] == '\n')
	{
		s[--len] = 0;
	}
	text(xx,yy,s);
	return len;
}

void FGAPI FGDrawBuffer::Resize(int dx, int dy)
{
	// test to new new size
	if (w + dx<1 || h + dy<1)
		return;
	if (dx==0 && dy==0 && image)
		return;
	lock();
	if (image) free(image);
	w += dx;
	h += dy;
	if (w*h)
		assert((image = (FGPixel *) malloc(areasize(w, h)))!=0);
	wwrk += dx;
	hwrk += dy;
	unlock();
}

/**
	Change the name of the object.
*/
void FGAPI FGDrawBuffer::SetName(const char *s)
{
	lock();
	if (name != empty_string) delete [] name;
	name = new char[strlen(s)+1];
	strcpy(name,s);
	unlock();
}

/**
 * Resize the image with ratio 'rx' and 'ry'.
 */
void FGAPI FGDrawBuffer::stretch(double rx, double ry)
{
	stretch((int)(w*rx), (int)(h*ry));
}

/**
 * Resize the image to the new size.
 */
void FGAPI FGDrawBuffer::stretch(int _w, int _h)
{
#ifndef INDEX_COLORS
		int xpos = 0, ypos = 0;
		unsigned pix;
		float r, g, b;
		int r1, g1, b1;
		int r2, g2, b2;
		int r3, g3, b3;
		int r4, g4, b4;
		int r0, g0, b0;
		float ratio_x;
		float ratio_y;
		float x1, y1;
		float xerr, yerr;
		float xfloat, yfloat;

		if (_w == w && _h == h)
			return;
		if (_w <1 || _h < 1)
			return;

		lock();

		FGPixel *dest = (FGPixel *)malloc(areasize(_w, _h)), *dest_pom=dest;

		ratio_x = (float)w / (float)_w;
		ratio_y = (float)h / (float)_h;

		for (ypos = 0; ypos < _h; ypos++)
				for (xpos = 0; xpos <  _w; xpos++) {
						xfloat = (xpos) * ratio_x;
						yfloat = (ypos) * ratio_y;

						x1 = (int)xfloat;
						y1 = (int)yfloat;

						xerr = 1.0 - (xfloat - (float)x1);
						yerr = 1.0 - (yfloat - (float)y1);
#define _getpixel(x,y,r,g,b) { pix = getpixel((int)x, (int)y); r=kRedComponent(pix); g=kGreenComponent(pix); b=kBlueComponent(pix); }
						_getpixel(x1, y1, r1, g1, b1);
						_getpixel(x1 + 1, y1, r2, g2, b2);
						_getpixel(x1 + 1, y1 + 1, r3, g3, b3);
						_getpixel(x1, y1 + 1, r4, g4, b4);
						r = (float)r1 * xerr +
							(float)r2 * (1.0 - xerr) +
							(float)r3 * (1.0 - xerr) +
							(float)r4 * xerr;

						r += (float)r1 * yerr +
							 (float)r2 * yerr +
							 (float)r3 * (1.0 - yerr) +
							 (float)r4 * (1.0 - yerr);

						r0 = int(r * 0.25);

						g = (float)g1 * xerr +
							(float)g2 * (1.0 - xerr) +
							(float)g3 * (1.0 - xerr) +
							(float)g4 * xerr;

						g += (float)g1 * yerr +
							 (float)g2 * yerr +
							 (float)g3 * (1.0 - yerr) +
							 (float)g4 * (1.0 - yerr);

						g0 = int(g * 0.25);

						b = (float)b1 * xerr +
							(float)b2 * (1.0 - xerr) +
							(float)b3 * (1.0 - xerr) +
							(float)b4 * xerr;

						b += (float)b1 * yerr +
							 (float)b2 * yerr +
							 (float)b3 * (1.0 - yerr) +
							 (float)b4 * (1.0 - yerr);

						b0 = int(b * 0.25);

#ifndef BPP32
							*dest_pom++ = (((r0<<8)&0xf800) | ((g0<<3)&0x7e0) | ((b0>>3)&0x1f));
#else
							*dest_pom++ = ((r0<<16) | (g0<<8) | b0);
#endif
		}
	unlock();
	Resize(_w-w, _h-h);
	lock();
	free(image);
	image = dest;
	unlock();
#endif
}

/**
	Fill a triangle with these three vertices.

	Coordinate [0:0] is top-left corner of the screen. INK is color
	index in 8-bit videomodes, or absolute color in 16/24 bit modes.
	Raster Operaration is the current value of ENUM_PPOP. The default
	is _GSET (i.e. simply copy) See set_ppop() for more details.
	Coordinate are clipped if needed before drawing.
	Function returns false if no drawing (i.e. object has been
	clipped-out completelly). The object XY-offset is added
	before drawing.
	@param x1 x position of first vertice
	@param y1 y position of first vertice
	@param x2 x position of second vertice
	@param y2 y position of second vertice
	@param x3 x position of third vertice
	@param y3 y position of third vertice
	@param color foreground color
*/
FGRect FGAPI FGDrawBuffer::ftriangle(int x1, int y1, int x2, int y2, int x3, int y3, unsigned color)
{
	FGPointArray points(3, &x1);
	return fpolygon(points, color);
}

void FGAPI FGDrawBuffer::turn_bitmap(int cnt, FGPixel *from, FGPixel *to)
{
	FGPixel *p = new FGPixel[w];

	for (int i = 0; i < cnt; i++)
	{
		memmove(p, to - (i * w), w*sizeof(FGPixel));
		memmove(to - (i * w), from + (i * w), w*sizeof(FGPixel));
		memmove(from + (i * w), p, w*sizeof(FGPixel));
	}
	delete p;
}

void FGAPI FGDrawBuffer::rotate(ENUM_ROTATE_DIRECTION direction)
{
	FGPixel *src, *dst, *dst2, pom;
	int foo;

	lock();
	switch(direction)
	{
		case ROTATE_90_CW:
			dst = dst2 = (FGPixel *) malloc(areasize(w,h));
			for(int j=0;j<w;j++)
			{
				for(int i=h-1;i>=0;i--)
				{
					*dst++ = *CalcAddr(j,i);
				}
			}
swapsize:
			foo = w;
			w   = h;
			h   = foo;
			foo = xoff;
			xoff= yoff;
			yoff= foo;
			foo = wwrk;
			wwrk= hwrk;
			hwrk= foo;
			free(image);
			image = dst2;
			break;
		case ROTATE_90_CCW:
			dst = dst2 = (FGPixel *) malloc(areasize(w,h));
			for(int j=w-1;j>=0;j--)
			{
				for(int i=0;i<h;i++)
				{
					*dst++ = *CalcAddr(j,i);
				}
			}
			goto swapsize;
		case ROTATE_180:
			src = CalcAddr(0, 0);
			dst = CalcAddr(w-1, h-1);
			for(int j=(w*h)/2;j>=0;j--)
			{
				pom    = *src;
				*src++ = *dst;
				*dst-- = pom;
			}
			break;
		case ROTATE_HORIZONTAL:
			for(int i=0;i<h;i++)
			{
				src = CalcAddr(0,i);
				dst = src + w-1;
				for(int j=0;j<w/2;j++)
				{
					pom = *src;
					*src++ = *dst;
					*dst-- = pom;
				}
			}
			break;
		case ROTATE_VERTICAL:
			turn_bitmap(h / 2, image, image + w * (h - 1));
			break;
	}
	unlock();
}

/*
 *  geometry is 0 .. w       in X
 *				0 .. h       in Y
 *	clipping is 0 .. wwrk in X
 *				0 .. hwrk in Y
 *	offset is xoff & yoff 0 .. w-1
 *
 *	NOTE: you must applicate offset first before clipping!
 */

#ifdef FG_NAMESPACE
}
#endif







