/**

  $Id: cfg.cpp,v 1.3 2004/02/26 21:41:28 majo Exp $

  $Log: cfg.cpp,v $
  Revision 1.3  2004/02/26 21:41:28  majo
  class FGInternalState
  char* -> const char*
  int ink = -1  -> unsigned ink = UNDEFINED_COLOR

  Revision 1.2  2004/02/23 20:08:01  majo
  all classes are with prefix FG* on now
  polygon functions uses FGPointArray from now
  class GuiEvent is renamed to FGEvent
  some by parameters overloaded methods was removed (class FGWindow)
  many other small changes

  Revision 1.1.1.1  2004/01/22 16:25:08  majo
  initial release


*/

#include "fastgl.h"
#include "cfg.h"
#include "_fastgl.h"

#ifdef FG_NAMESPACE
namespace fgl {
#endif

Config *cCfg;

//
// Configuration file handling	   (for fastgl)
//
int  Config::find(char * s)
{
	for (int i=0;i<entries;i++)
	{
		if (!strcmp(pole[i].name,s)) return i;
	}
	return -1;
}

void Config::CreateEmpty(void)
{
	char s[100] = "; ";
	strcat(s,name);
	WriteComment(s);
	WriteComment(";<.......name of item..........> t <...body of item... (LF)");
	WriteComment("; where 'name' is string up to 32 chars length");
	WriteComment("; where 't' is one of 's'(tring), 'i(nt)', 'b(oolean)' or 'd'(ouble)");
	WriteComment("; Warning! don't use TABs! & STRING MUST be enclosed in double \"\"");
	WriteComment("|..............................| T |.............................");
}

/**
	Open a '*.rc' config file with our formats.
	If the file doesn't exists, create an empty one.
	@param s filename
	@param write set to true if you want write to the file
*/
Config::Config(char * s, int write)
{
	char nn[128];
#if defined(__MSDOS__) || defined (_WIN32)
	sprintf(nn, "%s.rc", s);
#else
	if (FGApp::homedir) sprintf(nn, "%s/.%s.rc", FGApp::homedir, s);
	else sprintf(nn, ".%s.rc", s);
#endif
	strncpy(name, nn, 63);
	change = 0;
	entries = 0;
	writable = write;
	if (writable && fgstate.verbose) ::printf("Opening config file '%s'\n", name);
	f = fopen(name,"r");
	if (f)
	{
		ReadConfig();
		fclose(f);
	}
	else
	{
		CreateEmpty();
	}
}

/**
	Write the file on the disc if it is writable.
*/
Config::~Config()
{
	if (writable) Sync();
}

/**
	Write the comment line to the config.
	If it already contains this comments, do nothing,
	else append the comment at the end of file.
*/
void   Config::WriteComment(char * s)
{
	int i;
	if (entries>=MAX_CONF_ITEMS) return;
	if ((i=find(s))==-1) pole[entries++].ModifyNode(s);
	else  pole[i].ModifyNode(s);
	change = 1;
}

/**
	Write item of type 'int' with name.
	@param s the name of variable
	@param i value that will be writen
*/
void   Config::WriteInt(char * s, int i)
{
	int ii;
	if (entries>=MAX_CONF_ITEMS) return;
	if ((ii=find(s))==-1) pole[entries++].ModifyNode(s,i);
	else  pole[ii].ModifyNode(s,i);
	change = 1;
}

/**
	Write item of type 'int' with name.
	@param s the name of variable
	@param i value that will be writen
*/
void   Config::WriteBool(char * s, bool i)
{
	int ii;
	if (entries>=MAX_CONF_ITEMS) return;
	if ((ii=find(s))==-1) pole[entries++].ModifyNode(s,i);
	else  pole[ii].ModifyNode(s,i);
	change = 1;
}

/**
	Write item of type 'char *' with name. This procedure
	write strings that have max. 127 chars length.
	@param s the name of variable
	@param i string that will be writen
*/
void   Config::WriteString(char * s, char * i)
{
	int ii;
	if (strlen(i)>127) return;
	if (entries>=MAX_CONF_ITEMS) return;
	if ((ii=find(s))==-1) pole[entries++].ModifyNode(s,i);
	else  pole[ii].ModifyNode(s,i);
	change = 1;
}

/**
	Write item of type 'byte array' with the name. This procedure
	write an binary data in hex form that have max. 128 bytes length.
	@param s the name of variable
	@param i bytes that will be writen
	@param c number of bytes
*/
void   Config::WriteHex(char * s, char * i, int c)
{
	int ii;
	if (c>128) return;
	if (entries>=MAX_CONF_ITEMS) return;
	if ((ii=find(s))==-1) pole[entries++].ModifyNode(s,i,c);
	else  pole[ii].ModifyNode(s,i,c);
	change = 1;
}

/**
	Write item of type 'double' with name 's'.
	@param s the name of variable
	@param i value that will be writen
*/
void   Config::WriteDouble(char * s, double i)
{
	int ii;
	if (entries>=MAX_CONF_ITEMS) return;
	if ((ii=find(s))==-1) pole[entries++].ModifyNode(s,i);
	else  pole[ii].ModifyNode(s,i);
	change = 1;
}

static int ChToI(char xCH)
{
	if ((xCH >= '0') & (xCH <= '9'))
		return ((int) xCH) - '0';
	if ((xCH >= 'a') & (xCH <= 'f'))
		return ((int) xCH) - 'a' + 10;
	if ((xCH >= 'A') & (xCH <= 'F'))
		return ((int) xCH) - 'A' + 10;
	return -1;
}

static int  hexdata(char *s)
{
	int a = ChToI(s[0]);
	int b = ChToI(s[1]);
	if (a!=-1 && b!=-1) return a*16 + b;
	return 256;
}

void Config::ReadConfig(void)
{
	char s[360], *p, n[33], *t, hex[128];
	int typ,i;
	double d;
	bool b;

	for(;entries<MAX_CONF_ITEMS;)
	{
		*s = 0;
		fgets(s,359,f);
		if (!*s) break;
		*(s+strlen(s)-1) = 0;
		p = s;
		t = n;
		if (!isalpha(s[0]) && *s != '_')
		{
			pole[entries].ModifyNode(s);
		}
		else
		{
			while(isalnum(*p) || *p=='_') (*t++) = (*p++);
			*t = 0;
			switch(typ = s[33])
			{
				case Node::_DOUBLE:
					d = atof(s+35);
					pole[entries].ModifyNode(n,d);
					break;
				case Node::_INT:
					i = atoi(s+35);
					pole[entries].ModifyNode(n,i);
					break;
				case Node::_BOOLEAN:
					b = !!stricmp(s+35, "no");
					pole[entries].ModifyNode(n,b);
					break;
				case Node::_STRING:
					t = s+36;
					while(*t!='"') t++;
					*t = 0;
					pole[entries].ModifyNode(n,s+36);
					break;
				case Node::_HEXA:
					t = s+36;
					i = 0;
					while((typ=hexdata(t))<256 && i<128)
					{
						hex[i++] = typ;
						t+=2;
					}
					pole[entries].ModifyNode(n,hex,i);
					break;
			}
		}
		entries++;
	}
}

int Config::Compare(const void* p1, const void* p2)
{
	const Node* n1 = (Node *)p1;
	const Node* n2 = (Node *)p2;
	return strcmp(n1->name, n2->name);
}

/**
	Sorts all lines in config file.
	@note all comments will be removed from file.
*/
void  Config::Sort(void)
{
	int removed=0;
	change = true;

	for(int i=0;i<entries;i++)
	{
		if (pole[i].type == Node::_COMMENT)
		{
			pole[i].type = Node::_NONE;
			pole[i].name[0] = -1;
			removed++;
		}
	}
	qsort(pole, entries, sizeof(pole[0]), Compare);
	entries -= removed;
}

/**
	Saves the config object to its filename.
*/
void  Config::Sync(void)
{
	char s[360], hex[260],pom[8];
	int j;

	if (change==0) return;
	if (fgstate.verbose) ::printf("Writing config file '%s'\n", name);
	errno = 0;
	f = fopen(name,"w");
	if (f)
	{
		for(int i=0;i<entries;i++)
		{
			switch(pole[i].type)
			{
				case Node::_COMMENT:
					fputs(pole[i].s,f);
					break;
				case Node::_STRING:
					sprintf(s, "%-32s %c \"%s\"", pole[i].name, pole[i].type, pole[i].s);
					fputs(s,f);
					break;
				case Node::_INT:
					sprintf(s, "%-32s %c %d", pole[i].name, pole[i].type, pole[i].i);
					fputs(s,f);
					break;
				case Node::_BOOLEAN:
					sprintf(s, "%-32s %c %s", pole[i].name, pole[i].type, pole[i].b?"yes":"no");
					fputs(s,f);
					break;
				case Node::_DOUBLE:
					sprintf(s, "%-32s %c %f", pole[i].name, pole[i].type, pole[i].d);
					fputs(s,f);
					break;
				case Node::_HEXA:
					*hex = 0;
					for(j=0;j<pole[i].len;j++)
					{
						sprintf(pom, "%02x", pole[i].s[j]&255);
						strcat(hex,pom);
					}
					sprintf(s, "%-32s %c \"%s\"", pole[i].name, pole[i].type, hex);
					fputs(s,f);
					break;
			}
			fputc('\n',f);
		}
		fclose(f);
	}
	if (errno) perror("Config::Sync() ");
}

int  Config::ReadInt(char * s)
{
	int	n =	find(s);
	if (n>=0) return pole[n].i;
	return 0;
}

char const *  Config::ReadString(char * s)
{
	int	n =	find(s);
	if (n>=0) return (char	const *)pole[n].s;
	return "";
}

double  Config::ReadDouble(char * s)
{
	int	n =	find(s);
	if (n>=0) return pole[n].d;
	return 0.0;
}

/**
	Read the variable value of type 'int' from the config.
	@param s name of the variable (max. size is 32 chars)
	@param p the reference to the variable that will be initialized wit reads value.
	@return true if value is assigned to variable, false when variable is not found in the config.
*/
int  Config::ReadInt(char	* s, int &p)
{
	int	n =	find(s);
	if (n>=0) return (p = pole[n].i), 1;
	return 0;
}

/**
	Read the variable value of type 'bool' from the config.
	@param s name of the variable (max. size is 32 chars)
	@param p the reference to the variable that will be initialized wit reads value.
	@return true if value is assigned to variable, false when variable is not found in the config.
*/
int	    Config::ReadBool(char	* s, bool &p)
{
	int	n =	find(s);
	if (n>=0) return (p = pole[n].b), 1;
	return 0;
}

/**
	Read the null terminated string from the config.
	@param s name of the variable (max. size is 32 chars)
	@param p the pointer to the variable that will be initialized with reads value. You
	must allocate this memory before call.
	@param max the max. number of bytes that will be reads (max 127 chars).
	@return true if value is assigned to variable, false when variable is not found in the config.
*/
int	    Config::ReadString(char * s,	char *p, int max)
{
	int	n =	find(s);
	if (max>128) max = 128;
	if (n>=0)
	{
		strncpy(p,pole[n].s,max);
		return 1;
	}
	return 0;
}

/**
	Read the null terminated string from the config.
	@param s name of the variable (max. size is 32 chars)
	@param p the pointer to the pointer to the variable that will be
	initialized with reads value. The string is not copied, i.e. you cannot allocate the memory at (**p).
	@return true if value is assigned to variable, false when variable is not found in the config.
*/
int	    Config::ReadString(char * s,	char **p)
{
	int	n =	find(s);
	if (n>=0)
	{
		 *p = pole[n].s;
		return 1;
	}
	return 0;
}

/**
	Read the variable value of type 'double' from the config.
	@param s name of the variable (max. size is 32 chars)
	@param p the reference to the variable that will be initialized with reads value.
	@return true if value is assigned to variable, false when variable is not found in the config.
*/
int	    Config::ReadDouble(char * s,	double &p)
{
	int	n =	find(s);
	if (n>=0) return (p = pole[n].d), 1;
	return 0;
}

/**
	Read the array of bytes from the config.
	@param s name of the variable (max. size is 32 chars)
	@param p the pointer to the variable that will be initialized with reads value.
	@param max the max. number of bytes that will be reads (max 128 bytes).
	@return true if value is assigned to variable, false when variable is not found in the config.
*/
int	    Config::ReadHex(char * s, char *p, int max)
{
	int	n =	find(s);
	if (max>128) max = 128;
	if (n>=0)
	{
		memcpy(p, pole[n].s, max);
		return 1;
	}
	return 0;
}

#ifdef FG_NAMESPACE
}
#endif


