/*
  $Id: cfg.h,v 1.1.1.1 2004/01/22 16:25:08 majo Exp $

  $Log: cfg.h,v $
  Revision 1.1.1.1  2004/01/22 16:25:08  majo
  initial release

*/

#if !defined(Cfg_H)
#define Cfg_H

#ifdef FG_NAMESPACE
namespace fgl {
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>

/**
	Open a configuration file with name (without the extension ".rc").
	If any not exists, system creates a new empty into the RAM.
	If you use switch APP_CFG, system automatically opens in
	the App constructor. Default config file is appname.rc.
	The pointer at this config saves to global variable
	cApp->cfg and its destructed (and saved) in App destructor.
	The maximal number of items into the config file is up to 512.
	The number of at once opens config files is unlimited. The maximal
	size of name of variable is up to 32 chars. The maximal size
	of saved text string or byte array is up to 127 chars + '\\0'.
*/
class Config
{
  class Node
  {
	enum itemtype
	{
	_DOUBLE='d',
	_INT='i',
	_STRING='s',
	_COMMENT='c',
	_HEXA='h',
	_BOOLEAN='b',
	_NONE=255
	};

	friend class Config;
	int type;
	char name[36];
	union {
		int	i;
		char  s[132];
		double d;
		bool b;
	};
	int len;
	void  ModifyNode(char * comment)	// Modify comment
	{
		type = _COMMENT;
		*name =	0;
		strncpy(s, comment, 128);
	}
	void  ModifyNode(char * meno, char *	val) // Modify string
	{
		type = _STRING;
		strncpy(name, meno, 32);
		strncpy(s, val, 128);
	}
	void  ModifyNode(char * meno, int val) // Modify integer
	{
		type = _INT;
		strncpy(name, meno, 32);
		i	 = val;
	}
	void  ModifyNode(char * meno, bool val) // Modify boolean
	{
		type = _BOOLEAN;
		strncpy(name, meno, 32);
		b	 = val;
	}
	void  ModifyNode(char * meno, double	val) // Modify double
	{
		type = _DOUBLE;
		strncpy(name, meno, 32);
		d	 = val;
	}
	void  ModifyNode(char * meno, char *	val, int kolko) // Modify string
	{
		type = _HEXA;
		strncpy(name, meno, 32);
		len = kolko>128? 128 : kolko;
		memcpy(s, val, len);
	}
  };

		enum { MAX_CONF_ITEMS=512 };
		Node 	pole[MAX_CONF_ITEMS];
		char 	name[128];
		FILE 	*f;
		int		entries;
		int		change;
		int		writable;

		int	  find(char *	s);
		void CreateEmpty(void);
		void ReadConfig(void);
		static int Compare(const void*, const void*);

	protected:
	public:
		void 	 WriteComment(char * s);
		void	 Sync(void);
		void	 Sort(void);
		//! deprecated
		int		 ReadInt(char	* s);
		int		 ReadBool(char	* s, bool &p);
		int		 ReadInt(char	* s, int &p);
		int		 ReadString(char * s,	char *p, int max);
		//! deprecated
		char	const*  ReadString(char * s);
		int		 ReadString(char * s,	char **p);
		//! deprecated
		double	 ReadDouble(char * s);
		int		 ReadDouble(char * s, double &p);
		int		 ReadHex(char * s, char *, int);

		void	 WriteBool(char * s, bool);
		void	 WriteInt(char * s, int);
		void	 WriteString(char	* s, char *);
		void	 WriteDouble(char	* s, double);
		void	 WriteHex(char * s, char *, int);

		Config(char	*,int write=1);
		~Config();
};

/**
	The pointer to the applicatons Config object
	@ingroup Globals
*/
extern Config *cCfg;

#ifdef FG_NAMESPACE
}
#endif

#endif // Unit1_H
