/*
  $Id: cfgstore.cpp,v 1.5 2004/02/23 20:08:01 majo Exp $

  $Log: cfgstore.cpp,v $
  Revision 1.5  2004/02/23 20:08:01  majo
  all classes are with prefix FG* on now
  polygon functions uses FGPointArray from now
  class GuiEvent is renamed to FGEvent
  some by parameters overloaded methods was removed (class FGWindow)
  many other small changes

  Revision 1.4  2004/02/22 16:47:09  majo
  fixed fgl namespace for the file

  Revision 1.3  2004/02/06 16:32:51  majo
  fixed crash witn no file

  Revision 1.2  2004/02/06 16:30:11  majo
  fixed crash if file not found

*/

#include <string>
#include <map>
#include <vector>
#ifndef MINIMAL_STORE
#include <set>
#endif
#include <stdio.h>
#include <stdlib.h>

#include "base.h"
#include "cfgstore.h"

#ifdef FG_NAMESPACE
namespace fgl
{
#endif

const char* const stype[] = { /*"ListStr", "ListInt", "ListFloat", "ListHex",*/
                              "VectorStr", "VectorInt", "VectorFloat", "VectorHex",
                              "SetStr", "SetInt", "SetFloat", "SetHex"
							  };

std::string CfgStore::node_delimiter = ", ";    // Oddelovac zaznamov


CfgStore::CfgStore(const char* filename)
: fname(filename)
{
    Load();
}


CfgStore::~CfgStore()
{
    // Uvolni store struktury ulozene v store
    for (TStoreIterator it = GetStoreBegin(); it != GetStoreEnd(); it++)
    {
        delete (*it).second;
    }
}


StoreBase* CfgStore::CreateStore(StoreType type, const char* name)
{
    // Ak uz existuje store s takym nazvom, tak nepridavam dalsie
    if (store.count(name)) return 0;

    StoreBase* node = 0;
    switch (type)
    {
        case STORE_VECTOR_STR   : node = new StoreVectorStr(); break;
#ifndef MINIMAL_STORE
        case STORE_VECTOR_INT   : node = new StoreVectorInt(); break;
        case STORE_VECTOR_FLOAT : node = new StoreVectorFloat(); break;
        case STORE_VECTOR_HEX   : node = new StoreVectorHex(); break;
        case STORE_SET_STR      : node = new StoreSetStr(); break;
        case STORE_SET_INT      : node = new StoreSetInt(); break;
        case STORE_SET_FLOAT    : node = new StoreSetFloat(); break;
        case STORE_SET_HEX      : node = new StoreSetHex(); break;
#endif
        default:
            return 0;
    }
    store.insert(TPair(name, node));
    return node;
}


StoreBase* CfgStore::GetStore(const char* name)
{
    std::string s;
    TStoreIterator i = GetStoreIter(name);
    if (i != store.end()) return (*i).second; else return 0;
}


bool CfgStore::DeleteStore(const char* name)
{
    TStoreIterator i = GetStoreIter(name);
    if (i != store.end())
    {
        delete (*i).second;
        store.erase(i);
        return true;
    } else return false;
}


void CfgStore::Clear()
{
    // Uvolni store struktury ulozene v store
    for (TStoreIterator it = GetStoreBegin(); it != GetStoreEnd(); it++)
    {
        delete (*it).second;
    }
    store.clear();
}


bool CfgStore::Load(std::string& error)
{
    Clear();
    Scanner scanner(this);
    if (scanner.IsValidStream()) return scanner.scan_parse(error); else
    {
        error = "Can not open file '" + fname + "'";
        return false;
    }
}


void CfgStore::Save()
{
    FILE* fout;
    fout = fopen(fname.c_str(), "w");
	if (fout == 0)  return;

	for (TStoreIterator it = GetStoreBegin(); it != GetStoreEnd(); it++)
    {
		// Konverzia store struktury na string a jej zapis do suboru
		std::string str;
		(*it).second->StoreToString(str);

        fprintf(fout, "%s %s ( %s );\n", stype[(*it).second->GetType()], ((*it).first).c_str(), str.c_str());
    }
    fprintf(fout, "\n");
    fflush(fout);
    fclose(fout);
}


//--------------------------------------------------------------------------------------------------------------
// SCANNER/PARSER pre nacitanie udajov
//--------------------------------------------------------------------------------------------------------------


CfgStore::Scanner::Scanner(CfgStore* pstore)
: line_no(1), char_no(1), ch(0), last_token(NO_TOKEN), last_token_int(0), last_token_float(0), last_token_str(""), cfg_store(pstore)
{
    memset(char_table, SPECIAL, sizeof(char_table));
    for (ch = '0'; ch <= '9'; ch++) char_table[ch] = DIGIT;
    for (ch = 'A'; ch <= 'Z'; ch++) char_table[ch] = LETTER;
    for (ch = 'a'; ch <= 'z'; ch++) char_table[ch] = LETTER;
    char_table['.'] = POINT;
    char_table['+'] = SIGN;
    char_table['-'] = SIGN;
    char_table['$'] = HEXID;
    char_table['\"'] = QUOTE;

    stream = fopen(cfg_store->GetFileName(), "r");
    getch();
}


CfgStore::Scanner::~Scanner()
{
    if (stream) fclose(stream);
}


std::string CfgStore::Scanner::info_to_str()
{
    char info[128];
    sprintf(info, "(Line %d, Char %d) ", line_no, char_no);
    return std::string(info);
}


void CfgStore::Scanner::get_token()
{
    skip_blank();
    last_token = NO_TOKEN;
    last_token_int = 0;
    last_token_float = 0;
    last_token_str = "";

    if (ch == -1)
    {
        last_token = END_OF_FILE;
        return;
    }

    switch (char_table[ch])
    {
        case LETTER   : get_word(); break;
        case DIGIT    : get_number(); break;
        case POINT    : get_number(); break;
        case SIGN     : get_number(); break;
        case QUOTE    : get_string(); break;
        case HEXID    : get_hex(); break;
        default:
            get_special();
            break;
    }
}

void CfgStore::Scanner::get_word()
{
    while ((char_table[ch] == LETTER) || (char_table[ch] == DIGIT))
    {
        last_token_str += ch;
        getch();
    }
    last_token = IDENTIFIER;
}


void CfgStore::Scanner::get_number()
{
    if (char_table[ch] == SIGN)
    {
        last_token_str += ch;
        getch();
    }
    bool flt = false;
    if (char_table[ch] == POINT)
    {
        flt = true;
        last_token_str += '0';
        last_token_str += ch;
        getch();
    }
    while (char_table[ch] == DIGIT)
    {
        last_token_str += ch;
        getch();
    }
    if (char_table[ch] == POINT)
    {
        flt = true;
        last_token_str += ch;
        getch();
        while (char_table[ch] == DIGIT)
        {
            last_token_str += ch;
            getch();
        }
    }
    if (flt)
    {
        last_token_float = atof(last_token_str.c_str());
        last_token = FLOAT;
    } else {
        last_token_int = atoi(last_token_str.c_str());
        last_token = INTEGER;
    }
}


void CfgStore::Scanner::get_hex()
{
    getch();
    while ((char_table[ch] == DIGIT) || ((char_table[ch] == LETTER) && (ch >= 'A') && (ch <= 'F')))
    {
        last_token_str += ch;
        getch();
    }
    last_token_int = strtol(last_token_str.c_str(), 0, 16);
    last_token = HEX;
}


void CfgStore::Scanner::get_string()
{
    getch();
    while (char_table[ch] != QUOTE)
    {
        last_token_str += ch;
        getch();
    }
    getch();
    last_token = STRING;
}

void CfgStore::Scanner::get_special()
{
    last_token_str = ch;
    switch (ch)
    {
        case '(' : last_token = LPAREN; break;
        case ')' : last_token = RPAREN; break;
        case ',' : last_token = NODEDIV; break;
        case ';' : last_token = SECTEND; break;
        case '\n': last_token = END_OF_LINE; line_no++; char_no = 1; break;
    }
    getch();
}


bool CfgStore::Scanner::scan_parse(std::string& error)
{
    error = "";
    get_token();
    skip_eoln();
    while (last_token != END_OF_FILE)
    {
        //parse: definicia typu zaznamu
        if (last_token != IDENTIFIER) { error = info_to_str() + "Expect definition identifier"; return false; }
        bool found = false;
        StoreType def_type = StoreType(0);
        while (!found && (def_type < LAST_MARK))
        {
            found = (strcmp(last_token_str.c_str(), stype[int(def_type)]) == 0);
            if (!found) def_type = StoreType(def_type + 1) ;
        }
        if (!found) { error = info_to_str() + "Unknown definition identifier '" + last_token_str + "'"; return false; }

        //parse: identifikator zaznamu
        get_token();
        skip_eoln();
        if (last_token != IDENTIFIER) { error = info_to_str() + "Expect identifier as value name, but '" + last_token_str + "' found"; return false; }

        //vytvor store
        StoreBase* store = cfg_store->CreateStore(def_type, last_token_str.c_str());

        //parse: zaciatok pola
        get_token();
        skip_eoln();
        if (last_token != LPAREN) { error = info_to_str() + "Expect begin of data '(', but '" + last_token_str + "' found"; return false; }

        //parse: data
        if (!store) while (last_token != RPAREN) get_token(); else
        {
            switch (def_type)
            {
                case STORE_VECTOR_STR   : if (!parse_STORE_VECTOR_STR(store, error)) return false; break;
#ifndef MINIMAL_STORE
                case STORE_VECTOR_INT   : if (!parse_STORE_VECTOR_INT(store, error)) return false; break;
                case STORE_VECTOR_FLOAT : if (!parse_STORE_VECTOR_FLOAT(store, error)) return false; break;
                case STORE_VECTOR_HEX   : if (!parse_STORE_VECTOR_HEX(store, error)) return false; break;
                case STORE_SET_STR      : if (!parse_STORE_SET_STR(store, error)) return false;  break;
                case STORE_SET_INT      : if (!parse_STORE_SET_INT(store, error)) return false;  break;
                case STORE_SET_FLOAT    : if (!parse_STORE_SET_FLOAT(store, error)) return false;  break;
                case STORE_SET_HEX      : if (!parse_STORE_SET_HEX(store, error)) return false;  break;
#endif
                default:
                    error = info_to_str() + "Unimplemented parse '" + stype[def_type] + "' data type";
                    return false;
            }
        }

        //parse: koniec zaznamu
        get_token();
        if (last_token != SECTEND) { error = info_to_str() + "Expect end of data ';', but '" + last_token_str + "' found"; return false; }

        //neser sa, daj dalsi
        get_token();
        skip_eoln();
    }
    return true;
}


bool CfgStore::Scanner::parse_STORE_VECTOR_STR(StoreBase* store, std::string& error)
{
    bool done = false;
    do
    {
        get_token();
        skip_eoln();
        if (last_token != STRING) { error = info_to_str() + "Expect string value"; return false; }
        ((StoreVectorStr*) store)->push_back(last_token_str);

        get_token();
        skip_eoln();
        switch (last_token)
        {
            case NODEDIV : break;
            case RPAREN  : done = true; break;
            default:
                error = info_to_str() + "Expect ',' as node divider or ')' as end of array";
                return false;
        }
    } while (!done);
    return true;
}

#ifndef MINIMAL_STORE
bool CfgStore::Scanner::parse_STORE_VECTOR_INT(StoreBase* store, std::string& error)
{
    bool done = false;
    do
    {
        get_token();
        skip_eoln();
        if (last_token != INTEGER) { error = info_to_str() + "Expect integer value"; return false; }
        ((StoreVectorInt*) store)->push_back(last_token_int);

        get_token();
        skip_eoln();
        switch (last_token)
        {
            case NODEDIV : break;
            case RPAREN  : done = true; break;
            default:
                error = info_to_str() + "Expect ',' as node divider or ')' as end of array";
                return false;
        }
    } while (!done);
    return true;
}


bool CfgStore::Scanner::parse_STORE_VECTOR_FLOAT(StoreBase* store, std::string& error)
{
    bool done = false;
    do
    {
        get_token();
        skip_eoln();
        if (last_token == INTEGER) last_token_float = last_token_int;
        else if (last_token != FLOAT) { error = info_to_str() + "Expect float value"; return false; }
        ((StoreVectorFloat*) store)->push_back(last_token_float);

        get_token();
        skip_eoln();
        switch (last_token)
        {
            case NODEDIV : break;
            case RPAREN  : done = true; break;
            default:
                error = info_to_str() + "Expect ',' as node divider or ')' as end of array";
                return false;
        }
    } while (!done);
    return true;
}


bool CfgStore::Scanner::parse_STORE_VECTOR_HEX(StoreBase* store, std::string& error)
{
    bool done = false;
    do
    {
        get_token();
        skip_eoln();
        if (last_token != HEX) { error = info_to_str() + "Expect hexadecimal value with '$' prefix"; return false; }
        ((StoreVectorHex*) store)->push_back(last_token_int);

        get_token();
        skip_eoln();
        switch (last_token)
        {
            case NODEDIV : break;
            case RPAREN  : done = true; break;
            default:
                error = info_to_str() + "Expect ',' as node divider or ')' as end of array";
                return false;
        }
    } while (!done);
    return true;
}


bool CfgStore::Scanner::parse_STORE_SET_STR(StoreBase* store, std::string& error)
{
    bool done = false;
    do
    {
        get_token();
        skip_eoln();
        if (last_token != STRING) { error = info_to_str() + "Expect string value"; return false; }
        ((StoreSetStr*) store)->insert(last_token_str);

        get_token();
        skip_eoln();
        switch (last_token)
        {
            case NODEDIV : break;
            case RPAREN  : done = true; break;
            default:
                error = info_to_str() + "Expect ',' as node divider or ')' as end of array";
                return false;
        }
    } while (!done);
    return true;
}


bool CfgStore::Scanner::parse_STORE_SET_INT(StoreBase* store, std::string& error)
{
    bool done = false;
    do
    {
        get_token();
        skip_eoln();
        if (last_token != INTEGER) { error = info_to_str() + "Expect integer value"; return false; }
        ((StoreSetInt*) store)->insert(last_token_int);

        get_token();
        skip_eoln();
        switch (last_token)
        {
            case NODEDIV : break;
            case RPAREN  : done = true; break;
            default:
                error = info_to_str() + "Expect ',' as node divider or ')' as end of array";
                return false;
        }
    } while (!done);
    return true;
}


bool CfgStore::Scanner::parse_STORE_SET_FLOAT(StoreBase* store, std::string& error)
{
    bool done = false;
    do
    {
        get_token();
        skip_eoln();
        if (last_token == INTEGER) last_token_float = last_token_int;
        else if (last_token != FLOAT) { error = info_to_str() + "Expect float or integer value"; return false; }
        ((StoreSetFloat*) store)->insert(last_token_float);

        get_token();
        skip_eoln();
        switch (last_token)
        {
            case NODEDIV : break;
            case RPAREN  : done = true; break;
            default:
                error = info_to_str() + "Expect ',' as node divider or ')' as end of array";
                return false;
        }
    } while (!done);
    return true;
}


bool CfgStore::Scanner::parse_STORE_SET_HEX(StoreBase* store, std::string& error)
{
    bool done = false;
    do
    {
        get_token();
        skip_eoln();
        if (last_token != HEX) { error = info_to_str() + "Expect hexadecimal value with '$' prefix"; return false; }
        ((StoreSetHex*) store)->insert(last_token_int);

        get_token();
        skip_eoln();
        switch (last_token)
        {
            case NODEDIV : break;
            case RPAREN  : done = true; break;
            default:
                error = info_to_str() + "Expect ',' as node divider or ')' as end of array";
                return false;
        }
    } while (!done);
    return true;
}

#endif

#ifdef FG_NAMESPACE
}
#endif

