/*
  $Id: cfgstore.h,v 1.5 2004/02/22 16:47:09 majo Exp $

  $Log: cfgstore.h,v $
  Revision 1.5  2004/02/22 16:47:09  majo
  fixed fgl namespace for the file

  Revision 1.4  2004/02/10 16:18:59  majo
  fixed compile warnings

  Revision 1.3  2004/02/06 16:32:51  majo
  fixed crash witn no file

*/
#ifndef _CFGSTORE_H_
#define _CFGSTORE_H_

#ifdef FG_NAMESPACE
namespace fgl
{
#endif

#ifdef __BORLANDC__
#pragma option push -b
#pragma option push -a4
#endif

#define MAX_STORE_STRING_LEN    8192    // maximalna velkost ukladaneho retazca do store

#define MINIMAL_STORE

enum StoreType { STORE_VECTOR_STR = 0,
#ifndef MINIMAL_STORE
		 STORE_VECTOR_INT=1, STORE_VECTOR_FLOAT, STORE_VECTOR_HEX,
                 STORE_SET_STR, STORE_SET_INT, STORE_SET_FLOAT, STORE_SET_HEX,
#endif
                 LAST_MARK
};


class StoreBase
{
    private:
        StoreType type;

    protected:
        inline std::string ValueToString(std::string value) { return '\"' + value + '\"'; }
        inline std::string ValueToString(int value) { char s[33]; sprintf(s, "%d", value); return std::string(s); }
        inline std::string ValueToString(double value) { char s[33]; sprintf(s, "%f", value); return std::string(s); }
        inline std::string ValueToString(unsigned short value) { char s[33]; sprintf(s, "$%04X", value); return std::string(s); }

    public:
        StoreBase(StoreType _type) : type(_type) {}
        virtual ~StoreBase() {}
        inline StoreType GetType() { return type; }
        virtual void StoreToString(std::string& str) = 0;
};


//--------------------------------------------------------------------------------------------------------------


class CfgStore
{
    class Scanner
    {
        private:
            enum TCharCode { LETTER, DIGIT, SIGN, POINT, HEXID, QUOTE, SPECIAL };
            enum TTokenCode { NO_TOKEN, END_OF_FILE, END_OF_LINE, IDENTIFIER, LPAREN, RPAREN, NODEDIV, SECTEND, STRING, INTEGER, FLOAT, HEX };

            unsigned int line_no; // cislo aktualneho riadoku suboru (len pre diagnostiku chyby)
            unsigned int char_no; // cislo aktualneho charu v danom riadku (len pre diagnostiku chyby)
            char ch; // aktualne spracovavany znak
            TTokenCode last_token; // posledne detekoveny token
            int last_token_int; // posledna dekodovana int hodnota
            double last_token_float; // posledna dekodovana float hodnota
            std::string last_token_str; // posledny dekodovany string

            CfgStore* cfg_store;
            FILE* stream;
            TCharCode char_table[256];
            std::string info_to_str();
            inline void getch() { if (stream) { ch = (char) fgetc(stream); char_no++; } else ch = END_OF_FILE; }
            inline void skip_blank() { while (ch == ' ') getch(); }
            inline void skip_eoln() { while (last_token == END_OF_LINE) get_token(); }
            void get_token();
            void get_word();
            void get_number();
            void get_hex();
            void get_string();
            void get_special();
            bool parse_STORE_VECTOR_STR(StoreBase* store, std::string& error);
#ifndef MINIMAL_STORE
            bool parse_STORE_VECTOR_INT(StoreBase* store, std::string& error);
            bool parse_STORE_VECTOR_FLOAT(StoreBase* store, std::string& error);
            bool parse_STORE_VECTOR_HEX(StoreBase* store, std::string& error);
            bool parse_STORE_SET_STR(StoreBase* store, std::string& error);
            bool parse_STORE_SET_INT(StoreBase* store, std::string& error);
            bool parse_STORE_SET_FLOAT(StoreBase* store, std::string& error);
            bool parse_STORE_SET_HEX(StoreBase* store, std::string& error);
#endif
        public:
            Scanner(CfgStore* pstore);
            ~Scanner();
            bool scan_parse(std::string& error);
            bool IsValidStream() { return (stream != 0); }
    };

    private:
        typedef std::pair<const std::string, StoreBase*> TPair;
        typedef std::map<std::string, StoreBase*> TStore;

        std::string fname;
        TStore store;

    public:
        typedef TStore::iterator TStoreIterator;

        CfgStore(const char* filename);
        ~CfgStore();

        StoreBase* CreateStore(StoreType type, const char* name);
        StoreBase* GetStore(const char* name);
        bool DeleteStore(const char* name);

        void Clear();
        bool Load(std::string& error);
        void Load() { std::string str; Load(str); }
        void Save();

        inline const char* GetFileName() { return fname.c_str(); }
        inline TStoreIterator GetStoreBegin() { return store.begin(); }
        inline TStoreIterator GetStoreEnd() { return store.end(); }
        inline TStoreIterator GetStoreIter(const char* name) { return store.find(std::string(name)); }

        // rezervovane/formatovacie sekvencie znakov
        static std::string node_delimiter;
};


//--------------------------------------------------------------------------------------------------------------

template <typename Type>
class StoreVector : public StoreBase
{
    protected:
        typedef std::vector<Type> TVector;
        TVector vector;
        std::string out_format;

    public:
        typedef typename TVector::iterator TVectorIterator;

        StoreVector(StoreType type) : StoreBase(type) {}
        virtual ~StoreVector() {}

        virtual void StoreToString(std::string& str)
        {
            int i = 0;
            //str.clear();
            str = "";
	    for (TVectorIterator it = GetVectorBegin(); it != GetVectorEnd(); it++, i++)
            {
                if (i) str += CfgStore::node_delimiter;
                str += ValueToString((*it));
            }
        }

        inline TVectorIterator GetVectorBegin() { return vector.begin(); }
        inline TVectorIterator GetVectorEnd() { return vector.end(); }
        inline Type value(unsigned int index) { return vector[index]; }
        inline void value(unsigned int index, Type value) { vector[index] = value; }
        inline void push_front(Type value) { vector.push_front(value); }
        inline void push_back(Type value) { vector.push_back(value); }
        inline void insert(unsigned int index, Type value)
        {
            TVectorIterator it = vector.begin();
            for (int i = 0; i < index; i++, it++);
            vector.insert(it, value);
        }
        inline void erase(unsigned int index)
        {
            TVectorIterator it = vector.begin();
            for (unsigned int i = 0; i < index; i++, it++);
            vector.erase(it);
        }
        inline void clear() { vector.clear(); }
        inline int size() { return vector.size(); }
        inline bool empty() { return vector.empty(); }
};

#ifndef MINIMAL_STORE

template <typename Type>
class StoreSet : public StoreBase
{
    protected:
        typedef std::set<Type> TSet;
        TSet set;
        std::string out_format;

    public:
        typedef typename TSet::iterator TSetIterator;

        StoreSet(StoreType type) : StoreBase(type) {}
        virtual ~StoreSet() {}

        virtual void StoreToString(std::string& str)
        {
            int i = 0;
            //str.clear();
            str = "";
	    for (TSetIterator it = GetSetBegin(); it != GetSetEnd(); it++, i++)
            {
                if (i) str += CfgStore::node_delimiter;
                str += ValueToString((*it));
            }
        }

        inline TSetIterator GetSetBegin() { return set.begin(); }
        inline TSetIterator GetSetEnd() { return set.end(); }
        inline void insert(Type value) { set.insert(value); }
        inline void erase(Type value) { set.erase(value); }
        inline void clear() { set.clear(); }
        inline int size() { return set.size(); }
        inline bool empty() { return set.empty(); }
};

#endif

//--------------------------------------------------------------------------------------------------------------

// Vektory roznych typov
class StoreVectorStr : public StoreVector<std::string>
{
    public:
        StoreVectorStr() : StoreVector<std::string>(STORE_VECTOR_STR) {}
};

#ifndef MINIMAL_STORE
class StoreVectorInt : public StoreVector<int>
{
    public:
        StoreVectorInt() : StoreVector<int>(STORE_VECTOR_INT) {}
};

class StoreVectorFloat : public StoreVector<double>
{
    public:
        StoreVectorFloat() : StoreVector<double>(STORE_VECTOR_FLOAT) {}
};

class StoreVectorHex : public StoreVector<unsigned short>
{
    public:
        StoreVectorHex() : StoreVector<unsigned short>(STORE_VECTOR_HEX) {}
};


// Mnoziny roznych typov
class StoreSetStr : public StoreSet<std::string>
{
    public:
        StoreSetStr() : StoreSet<std::string>(STORE_SET_STR) {}
};

class StoreSetInt : public StoreSet<int>
{
    public:
        StoreSetInt() : StoreSet<int>(STORE_SET_INT) {}
};

class StoreSetFloat : public StoreSet<double>
{
    public:
        StoreSetFloat() : StoreSet<double>(STORE_SET_FLOAT) {}
};

class StoreSetHex : public StoreSet<unsigned short>
{
    public:
        StoreSetHex() : StoreSet<unsigned short>(STORE_SET_HEX) {}
};
#endif

#ifdef __BORLANDC__
#pragma option pop /* pop -a switch */
#pragma option pop /* pop -b */
#endif

#ifdef FG_NAMESPACE
}
#endif

#endif

