/*
	OpenGUI - Drawing & Windowing library

	Copyright (C) 1996,2004  Marian Krivos

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Library General Public
	License as published by the Free Software Foundation; either
	version 2 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Library General Public License for more details.

	You should have received a copy of the GNU Library General Public
	License along with this library; if not, write to the Free
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	nezmar@atlas.sk

	fastgl.h   - declarations of many objects for FastGL

  $Id: fastgl.h,v 1.32 2004/09/27 06:59:57 majo Exp $
*/

#ifndef	__WINDOW_H
#define	__WINDOW_H

#include "fgcolor.h"
#include "base.h"
#include "cfg.h"


#ifdef FG_NAMESPACE
namespace fgl {
#endif

/* HJH modification: protect compiler options for structure alignment and enum
 * size if the compiler is Borland C++ */
#ifdef __BORLANDC__
#pragma option push -b
#pragma option push -a4
#endif

enum
{
	BNORMAL	= 7,
	BPUSHED	= 3,
	BDISABLED =	5,
	BUNVISIBLE = 6
};

#define	FGFONT_TITLE	FONTSYS
#define	FGFONT_MENU		FONTSYSLIGHT
#define	FGFONT_BUTTON	FONTSYSLIGHT

/**
	@addtogroup Defines
	@{
*/
//! predefined FGWindow foreground color
#define	IM		 	CScheme->window_fore // CBLACK standard menu fore & background colors
//! predefined FGWindow background color
#define	PM			CScheme->window_back

//! The FGWindow has frame.
const unsigned  WFRAMED =		0x0001;
//! The FGWindow has caption.
const unsigned 	WTITLED	=	0x0002;
//! The FGWindow will be shown as modal.
const unsigned 	WMODAL =		0x0004;
//! The FGWindow sends EVENTNOTIFY to its handler when it is destroyed.
const unsigned 	WNOTIFY	=	0x0008;
//! internal
const unsigned 	WHIDEN =		0x0010;
//! The FGWindow is unable to move/change its position.
const unsigned 	WUNMOVED =	0x0020;
//! The FGWindow has a line menu.
const unsigned 	WMENU =		0x0040;
//! The FGWindow is without pictograms (to minimize and to hide it).
const unsigned 	WNOPICTO =	0x0080;
//! internal
const unsigned 	WBITMAP	=	0x0100;
//! The FGWindow emit CLICKLEFTEVENT and CLICKRIGHTEVENT when you clicks with mouse on it.
const unsigned 	WCLICKABLE =	0x0200;
//! internal
const unsigned 	WDIRTY =		0x0400;
//! internal
const unsigned 	WTEMPHIDEN =	0x0800;
//! internal
const unsigned 	WLOCKED	=	0x1000;
//! internal
const unsigned 	WTRIGGER =	0x2000;
//! The FGWindow don't get focus on create.
const unsigned 	WLASTFOCUS =	0x4000;
//! The FGWindow is able to change its size.
const unsigned 	WSIZEABLE =	0x8000;
//! The FGWindow sets its size and position from WindowDatabase if any on create time.
const unsigned 	WUSELAST =	0x00010000;
//! Show the minimize pictogram in the FGWindow title.
const unsigned 	WMINIMIZE =	0x00020000;
//! The FGWindow has a statusbar in the bottom.
const unsigned 	WSTATUSBAR =	0x00040000;
//! internal
const unsigned 	WEXIST =		0x00080000;
//! internal
const unsigned 	WGLCONTEXT =	0x00100000;
/**
The FGWindow uses notebooks. Use this flag to create multiple page dialog
or tabbed notebook. It displays multiple overlapping pages that are TabPage
objects. The user selects selects a page by clicking the page's tab that appears
at the top of control.
@see FGWindow::AddTabPage()
*/
const unsigned 	WNOTEBOOK =	0x00200000;
//! the FGWindow has controlled layout
const unsigned 	WLAYOUT =	0x00400000;
//! Show the FGWindow in the center of screen.
const unsigned 	WCENTRED =	0x00800000;
//! Close the FGWindow when ESCAPE key is pressed.
const unsigned 	WESCAPE	 =	0x01000000;
//! Alligne the FGWindow to 32 bit sizes
const unsigned 	WALIGNED =	0x02000000;
//! Enable fast move of the FGWindow, i.e. don't redraw opaque.
const unsigned 	WFASTMOVE =	0x04000000;
//! internal
const unsigned 	WNOICON	=	0x10000000;
//! internal
const unsigned 	WNODRAWFROMCONSTRUCTOR = 0x20000000;
//! internal
const unsigned 	WICONWINDOWTYPE	=   0x40000000;
//! internal
const unsigned 	WMENUWINDOWTYPE	=   0x80000000;
//! standard predefined values.
const unsigned 	WSTANDARD =	(WCLICKABLE|WTITLED|WFRAMED);

//! Enable to remember FGWindow's positions and sizes.
const unsigned 	APP_WINDOWDATABASE =		1;
//! Enable Configuration file in form EXEFILENAME.rc
const unsigned 	APP_CFG	=				2;
//! Enable The ALT+X hotkey to terminates your application.
const unsigned 	APP_ENABLEALTX =			4;
//! Enable the rectangle selections by mouse.
const unsigned 	APP_MAGNIFIER =			8;
//! Enable and initialize the root FGWindow, i.e. background.
const unsigned 	APP_ROOTWINDOW =			16;
//! Enable visual efects.
const unsigned 	APP_ENABLEEFECTS =		0x20;
//! Enable all the above.
const unsigned APP_ALL =					0x1F;
//! Use local application directory for Config files instead of HOMEDIR.
const unsigned 	APP_LOCALDIR =			0x80000000;

/**	@} */

/**
	The current Operating System code.
	@ingroup Enums
*/
enum ENUM_OS { OSTYPE_LINUX=1, OSTYPE_MSDOS, OSTYPE_WIN32, OSTYPE_QNX };

/**
	Return codes for FGApp::ShowModal()
	@ingroup Enums
*/
enum MODAL_RETURN
{	mrNone   = 	-1,
	mrOk	 = 	-2,
	mrCancel =	-3,
	mrYes	 =	-2,
	mrNo	 =	-3,
	mrAll	 =	-4,
	mrIgnore =	-5,
	mrClose	 =	-6,
	mrQuit	 =	-7,
	mrRetry =	-8
};

class FGBaseGui;
class FGFileDialog;
class FGColorDialog;
class FGControl;
class FGWindow;
class FGMenuWindow;
class FGBitmap;
class FGEditBox;
class FGButtonGroup;
class FGListBox;
class FGPushButton  ;
class FGPointButton;
class FGCheckButton;
class FGBaseMenuItem;
class FGPushButton;
class FGSlider;
class FGTabPage;

/**
	Show the error dialog with defined string.
	@ingroup Misc
	@param s a string that will be shown
	@param flag 0 for warning or for fatal error (will exit after presskey). If you want MODAL dialog, you must add to the flag the value 2.
*/
extern "C" void	IError(char	*s,	int	flag);

/**
Provides simply and easy theme like object.
The library creates & initializes an instance of this structure on the startup
and uses it for a default settings for Controls and Windows that are created then.
*/
struct FGColorScheme
{
	/** a default FGWindow background color */
	int		window_back;
	/** a default FGWindow foreground color */
	int		window_fore;
	/** an active FGWindow title color */
	int		active_title;
	/** an inactive FGWindow title color */
	int		inactive_title;
	/** a default FGWindow shaded frame color 1 */
	int		wnd_bord1;
	/** a default FGWindow shaded frame color 2 */
	int		wnd_bord2;
	/** a default FGWindow shaded frame color 3 */
	int		wnd_bord3;
	/** a default FGWindow statusbar color */
	int		statusbar;
	/** a default base menu background color */
	int		menu_back;
	/** a default base menu foreground color */
	int		menu_fore;
	/** a default base menu background color when active */
	int		menu_back_active;
	/** a default base menu foreground color when active */
	int		menu_fore_active;
	/** a default push button foreground color */
	int		button_fore;
	/** a default push button background color */
	int		button_back;
	/** a default push button foreground color when pushed */
	int		button_fore_pushed;
	/** a default push button background color when pushed */
	int		button_back_pushed;
	/** a default push button border 1 */
	int		button_bord1;
	/** a default push button border 2 */
	int		button_bord2;
	/** a default push button border 3 */
	int		button_bord3;
	/** a default edit box background color if not activated */
	int		edit_background_active;
	/** a default edit box foreground color */
	int		edit_foreground;
	/** a default edit box border 1 */
	int		edit_border;
	/** a default edit box border 2 */
	int		edit_background;
	/** a default slider color */
	int		slider;
	/** a default pulldown menu background color */
	int		menuwindow_back;
	/** a default pulldown menu foreground color */
	int		menuwindow_fore;
	/** a default pulldown menu item background color */
	int		pdmenu_back_active;
	/** a default pulldown menu item foreground color */
	int		pdmenu_fore_active;
	/** a default pulldown menu item color when disabled */
	int		pdmenu_gray;
	/** a default edit box color when disabled */
	int		edit_fore_disable;
	/** a default notebook color (current item)*/
	int		notebook_active;
	/** a default slider color when disabled */
	int		slider_disable;
  public:
    FGColorScheme();
};

// subbitmap

extern FGColorScheme * CScheme;

/**
	The TIMER procedure callback type.
	@ingroup Types
*/
typedef unsigned int (*TimerCall)(unsigned int);

#define TIMER_RESOLUTION	10	/* Experimentally determined */

/**
	Some timing class.
*/
class FGTimer : public FGConnector
{
	protected:
		friend void  			HandleAlarm(int sig);
		int						state;
		static void				StartTicks(void);
	public:
		FGTimer(unsigned int val, TimerCall cb)
		{
			state = SetTimer(val, cb);
		}
		FGTimer(TimerCall cb)
		{
			state = SetTimer(0, cb);
		}
		virtual ~FGTimer()
		{
			TimerQuit();
		}
		static int			SetTimer(unsigned int val, TimerCall cb);
		static unsigned int GetTicks(void);
		static void			Delay(unsigned int);
		static int			TimerInit(void);
		static void			TimerQuit(void);
};

/**
	System event object. The events are presents in many types. By example
	event is key pressing, mouse move, click on the mouse button, window move etc.
*/
class FGEvent
{
		friend class FGBaseGui;
		friend class FGApp;
		int		type;
		int		key;
		int		buttons, x,	y, w, h;
		/**
		This is a public member of FGEvent class and it is assigned for ACCELEVENT only.
		It contains the type of controls. Normally it is unnecessary. The types are
		enum ObjectType {WINDOW, CHECKBUTTON, e.g.
		*/
		int		guiType;

		static	const char *eventName[];
	protected:
	public:
    	/**
		This is a public member of FGEvent class and it is initialised for
        window procedures only. It contains the pointer at the target window.
		It is useful when one procedure is assigned to more windows.
		Using of this member is in general better, than global pointers.
		*/
		FGWindow	*wnd;
		/**
		This is assigned for ACCELEVENT only.
		It contains the type of controls. Normally it is unnecessary. The types are
		enum ObjectType { WINDOW, CHECKBUTTON, e.g.
		*/
		int		GetObjectType(void) const { return guiType; }
		/**
		This is a public member of FGEvent class and it is initialised for
		window procedures only and ACCELEVENT only. It contains the pointer
		at the Controls that emit this event.
		*/
		FGControl *accel;
		/**
		Create the event of appropriate type. In the most cases, you will parse
		system events only. Creating of the events by the hand and sending them
		to other windows is less frequently. The response to the system events
		is adjusted into separate window procedures, each per window. This procedure
		is fact a one big switch statement with many cases.
		*/
		FGEvent(int T=NOEVENT,	int	K=0, int X=0, int Y=0, int W=0, int H=0)
			: type(T), key(K), x(X), y(Y), w(W), h(H), guiType(0), wnd(0), accel(0)
		{
		}
		//! Returns the value of key member of class FGEvent. As for event KEYEVENT, it return a key code.
		int		GetKey(void)  const { return key;	}
		//! Return a text string, which identified the name of events, or string "NONSENSE".
		const char	*GetName(void)  const { if (type<0 || type>=LASTEVENT) return eventName[LASTEVENT]; return eventName[type]; }
		//! Returns type of the event.
		int		GetType(void)  const { return	type; }
		//! Returns symbolic name of keycode as ASCIIZ string.
		static const char* GetKeySymbolName(unsigned int code);
		/**
		Returns an OR-ed value of the all mouse's buttons that are pushed.
		There are these symbolic values for buttons:
			-	FG_BUTTON_LEFT
			-	FG_BUTTON_RIGHT
			-	FG_BUTTON_MIDDLE
		*/
		int		GetButtons(void)  const {	return buttons;	}
		//! Returns x co-ordination for events that initialise this member (MOVEEVENT, CLICKLEFTEVENT).
		int		GetX(void)  const { return	x; }
		//! Returns y co-ordination for events that initialise this member (MOVEEVENT, CLICKLEFTEVENT).
		int		GetY(void)  const { return	y; }
		//! Returns a width.
		int		GetW(void)  const { return	w; }
		//! Returns a width.
		int		GetH(void)  const { return	h; }
};

/**
	@addtogroup Types
	@{
*/
/**
The type of an FGApp::SetTimerProc procedure callback.
*/
typedef	void (*TimerProc)(int);
/**
The type of a FGWindow handler. This procedure is called when the FGWindow EVENT (of type FGEvent) is occured.
*/
typedef	void (*GuiHwnd)(FGEvent *);
/**
The type of an Application handler. This procedure is called when system EVENT (of type FGEvent) is occured.
*/
typedef	int	(*MainHwnd)(FGEvent *);
/**
The pointer to a FGControl type.
*/
typedef	FGControl * CallBack;
/**
The prototype for a LISTBOX container callback procedure.
*/
typedef void (*FGListBoxCallBack)(int flag, FGWindow *wnd, int x, int y, void *data, int index);
/**
The prototype for a FGFileDialog container callback procedure.
*/
typedef void (*FileDialogCallBack)(char * filename, FGFileDialog *);
/**
The prototype for a ColorDialog container callback procedure.
*/
typedef void (*ColorDialogCallBack)(FGPixel color);
/**
The prototype for a ColorDialog container callback procedure.
*/
typedef void (*FontDialogCallBack)(int f);

/** @} Types */

/**
 @internal.
*/
typedef	void (*ControlCall)(FGControl *);


//!	base objects for all graphics objects
/**
	This class is base parent of all graphics objects into OpenGUI library
	and encapsulates all basic methods. These methods are used in FGWindow,
	FGEditBox, FGPointButton and other GUI classes.
	It implements a linked list of objects also. This is used to release memory
	when FGWindow is destroyed.
*/
class FGBaseGui : public FGDrawBuffer
{
		friend class FGWindow;
	    friend class FGControl;
		friend class FGTabPage;

		FGMutex	    ListMutex;
		FGBaseGui*  last;
		FGBaseGui*  next;
		FGBaseGui* 	OverSprite(FGBaseGui * od, int	x, int y, int w, int h);
		FGBaseGui* 	BaseGuiFind(FGEvent *e);
	protected:
		enum bs	{
            EBOX_LOW=1, EBOX_UPR, BDOWN, BUP,
            WHIDE, WVISIBLE, WDEACTIVE, WACTIVE };

		int		 	x;
		int		 	y;
		FGBaseGui*  tab_page_controls;
		void*		private_data;
		void*		user_data;
		unsigned    ink;
		unsigned	paper;

		FGBaseGui(const FGBaseGui &)
		{
			::printf("No COPY Constructor defined!\n");
			exit(-1);
		}
		void	FGAPI Resize(int dx, int dy);
		inline void	FGAPI SetLast(FGBaseGui *l) { last	= l; }
		inline void	FGAPI SetNext(FGBaseGui *l) { next	= l; }
		inline FGBaseGui* FGAPI GetNext(void)  const {	return next; }
		inline FGBaseGui* FGAPI GetLast(void)  const {	return last; }
		virtual		void MoveIn(void) {}
		virtual		void MoveOut(void) {}

	public:
		FGBaseGui(int	xs,	int	ys,	int	ws,	int	hs,	const char *nm, ObjectType typ, FGPixel i, FGPixel p, long flag);
		virtual		~FGBaseGui();

		//! Returns a foreground color of the object.
		unsigned  	FGAPI GetInk(void)  const { return ink; }
		//! Retutns a background color of the object.
		unsigned   	FGAPI GetPaper(void)  const { return paper; }

		FGBaseGui	* 	FindId(int	idw);
		/**
		Returns vivibility of the object.
		*/
		bool	  	IsVisible(void)  const	{ return !(status&(WHIDEN|WTEMPHIDEN));	}
		/**
		Set the object's color as default for drawing.
		*/
		void        SetColors(void)
		{
			set_colors(ink, paper);
		}
		/**
		Returns the 'x' coordinates of the object.
		*/
		int		  	GetX(void)  const { return	x; }
		/**
		Returns the 'y' coordinates of the object.
		*/
		int		  	GetY(void)  const { return	y; }
		/**
		Callback on focus get.
		*/
		virtual void OnFocus(void) { }
		/**
		Callback on focus lost.
		*/
		virtual void OnLostFocus(void) { }
        //! @internal
		void*		GetPrivateData(void) const { return private_data; }
        //! @internal
		void		SetPrivateData(void* val) { private_data = val; }
		//! Get user data pointer
		void*		GetUserData(void) const { return user_data; }
		//! Set user data pointer
		void		SetUserData(void* val) { user_data = val; }
};

#ifndef __GNUC__
#pragma	pack(1)
#endif

//!	Used for an image file manipulating (load/save of various formats is implemented).
class FGBitmap : public FGDrawBuffer
{
	public:
        //! return values from LoadImage
        enum ReturnValues { IMAGE_OK, IMAGE_FILE_NOT_FOUND, IMAGE_UNKNOWN_FORMAT,
		    IMAGE_UNSUPPORTED_FORMAT, IMAGE_FILE_WRITE_ERROR };
    private:
    	struct RGB3
	    {
		    char bBlue;				 /* Blue component of the color definition */
    		char bGreen;			 /* Green component of the color definition*/
	    	char bRed;				 /* Red component of the color definition  */
    	};
	    struct RGB4
    	{
	    	char bBlue;				 /* Blue component of the color definition */
		    char bGreen;			 /* Green component of the color definition*/
    		char bRed;				 /* Red component of the color definition  */
	    	char fcOptions;			 /* Reserved, must be zero                 */
    	};
	    struct bitmap_file_header
    	{
			unsigned short	usType;
		    unsigned short	cbSizeLow, cbSizeHigh;
    		short			reserve1, reserve2;
	    };
    	struct bitmap_info
	    {
    		unsigned short	imageOffset;
	    	short			reserve;
    	};
    	struct bitmap_info_header1
	    {
    	  unsigned long	 size;
	      unsigned short cx;			   /* Bit-map width in pels                  */
    	  unsigned short cy;			   /* Bit-map height in pels                 */
	      unsigned short cPlanes;		   /* Number of bit planes                   */
    	  unsigned short cBitCount;		   /* Number of bits per pel within a plane  */
    	  unsigned long	 ulCompression;	   /* Compression scheme used to store the bitmap */
    	  unsigned long	 cbImage;		   /* Length of bit-map storage data in chars*/
    	  unsigned long	 cxResolution;	   /* x resolution of target device          */
    	  unsigned long	 cyResolution;	   /* y resolution of target device          */
    	  unsigned long	 cclrUsed;		   /* Number of color indices used           */
    	  unsigned long	 cclrImportant;	   /* Number of important color indices      */
    	};
    	struct bitmap_info_header2
    	{
    	  unsigned long	 size;
    	  unsigned long	 cx;			   /* Bit-map width in pels                  */
    	  unsigned long	 cy;			   /* Bit-map height in pels                 */
    	  unsigned short cPlanes;		   /* Number of bit planes                   */
    	  unsigned short cBitCount;		   /* Number of bits per pel within a plane  */
    	  unsigned long	 ulCompression;	   /* Compression scheme used to store the bitmap */
    	  unsigned long	 cbImage;		   /* Length of bit-map storage data in chars*/
    	  unsigned long	 cxResolution;	   /* x resolution of target device          */
    	  unsigned long	 cyResolution;	   /* y resolution of target device          */
    	  unsigned long	 cclrUsed;		   /* Number of color indices used           */
    	  unsigned long	 cclrImportant;	   /* Number of important color indices      */
    	};

		void init(void);
		bitmap_file_header	bmfh;
		bitmap_info			bmi;
		union {
			bitmap_info_header1	bmih1;
			bitmap_info_header2	bmih2;
		} bmih;
		union {
			RGB3 rgb3[256];
			RGB4 rgb4[256];
		} rgb;

		int*		cUsedTable;
		int			format;			    // IMAGE_*
    	char	    szLastError[256];	//debugging

		void		DoPalette(int);
		void 		RGBtoBGR(unsigned char *buffer, unsigned char *input, int length);

		ReturnValues pcx(char *n, int& one, unsigned int& bits, unsigned int& c);
		ReturnValues bmp(char *n, int& one, unsigned int& bits, unsigned int& c);
		ReturnValues gif(char *n, int& one, unsigned int& bits, unsigned int& c);
		ReturnValues jpg(char *n, int& one, unsigned int& bits, unsigned int& c);
		ReturnValues png(char *n, int& one, unsigned int& bits, unsigned int& c);
		ReturnValues tga(char *n, int& one, unsigned int& bits, unsigned int& c);
		ReturnValues tiff(char *n, int& one, unsigned int& bits, unsigned int& c);

		int detect(const char *);

	public:
		enum ImageFormats { IMAGE_AUTO, IMAGE_BMP, IMAGE_GIF, IMAGE_TGA,
			IMAGE_PNG, IMAGE_JPG, IMAGE_TIF, IMAGE_PCX, IMAGE_UNKNOWN,
			IMAGE_LAST };
		/**
			Creates an empty bitmap image object.
			You must use LoadImage() to load from file.
		*/

		FGBitmap();
		/**
		Creates "in memory" object from the bitmap file with name.
		The file must be one from supported and enabled image formats.
		If an operation is OK, then data member type is != BMP_NONE,
		else file doesn't exist or isn't a valid image file.
		@deprecated
		*/
		FGBitmap(const char	*name);
		FGBitmap(FGDrawBuffer *);
		FGBitmap(int,int,FGPixel=CBLACK);
		FGBitmap(FGBitmap& old) : FGDrawBuffer(old)
		{
			bmfh = old.bmfh;
			bmi = old.bmi;
			bmih = old.bmih;
			rgb = old.rgb;
			cUsedTable = 0;
#ifdef INDEX_COLORS
			if (old.cUsedTable)
			{
				cUsedTable = new int[256];
				memcpy(cUsedTable, old.cUsedTable, sizeof(long)*256);
			}
#endif
		}
		virtual ~FGBitmap();
        //! @deprecated
		ReturnValues BitmapSave(const char *, bool convert=false);
		ReturnValues JpegSave(const char *name);
		ReturnValues PngSave(const char *name);

		ReturnValues LoadImage(const char* name, int imagefmt = IMAGE_AUTO);
		ReturnValues SaveImage(const char* name, int imagefmt);

/*
		int		BitmapSave(void)
		{
			if (name && name[0] > ' ')
				return BitmapSave(name);
			else return 0;
		}
*/
};
#ifndef __GNUC__
#pragma	pack()
#endif

/**
When you write the code to manage some menus, sometimes you need program
the group of buttons, that would worked as welded. When you choose one,
previous goes to the inactive state - this is called a FGButtonGroup -
and make possible to you easy aggregate items like CheckButtons,
PointButtons, or FGPushButton (the word or is at a right place,
because you can't mix these types together!).
The default number of items is 32.
*/
class FGButtonGroup
{
		friend class FGControl;
		FGControl* 	curr;
		FGControl**	array;
		int			count;
		int			size;
		int			type; // 1-check,point; 2-push; 0 undefined

	public:
		/**
		Creates an empty group of the FGControl. Default number of items is 32. You can pass your own value.
		*/
		FGButtonGroup(int maximum=32)
		{
			size = maximum;
			count = type = 0;
			curr = 0;
			array = new FGControl *[size];
		}
		~FGButtonGroup() { delete [] array; }
		/**
		Add FGControl object to the FGButtonGroup. The argument active is
		not mandatory (default = false) and if you set this to true,
		then object will have the active state at the start.
		If you mix no acceptable items together, or if you add other
		than allowed types, an error dialog will be displayed.
		*/
		void FGAPI AddToGroup(FGControl *p, int activ = 0);
		/**
			Reset the group. You can add controls now again.
		*/
		void FGAPI Clear(void) { type = count = 0; curr = 0; }
        //! Redraw all the group completely.
		void FGAPI RefreshGroup(FGControl *c);
		//! Disable all the group completely.
		void FGAPI DisableGroup(void);
		//! Enable all the group completely.
		void FGAPI EnableGroup(void);
};

//! The predecessor for the all widgets.
/**
This family of graphics controls contains the class FGPushButton, FGCheckButton,
FGEditBox and Base menu. It is child objects of class Controls, which is
a child object of base class FGBaseGui. There is a more information.
NOTE: All of the FGControl class objects and its parents are on the create
time automatically attached with its FGWindow. When a FGWindow is destroyed,
Controls are destroyed either automatically. Don't delete it explicitly!
@code
void call_back_procedure(CallBack cb)
{
	... // your code
}
@endcode
*/
class FGControl :	public FGBaseGui
{
		friend class FGWindow;
		friend class FGButtonGroup;
		friend class FGApp;

		ControlCall	GetHandler(void)  const { return fnc; }

		void		RegisterToGroup(FGButtonGroup *g) { grp = g;}
		void		RefreshGroup(void);
		void		RemoveFromGroup(void) { grp = 0; }
		static 		FGControl * ButtonFind(FGEvent *e,int	n);

	protected:
		int			local_id;
		int			key;		   	// hotkey
		ControlCall	fnc;			// handler
		FGWindow*   owner;			// owner of this object
		FGButtonGroup *grp;			// if is group members

		FGControl(int	xs,	int	ys,	int	ws,	int	hs,	const char *nm, ObjectType typ, int key, int flag, FGWindow *, ControlCall f=0, FGPixel p=CGRAY1, FGPixel i=CDARK);
		virtual		~FGControl();

		int	   		GetXr(void);
		int	   		GetYr(void);
		void		Underscore(int xo, int yo, int c);
		void 		frame(void);
		void 		update_owner(void);
		virtual		void ClickDown(void) {}
		virtual		void WheelSpin(int) {}	// -1 up, 1 down
		virtual     void GetHints(int& xx, int& yy, int& ww, int& hh) { xx = x; yy = y; ww = w; hh = h; }
		bool        IsYourTabPage(void);
	public:
		/**
		Redraws whole the object.
		*/
		virtual		void draw(void)=0;
		/**
		Change the name of the object and redraws one.
		*/
		virtual		void SetName(const char *s) { FGDrawBuffer::SetName(s); draw(); }
		/**
		Callback on the FGControl activation by user by click or hotkey pressing.
		You can override this in child object.
		*/
		virtual		void OnActivate(void) { RunSignal(); }
		/**
		Change the 'hotkey' of the object.
		*/
		void		SetKey(int k) {	if (k>='a' && k<='z') k	= toupper(k); key=k; draw(); }
		/**
			Returns current 'hotkey'.
		*/
		int			GetKey(void) const { return key; }
		/**
		Changes the state of the object (draws it as normal and allows activate one).
		*/
		void		Enable(void);
		/**
		Changes the state of the object (draws it as grayed and don't allows activate one).
		*/
		void		Disable(void);
		/**
		Change the state of the object to state "active" (as you as click
		with mouse at one). It is a good choice if you want input
		to the FGEditBox immediately without clicking at one.
		*/
		virtual		void ClickUp(int a);  // if TRUE, so call handler
		/**
		With this procedure you can change the state of the two-state (ON/OFF) switches.
		*/
		void		SetTrigger(int a) {	if (a) status |= WTRIGGER;	else	status &= (~WTRIGGER);	draw();}
		/**
		Returns the state of the FGControl (TRUE or FALSE).
		*/
		int	   		GetTrigger(void) const { return status&WTRIGGER?1:0;	}
		/**
		Returns the parent Widget, i.e. window that contains this FGControl.
		*/
		FGWindow		*GetOwner(void) const { return owner;	}
		/**
		Each FGControl in the FGWindow has local ID. You can use this ID with handling ACCELEVENT by example.
		The all controls have local id. This value goes from 0 to .., for each
		window and its control items. For example: when window contains 5 buttons,
		then these buttons have local id values from 0 to 4,according to the order
		of its creating. This is useful for fast and easy testing which button
		(when you use one call-back procedure for more buttons i.e.).
		*/
		int			GetLocalId() const { return local_id; }
		/**
		Set new foreground color and redraws object.
		*/
		void		SetInk(FGPixel i) { state._ink = ink = i; draw(); }
		/**
		Set new background color and redraws object.
		*/
		void		SetPaper(FGPixel p) { state._paper = paper = p; draw(); }
		/**
		Set new background color and redraws object.
		*/
		void		SetFont(unsigned f) { set_font(f); draw(); }
		/**
		Set a new handler for object.
		*/
		void		SetHandler(ControlCall w) {	fnc=w; }
		/**
		You can use this static method as predefined callback when you can
		close the parent FGWindow.
		@code
			my_window->AddBaseMenu("Close the window", 'C', FGControl::Close);
		@endcode
		*/
		static		void Close(CallBack cb);
		/**
		You can use this static method as predefined callback when you can
		close the parent FGWindow.
		@code
			my_window->AddBaseMenu("Quit the Application", 'C', Quit::Close);
		@endcode
		*/
		static		void Quit(CallBack cb);
};

/**
	Use Label to add text that user can't edit to a FGWindow. This text can be
	used to label another FGControl object. To add an object to a FGWindow that displays text that
	a user can edit use FGEditBox.

	This button can be grouped into the set of mutually exclussive options
	to the user - that is, only one radio button in a set can be selected
	at a time. When the user selects a button, the previously selected
	button becomes unselected. Buttons are frequently grouped in
	a button group box (FGButtonGroup). Add the group box to the FGWindow first,
	then put the buttons into the group box.
	For example, two buttons on a FGWindow can be checked at the same time only if they are
	contained in separate containers, such as two different group boxes.

	Use FGWindow::AddLabel() instead.
*/
class FGLabel : public FGControl
{
	protected:
		void ClickUp(int a);
		unsigned int is_transparent:1, high_light:1;
	public:
		// redraws object completely
		virtual	void draw(void);
		FGLabel(int xs, int ys, const char *nm, int key, FGWindow *w, ControlCall f, unsigned i, unsigned p);
		void SetTransparent(void) { is_transparent = 1; draw(); }
		void MoveIn(void)
		{
			high_light = 1;
			draw();
		}
		void MoveOut(void)
		{
			high_light = 0;
			draw();
		}
};

/**
* A 'decorate' class. This is not a real GUI FGControl item. It is
* intended for visual purposes only. There was an old method WindowPanel()
* that not success because of it's static character - when FGWindow was overdrawed
* the all Controls redraws correctly but static text and - WindowPanel was washed out.
* This new class is redrawed automagically (as child of FGControl class).
*/
class FGPanel : public FGControl
{
	protected:
		FGPixel	ink1, ink2;
	public:
		// redraws object completely
		virtual	void draw(void);
		FGPanel(int xs, int ys, int ws, int hs, const char *nm, FGWindow *w, int i1, int i2, int p);
};

/**
* A 'decorate' class. This is not a real GUI FGControl item. It is
* intended for visual purposes only - i.e. use with TabPages.
* This new class is redrawed automagically (as child of FGControl class).
* @see Window::AddImage()
*/
class FGImage : public FGControl
{
	protected:
	    FGDrawBuffer*   bitmap;
	public:
		//! redraws object completely
		virtual	void draw(void);
		//! create FGImage object and attach FGDrawImage object to the one.
		FGImage(int xs, int ys, FGDrawBuffer* img, FGWindow *w);
};

/**
* A 'decorate' class. This is not a real GUI FGControl item. It is
* intended for visual purposes only. When FGWindow was overdrawed
* the all Controls redraws correctly but static text and - WindowPanel was washed out.
* This new class is redrawed automagically (as child of FGControl class).
*/
class FGText : public FGControl
{
	protected:
	public:
		FGText(int xs, int ys, const char *nm, FGWindow *w, int i, int p);
		// redraws object completely
		virtual	void draw(void);
};

/**
	Use FGPushButton to put a standard push button on a FGWindow. FGPushButton
	introduces several properties to control its behaviour in a dialog box settings.
	Users choose button controls to initiate actions.

	@image html pushbutt.png

	This button can be grouped into the set of mutually exclussive options
	to the user - that is, only one radio button in a set can be selected
	at a time. When the user selects a button, the previously selected
	button becomes unselected. Buttons are frequently grouped in
	a button group box (FGButtonGroup). Add the group box to the FGWindow first,
	then put the buttons into the group box.
	For example, two buttons on a FGWindow can be checked at the same time only if they are
	contained in separate containers, such as two different group boxes.

	To add this FGControl use FGWindow::AddPushButton()
*/
class FGPushButton : public FGControl {
		FGDrawBuffer*   icon;
		FGDrawBuffer*   pushed;
		FGDrawBuffer*   disabled;
	protected:
	public:
		FGPushButton(int xs, int ys, int ws, int hs, const char	*nm, int key, int flag,	FGWindow *w, ControlCall f=0)
			: FGControl(xs, ys, ws, hs, nm, PUSHBUTTON, key, flag, w,	f)
			{
				set_font(FGFONT_BUTTON);
				icon = 0;
				pushed = disabled = 0;
				draw();
			}
		FGPushButton(int xs, int ys, int key,	int	flag, FGWindow *w, FGDrawBuffer	*bm, ControlCall f=0)
			: FGControl(xs, ys, bm->GetW() + 4, bm->GetH() + 4, "", PUSHBUTTON, key, flag|WBITMAP, w, f)
			{
				set_font(FONTSYS);
				icon = bm;
				pushed = disabled = 0;
				draw();
			}
		FGPushButton(int xs, int ys, int ws, int hs, const char* nm, int key,	int	flag, FGWindow *w, FGDrawBuffer	*bm, ControlCall f=0)
			: FGControl(xs, ys, ws, hs, nm, PUSHBUTTON, key, flag|WBITMAP, w, f)
			{
				set_font(FONTSYS);
				icon = bm;
				pushed = disabled = 0;
				draw();
			}
		void ClickDown(void);
		void ClickUp(int);
		void MoveIn(void);
		void MoveOut(void);
		void Push(void);
		void Release(void);
		// redraws object completely
		virtual	void draw(void);
		void AddBitmaps(FGDrawBuffer* push=0, FGDrawBuffer* disable=0)
		{
			pushed = push;
			disabled = disable;
			draw();
		}
};

/**
	FGCheckButton represents a check box that can be on (checked)
	or off (unchecked). It differents from FGPointButton only visually.

	@image html check.png

	This button can be grouped into the set of mutually exclussive options
	to the user - that is, only one radio button in a set can be selected
	at a time. When the user selects a button, the previously selected
	button becomes unselected. Buttons are frequently grouped in
	a button group box (FGButtonGroup). Add the group box to the FGWindow first,
	then put the buttons into the group box.
	For example, two buttons on a FGWindow can be checked at the same time only if they are
	contained in separate containers, such as two different group boxes.

	To add this FGControl use FGWindow::AddCheckButton() .
*/
class FGCheckButton :	public FGControl {
	protected:
		int		mask;
		int		* variable;
		void MoveIn(void);
		void MoveOut(void);
	public:
		FGCheckButton(int m, int	xs,	int	ys,	const char *nm, int key, int flag, FGWindow	*w,	int	fg,	int	bg,	ControlCall f=0, int *var=0)
			: FGControl(xs, ys, strlen(nm)?strlen(nm)*8+20:16, 16, nm, CHECKBUTTON,	key, flag, w, f,bg,fg)
			{
				set_font(FGFONT_BUTTON);
				variable = var;
				mask = m;
				if (variable) SetTrigger(*variable&mask); else draw();
			}
		FGCheckButton(int	xs,	int	ys,	const char *nm, int key, int flag, FGWindow	*w,	int	fg,	int	bg,	ControlCall f=0, bool *var=0)
			: FGControl(xs, ys, strlen(nm)?strlen(nm)*8+20:16, 16, nm, CHECKBUTTON,	key, flag, w, f,bg,fg)
			{
				set_font(FGFONT_BUTTON);
				variable = (int *)var;
				mask = 1;
				if (variable) SetTrigger(*variable&mask); else draw();
			}
		void ChangeItem(int *v) { variable = v; SetTrigger(*v&mask); draw(); }
		// redraws object completely
		virtual	void draw(void);
};

/**
	FGPointButton represents a radio button that can be on (checked)
	or off (unchecked). It differents from FGCheckButton only visually.

	@image html point.png

	This button can be grouped into the set of mutually exclussive options
	to the user - that is, only one radio button in a set can be selected
	at a time. When the user selects a button, the previously selected
	button becomes unselected. Buttons are frequently grouped in
	a button group box (FGButtonGroup). Add the group box to the FGWindow first,
	then put the buttons into the group box.
	For example, two buttons on a FGWindow can be checked at the same time only if they are
	contained in separate containers, such as two different group boxes.

	To add this FGControl use FGWindow::AddCheckButton() .
*/
class FGPointButton :	public FGControl {
	protected:
		int		mask;
		int		* variable;
		void MoveIn(void);
		void MoveOut(void);
	public:
		FGPointButton(int m, int	xs,	int	ys,	const char *nm, int key, int flag, FGWindow	*w,	int	fg,	int	bg,	ControlCall f=0, int *var=0)
			: FGControl(xs, ys, strlen(nm)?strlen(nm)*8+20:16, 16, nm, POINTBUTTON,	key, flag, w, f,bg,fg)
			{
				set_font(FGFONT_BUTTON);
				variable = var;
				mask = m;
				if (variable) SetTrigger(*variable&mask); else draw();
			}
		void ChangeItem(int *v) { variable = v; SetTrigger(*v&mask); draw(); }
		// redraws object completely
		virtual	void draw(void);
};

/**
	The two state (like FGCheckButton) FGPushButton.
*/
class FGTwoStateButton : public FGPointButton
{
	protected:
	public:
		FGTwoStateButton(int xs, int ys, int ws, int hs, const char *nm, int key, FGWindow *wnd, int *var=0, ControlCall f=0) :
			FGPointButton(1, xs,ys,"", key, BNORMAL, wnd, 0, 15, f, var)
		{
			Resize(ws-w, hs-h);
			SetName(nm); // ma v sebe draw
		}
		// redraws object completely
		virtual	void draw(void);
};

/**
You can use this object to get any text or number from user.
When you activate such object, you can use keyboard to edit
the text (HOME, DEL, BACKSPACE, ARROWS and ESC work properly).
The input is passed to your program when the user hits
the 'ENTER' key. The ESC key pressing terminates the input without change.
You can type input line up to 127 characters long. The special shortcuts
CTRL+INS and SHIFT+INS are used to standard COPY&PASTE mechanism.

@image html editbox.png

*/
class FGEditBox :	public FGControl
{
	    //! max. size of the edited text string
	    static const int bufsize = 252;

		friend class FGUpDown;
		friend class FGWindow;

		enum dtype { EDIT_INT, EDIT_STRING, EDIT_DOUBLE};

		void init(void);
		void UpDown(double step)
		{
			if (data_type==EDIT_INT)
			{
				 *(int	*)ptr += int(step);
			}
			else if	(data_type==EDIT_DOUBLE)
			{
				 *(double *)ptr += step;
			}
			TestRange();
		}
		void TestRange(void)
		{
			if (check_range==0) return;
			if (data_type==EDIT_INT)
			{
				int	dato = *(int *)ptr;
				if (dato<min) *(int	*)ptr =	min;
				else if	(dato>max) *(int *)ptr = max;
			}
			else if	(data_type==EDIT_DOUBLE)
			{
				double dato	= *(double *)ptr;
				if (dato<mind) *(double	*)ptr =	mind;
				else if	(dato>maxd)	*(double *)ptr = maxd;
			}
		}
		int	inputproc(unsigned);
		void input(void);
		void FlushInput(bool reset=false)
		{
			if (isinput)				   // if input, force ESC
			{
				inputproc(reset ? ESC : CR);
			}
		}
		virtual	void WheelSpin(int direction)
		{
			if (direction<0) UpDown(-1);
			else UpDown(1);
			draw();
			if (fnc)
				fnc(this);
		}
	protected:
		int		w1,w2,min,max;
		dtype   data_type;
		unsigned first:1,passwd:1,check_range:1,hex:1,caps:2,isinput:1;
		double	mind,maxd;
		void  *	ptr;
		char	buf[bufsize];
		static	char clip[bufsize];
		static	int is_clip;
		int		pos, maxpos, iline, icol, offset, size;

		void MoveIn(void);
		void MoveOut(void);
		//! draw in input modet
		virtual	void draw(bool cursor, bool first_time=false);
	public:
		// redraws object completely
		virtual	void draw(void);
		virtual ~FGEditBox();
		/**
		Switch the inputbox into input mode (shows cursor and redirect all FGWindow's keypress into one).
		*/
		void ClickUp(int state=1);
		/**
		Sets password mode. All typed characters will be shown as '*'.
		*/
		void PasswdMode(int m) { passwd = m; draw(); }
		/**
		The input will parsed as a hexadecimal number (without '0x' prefix).
		*/
		void HexMode(int m) { hex = m; draw(); }
		/**
		All letters will be converted to the capitals.
		*/
		void CapsMode(int m);
		/**
		Change the data that are edited by editbox.
		You must call this methods to apropriate FGEditBox type - this is for an integer type of the object.
		*/
		void ChangeItem(int	*p)
		{
			assert(data_type==EDIT_INT);
			ptr	= p;
			draw();
		}
		/**
		Change the data that are edited by editbox.
		You must call this methods to apropriate FGEditBox type - this is for an string type of the object.
		*/
		void ChangeItem(char *p)
		{
			assert(data_type==EDIT_STRING);
			ptr= p;
			draw();
		}
		/**
		Change the data that are edited by editbox.
		You must call this methods to apropriate FGEditBox type - this is for an double type of the object.
		*/
		void ChangeItem(double *p)
		{
			assert(data_type==EDIT_DOUBLE);
			ptr	= p;
			draw();
		}
		/**
		Change the range for edited values.
		You must call this methods to apropriate FGEditBox type - this one is for a double type of the object.
		*/
		void SetRange(double mi, double ma)
		{
			assert(data_type==EDIT_DOUBLE);
			mind = mi;
            maxd = ma;
            TestRange();
			draw();
		}
		/**
		Change the range for edited values.
		You must call this methods to apropriate FGEditBox type - this one is for a int type of the object.
		*/
		void SetRange(int mi, int ma)
		{
			assert(data_type==EDIT_INT);
			min = mi;
            max = ma;
            TestRange();
			draw();
		}
		/**
		Sets a new size of the input string. Default is calculated from the objects width. You can set values from 1 to 127.
		*/
		void SetSize(int ns)
		{
			pos = offset = 0;
			if(ns>0 && ns<bufsize)
			{
				size = ns;
				draw();
			}
		}
		/**
		Override the disable methods.
		*/
		void Disable(void);
		/**
		Create an editbox for an integer number input. Don't call this directly. Rather use FGWindow::AddEditBox() methods instead.
		*/
		FGEditBox(int xs,	int ys,	int ws1, int ws2, const char *nm, int	key, FGWindow *w,	int *pt, int ink, int paper, ControlCall f, int mn, int mx, int check);
		/**
		Create an editbox for an text input. Don't call this directly. Rather use FGWindow::AddEditBox() methods instead.
		*/
		FGEditBox(int sz, int	xs,	int	ys,	int	ws1, int ws2, const char *nm,	int	key, FGWindow	*w,	char *pt, int ink, int paper, ControlCall f);
		/**
		Create an editbox for an double precission number input. Don't call this directly. Rather use FGWindow::AddEditBox() methods instead.
		*/
		FGEditBox(int xs,	int ys,	int ws1, int ws2, const char *nm, int	key, FGWindow *w,	double *pt, int	ink, int paper,	ControlCall f, double mn, double mx, int check);
};

/**
When your application performs a time-consuming operation, you can use a progress bar
to show how much of the task is completed. A progress bar displays a dotted line
that grows from left to right.

@image html progress.png

*/
class FGProgressBar : public FGControl
{
	protected:
		int		steps, value;
		int		sirka;
	public:
		void	draw(void);
		/**
		    Set new value of object in range <0,steps>
		*/
		void	FGAPI setProgress(int a) { if (a>=0 && a<=steps) value = a; draw(); }
		/**
		    Returns current value of object.
		*/
		int	FGAPI progress(void) const { return value; }
		/**
		    Creates a progress bar.
		    @param w parent FGWindow
		    @param xx x coord
		    @param yy y coord
		    @param ww width in pixels
		    @param hh height in pixels
		    @param s step
		*/
		FGProgressBar(FGWindow *w, int xx, int yy, int ww, int hh, int s);
};

/**
	Use FGBaseMenu to specify the apperance and behaviour of an item
	in a menu. Each FGWindow can contain multiple menu items.
    @see FGWindow::AddMenu().
*/
class FGBaseMenu : public	FGControl
{
	public:
		// redraws object completely
		virtual	void draw(void);
		void MoveIn(void);
		void MoveOut(void);
		virtual	void ClickUp(int a);  // if TRUE, so call handler
		FGBaseMenu(const char *nm, int key, FGWindow *w, ControlCall f, int font);
};

/**
	Use FGBaseMenuItem to specify the apperance and behaviour of an item
	in a menu. Each FGMenuWindow can contain multiple menu items.
*/
class FGBaseMenuItem : public	FGControl	{
	public:
		// redraws object completely
		virtual	void draw(void);
		void MoveIn(void);
		void MoveOut(void);
		virtual	void ClickUp(int a);  // if TRUE, so call handler
		FGBaseMenuItem(const char *nm, int key, FGMenuWindow *w, ControlCall f, int font);
};

/**
    The parent for all Sliders.
*/
class FGSlider  : public FGControl
{
	protected:
		virtual void draw(void) {}
		int		minv,maxv,*val,steps,smer; // 0 je vodorovny, 1 zvysly
		void 	frame(int,int,int,int,int f=0);
		virtual void MoveIn(void);
		virtual void MoveOut(void);
		virtual void ClickDown(void);
		virtual	void WheelSpin(int direction)
		{
			if (direction<0) StepDown();
			else if (direction>0) StepUp();
		}
		void StepDown(void)
		{
			modify(*val-minv-steps);
		}
		void StepUp(void)
		{
			modify(*val-minv+steps);
		}
	public:
		void	modify(int);
		void    redraw(void)
		{
			ControlCall tmp = fnc;
			fnc = 0;
			draw();
			fnc = tmp;
		}
		FGSlider(FGWindow *win, int xs, int ys, int ws, int hs, int minimal, int maximal, int s, int *v, ControlCall f) :
			FGControl(xs, ys, ws, hs, 0, SLIDEBAR, 0, BNORMAL, win, f), minv(0), maxv(0)
		{
			if (minimal>maximal)
			{
				minv = maximal;
				maxv = minimal;
			}
			else
			{
				minv = minimal;
				maxv = maximal;
			}
			if (*v<minv)
				*v=minv;
			if (*v>maxv)
				*v=maxv;
			val  = v;
			steps= s;
			smer = h>w;
		}
};

/**
	A horizontal FGSlider.
	@image html sliderh.png
*/
class FGSlideBarH : public FGSlider
{
// 3 - 16 - val - 3 - 38
	public:
		FGSlideBarH(int x, int y, int min, int max, int step, int *val, FGWindow *win, ControlCall f=0):
			FGSlider(win,x,y,(max-min)/step+16+12+38,16,min,max,step,val,0)
		{
			draw();
            fnc = f;
		}
		// redraws object completely
		virtual	void draw(void);
};

/**
	A vertical FGSlider.
	@image html sliderv.png
*/
class FGSlideBarV : public FGSlider
{
// 3 - 16 - val - 3 - 38
	public:
		FGSlideBarV(int x, int y, int min, int max, int step, int *val, FGWindow *win, ControlCall f=0):
			FGSlider(win,x,y,16,(max-min)/step+16+12+38,min,max,step,val,0)
		{
			draw();
            fnc = f;
		}
		virtual void draw(void);
};

/**
	It is a rectangle space at the screen that is intended to the communication
	with the user. Its look & feel is derived from the flags assigned
	on the create time. The maximal size of the window is limited
	by current video mode resolutions only. It's have own fore
	and back color, title (or no). You can use many windows
	at once (it is limited with free memory only), but only and just
	only one may be in the active state. You can draw your windows everywhere,
	but if you go out of bounds of the screen, then graphics system
	adjust the size and position at the right values. You can set
	that windows to remember its position and size. The each window is active
	immediate after it is created. To the active window goes user's input,
	key pressing, mouse events and other system events. You can serve
	these events by defining your own FGWindow Handler. It is a single
	procedure with type GuiHwnd and it is described below. Because using
	FGWindow methods to a not existing object, library has implemented
	safe mechanism (it is not mandatory but it is a good practise).
	@callgraph
	@image html wnd.png
*/
class FGWindow : public FGBaseGui
{
		friend	class FGControl;
		friend	class FGPanel;
		friend	class FGEditBox;
		friend  class FGBaseMenu;
		friend  class FGApp;

		FGWindow(FGWindow **, FGPixel);	// root window
		FGControl		*oldidm;
		int			curr_ctrl;			// default button
		int			c_menupos;
		unsigned 	first_draw:1, changed:1, nControls:16; // pocet buttonov priradenych oknu
		int			xpos, ypos;
		void 		(*handler)(FGEvent	*);
		// for locking dynamic (WindowMove(), WindowLine() ... )
		FGMutex 	StateMutex;
		void*       tabpage_list;   // list of assigned TabPages
		int			optimize;
		FGPixel		statuscolor;
		FGPixel		optimize_color;
		FGEditBox 	*ininput;
		FGControl 	*call_handler; // set if need call handler (in translateevent)
		int		line_x1, line_y1, line_x2, line_y2;
		const   char * current_tab_page;
		int			icon_x, icon_y;
		FGWindow	 	**itself;
		FGControl		*Buttony;
		FGWindow		*iconized;
		void	(*CallWindowHook)(int x, int y, int& nx, int& ny, int);

		static	FGWindow *current, *first;  // last in list of all windows
		static	int	nWnd;

		FGWindow* FGAPI OdkryteOkno(FGWindow *pokial, int x, int y, int w, int h);
		virtual int iconize(int xs, int ys);
		int		FGAPI Rozporcuj(int X1, int Y1, int X2, int Y2, int *in, int *out);
		void	FGAPI intersect(int, int, int, int);
		static  FGAPI int	_over(int, int, int, int, int xx, int yy, int ww, int hh);
		void	FGAPI Relocate(int xx, int yy);
		static 	void FGAPI DestroyWindow(void);
		static	void FGAPI nextwnd(void);
		FGEditBox* FGAPI FindNextEditBox(void);
		int	 	FGAPI TitleFind(FGEvent	*e);
		void 	FGAPI WindowUpdate(int dx, int dy, int all=0);
		FGBaseGui* FGAPI GetLastButton(void); // last in list

		void	FGAPI ResetActive(void)
		{
			call_handler = 0;
		}
		FGControl * FGAPI IsActive(void)
		{
			return call_handler;
		}
		int		FGAPI IsInInput(void) const { return !!ininput; }
		void 	FGAPI SetInInput(FGEditBox *e)
		{
			ininput = e;
		}
		void	FGAPI WindowInputChar(unsigned c)
		{
			ininput->inputproc(c);
		}
		void	FGAPI SetDefaultControl(int _id)
		{
			curr_ctrl = _id;
		}
		void	FGAPI WindowUpdateBlock(int x, int y, int w, int h);
		void	FGAPI SetNextTabPage(void);
		void	FGAPI DrawTabPages(void);
		int		FGAPI ClickTabPage(int, int);
		void	FGAPI RelocateTabPage(int xx, int yy);
		void 	FGAPI puts(const char *s) { while(*s) wputc(*s++);	}
		void	FGAPI wputc(int);
		void	FGAPI move(int x=0,	int	y=0) { wputc(-1); xpos=x; ypos=y; }
		//! internal
		int		FGAPI GetXM(int);
		//! internal
		int		FGAPI GetYM(void);
	protected:
		//! @deprecated
		void	WindowHide(void);
		//! @deprecated
		void	WindowShow(void);
		//! @deprecated
		void	WindowRecopy(void);

		FGDrawBuffer 	*icon;

		virtual void	FGAPI RepaintBlock(int x,	int	y, int w, int h);
		void	FGAPI init(void);
		void	FGAPI default_clipping(void);
		long	FGAPI WindowStatus(long status);
		virtual void WindowPiktograms(int x, int y);
		virtual void WindowDrawTitle(int color); // old virtual
	public:
		// redraws object completely
		virtual	void draw(void);
		//! internal
		void	InstallWindowHook(void(*hook)(int, int, int&, int&, int))
		{
			CallWindowHook = hook;
		}
		//! internal
		void	SetActive(FGControl *c);
		//! internal
		void	RunHandler(void);
		//! @deprecated
		static	void HideAll(void);
		//! @deprecated
		static	void ShowAll(void);
		static 	void FGAPI WindowVirtualRoot(int, int, int, int);
		//! Returns current focussed FGWindow if any.
		static	FGWindow* FGAPI GetCurrent(void) { return current; }
		static 	FGWindow* FGAPI WindowFind(FGEvent	*e);
		//! Finds a FGWindow by ID.
		static  FGWindow* FGAPI WindowFind(int idw)
		{
			return (FGWindow *) (first->FindId(idw));
		}
		void 	FGAPI BroadcastMessage(FGEvent *e);
		//! Lock the FGWindow to access from other threads.
		void	StateLock(void)
		{
			StateMutex.Lock();
		}
		//! Unlock the FGWindow to access from other threads.
		void	StateUnlock(void)
		{
			StateMutex.Unlock();
		}
		// the window notify interface
		//! Marks the FGWindow state as 'CHANGED'. Works together with notifycation on window close.
		void	FGAPI SetChange(void) { changed = 1; }
		//! Marks the FGWindow state as 'NOT CHANGED'. Works together with notifycation on window close.
		void	FGAPI ResetChange(void) { changed = 0; }
		bool	FGAPI ShowNotify(void);
		void	AbsToWindowPosition(int& _x, int& _y) { _x -= x; _y -= y; }
		//! Sets the icon position.
		void	FGAPI SetIconPosition(int x, int y) { icon_x = x; icon_y = y; }
		//! Returns true/false if the FGWindow is iconized.
		int		FGAPI IsIconized(void) const { return !!(int)iconized; }
		//! Iconize the FGWindow at its icon position or if not available, at the mouse position.
		int 	FGAPI WindowIconize(void);
		//! Iconize the FGWindow at the exact position.
		int 	FGAPI WindowIconize(int, int);
		//! Attaches the image as icon to the FGWindow.
		void 	FGAPI WindowAttachIcon(FGDrawBuffer *ico) { icon = ico; }

		//! Returns the pointer to the first FGControl in the FGWindow.
		FGControl* FGAPI GetCurrentControls(void) { return	Buttony; }

		void	FGAPI WindowRepaint(int	xr,	int	yr,	int	w, int h);
		void	FGAPI WindowRepaintUser(int	xr,	int	yr,	int	w, int h);

		FGWindow(FGWindow **, int xs, int ys, int ws, int hs, const char *nm,	GuiHwnd=0, int i=IM, int p=PM, int flag=WTITLED|WFRAMED|WCLICKABLE);
		virtual ~FGWindow();
		bool	WindowClose(void);
		/**
			Flushes the input line if one is currently edited.
			@return true if any input flushed or false if there is nothing to do.
		*/
		bool	FGAPI WindowFlushInput(void)
		{
			if (ininput)
			{
				WindowInputChar(CR);
				ininput = 0;
				return true;
			}
			return false;
		}

		FGPushButton*       AddPushButton(int xs, int ys, int ws, int hs, const char *nm, int key=0, ControlCall f=0, int flag=BNORMAL);
		FGPushButton*       AddPushButton(int xs, int ys, int ws, int hs, const char *nm, int key, FGDrawBuffer *bm, ControlCall f, int flag=BNORMAL);
		FGPushButton*       AddPushButton(int xs, int ys, int key, FGDrawBuffer *bm, ControlCall f, int flag=BNORMAL);
		FGSlideBarH*        AddSlideBarH(int xs, int ys, int min, int max, int step, int *val, ControlCall f=0);
		FGSlideBarV*        AddSlideBarV(int xs, int ys, int min, int max, int step, int *val, ControlCall f=0);
		FGCheckButton*      AddCheckButton(int	xs,	int	ys,	const char *nm, int key=0, int * variable=0, ControlCall=0, int flag=BNORMAL);
		FGCheckButton*      AddCheckButton(int	xs,	int	ys,	const char *nm, int key, bool * variable, ControlCall=0, int flag=BNORMAL);
		FGPointButton*      AddPointButton(int	xs,	int	ys,	const char *nm, int key=0, int * variable=0, ControlCall=0, int flag=BNORMAL);
		FGCheckButton*      AddCheckButtonMask(int mask, int xs, int ys,	const char *nm, int key=0, int * variable=0, ControlCall=0, int flag=BNORMAL);
		FGPointButton*      AddPointButtonMask(int mask, int xs, int ys,	const char *nm, int key=0, int * variable=0, ControlCall=0, int flag=BNORMAL);
		FGTwoStateButton*   AddTwoStateButton(int	xs,	int	ys,	int w, int h, const char *nm, int key=0, int * variable=0, ControlCall=0);
		FGEditBox*          AddEditBox(int	xs,	int	ys,	int	ws1, int ws2, const char *nm,	int	key, int *p, ControlCall f, int min, int max);
		FGEditBox*          AddEditBox(int	xs,	int	ys,	int	ws1, int ws2, const char *nm,	int	key, int *p, ControlCall f=0);
		FGEditBox*          AddEditBox(int	xs,	int	ys,	int	ws1, int ws2, const char *nm,	int	key, char *p, ControlCall f=0);
		FGEditBox*          AddEditBox(int sz, int	xs,	int	ys,	int	ws1, int ws2, const char *nm,	int	key, char *p, ControlCall f=0);
		FGEditBox*          AddEditBox(int	xs,	int	ys,	int	ws1, int ws2, const char *nm,	int	key, double *p, ControlCall f, double min, double max);
		FGEditBox*          AddEditBox(int	xs,	int	ys,	int	ws1, int ws2, const char *nm,	int	key, double *p, ControlCall f=0);
		FGBaseMenu*         AddBaseMenu(const char *nm,	int	key=0, ControlCall f=0);
		FGProgressBar*      AddProgressBar(int xs, int ys, int ws, int hs, int size);
		FGListBox* 	        AddListBox(int xs, int ys, int w, int h, int dropdown, FGListBoxCallBack drawone=0, void* userdata = 0);
		FGLabel*	        AddLabel(int x, int y, const char *s, int key=0, ControlCall f=0, unsigned  i = UNDEFINED_COLOR, unsigned p = UNDEFINED_COLOR);
   		FGPanel*	        AddPanel(int, int, int, int, const char *s=0, unsigned paper = UNDEFINED_COLOR, unsigned ink=CWHITED, unsigned ink2=CDARK);
		FGText*		        AddText(int x, int y, const char *s, unsigned ink = UNDEFINED_COLOR, unsigned paper = UNDEFINED_COLOR);
		FGImage*            AddImage(int x, int y, FGDrawBuffer* img);

		void 		FGAPI WindowText(int x, int y, const char *s, unsigned Ink = UNDEFINED_COLOR, unsigned Paper = UNDEFINED_COLOR);

		void 		FGAPI WindowText(int x, int y, const char *s, FGFontProperty * f, unsigned color, unsigned bk);
		void 		FGAPI WindowTextUTF8(int x, int y, const char *s, FGFontProperty * f, unsigned color, unsigned bk);
		void 		FGAPI WindowTextUnicode(int x, int y, const unsigned short *s, FGFontProperty * f, unsigned color, unsigned bk);

		void 		FGAPI WindowPanel(int, int, int, int, const char *s=0, int unsigned = CWHITED, unsigned ink2=CDARK);

		void 		FGAPI WindowBox(int x, int y, int a, int b, unsigned color = UNDEFINED_COLOR);

		void 		FGAPI WindowPixel(int x,	int	y, unsigned color = UNDEFINED_COLOR);

		void 		FGAPI WindowRect(int	x, int y, int a, int b, unsigned color = UNDEFINED_COLOR);

		void 		FGAPI WindowPatternRect(int	x, int y, int a, int b, FGPattern *);

		void 		FGAPI WindowLine(int	x, int y, int a, int b,	unsigned color = UNDEFINED_COLOR);
		void 		FGAPI WindowPatternLine(int	x, int y, int a, int b, FGPattern *p=&PatternDot);

		void 		FGAPI WindowDrawCircle(int x, int y, int	r, unsigned color = UNDEFINED_COLOR);
		void 		FGAPI WindowFillCircle(int x, int y, int	r, unsigned color = UNDEFINED_COLOR);

		void 		FGAPI WindowDrawEllipse(int x, int y, int	rx, int ry, unsigned color = UNDEFINED_COLOR);
		void 		FGAPI WindowFillEllipse(int x, int y, int	rx, int ry, unsigned color = UNDEFINED_COLOR);

		void 		FGAPI WindowDrawArc(int x, int y, double ang1, double ang2, int r, unsigned color = UNDEFINED_COLOR);

		FGRect 		FGAPI WindowFillPolygon(const FGPointArray& , unsigned col = UNDEFINED_COLOR);
		void 		FGAPI WindowDrawPolygon(const FGPointArray&, unsigned col = UNDEFINED_COLOR);

		FGRect 		FGAPI WindowFillTriangle(int x1, int y1, int x2, int y2, int x3, int y3, unsigned col = UNDEFINED_COLOR);

		void 		FGAPI WindowSpline(int points[8], unsigned color = UNDEFINED_COLOR);

		void 		FGAPI WindowScrollDown(int xx, int yy, int ww, int hh ,int about);
		void 		FGAPI WindowScrollUp(int xx, int yy, int ww, int hh ,int about);
		void		FGAPI WindowScrollLeft(int xx, int yy, int ww, int hh ,int about);
		void		FGAPI WindowScrollRight(int xx, int yy, int ww, int hh ,int about);

		void 		FGAPI WindowStatusBar(int, const char *s, int c=CBLACK);

		int	 		printf(const char * , ... );
		int	 		printf(int,int,const char *, ... );

		//! The helper for event sending to the FGWindow.
		void 		FGAPI SendEvent(int event, int key=0, int x=0, int y=0)
		{
			FGEvent e(event,key,x,y);
			SendToWindow(&e);
		}
		//! The helper for event sending to the FGWindow.
		void 		FGAPI SendEvent(int event, int x, int y, int ww, int hh)
		{
			FGEvent e(event,0,x,y,ww,hh);
			SendToWindow(&e);
		}
		//! Send the event to the FGWindow handler.
		void 		FGAPI SendToWindow(FGEvent *p);

		/**
			Lock the FGWindow for the immediate update to the screen. This lock disable
			the FGWindow redrawing on the screen until WindowUnlock() is called.
			Useful when you draw many small objects at once and you want optimize for speed.
			@note this is not related to threading
		*/
		void 		FGAPI WindowLock(void) {	status|=WLOCKED; }
		/**
			Unlock (and update on the screen if needed) the Locked ( WindowLock() ) FGWindow.
			@note this is not related to threading
		*/
		void 		FGAPI WindowUnLock(void);
		void 		FGAPI WindowResize(int, int);
		void 		FGAPI WindowMove(int,int);
		void 		FGAPI WindowFocus(void);
		void 		FGAPI WindowShape(int, int, int, int);
		void 		FGAPI WindowPutBitmap(int x,	int	y, int xs, int ys,	int	w, int h, FGDrawBuffer *p);
		#define		WindowCopyFrom WindowPutBitmap
		void 		FGAPI WindowGetBitmap(int x,	int	y,	int	xs,	int	ys,	int	w, int h, FGDrawBuffer *p);
		MODAL_RETURN FGAPI ShowModal(void);
		bool 		FGAPI WindowSetWorkRect(int,int,int,int);
		//! Removes all the controls from window (buttons, listoboxes, editboxes etc.).
		void 		FGAPI RemoveControls(void);
		//! Disables all the controls in the FGWindow at once.
		void		FGAPI DisableControls(void);
		//! Enables all the controls in the FGWindow at once.
		void		FGAPI EnableControls(void);
		//! Sets the FGWindow foreground color for layout and default foreground color for drawing.
		void		FGAPI SetInk(FGPixel i)
		{
			state._ink = ink = i;
		}
		//! Sets the FGWindow background color for layout and default background color for drawing.
		void		FGAPI SetPaper(FGPixel p)
		{
			state._paper = paper = p;
		}
		//! Sets the new caption for the FGWindow.
		void 		FGAPI SetName(const char *s)
		{
			FGDrawBuffer::SetName(s);

			if (this==current)
                WindowStatus(WACTIVE);
			else
                WindowStatus(WDEACTIVE);
		}
		/**
		Adds a TabPage to the FGWindow. This Tab will contens all controls added to the FGWindow before this.
		From now, you can add controls to the next Tab.
		@param name caption / id string
		@return returns false if the name is already used (and all current controls are dropped!)
		@note that the call to this function creates a tab with controls that
		are ALREADY in the window but not with controls that you will add after this!
		@see SetTabPage() GetCurrentPage()
		*/
		bool		FGAPI AddTabPage(const char *name);
		/**
		Returns current tab page name or null.
		@see SetTabPage()
		*/
		const char* FGAPI GetCurrentTabPage(void)
		{
			return current_tab_page;
		}
		/**
			Call before adding new controls.
		*/
		void PrepareForAddTabPage(void)
		{
			Buttony = 0;
		}
		/**
		Sets the new current tab page.
		@param name caption / id string
		@see GetCurrentPage()
		*/
		FGTabPage* FGAPI SetTabPage(const char *name);
		//! Removes all tab pages from the window.
		void FGAPI DeleteTabPages(void);
		//! Removes tab page from the window.
		void FGAPI DeleteTabPage(const char *name);
		//! Returns the number of attached TabPages.
		int	 GetNumberOfTabPages(void);
		// the virtual callback interface

		//! override this member when you want your own processing of KEYPRESS
		virtual void OnKeyPress(int akey) { }
		//! override this member when you want your own processing of MOUSE MOVING
		virtual void OnMouseMove(int x, int y) { }
		//! override this member when you want your own processing of MOUSE LEFT CLICK
		virtual void OnClick(int x, int y) { }
		//! overload this member when you want catch the click on middle mouse's button event.
		virtual void OnMiddleButton(int, int) { }
		//! override this member when you want your own processing of MOUSE RIGHT CLICK
		virtual void OnContextPopup(int x, int y) { }
		//! overload this member when you want catch the mouse's wheel spin.
		virtual void OnWheel(int x, int y, int delta) { }
		//! override this member when you want your own processing of REPAINT event.
		virtual void OnPaint(void) { }
		//! override this member when you want your own processing of DRAG&DROP
		virtual void OnStartDrag(int , int, int) {}
		//! override this member when you want your own processing of DRAG&DROP
		virtual void OnEndDrag(int ,int, int, int, int) {}
		//! override this member when you want your own processing of WINDOW MOVE
		virtual void OnMove(int dx, int dy) { }
		//! override this member when you want your own processing of WINDOW RESIZE
		virtual void OnResize(int dx, int dy) { }
		//! override this member when you want your own processing of WINDOW ICONIZE
		virtual void OnIconize(void) { }
		//! override this member when you want your own processing of ACCELERATORS
		virtual void OnAccelerator(FGControl *which) { }
		//! override this member when you want to handle the TabPage switch
        virtual void OnTabSwitch(const char *) { }
};

/**
	The instance of this class encapsulates the application.
	You have to declare this construct at the start of your
	program. See the "Hello World" example.
*/
class FGApp : public FGConnector
{
		static const int pathsize = 256;
		friend class FGWindow;
		int mouseflag;
		static int mx, my;
		static  int 	fulldrag;
		int mb, oldx, oldy, oldb, drag, action;
		int XX, YY;
		bool locked;

		FGInputDevice *user_input_drv;

		static int __fg_fps, __fg_fps_total;

		int mousefirst, resize, tmp_event,
			resizex, resizey,
			dragx, dragy,
			clickx,	clicky,	clickw,	clickh,
			vyrez;

		FGControl *idcb;
		FGControl *idc2;
		FGWindow *drag_okno;
		static int repeat_status, repeat_delay1, repeat_delay2;
		FGEvent *hold_event;
		static bool is_ctrl, is_alt, is_shift;

		enum { MAX_EVENT_QUEQUE=32	}; // numbers of events to app per one user event

		void blue_rect(int x, int y, int w, int h);
		void blue_rect2(int x, int y, int w, int h);
		void AutoRepeatStart(FGControl *);
		void AutoRepeatDo(void);
		void AutoRepeatEnd(void);
		void DoubleClick(int& event);

		static	FGEvent queque[MAX_EVENT_QUEQUE];
		static	void (*DelayProc)(void);
		static	void (*OnEverySecond)(int);
		static	int	quequeIndex;
		static  FGWindow *Root;
		static	MainHwnd appHandler;
		static  unsigned long ticks;
		static  int ctrl_break;
		static  int startXdrag;
		static  int startYdrag;
		static  int endXdrag;
		static  int endYdrag;
		static void (*DragShape)(int , int ,int	, int );
		static FGPixel __p4[8];
		static FGPattern __patt;
		void CallDelayProc(void)
		{
			if (DelayProc)
			{
				DelayProc();
				return;
			}
			else if (OnIdle() == true)
			{
				return ;
			}
			// wait 10 msec if nothing to do
			delay(10);
		}
		void GetUserEvent(FGEvent& e, int);
		void TranslateUserEvent(FGEvent& e);
		void Timer(void);
		int isgmkey(int	k);
	protected:
	public:
		int  RemoveMousePointer(void);
		void ShowMousePointer(void);
		int get_key(void);
		static  const FGMouseCursor *__fg_cursor;
		//! Internal flags.
		static	int		flags;
		//! Videomode number that is used for application.
		static	int		video;
		//! The color of application background.
		static	int		background;
		//! Standard C 'argc' on startup.
		static	int		Argc;
		//! Standard C 'argv' on startup.
		static	char	**Argv;
		//! Name of the application exefile.
		static	char	* name;
		//! Home directory (e.g. '/home/user' or '/root' ).
		static	char	*homedir;
		//! The application's starts directory.
		static  char	currdir[pathsize];
		//! true/false if TrueType fonts are supported.
		static	int 	ttf_support;
		/**
		Set the procedure that will be call many times per second when application will be idle.
		When the system is waiting to user input, it waits into inner loop.
		Because this time, I add to the library little hack. You can define
        some very little bit of code, procedure that will be called
		many times per second. Call with parameter 0 to switch this feature off.
		@param fnc your callback procedure of type void (*fnc)(void).
		*/
		void SetDelayProc(void (*fnc)(void))
		{
			DelayProc =	fnc;
		}
		void UpdateMousePointer(void);
		static void SaveScreen(void);
		/**
		By default, after <CTRL+C> hitting application normally shutdowns.
		You can switch this behaviour on/off by single calling this method.
		*/
		static void DisableCtrlBreak(void)
		{
			ctrl_break ^= 1;
		}
        /** To change window drawing when WindowMove event is performed.
			@param a true for opaque, false for window frame drawing only (fast).
		*/
		static void SetWindowMoveStyle(int a)
		{
			fulldrag = a;
		}
		FGApp(int m, int& argc, char **& argv, int bck, int appFlags=APP_ENABLEALTX);
		virtual ~FGApp();
		/**
		When you define the application procedure also,
		you can send to this procedure any event by this method.
		*/
		static	void FGAPI SendToApp(FGEvent	*x);
		void	FGAPI Run(MainHwnd hwnd=0);
		void	FGAPI FGYield(void);
		/**
		Call this methods, for example, from menu item "EXIT".
		When it is once called, it will causes jump out of the main
		application loop - return from Run().
		*/
		static	void AppDone(void)
		{
			FGEvent e(QUITEVENT);
			SendToApp(&e);
		}
		/**
		    Close the inner most FGApp loop only.
		*/
		static	void AppClose(void)
		{
			FGEvent e(CLOSEEVENT);
			SendToApp(&e);
		}
		/**
		This is another call-back procedure that will be call per each 't' msec.
		Your user-defined procedure expects one int parameter - number
		of second from the start of your program.
		Call with parameter 0 switch this feature off.
		NOTE! resolution is cca. 20ms
		@param p the callback routine
		@param t number of milisec. between two calls the callback (default is 1 sec.)
		*/
		void	SetTimerProc(TimerProc p, int t=1000)
		{
			OnEverySecond =	p;
			ticks = t;
		}
		/**
		   Returns pointer to the ROOT window or 0 if one is not initialized.
		   Be sure before using this function, that you add switch APP_ROOTWINDOW
		   to the FGApp constructor else program will be aborted.
		   You can draw using this pointer with standard window stuff (WindowLine (),  WindowText () etc.)
		   NOTE: You can't use any Controls into this window, it is only for drawing!
		   You can initialize one in FGApp constructor time by passing APP_ROOTWINDOW as flags.
		   @return pointer to whole screen window ROOT.
		*/
		static FGWindow * GetRootWindow(void)
		{
			return Root;
		}
		MODAL_RETURN FGAPI RunModal(FGWindow *which);
		//! Returns current mouse coordinate in X axis.
		static int	GetMouseX(void)
		{
			return mx;
		}
		//! Returns current mouse coordinate in Y axis.
		static int	GetMouseY(void)
		{
			return my;
		}
        //! @return true if the key is pressed at the time.
		static bool IsShift(void);
        //! @return true if the key is pressed at the time.
		static bool IsCtrl(void);
        //! @return true if the key is pressed at the time.
		static bool IsAlt(void);

	    static void set_mmx(void);
		static void reset_mmx(void);
		static int test_mmx(void);

		static void SetCaption(char *new_name);

		const FGMouseCursor * CursorLoad(const FGMouseCursor *cur);

		int EnableBuffering(int mode);
		int Flip(void);
		void DisableBuffering(void);

		ENUM_OS GetOS(void);

        /**
        Set variables with actual data. It is dependent on the app flag APP_MAGNIFIFIER.
		@param x position of top left corner
        @param y position of top left corner
		@param w the width of selected rectangle that is showed when you drag mouse with button hold-down.
		@param h the height of selected rectangle that is showed when you drag mouse with button hold-down.
        */
		static void	GetDragVector(int &x, int &y, int &w, int &h);
        /**
        Set procedure that will be called when you drag&drop to draw out dragging object.
        This procedures has four argument x,y for position in the current window
        and offset_y, offset_x are current offset from [x,y]. To restore default,
        use with argument 0. For more see the RAD project - source rad_prj.cpp.
        */
		static void	SetDragShape(void(*a)(int,int,int,int)=0);
		static void SetRepeatDelay(int	c1,	int	c2);

		void SetInputDevice(FGInputDevice *idev) { user_input_drv = idev; }
		void ResetInputDevice(void) { user_input_drv = 0; }

		void LockInput(void) { locked = true; }
		void UnlockInput(void) { locked = false; }

		// Object-oriented API via overloaded methods
		//! overload this member when you want catch any keypress event.
		virtual void OnKeyPress(int) { }
		//! overload this member when you want catch any mousemove event.
		virtual void OnMouseMove(int , int) { }
		//! overload this member when you want catch the click on left mouse's button event.
		virtual void OnClick(int, int) { }
		//! overload this member when you want catch the click on middle mouse's button event.
		virtual void OnMiddleButton(int, int) { }
		//! overload this member when you want catch the click on right mouse's button event.
		virtual void OnContextPopup(int, int) { }
		//! overload this member when you want catch the mouses wheel spin.
		virtual void OnWheel(int x, int y, int delta) { }
		//! overload this member when you want catch the start of mouse drag.
		virtual void OnStartDrag(int , int, int) {}
		//! overload this member when you want catch the end of mouse drag.
		virtual void OnEndDrag(int ,int, int, int, int) {}
		//! overload this member when you want catch the cursor arrives screen boundary.
		virtual void OnCursorOut(int)
		{
		}
		/**
		The wrapper for an application idle proc.
		@return If you override this on in your class and you call
		some code from it you must return true from this callback to know that the system
		will not wait. Elsewhere your system may go slowdown.
		@note if call SetDelayProc() with non-null argument then this callback will not be called!
		@see SetDelayProc()
		*/
		virtual bool OnIdle(void)
		{
			return false;	// returns true if do some work
		}
		/**
		The wrapper for application timer proc.
		@note if call SetTimerProc() with non-null first argument then this callback will not be called!
		The default time slice is 1000 msec. You can change this by call SetTimerProc() fith first argument
		null and second with number of msecs for a new timeslice.
		@param secs in secs from the start of app.
		@see SetTimerProc()
		*/
		virtual void OnTimer(int secs)
		{
		}
		//! override this member when you want to handle the TabPage switch
        virtual void OnTabSwitch(const char *) { }
};

/**
	The pointer to the current FGApp object
	@ingroup Globals
*/
extern FGApp *cApp;

/**
	Used for Pop-up menus.
	@see class FGWindow
*/
class FGMenuWindow : public FGWindow
{
		friend class FGWindow;
		friend class FGBaseMenuItem;

		static GuiHwnd Proc;
		static FGMenuWindow *	CurrentMenu;
		static void	MenuWindowHandler(FGEvent *p);
		int		offset;
		void _initmw(GuiHwnd proc)
		{
			offset = 0;
			Proc = proc;
			CurrentMenu	= this;
			set_font(FGFONT_MENU);
		}
		int	GetXM(void)	const
		{
			return 4;
		}
		int	GetYM(int a)
		{
			return (offset+=a,offset-a);
		}
	public:
		/**
		Creates pulldown menu.
		@param x x coordinate
		@param y y coordinate
		@param w a width of menu
		@param h a height of menu
		@param proc an optional window callback procedure
		*/
		FGMenuWindow(int x, int y, int w,	int	h, GuiHwnd proc=0):
			FGWindow(0,x,y,w,h,"", MenuWindowHandler, CScheme->menuwindow_fore, CScheme->menuwindow_back,WFRAMED|WCLICKABLE|WNOPICTO|WMENUWINDOWTYPE|WNODRAWFROMCONSTRUCTOR)
		{
			draw();
			_initmw(proc);
		}
		/**
		Creates pulldown menu at current mouse position.
		@param w a width of menu
		@param h a height of menu
		@param proc an optional window callback procedure
		*/
		FGMenuWindow(int w, int h, GuiHwnd proc=0);
		virtual ~FGMenuWindow()
		{
			CurrentMenu	= 0;
		}
		FGBaseMenuItem* AddMenu(char *nm,	int	key=0, ControlCall f=0);
		FGLabel*        AddLabel(char *nm, int key=0, ControlCall f=0, unsigned i = UNDEFINED_COLOR, unsigned p = UNDEFINED_COLOR);
		FGText*		    AddText(const char *s, unsigned ink = UNDEFINED_COLOR, unsigned paper = UNDEFINED_COLOR);
		FGCheckButton*  AddCheckButton(char *nm, int key=0, int * variable=0, ControlCall f=0, int flag=BNORMAL);
		FGCheckButton*  AddCheckButton(char *nm, int key=0, bool * variable=0, ControlCall f=0, int flag=BNORMAL);
		FGPointButton*  AddPointButton(char *nm, int key=0, int * variable=0, ControlCall f=0, int flag=BNORMAL);
		FGCheckButton*  AddCheckButtonMask(int, char *nm, int key=0, int * variable=0, ControlCall f=0, int flag=BNORMAL);
		FGPointButton*  AddPointButtonMask(int, char *nm, int key=0, int * variable=0, ControlCall f=0, int flag=BNORMAL);
		FGEditBox*      AddEditBox(int	ws1, int ws2, char *nm,	int	key, int *p,ControlCall f, int min, int max);
		FGEditBox*      AddEditBox(int	ws1, int ws2, char *nm,	int	key, int *p, ControlCall f=0);
		FGEditBox*      AddEditBox(int	ws1, int ws2, char *nm,	int	key, char *p, ControlCall f=0);
		FGEditBox*      AddEditBox(int sz, int	ws1, int ws2, char *nm,	int	key, char *p, ControlCall f=0);
		FGEditBox*      AddEditBox(int	ws1, int ws2, char *nm,	int	key, double	*p,	ControlCall f, double min, double max);
		FGEditBox*      AddEditBox(int	ws1, int ws2, char *nm,	int	key, double	*p,	ControlCall f=0);
		void            Separator(void);
};

/**
* Object intended for easy error dialogs.
*/
class FGDialog
{
		FGWindow* wnd;
		FGControl* ctrl;
		unsigned width;

	public:
		FGDialog(char* format, ...);
		FGDialog(unsigned ink, unsigned paper, const char* title, char* format, ...);
		~FGDialog()
		{
			delete wnd;
		}
		FGWindow* GetWindow(void) { return wnd; }
		MODAL_RETURN ShowOk(void);
		MODAL_RETURN ShowYesNo(void);
		MODAL_RETURN ShowRetryIgnoreCancel(void);
};

unsigned int CalculateCRC(unsigned StartCRC, void *Addr, unsigned Size);

#ifdef FG_NAMESPACE
}
#endif

#ifdef __BORLANDC__
#pragma option pop /* pop -a switch */
#pragma option pop /* pop -b */
#endif

#endif
