# Intel C++ 7.0 compiler
#CC	= icc
#GNU C++
CC	= g++

CFLAGS  = -g -pipe -c -O2 -fno-exceptions -march=pentium -DINTO_FGL

HDRS    = base.h _fastgl.h config.h cfg.h internal.h

LIBS	= -ldl -lm -lstdc++
INCPATH	= -I/usr/X11R6/include -I/usr/include/freetype2
LIBPATH	= -L/usr/X11R6/lib

OBJS	=   fastgl.o base.o cfg.o resource.o widgets.o user.o linkeyb.o dll.o linux.o editor.o bitmap.o timer.o drivers.o detect.o listbox.o pane.o cfgstore.o fgmatrix.o fgcolor.o
LOBJS	=   fastgl.lo base.lo cfg.lo resource.lo widgets.lo user.lo linkeyb.lo dll.lo linux.lo editor.lo bitmap.lo timer.lo drivers.lo detect.lo listbox.lo pane.lo cfgstore.lo fgmatrix.lo fgcolor.lo
BASE	=   /usr
INSTDIR =   $(BASE)/include/fastgl


ifdef X11SUPPORT
OPTIONS0	= -DX11_DRIVER
endif

ifdef DGASUPPORT
OPTIONS1	= -DDGA_DRIVER
endif

ifdef PNGSUPPORT
OPTIONS2	= -DFG_PNG
endif

ifdef JPGSUPPORT
OPTIONS3	= -DFG_JPEG
endif

ifdef TTFSUPPORT
OPTIONS4	= -DFG_TTF
OBJS2		= fgttf.o
LOBJS2		= fgttf.lo
SRCS2		= fgttf.cpp
endif

ifdef THREADSUPPORT
OPTIONS5	= -DFG_THREADED
endif

ifdef TIFFSUPPORT
OPTIONS6	= -DFG_TIFF
endif

CFLAG	= $(CFLAGS) $(COLORDEPTH) $(OPTIONS0) $(OPTIONS1) $(OPTIONS2) $(OPTIONS3) $(OPTIONS4) $(OPTIONS5) $(OPTIONS6) $(INCPATH)

all: $(STARGET) $(DTARGET).4 instd insts insth 
    	
alls: $(STARGET) insts insth

insth:
	-rm -Rf /usr/local/include/fastgl
	-rm -f /usr/include/fastgl/*.h
	install -d -g root -o root -m 755 $(INSTDIR)
	install -g root -o root -m 0644 fastgl.h $(INSTDIR)
	install -g root -o root -m 0644 base.h $(INSTDIR)
	install -g root -o root -m 0644 cfg.h $(INSTDIR)
	install -g root -o root -m 0644 widgets.h $(INSTDIR)
	install -g root -o root -m 0644 pane.h $(INSTDIR)
	install -g root -o root -m 0644 fgmatrix.h $(INSTDIR)
	install -g root -o root -m 0644 fgcolor.h $(INSTDIR)
	install -g root -o root -m 0644 cfgstore.h $(INSTDIR)
	install -g root -o root -m 0644 internal.h $(INSTDIR)
	install -g root -o root -m 0644 listbox.h $(INSTDIR)
	install -g root -o root -m 0644 fglmesa.h $(INSTDIR)

instd:
	-rm -f /usr/local/lib/$(DTARGET).4
	-rm -f /usr/lib/$(DTARGET).3
	install -d -g root -o root -m 755 $(BASE)/lib
	install -g root -o root -m 0755 $(DTARGET).4 /usr/lib
	ldconfig

insts:
	-rm -f /usr/local/lib/$(STARGET)
	install -d -g root -o root -m 755 $(BASE)/lib
	install -g root -o root -m 0644 $(STARGET) $(BASE)/lib

clean:
	rm -f *.a	
	rm -f *.o *.lo
	rm -f *.obj
	rm -f *.lib
	rm -f *.err
	rm -f *.so.?
	rm -f *.tds *.rc *.dsk *.desk core *.exe *.gch

$(STARGET): $(OBJS) $(OBJS2)
	ar rcs $(STARGET) $(OBJS) $(OBJS2) 

$(DTARGET).4: $(LOBJS) $(LOBJS2)
	$(CC) -g -fnostdinc -shared -Wl,-soname,$(DTARGET) -o $(DTARGET).4 $(LOBJS) $(LOBJS2) $(LIBS) $(LIBPATH) $(X11SUPPORT) $(DGASUPPORT) $(JPGSUPPORT) $(PNGSUPPORT) $(TTFSUPPORT) $(THREADSUPPORT) $(TIFFSUPPORT)

fastgl.o:  $(HDRS)     fastgl.cpp fastgl.h
	$(CC)  $(CFLAG)  fastgl.cpp 

bitmap.o:  $(HDRS)     bitmap.cpp fastgl.h
	$(CC)  $(CFLAG)  bitmap.cpp

widgets.o:  $(HDRS)     widgets.cpp widgets.h fastgl.h
	$(CC)  $(CFLAG)  widgets.cpp

editor.o:  $(HDRS)     editor.cpp widgets.h fastgl.h
	$(CC)  $(CFLAG)  editor.cpp

linux.o:  $(HDRS)     linux.cpp linux.h
	$(CC)  $(CFLAG)  linux.cpp

drivers.o:  $(HDRS)     drivers.cpp linux.h 
	$(CC)  $(CFLAG)  drivers.cpp

detect.o:  $(HDRS)     detect.cpp
	$(CC)  $(CFLAG)  detect.cpp

base.o:  $(HDRS)   base.cpp
	$(CC)  $(CFLAG)  base.cpp 

cfg.o:  $(HDRS)   cfg.cpp
	$(CC)  $(CFLAG)  cfg.cpp 

dll.o:  $(HDRS)   dll.cpp dll.h
	$(CC)  $(CFLAG)  dll.cpp 

user.o:  $(HDRS)   user.cpp fastgl.h
	$(CC)  $(CFLAG)  user.cpp

timer.o:  $(HDRS)   timer.cpp fastgl.h
	$(CC)  $(CFLAG)  timer.cpp

resource.o:  $(HDRS)   resource.cpp
	$(CC)  $(CFLAG)  resource.cpp

linkeyb.o:  linkeyb.h  linkeyb.cpp
	$(CC)  $(CFLAG)  linkeyb.cpp

fgstring.o:  fgstring.h fgstring.cpp
	$(CC)  $(CFLAG)  fgstring.cpp

listbox.o:  listbox.h  listbox.cpp
	$(CC)  $(CFLAG)  listbox.cpp

fgttf.o:  fgttf.h  fgttf.cpp
	$(CC)  $(CFLAG)  fgttf.cpp

cfgstore.o:  $(HDRS)   cfgstore.cpp
	$(CC)  $(CFLAG)  cfgstore.cpp

pane.o:  $(HDRS)   pane.cpp cfgstore.h
	$(CC)  $(CFLAG)  pane.cpp

fgmatrix.o:  $(HDRS)   fgmatrix.cpp fgmatrix.h base.h
	$(CC)  $(CFLAG)  fgmatrix.cpp

fgcolor.o:  $(HDRS)   fgcolor.cpp fgcolor.h base.h
	$(CC)  $(CFLAG)  fgcolor.cpp


fastgl.lo:  $(HDRS)     fastgl.cpp fastgl.h
	$(CC)  $(CFLAG) -fPIC -DPIC -o $@    fastgl.cpp

bitmap.lo:  $(HDRS)     bitmap.cpp fastgl.h
	$(CC)  $(CFLAG) -fPIC -DPIC -o $@    bitmap.cpp

widgets.lo:  $(HDRS)     widgets.cpp widgets.h fastgl.h
	$(CC)  $(CFLAG) -fPIC -DPIC -o $@    widgets.cpp

editor.lo:  $(HDRS)     editor.cpp widgets.h fastgl.h
	$(CC)  $(CFLAG) -fPIC -DPIC -o $@    editor.cpp

linux.lo:  $(HDRS)     linux.cpp linux.h
	$(CC)  $(CFLAG) -fPIC -DPIC -o $@    linux.cpp

drivers.lo:  $(HDRS)     drivers.cpp linux.h 
	$(CC)  $(CFLAG) -fPIC -DPIC -o $@    drivers.cpp

detect.lo:  $(HDRS)     detect.cpp
	$(CC)  $(CFLAG) -fPIC -DPIC -o $@    detect.cpp

base.lo:  $(HDRS)   base.cpp
	$(CC)  $(CFLAG) -fPIC -DPIC -o $@    base.cpp 

cfg.lo:  $(HDRS)   cfg.cpp
	$(CC)  $(CFLAG) -fPIC -DPIC -o $@    cfg.cpp 

dll.lo:  $(HDRS)   dll.cpp dll.h
	$(CC)  $(CFLAG) -fPIC -DPIC -o $@    dll.cpp 

user.lo:  $(HDRS)   user.cpp fastgl.h
	$(CC)  $(CFLAG) -fPIC -DPIC -o $@    user.cpp

timer.lo:  $(HDRS)   timer.cpp fastgl.h
	$(CC)  $(CFLAG) -fPIC -DPIC -o $@    timer.cpp

resource.lo:  $(HDRS)   resource.cpp
	$(CC)  $(CFLAG) -fPIC -DPIC -o $@    resource.cpp

linkeyb.lo:  linkeyb.h  linkeyb.cpp
	$(CC)  $(CFLAG) -fPIC -DPIC -o $@    linkeyb.cpp

fgstring.lo:  fgstring.h fgstring.cpp
	$(CC)  $(CFLAG) -fPIC -DPIC -o $@    fgstring.cpp

listbox.lo:  listbox.h  listbox.cpp base.h fastgl.h
	$(CC)  $(CFLAG) -fPIC -DPIC -o $@    listbox.cpp

fgttf.lo:  fgttf.h  fgttf.cpp base.h fastgl.h
	$(CC)  $(CFLAG) -fPIC -DPIC -o $@    fgttf.cpp

cfgstore.lo:  $(HDRS)   cfgstore.cpp
	$(CC)  $(CFLAG) -fPIC -DPIC -o $@   cfgstore.cpp

pane.lo:  $(HDRS)   pane.cpp cfgstore.h
	$(CC)  $(CFLAG) -fPIC -DPIC -o $@   pane.cpp

fgmatrix.lo:  $(HDRS)   fgmatrix.cpp fgmatrix.h base.h
	$(CC)  $(CFLAG) -fPIC -DPIC -o $@   fgmatrix.cpp

fgcolor.lo:  $(HDRS)   fgcolor.cpp fgcolor.h base.h
	$(CC)  $(CFLAG) -fPIC -DPIC -o $@   fgcolor.cpp

