/*
    OpenGUI - Drawing & Windowing library

    Copyright (C) 1996,2004  Marian Krivos

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
	License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    nezmar@atlas.sk
	Get & translate user events


  $Id: user.cpp,v 1.14 2004/09/28 08:20:41 majo Exp $

  $Log: user.cpp,v $
  Revision 1.14  2004/09/28 08:20:41  majo
  Calibration stuff

  Revision 1.13  2004/09/27 13:26:06  majo
  cal4

  Revision 1.12  2004/09/27 12:53:45  majo
  calibration 4

  Revision 1.11  2004/09/27 10:53:42  majo
  calibrating part III.

  Revision 1.10  2004/09/27 09:45:22  majo
  Calibration part II.

  Revision 1.9  2004/09/27 07:01:17  majo
  *

  Revision 1.8  2004/09/27 06:59:57  majo
  SetParameter(void *)
  Calibration part I.

  Revision 1.7  2004/08/25 21:16:10  majo
  priebezne zmeny

  Revision 1.6  2004/04/16 08:56:26  majo

  fixed error with MOVEEVENT when mouse dragging was performed

  Revision 1.5  2004/03/18 16:05:27  majo
  readded changes to branch

  Revision 1.4  2004/03/03 20:12:39  majo
  added class FGColor
  fixed Solaris compiling

  Revision 1.3  2004/02/26 21:41:29  majo
  class FGInternalState
  char* -> const char*
  int ink = -1  -> unsigned ink = UNDEFINED_COLOR

  Revision 1.2  2004/02/23 20:08:01  majo
  all classes are with prefix FG* on now
  polygon functions uses FGPointArray from now
  class GuiEvent is renamed to FGEvent
  some by parameters overloaded methods was removed (class FGWindow)
  many other small changes

  Revision 1.1.1.1  2004/01/22 16:25:08  majo
  initial release

*/

/* TODO :
	Zakladna schema je vygenerovat nejaku udalost, tu posla� oknu,
	pripadne i aplikacii via handler ak nejaky existuje.
	Nova schema by mala dodrzat staru a pridat pretazenie
	urcitych definovanych sprav. Tieto by mali rozne
	parametre. Pretazenie by zaroven odstavilo posielanie
	konkretnej spravy via handler. Podobne by sa mal spravat
	i aplikacny handler.
*/

#include <signal.h>

#if defined( __linux__) || defined( __rtems__) || defined(__sun__)
#include <sched.h>
#ifdef __rtems__
#include <rtems/kd.h>
#else
#include <sys/kd.h>
#endif
#include <sys/ioctl.h>
#endif

#include "config.h"
#include "fastgl.h"
#include "_fastgl.h"
#include "drivers.h"

#ifdef FG_NAMESPACE
namespace fgl {
#endif

#ifdef _WIN32
extern void sched_yield(void);
#endif

#define		REPEAT_DELAY1			22
#define		REPEAT_DELAY2			4
#define		DRAG_TRESHOLD		    3

int FGApp::fulldrag = 1;
int FGApp::startXdrag =	0;
int FGApp::startYdrag =	0;
int FGApp::endXdrag	= 0;
int FGApp::endYdrag	= 0;
int FGApp::mx=0, FGApp::my=0;
int FGApp::__fg_fps=0, FGApp::__fg_fps_total=0;
const FGMouseCursor * FGApp::__fg_cursor = idc_normal;
int FGApp::repeat_status=0, FGApp::repeat_delay1=REPEAT_DELAY1, FGApp::repeat_delay2=REPEAT_DELAY2;
static FGPixel m_save[MSZ*MSZ];
void (*FGApp::DragShape)(int , int ,int	, int ) = 0;
int FGApp::ctrl_break=0;
FGPixel FGApp::__p4[8] = {0, 0, 0, 0, 1, 1, 1, 1};
FGPattern FGApp::__patt(__p4, 8, 1);

const char *FGEvent::eventName[]=
{
	  "NOEVENT", "KEYEVENT", "MOUSEEVENT", "QUITEVENT", "ACCELEVENT",
	  "TERMINATEEVENT", "MOVEEVENT", "CLICKRIGHTEVENT",
	  "CLICKLEFTEVENT", "CLICKLMIDDLEEVENT", "DRAGLEFTEVENT", "DRAGRIGHTEVENT",
	  "GETFOCUSEVENT", "LOSTFOCUSEVENT", "INITEVENT",
	  "BUTTONHOLDEVENT", "WINDOWRESIZEEVENT","WINDOWMOVEEVENT",
	  "STARTDRAGLEFTEVENT", "STARTDRAGRIGHTEVENT", "REPAINTEVENT",
	  "NOTIFYEVENT", "ICONIZEEVENT", "CURSOROUTEVENT", "DBLCLICKRIGHTEVENT",
	  "DBLCLICKLEFTEVENT", "DBLCLICKMIDDLEEVENT", "MOUSEWHEELEVENT", "TABSWITCHEVENT",
// must by at the end
	  "NONSENSE"
};

//! Set caption when application running in windowed mode (X11 or WIN32 GDI).
void FGApp::SetCaption(char *new_name)
{
	__fg_driver->SetCaption(new_name);
}


/**
	Save the whole screen to the image file in the current directory,
	name is created as current time in HEX form.
	This function is called by application automagically when
	you press PRINT_SCREEN key.
*/
void FGApp::SaveScreen(void)
{
	FGBitmap *a =	new FGBitmap(X_width, Y_width);
	char s[20];
	time_t num = time(0);
	get_block(0,0,X_width,Y_width,a->GetArray());
	sprintf(s,"%x",(unsigned int)num);
	if (fgstate.verbose) printf("Bitmap '%s' saved\n", s);
	a->BitmapSave(s, true);
	delete a;
}

/**
	Use this for fine calibration for HOLD_EVENT event.
*/
void FGApp::SetRepeatDelay(int c1, int c2)
{
	if (c1>0 &&	c1<2000)
		repeat_delay1 = c1;
	if (c2>0 &&	c2<500)
		repeat_delay2 = c2;
}

/**
	Test keyboard immediately.
	Try peek any KEY from the INPUT and	return 0 if none, or its keycode.
*/
int FGApp::get_key(void)
{
	FGEvent ev(0);
	GetUserEvent(ev, 1);
	if (ev.GetType()==KEYEVENT)
		return ev.GetKey();
	else return 0;
}

// start autorepeat routine
void FGApp::AutoRepeatStart(FGControl *id)
{
	hold_event->key=id->GetId();
	hold_event->accel=id;
	hold_event->guiType=id->GetType(); // od buttonu typ ...
	repeat_status =	1;
	AutoRepeatDo();
	repeat_status =	REPEAT_DELAY1;
}

// end autorepeat routine
void FGApp::AutoRepeatEnd(void)
{
	repeat_status =	0;
}

// process -||-
void FGApp::AutoRepeatDo(void)
{
	if (--repeat_status	<= 0)
	{
		FGWindow::GetCurrent()->SendToWindow(hold_event);
		repeat_status =	REPEAT_DELAY2;
	}
}

/**
	Return coordinates of last DRAG&DROP mouse operation.
*/
void FGApp::GetDragVector(int &a, int &b, int &c, int &d)
{
	a =	startXdrag;
	b =	startYdrag;
	c =	endXdrag;
	d =	endYdrag;
}

/**
	Override current shape that is drawed on the screen
	when D&D operation is performed.
*/
void FGApp::SetDragShape(void (*fnc)(int, int, int, int))
{
	DragShape = fnc;
}

void FGApp::GetUserEvent(FGEvent& EVENT, int waits)
{
	int	xx2, yy2, rc = 0;
	xx2	= XX;
	yy2	= YY;

	EVENT.type = 0;

	if (mouseflag == 1)
		vector_mouse_cursor(XX, YY, TRUE, mouseflag);

	for (; waits; waits--)
	{
		if (locked == false)
		{
			rc = __fg_driver->get_event(EVENT.type,EVENT.key,EVENT.x,EVENT.y,EVENT.buttons);
			// try extra input if available
			if (rc == 0)
				if (user_input_drv)
				{
					rc = user_input_drv->GetInputEvent(EVENT.type,EVENT.key,EVENT.x,EVENT.y,EVENT.buttons);
					if (rc)
						user_input_drv->Scale(EVENT.x, EVENT.y);
//printf("(%d) t = %d, x = %d, y = %d, butt = %d\n", rc, EVENT.type, EVENT.x, EVENT.y, EVENT.buttons);
				}
		}

#ifdef FG_MOUSEKEYS
		if (EVENT.type == KEYEVENT)
		{
			static int last=0, akcel=2, cbut=0;
			int doit=0;

			if (last != EVENT.key)	// reset accel
				akcel = 2;

			switch(EVENT.key)
			{
				case FG_MOUSEKEYS_UP:
					EVENT.type = MOUSEEVENT;
					EVENT.y = YY-accel;
					EVENT.x = XX;
					doit = 1;
					break;
				case FG_MOUSEKEYS_DOWN:
					EVENT.type = MOUSEEVENT;
					EVENT.y = YY+accel;
					EVENT.x = XX;
					doit = 1;
					break;
				case FG_MOUSEKEYS_LEFT:
					EVENT.type = MOUSEEVENT;
					EVENT.x = XX-accel;
					EVENT.y = YY;
					doit = 1;
					break;
				case FG_MOUSEKEYS_RIGHT:
					EVENT.type = MOUSEEVENT;
					EVENT.x = xx+accel;
					EVENT.y = YY;
					doit = 1;
					break;
				case FG_MOUSEKEYS_BUTTON1:
					EVENT.type = MOUSEEVENT;
					EVENT.x = XX;
					EVENT.y = YY;
					cbut ^= 1;
					doit = 1;
					break;
				case FG_MOUSEKEYS_BUTTON2:
					type = MOUSEEVENT;
					EVENT.x = XX;
					EVENT.y = YY;
					cbut ^= 2;
					doit = 1;
					break;
			}
			if (doit)
			{
				if (EVENT.akcel<16)		// max accel diff
					EVENT.akcel += 2;
				// round up co-ordinates
				if (EVENT.x<0)
					EVENT.x = 0;
				if (EVENT.x>(X_width - 1))
					EVENT.x = (X_width - 1);
				if (EVENT.y<0)
					EVENT.y = 0;
				if (EVENT.y>Y_max)
					EVENT.y = Y_max;
				EVENT.buttons = cbut;
			}
			last = EVENT.key;
		}
#endif
		if (rc==0)
		{
			sched_yield();
		}
		else
		{
			if (EVENT.type==KEYEVENT && EVENT.key==PRTRSC)
			{
				SaveScreen();
				continue;
			} // PrtScr
#if defined(__linux__) || defined(__rtems__) || defined(_WIN32)
			else if (EVENT.type==KEYEVENT && EVENT.key==0x0003 && ctrl_break==0)
			{
				raise(SIGINT);
				exit(2);
			}
#endif // LINUX
			XX = EVENT.x;
			YY = EVENT.y;
			break;
		}
		if (waits != 1)
		{
			FGApp::CallDelayProc();			// call delay handler
			FGApp::Timer();                   // call timer handler
			if (repeat_status>0)
				AutoRepeatDo();
		}
	}
//	printf("typ %d, key %x, x %d, y %d b %d\n", EVENT.type,EVENT.key,EVENT.x,EVENT.y,EVENT.buttons);
	if (mouseflag == 1)
		vector_mouse_cursor(xx2, yy2,FALSE, mouseflag);
	if (EVENT.type ==	KEYEVENT)
		mouseflag =	0;
	else
		mouseflag =	1;
}

/**
	Redraw mouse pointer if needed.
*/
void FGApp::UpdateMousePointer(void)
{
	if (mouseflag == 1)
		vector_mouse_cursor(mx, my, TRUE, mouseflag);
}

/**
	Hide mouse pointer if needed.
*/
int FGApp::RemoveMousePointer(void)
{
	if (mouseflag == 1)
		vector_mouse_cursor(mx, my, FALSE, mouseflag);
	mouseflag =	0;
	return mouseflag;
}

/**
	Redraw mouse pointer unconditionally.
*/
void FGApp::ShowMousePointer(void)
{
	mouseflag =	1;
	vector_mouse_cursor(mx, my, TRUE, mouseflag);
}

void FGApp::blue_rect(int x, int y,	int	w, int h)
{
	draw_pattern_box(x,y, w-1, h-1, &__patt, _GNOT);
}

void FGApp::blue_rect2(int x, int y, int w, int h)
{
	if (!DragShape) blue_rect(x,y,w,h);
	else DragShape(x,y,w,h);
}

int FGApp::isgmkey(int k)
{
	FGWindow *c =	FGWindow::GetCurrent();

	switch (k)
	{
		case CTRL_TAB:			// ctrl+TAB
	 		if (c) FGWindow::nextwnd();
			return 1;
		case ALT_X:			// alt+X
			if (flags	& APP_ENABLEALTX)
			{
				AppDone();
				return 1;
			}
			else return	0;
		case TAB:
			if (c) if (c->GetStatus() & WNOTEBOOK)
			{
				c->SetNextTabPage();
			}
			break;
		case ESC:
			if (c) if (!(c->GetStatus() & WESCAPE))
				break;
		case ALT_F04:			// alt+F4
			if (c)
				c->WindowClose();
			return 1;
		case ALT_F02:			// alt+F12
			if (c)
			{
				if (c->FGWindow::WindowIconize()) return 1;
			}
			return 0;
	}
	return 0;
}

void FGApp::DoubleClick(int& event)
{
	static int last=0, tmout=0;
	int now = FGClock();
	if (event == last && (now-tmout) <= FGDBLCLICKTIME)
		event += (DBLCLICKLEFTEVENT-CLICKLEFTEVENT);
	else
	{
		last = event;
		tmout = now;
	}
}

// event handling core
void FGApp::TranslateUserEvent(FGEvent& EVENT)
{
	FGWindow *n =	0, *curr = FGWindow::GetCurrent(), *ecurr=0, *idw=0;
	FGControl *idm=0;
	int	magn = flags&APP_MAGNIFIER;
	FGEditBox *ebox=0;
	FGEvent original(EVENT);
	EVENT.guiType = 0;

	FGEvent app(NOEVENT); // events to  application

	if (curr)
		switch (EVENT.type)
		{
			case KEYEVENT:
				if (curr && !curr->IsInInput())	//ai nieje input
				{
					if (isgmkey(EVENT.key))
						return ;
					idc2 = FGControl::ButtonFind(&EVENT, 7);
					if (curr->IsVisible() && idc2)	//tak zisti buttony
					{
						idc2->ClickDown();
						EVENT.type = ACCELEVENT;	//bude to sprava
						EVENT.guiType	= idc2->GetType(); // od buttonu typ ...
						EVENT.accel = idc2;
						EVENT.key	= idc2->GetId();
						delay(100);
						Puk();
						idc2->ClickUp(TRUE);
						curr = FGWindow::GetCurrent();
					}
					else // plain KEY - pass to AppProc
					{
						app.type = KEYEVENT;
						app.key  = EVENT.key;
					}
				}
				else
				{
					if (EVENT.key==TAB)
					{
						ebox = curr->FindNextEditBox();
						ecurr = curr;
						if (curr)
							curr->WindowInputChar(CR);
					}
					else
						if (curr)
							curr->WindowInputChar(EVENT.key);
					EVENT.type = NOEVENT;
				}
				break;
			case MOUSEEVENT:
				if (mousefirst)
				{
					mousefirst = 0;
					mx = oldx =	EVENT.x;
					my = oldy =	EVENT.y;
					mb = oldb =	EVENT.buttons;
				}
				mx = EVENT.x;
				my = EVENT.y;
				mb = EVENT.buttons;
				tmp_event = EVENT.type;
				EVENT.type = NOEVENT;
				// second phase for windowresize
				if (resize && ((mb & 1)	== 1))
				{
					if (fulldrag && !(curr->GetStatus()&WFASTMOVE))
					{
						curr->WindowResize(mx -	oldx, my - oldy);
					}
					else
					{
						blue_rect(curr->GetX(),	curr->GetY(), curr->GetW() + resizex, curr->GetH() + resizey);
						resizex	+= (mx - oldx);
						resizey	+= (my - oldy);
						blue_rect(curr->GetX(),	curr->GetY(), curr->GetW() + resizex, curr->GetH() + resizey);
					}
					resize++;
					break;
				}
				// last phase for windowresize
				else if	(resize	&& (!mb))
				{
					if (!fulldrag || curr->GetStatus()&WFASTMOVE)
					{
						blue_rect(curr->GetX(),	curr->GetY(), curr->GetW() + resizex, curr->GetH() + resizey);
						if (resizex	+ resizey)
						{
							curr->WindowResize(resizex,	resizey);
						}
					}
					resize = 0;
				}
				// second phase for windowmove
				else if	(drag && ((mb &	1) == 1))
				{
					if (fulldrag && !(curr->GetStatus()&WFASTMOVE))
					{
						curr->WindowMove(mx	- oldx,	my - oldy);
					}
					else
					{
						blue_rect(curr->GetX() + dragx,	curr->GetY() + dragy, curr->GetW(),	curr->GetH());
						dragx += (mx - oldx);
						dragy += (my - oldy);
						blue_rect(curr->GetX() + dragx,	curr->GetY() + dragy, curr->GetW(),	curr->GetH());
					}
					drag++;
				}
				// last phase for windowmove
				else if	(drag && (!mb))
				{
					if (!fulldrag || curr->GetStatus()&WFASTMOVE)
					{
						blue_rect(curr->GetX() + dragx,	curr->GetY() + dragy, curr->GetW(),	curr->GetH());
						if (dragx +	dragy)
							curr->WindowMove(dragx,	dragy);
					}
					drag = 0;
				}
				// release mousebutton on control
				else if	(action	&& (!mb))
				{
					action = 0;
					idcb->ClickUp(TRUE);
					AutoRepeatEnd();
					curr = FGWindow::GetCurrent();
					if (curr->IsVisible() && ((idc2->ButtonFind(&EVENT, 3)) == idcb))
					{
						EVENT.type = ACCELEVENT;	//bude to sprava
						EVENT.guiType	= idcb->GetType();
						EVENT.accel = idcb;
					}
					else
						EVENT.type = NOEVENT;
					EVENT.key	= idcb->GetId();
					break;
				}
				// push mousebutton on control
				else if	(action	&& ((mb	& 1) ==	1))
				{
					int tt = idcb->GetType();
					if (curr->IsVisible() && ((idc2	= FGControl::ButtonFind(&EVENT,3)) != idcb) && tt!=FGControl::SLIDEBAR)
					{
						action = 0;
						AutoRepeatEnd();
						if (tt!=FGControl::POINTBUTTON && tt!=FGControl::CHECKBUTTON)
							idcb->ClickUp(FALSE);	// nevolat handler !!!!
						break;
					}
					else
					{
						idcb->ClickDown();
					}
				}
				else if	(vyrez && (!mb)) // vyrez end
				{
					if (drag_okno) // right click to wnd
					{
						startXdrag = clickx	- curr->GetX() - curr->GetXW();
						startYdrag = clicky	- curr->GetY() - curr->GetYW();
						endXdrag = clickw;
						endYdrag = clickh;
					}
					else
					{
						startXdrag = clickx;
						startYdrag = clicky;
						endXdrag = clickw;
						endYdrag = clickh;
					}
					if (vyrez > DRAG_TRESHOLD+1)
						if (magn) blue_rect2(clickx, clicky, clickw, clickh);
					if (clickw<0) {	clickx += clickw; clickw = -clickw;	}
					if (clickh<0) {	clicky += clickh; clickh = -clickh;	}
					if (drag_okno) // right click to wnd
					{
						if (vyrez >DRAG_TRESHOLD && oldb == 2)
							EVENT.type = DRAGRIGHTEVENT;
						else if	(vyrez <= DRAG_TRESHOLD	&& oldb	== 2)
						{
							DoubleClick(EVENT.type = CLICKRIGHTEVENT);
						}
						else if	(vyrez > DRAG_TRESHOLD && oldb == 1)
						{
							EVENT.type =	DRAGLEFTEVENT;
						}
						else if	(vyrez <= DRAG_TRESHOLD	&& oldb	== 1)
						{
							DoubleClick(EVENT.type = CLICKLEFTEVENT);
						}
						EVENT.w =	clickw;
						EVENT.h =	clickh;
						curr = drag_okno; // redirect it to right window
						EVENT.x =	clickx - curr->GetX() - curr->GetXW();
						EVENT.y =	clicky - curr->GetY() - curr->GetYW();

						app.type = EVENT.type;
						app.x =	clickx;
						app.y =	clicky;
						app.w =	clickw;
						app.h =	clickh;
					}
					else  // right click to background
					{
						if (vyrez > DRAG_TRESHOLD && oldb == 2)	app.type = DRAGRIGHTEVENT;
						else if	(vyrez <= DRAG_TRESHOLD && oldb	== 2)
						{
							DoubleClick(app.type = CLICKRIGHTEVENT);

						}
						else if	(vyrez > DRAG_TRESHOLD && oldb == 1) app.type =	DRAGLEFTEVENT;
						else if	(vyrez <= DRAG_TRESHOLD && oldb	== 1)
						{
							DoubleClick(app.type = CLICKLEFTEVENT);
						}
						app.x =	clickx;
						app.y =	clicky;
						app.w =	clickw;
						app.h =	clickh;
					}
					vyrez =	0;
					break;
				}
				else if	(vyrez && (mb &	3))	// vyrez continue
				{
					if (vyrez == DRAG_TRESHOLD+1)	if (magn) blue_rect2(clickx,	clicky,	clickw,	clickh);
					if (vyrez > DRAG_TRESHOLD) if (magn) blue_rect2(clickx, clicky, clickw, clickh);
					clickw += (mx -	oldx);
					clickh += (my -	oldy);
					if (vyrez > DRAG_TRESHOLD) if (magn) blue_rect2(clickx, clicky, clickw, clickh);
					vyrez++;
					if (vyrez == DRAG_TRESHOLD+1)
					{
						int type2 =	mb==2?STARTDRAGRIGHTEVENT:STARTDRAGLEFTEVENT;
						if (drag_okno)
						{
							EVENT.type = type2;
							EVENT.x =	clickx - curr->GetX() - curr->GetXW();
							EVENT.y =	clicky - curr->GetY() - curr->GetYW();
						}
// 29.7.2004 send to App too!
//						else
						{
							app.type = type2;
							app.x = clickx;
							app.y = clicky;
						}
					}
					else // mouse move only!
					{
						// when ptr is over the window, its receive relative pos
						// app proc receive EACH movement as absolute pos
						app.x = EVENT.x;
						app.y = EVENT.y;
						app.key = 0;

						if (drag_okno)
						{
							EVENT.type = MOVEEVENT;
							EVENT.x = EVENT.x - curr->GetX() - curr->GetXW();
							EVENT.y = EVENT.y - curr->GetY() - curr->GetYW();
//printf("%d %d [%d %d %d]\n", EVENT.x, EVENT.y, app.x, curr->GetX(), curr->GetXW());
						}

						app.type = MOVEEVENT;
					}
				}
				// most first push button - what next?
				else if	((!action) && (!drag) && (!vyrez) && (!resize) && (mb == 1))
				{
					int	t;

					t =	curr->TitleFind(&EVENT);
					if (curr->IsVisible()
					   && ((idcb = FGControl::ButtonFind(&EVENT, 7)) !=	0) && t	!= 4)
					{
						action = 1;
						idcb->ClickDown();
						AutoRepeatStart(idcb);
					}
					else if	(t)
					{
						// windowmove start
						if (t == 1 && (!oldb) && !(curr->GetStatus() & WUNMOVED))
						{
							drag = 1;
							dragx =	dragy =	0;
							if (!fulldrag || curr->GetStatus()&WFASTMOVE)
							{
								blue_rect(curr->GetX() + dragx,	curr->GetY() + dragy, curr->GetW(),	curr->GetH());
							}
						}
						// close window
						else if	(t == 2)
						{
							if (!curr->WindowClose())
							{
								// return if deleted
								return ;
							}
							curr = FGWindow::GetCurrent();
						}
						// hide
						else if	(t==3) curr->WindowIconize();
						// resize
						else if	(t == 4)
						{
							resize = 1;
							resizex	= resizey =	0;
							if (!fulldrag || curr->GetStatus()&WFASTMOVE)
							{
								blue_rect(curr->GetX(),	curr->GetY(), curr->GetW() + resizex, curr->GetH() + resizey);
							}
						}
					}
					else		// tuklo sa vedla titulku
					{
						n =	FGWindow::WindowFind(&EVENT);
						// do okna
						// prepni na toto okno ak nieje aktivne
						if ((n != curr)	&& (n))
						{
							if (!(curr->GetStatus()	& WMODAL))
							{
								n->WindowFocus();
								EVENT.type = tmp_event;
								TranslateUserEvent(EVENT);
								curr = FGWindow::GetCurrent();
								if (curr != n || !n) return; // window destroyed !!
							}
							else Snd(100,70);
						}
						else
						{
							vyrez =	1;
							drag_okno =	FGWindow::WindowFind(&EVENT); // 0-back, 1-wnd
							clickx = EVENT.x;
							clicky = EVENT.y;
							clickw = clickh	= 0;
						}
					}
				}
				else if	(oldb == FG_BUTTON_NONE && mb == FG_BUTTON_RIGHT)	// right click down
				{
					vyrez =	1;
					drag_okno =	FGWindow::WindowFind(&EVENT);  // find window for click
					if ((drag_okno != curr) && drag_okno)
					{
						if (!(curr->GetStatus()	& WMODAL))
						{
							drag_okno->WindowFocus();
							EVENT.type = tmp_event;
							TranslateUserEvent(EVENT);
							curr = FGWindow::GetCurrent();
							if (curr != drag_okno || !drag_okno) return; // window destroyed !!
						}
						else Snd(100,70);
					}
					if (curr && curr->IsInInput())  // if input, force ESC
					{
						curr->WindowInputChar(ESC);
					}
					clickx = EVENT.x;
					clicky = EVENT.y;
					clickw = clickh	= 0;
				}
				else if	(mb == FG_BUTTON_NONE && oldb == FG_BUTTON_NONE)	// only mouse move
				{
					EVENT.type = MOVEEVENT;
					app.type = MOVEEVENT;
					app.x = EVENT.x;
					app.y = EVENT.y;
					app.key = 0;

					idm	= FGControl::ButtonFind(&EVENT, 3);

					if (idm	&& idm->GetOwner() != curr)
					{
						break;
					}
					if (idm)
					{
						EVENT.key = idm->GetId();
						if (idm->GetTrigger()) idm = 0;
					}
					else if	((idw =	FGWindow::WindowFind(&EVENT)) != 0)
					{
						EVENT.key	= idw==curr?0:-1;
						curr = idw;		// sender is changed !!!
						EVENT.x =	EVENT.x-idw->GetX()-idw->GetXW();
						EVENT.y =	EVENT.y-idw->GetY()-idw->GetYW();
					}
					else
					{
						EVENT.type = NOEVENT;
						app.type = MOVEEVENT;
						app.x = EVENT.x;
						app.y = EVENT.y;
						app.key = 0;
					}
					if (curr)
					{
						if ((!curr->oldidm && idm) || ((idm && curr->oldidm) &&	(idm !=	curr->oldidm)))
							idm->MoveIn();
						if ((curr->oldidm && !idm) || ((idm && curr->oldidm) &&	(idm !=	curr->oldidm)))
						{
							curr->oldidm->MoveOut();
							curr->oldidm = 0;
						}
					}
					break;
				}
				else if	(oldb == FG_BUTTON_MIDDLE && mb == FG_BUTTON_NONE)	// middle click down
				{
					if (FGWindow::WindowFind(&EVENT))
					{
						EVENT.x =	clickx - curr->GetX() - curr->GetXW();
						EVENT.y =	clicky - curr->GetY() - curr->GetYW();
						EVENT.w = EVENT.h = 0;
						DoubleClick(EVENT.type = CLICKMIDDLEEVENT);
					}
					else
					{
						app.x = clickx;
						app.y = clicky;
						app.w = app.h=0;
						DoubleClick(app.type = CLICKMIDDLEEVENT);
					}
				}
				else if	((mb == FG_BUTTON_WHEEL_UP || mb == FG_BUTTON_WHEEL_DOWN) && oldb == FG_BUTTON_NONE)	// middle click down
				{
					idm	= FGControl::ButtonFind(&original, 7);
					int direction = (mb == FG_BUTTON_WHEEL_UP)?-1:1;
					EVENT.buttons = app.buttons = direction;
					if (FGWindow::WindowFind(&EVENT))
					{
						EVENT.x =	mx - curr->GetX() - curr->GetXW();
						EVENT.y =	my - curr->GetY() - curr->GetYW();
						EVENT.w = EVENT.h = 0;
						EVENT.type = MOUSEWHEELEVENT;
						if (curr && idm)
						{
							idm->WheelSpin(direction);
						}
					}
					app.x = mx;
					app.y = my;
					app.w = app.h=0;
					app.type = MOUSEWHEELEVENT;
				}
		}
	else
	{
		switch(EVENT.type)  // send & mouse keys only
		{
			case KEYEVENT:
				if (isgmkey(EVENT.key))
					break;	// exit when no window are
			case MOVEEVENT:
				FGApp::SendToApp(&EVENT);
		}
		return ;
	}							// no window exists

	if (app.type==MOVEEVENT)
	{
		app.wnd = idw;
		EVENT.accel = idm;
    }

	if (!action && !drag && !vyrez && !resize && (EVENT.type==MOVEEVENT || app.type==MOVEEVENT))
	{
		int value=0;
		if (oldx==mx)
		{
			if (oldx==0) value |= 1;
			else if (oldx>=(X_width - 1)-1) value |= 2;
		}
		if (oldy==my)
		{
			if (oldy==0) value |= 4;
			if (oldy>=(Y_width-1)-1) value |= 8;
		}
		if (value)
		{
			app.type = CURSOROUTEVENT;
			app.key  = value;
		}
	}
	oldx = mx;
	oldy = my;
	oldb = mb;
	if (idm && curr) curr->oldidm = idm;

	// there are it's magic place where an events goes into the right place
	if (EVENT.type == ACCELEVENT) app = EVENT;

	int ii = curr ? curr->GetId() : -1;

	if (app.type !=	NOEVENT)
		FGApp::SendToApp(&app); // send to app if needed

	if (EVENT.type != NOEVENT)   // send to FGWindow if needed
	{
		if (EVENT.type == ACCELEVENT && curr->IsActive())
			curr->IsActive()->RefreshGroup();

		if (EVENT.type==MOVEEVENT && EVENT.key==-1)
		{
			EVENT.key = 0;
			if (idw) idw->SendToWindow(&EVENT);
		}
		else if ((curr=FGWindow::GetCurrent()) != 0)
		{
			curr->SendToWindow(&EVENT);
		}
	}

	curr = FGWindow::WindowFind(ii);

	if (curr)
		curr->RunHandler(); // call to control handler also if window still exist

	// at this point, call_handler is already NULL

	// switch to the next Editbox
	if (ebox && ecurr && curr==ecurr)
		ebox->ClickUp(1);
}

/**
	Load & set the new mouse pointer.
*/
const FGMouseCursor * FGApp::CursorLoad(const FGMouseCursor *cur)
{
	if (cur == __fg_cursor)
		return cur;
	const FGMouseCursor *old = __fg_cursor;
	int oldf = RemoveMousePointer();
	__fg_cursor = cur;
	if (oldf)
		ShowMousePointer();
	return old;
}

int	__MouseDraw(int x, int y, int a, int& mouseflag)
{
	int	old=mouseflag;
	int	i;
	FGPixel *maska, *mapa;
	static FGPixel work[sizeof(m_save)], tmp[sizeof(m_save)];

	mouseflag =	a;
	if (a) // draw cursor
	{
		VideoToRam(m_save, x-cApp->__fg_cursor->xoff, y-cApp->__fg_cursor->yoff, 0, 0, cApp->__fg_cursor->w, cApp->__fg_cursor->h, cApp->__fg_cursor->w, cApp->__fg_cursor->h);
		memcpy(work, m_save, sizeof(m_save));
		maska =	cApp->__fg_cursor->mask;
		mapa  = cApp->__fg_cursor->bitmap;
		for	(i = 0;	i <	(cApp->__fg_cursor->w * cApp->__fg_cursor->h); i++)
		{
			work[i]	&= *maska;
			work[i]	|= ((*mapa++) &	(FGPixel)(~*maska++));
		}
		RamToVideo(work, 0, 0, x-cApp->__fg_cursor->xoff, y-cApp->__fg_cursor->yoff, cApp->__fg_cursor->w, cApp->__fg_cursor->h, cApp->__fg_cursor->w, cApp->__fg_cursor->h, BITBLIT_COPY, _GSET);
	}
	else // restore background
	{
		VideoToRam(tmp, x-cApp->__fg_cursor->xoff, y-cApp->__fg_cursor->yoff, 0, 0, cApp->__fg_cursor->w, cApp->__fg_cursor->h, cApp->__fg_cursor->w, cApp->__fg_cursor->h);
		if (memcmp(tmp,	work, sizeof(m_save)) != 0) // yes, changed
		{
			for (i = 0; i < (int)sizeof(m_save)/bpp; i++)
			{
				if (tmp[i] != work[i])
				{
					m_save[i] = tmp[i];
				}
			}
		}
		RamToVideo(m_save, 0, 0, x-cApp->__fg_cursor->xoff, y-cApp->__fg_cursor->yoff, cApp->__fg_cursor->w, cApp->__fg_cursor->h, cApp->__fg_cursor->w, cApp->__fg_cursor->h, BITBLIT_COPY, _GSET);
	}
	return old;
}

// mouse cursor drawer
mouser_t   vector_mouse_cursor= __MouseDraw;

/**
	Enable drawing into the off-screen video RAM.
	This feature is intended for flickerless full screen
	animation. Parameter mode is one of these predefined:

	enum { FG_DOUBLEBUFFER, FG_TRIPLEBUFFER, FG_QUADBUFFER };

	More buffers may help for speed but is X-times memory
	hungry and not the all videocards supports it.
	On exit, return number of true allocated buffers (0 if error).
	You can still use BUFFERING mode when function fail but
	screen will flicker - all drawing goes to the visible screen.
*/
int FGApp::EnableBuffering(int mode)
{
	if (__fg_driver)
		return __fg_driver->EnableBuffering(mode);
	return (ENUM_BUFFERING)false;
}

/**
	Call this routine when all objects of the scene are
	drawn. The buffer will be showed at the next screen
	vertical retrace. Return 0 if fail - no buffering is
	allowed/initialized.
*/
int FGApp::Flip(void)
{
	if (__fg_driver)
	{
		__fg_fps++;
		if (fgstate.verbose)
		{
			gprintf(CWHITE, CBLACK,16,16,"fps %3d", __fg_fps_total);
		}
		UpdateMousePointer();
		return __fg_driver->Flip();
	}
	return 0;
}

/**
	Disable video buffering and release the used video memory.
*/
void FGApp::DisableBuffering(void)
{
	if (__fg_driver)
		__fg_driver->DisableBuffering();
}

/**
	Return current Operating system code.
	Code is one from these OS types.
*/
ENUM_OS FGApp::GetOS(void)
{
#ifdef __QNX__
	return OSTYPE_QNX;
#endif
#ifdef __linux__
	return OSTYPE_LINUX;
#endif
#ifdef __MSDOS__
	return OSTYPE_MSDOS;
#endif
#ifdef _WIN32
	return OSTYPE_WIN32;
#endif
}

struct KeyPair
{
	const char* string;
	unsigned int key;
} keys[] = {
{ "CR", 13 },
{ "ESC", 27 },
{ "TAB", 9 },
{ "LF", 10 },
{ "PRTRSC", PRTRSC },
{ "BACKSP", BACKSP },
{ "KUP", KUP },
{ "KDOWN", KDOWN },
{ "KRIGHT", KRIGHT },
{ "KLEFT", KLEFT },
{ "INSERT", INSERT },
{ "DEL", DEL },
{ "HOME", HOME },
{ "END", END },
{ "PGUP", PGUP },
{ "PGDOWN", PGDOWN},

{ "ALT_A", ALT_A },
{ "ALT_B", ALT_B },
{ "ALT_C", ALT_C },
{ "ALT_D", ALT_D },
{ "ALT_E", ALT_E },
{ "ALT_F", ALT_F },
{ "ALT_G", ALT_G },
{ "ALT_H", ALT_H },
{ "ALT_I", ALT_I },
{ "ALT_J", ALT_J },
{ "ALT_K", ALT_K },
{ "ALT_L", ALT_L },
{ "ALT_M", ALT_M },
{ "ALT_N", ALT_N },
{ "ALT_O", ALT_O },
{ "ALT_P", ALT_P },
{ "ALT_Q", ALT_Q },
{ "ALT_R", ALT_R },
{ "ALT_S", ALT_S },
{ "ALT_T", ALT_T },
{ "ALT_U", ALT_U },
{ "ALT_V", ALT_V },
{ "ALT_W", ALT_W },
{ "ALT_X", ALT_X },
{ "ALT_Y", ALT_Y },
{ "ALT_Z", ALT_Z },

{ "F01", F01},
{ "F02", F02},
{ "F03", F03},
{ "F04", F04},
{ "F05", F05},
{ "F06", F06},
{ "F07", F07},
{ "F08", F08},
{ "F09", F09},
{ "F10", F10},
{ "F11", F11},
{ "F12", F12},

{ "ALT_F01", ALT_F01},
{ "ALT_F02", ALT_F02},
{ "ALT_F03", ALT_F03},
{ "ALT_F04", ALT_F04},
{ "ALT_F05", ALT_F05},
{ "ALT_F06", ALT_F06},
{ "ALT_F07", ALT_F07},
{ "ALT_F08", ALT_F08},
{ "ALT_F09", ALT_F09},
{ "ALT_F10", ALT_F10},
{ "ALT_F11", ALT_F11},
{ "ALT_F12", ALT_F12},

{ "CTRL_F01", CTRL_F01},
{ "CTRL_F02", CTRL_F02},
{ "CTRL_F03", CTRL_F03},
{ "CTRL_F04", CTRL_F04},
{ "CTRL_F05", CTRL_F05},
{ "CTRL_F06", CTRL_F06},
{ "CTRL_F07", CTRL_F07},
{ "CTRL_F08", CTRL_F08},
{ "CTRL_F09", CTRL_F09},
{ "CTRL_F10", CTRL_F10},
{ "CTRL_F11", CTRL_F11},
{ "CTRL_F12", CTRL_F12},

{ "ALT_UP", ALT_UP},
{ "ALT_DOWN", ALT_DOWN},
{ "ALT_RIGHT", ALT_RIGHT},
{ "ALT_LEFT", ALT_LEFT},
{ "ALT_TAB", ALT_TAB},
{ "ALT_INSERT", ALT_INSERT},
{ "ALT_DEL", ALT_DEL},
{ "ALT_HOME", ALT_HOME},
{ "ALT_END", ALT_END},
{ "ALT_PGUP", ALT_PGUP},
{ "ALT_PGDOWN", ALT_PGDOWN},

{ "CTRL_UP", CTRL_UP},
{ "CTRL_DOWN", CTRL_DOWN},
{ "CTRL_RIGHT", CTRL_RIGHT},
{ "CTRL_LEFT", CTRL_LEFT},

{ "CTRL_TAB", CTRL_TAB },
{ "CTRL_INSERT", CTRL_INSERT },
{ "CTRL_DEL", CTRL_DEL },
{ "CTRL_HOME", CTRL_HOME },
{ "CTRL_END", CTRL_END },
{ "CTRL_PGUP", CTRL_PGUP },
{ "CTRL_PGDOWN", CTRL_PGDOWN },
{0, 0},
};

const char* FGEvent::GetKeySymbolName(unsigned int code)
{
	struct KeyPair* ptr = keys;

	if (code>=' ' && code<127)
		return "Character";

	while(ptr->key)
	{
		if (ptr->key == code)
			return ptr->string;
		ptr++;
	}
	return "Undefined";
}

FGInputDevice::FGInputDevice()
{
	working = false;
	init(GetXRes(), GetYRes());
	load_values();
	setCalibrationMatrix(imaginary, real, &matrix);
}

/*
 *
 *   Copyright (c) 2001, Carlos E. Vidales. All rights reserved.
 *
 *   This sample program was written and put in the public domain
 *    by Carlos E. Vidales.  The program is provided "as is"
 *    without warranty of any kind, either expressed or implied.
 *   If you choose to use the program within your own products
 *    you do so at your own risk, and assume the responsibility
 *    for servicing, repairing or correcting the program should
 *    it prove defective in any manner.
 *   You may copy and distribute the program's source code in any
 *    medium, provided that you also include in each copy an
 *    appropriate copyright notice and disclaimer of warranty.
 *   You may also modify this program and distribute copies of
 *    it provided that you include prominent notices stating
 *    that you changed the file(s) and the date of any change,
 *    and that you do not charge any royalties or licenses for
 *    its use.
 *
 *
 *
 *   File Name:  calibrate.c
 *
 *
 *   This file contains functions that implement calculations
 *    necessary to obtain calibration factors for a touch screen
 *    that suffers from multiple distortion effects: namely,
 *    translation, scaling and rotation.
 *
 *   The following set of equations represent a valid display
 *    point given a corresponding set of touch screen points:
 *
 *
 *                                              /-     -\
 *              /-    -\     /-            -\   |       |
 *              |      |     |              |   |   Xs  |
 *              |  Xd  |     | A    B    C  |   |       |
 *              |      |  =  |              | * |   Ys  |
 *              |  Yd  |     | D    E    F  |   |       |
 *              |      |     |              |   |   1   |
 *              \-    -/     \-            -/   |       |
 *                                              \-     -/
 *
 *
 *    where:
 *
 *           (Xd,Yd) represents the desired display point
 *                    coordinates,
 *
 *           (Xs,Ys) represents the available touch screen
 *                    coordinates, and the matrix
 *
 *           /-   -\
 *           |A,B,C|
 *           |D,E,F| represents the factors used to translate
 *           \-   -/  the available touch screen point values
 *                    into the corresponding display
 *                    coordinates.
 *
 *
 *    Note that for practical considerations, the utilitities
 *     within this file do not use the matrix coefficients as
 *     defined above, but instead use the following
 *     equivalents, since floating point math is not used:
 *
 *            A = An/Divider
 *            B = Bn/Divider
 *            C = Cn/Divider
 *            D = Dn/Divider
 *            E = En/Divider
 *            F = Fn/Divider
 *
 *
 *
 *    The functions provided within this file are:
 *
 *          setCalibrationMatrix() - calculates the set of factors
 *                                    in the above equation, given
 *                                    three sets of test points.
 *               getDisplayPoint() - returns the actual display
 *                                    coordinates, given a set of
 *                                    touch screen coordinates.
 * translateRawScreenCoordinates() - helper function to transform
 *                                    raw screen points into values
 *                                    scaled to the desired display
 *                                    resolution.
 *
 *
*/

/**
 *
 *     Function: setCalibrationMatrix()
 *
 *  Description: Calling this function with valid input data
 *                in the display and screen input arguments
 *                causes the calibration factors between the
 *                screen and display points to be calculated,
 *                and the output argument - matrixPtr - to be
 *                populated.
 *
 *               This function needs to be called only when new
 *                calibration factors are desired.
 *
 *
 *  Argument(s): displayPtr (input) - Pointer to an array of three
 *                                     sample, reference points.
 *               screenPtr (input) - Pointer to the array of touch
 *                                    screen points corresponding
 *                                    to the reference display points.
 *               matrixPtr (output) - Pointer to the calibration
 *                                     matrix computed for the set
 *                                     of points being provided.
 *
 *
 *  From the article text, recall that the matrix coefficients are
 *   resolved to be the following:
 *
 *
 *      Divider =  (Xs0 - Xs2)*(Ys1 - Ys2) - (Xs1 - Xs2)*(Ys0 - Ys2)
 *
 *
 *
 *                 (Xd0 - Xd2)*(Ys1 - Ys2) - (Xd1 - Xd2)*(Ys0 - Ys2)
 *            A = ---------------------------------------------------
 *                                   Divider
 *
 *
 *                 (Xs0 - Xs2)*(Xd1 - Xd2) - (Xd0 - Xd2)*(Xs1 - Xs2)
 *            B = ---------------------------------------------------
 *                                   Divider
 *
 *
 *                 Ys0*(Xs2*Xd1 - Xs1*Xd2) +
 *                             Ys1*(Xs0*Xd2 - Xs2*Xd0) +
 *                                           Ys2*(Xs1*Xd0 - Xs0*Xd1)
 *            C = ---------------------------------------------------
 *                                   Divider
 *
 *
 *                 (Yd0 - Yd2)*(Ys1 - Ys2) - (Yd1 - Yd2)*(Ys0 - Ys2)
 *            D = ---------------------------------------------------
 *                                   Divider
 *
 *
 *                 (Xs0 - Xs2)*(Yd1 - Yd2) - (Yd0 - Yd2)*(Xs1 - Xs2)
 *            E = ---------------------------------------------------
 *                                   Divider
 *
 *
 *                 Ys0*(Xs2*Yd1 - Xs1*Yd2) +
 *                             Ys1*(Xs0*Yd2 - Xs2*Yd0) +
 *                                           Ys2*(Xs1*Yd0 - Xs0*Yd1)
 *            F = ---------------------------------------------------
 *                                   Divider
 *
 *
 *       Return: OK - the calibration matrix was correctly
 *                     calculated and its value is in the
 *                     output argument.
 *               NOT_OK - an error was detected and the
 *                         function failed to return a valid
 *                         set of matrix values.
 *                        The only time this sample code returns
 *                        NOT_OK is when Divider == 0
 *
 *
 *
 *                 NOTE!    NOTE!    NOTE!
 *
 *  setCalibrationMatrix() and getDisplayPoint() will do fine
 *  for you as they are, provided that your digitizer
 *  resolution does not exceed 10 bits (1024 values).  Higher
 *  resolutions may cause the integer operations to overflow
 *  and return incorrect values.  If you wish to use these
 *  functions with digitizer resolutions of 12 bits (4096
 *  values) you will either have to a) use 64-bit signed
 *  integer variables and math, or b) judiciously modify the
 *  operations to scale results by a factor of 2 or even 4.
 *
 *
 */
int FGInputDevice::setCalibrationMatrix(POINT * displayPtr, POINT * screenPtr, MATRIX * matrixPtr)
{
	int retValue = true;

	matrixPtr->Divider = (((double)screenPtr[0].x - (double)screenPtr[2].x) * ((double)screenPtr[1].y - (double)screenPtr[2].y)) -
		(((double)screenPtr[1].x - (double)screenPtr[2].x) * ((double)screenPtr[0].y - (double)screenPtr[2].y));

	if (matrixPtr->Divider == 0)
	{
		retValue = false;
	}
	else
	{
		matrixPtr->An = (((double)displayPtr[0].x - displayPtr[2].x) * (screenPtr[1].y - screenPtr[2].y)) -
			(((double)displayPtr[1].x - displayPtr[2].x) * ((double)screenPtr[0].y - screenPtr[2].y));

		matrixPtr->Bn = (((double)screenPtr[0].x - screenPtr[2].x) * ((double)displayPtr[1].x - displayPtr[2].x)) -
			(((double)displayPtr[0].x - displayPtr[2].x) * ((double)screenPtr[1].x - screenPtr[2].x));

		matrixPtr->Cn = ((double)screenPtr[2].x * displayPtr[1].x - (double)screenPtr[1].x * displayPtr[2].x) * screenPtr[0].y +
			((double)screenPtr[0].x * displayPtr[2].x - (double)screenPtr[2].x * displayPtr[0].x) * screenPtr[1].y +
			((double)screenPtr[1].x * displayPtr[0].x - (double)screenPtr[0].x * displayPtr[1].x) * screenPtr[2].y;

		matrixPtr->Dn = (((double)displayPtr[0].y - displayPtr[2].y) * ((double)screenPtr[1].y - screenPtr[2].y)) -
			(((double)displayPtr[1].y - displayPtr[2].y) * ((double)screenPtr[0].y - screenPtr[2].y));

		matrixPtr->En = (((double)screenPtr[0].x - screenPtr[2].x) * ((double)displayPtr[1].y - displayPtr[2].y)) -
			(((double)displayPtr[0].y - displayPtr[2].y) * ((double)screenPtr[1].x - screenPtr[2].x));

		matrixPtr->Fn = ((double)screenPtr[2].x * displayPtr[1].y - screenPtr[1].x * (double)displayPtr[2].y) * screenPtr[0].y +
			((double)screenPtr[0].x * displayPtr[2].y - screenPtr[2].x * (double)displayPtr[0].y) * screenPtr[1].y +
			((double)screenPtr[1].x * displayPtr[0].y - screenPtr[0].x * (double)displayPtr[1].y) * screenPtr[2].y;
	}

	return (retValue);
}

/**
 *
 *     Function: getDisplayPoint()
 *
 *  Description: Given a valid set of calibration factors and a point
 *                value reported by the touch screen, this function
 *                calculates and returns the true (or closest to true)
 *                display point below the spot where the touch screen
 *                was touched.
 *
 *
 *
 *  Argument(s): displayPtr (output) - Pointer to the calculated
 *                                      (true) display point.
 *               screenPtr (input) - Pointer to the reported touch
 *                                    screen point.
 *               matrixPtr (input) - Pointer to calibration factors
 *                                    matrix previously calculated
 *                                    from a call to
 *                                    setCalibrationMatrix()
 *
 *
 *  The function simply solves for Xd and Yd by implementing the
 *   computations required by the translation matrix.
 *
 *                                              /-     -\
 *              /-    -\     /-            -\   |       |
 *              |      |     |              |   |   Xs  |
 *              |  Xd  |     | A    B    C  |   |       |
 *              |      |  =  |              | * |   Ys  |
 *              |  Yd  |     | D    E    F  |   |       |
 *              |      |     |              |   |   1   |
 *              \-    -/     \-            -/   |       |
 *                                              \-     -/
 *
 *  It must be kept brief to avoid consuming CPU cycles.
 *
 *
 *       Return: OK - the display point was correctly calculated
 *                     and its value is in the output argument.
 *               NOT_OK - an error was detected and the function
 *                         failed to return a valid point.
 *
 *
 *
 *                 NOTE!    NOTE!    NOTE!
 *
 *  setCalibrationMatrix() and getDisplayPoint() will do fine
 *  for you as they are, provided that your digitizer
 *  resolution does not exceed 10 bits (1024 values).  Higher
 *  resolutions may cause the integer operations to overflow
 *  and return incorrect values.  If you wish to use these
 *  functions with digitizer resolutions of 12 bits (4096
 *  values) you will either have to a) use 64-bit signed
 *  integer variables and math, or b) judiciously modify the
 *  operations to scale results by a factor of 2 or even 4.
 *
 *
 */
int FGInputDevice::getDisplayPoint(POINT * displayPtr, POINT * screenPtr, MATRIX * matrixPtr)
{
	int retValue = true;


	if (matrixPtr->Divider != 0)
	{
		/* Operation order is important since we are doing integer */
		/*  math. Make sure you add all terms together before      */
		/*  dividing, so that the remainder is not rounded off     */
		/*  prematurely.                                           */

		displayPtr->x = (int) (((matrixPtr->An * screenPtr->x) + (matrixPtr->Bn * screenPtr->y) + matrixPtr->Cn) / matrixPtr->Divider);
		displayPtr->y = (int) (((matrixPtr->Dn * screenPtr->x) + (matrixPtr->En * screenPtr->y) + matrixPtr->Fn) / matrixPtr->Divider);
	}
	else
	{
		retValue = false;
	}

	return (retValue);

}

void FGInputDevice::init(int xmax, int ymax)
{
	imaginary[0].x = xmax / 100 * 15;
	imaginary[0].y = ymax / 100 * 15;

	imaginary[1].x = xmax / 100 * 50;
	imaginary[1].y = ymax / 100 * 80;

	imaginary[2].x = xmax / 100 * 85;
	imaginary[2].y = ymax / 100 * 50;

	real[0] = imaginary[0];
	real[1] = imaginary[1];
	real[2] = imaginary[2];
}

void FGInputDevice::load_values(void)
{
	Config* cfg = new Config("calibration_data");

	cfg->ReadInt("im_x0", imaginary[0].x);
	cfg->ReadInt("im_y0", imaginary[0].y);
	cfg->ReadInt("im_x1", imaginary[1].x);
	cfg->ReadInt("im_y1", imaginary[1].y);
	cfg->ReadInt("im_x2", imaginary[2].x);
	cfg->ReadInt("im_y2", imaginary[2].y);

	cfg->ReadInt("re_x0", real[0].x);
	cfg->ReadInt("re_y0", real[0].y);
	cfg->ReadInt("re_x1", real[1].x);
	cfg->ReadInt("re_y1", real[1].y);
	cfg->ReadInt("re_x2", real[2].x);
	cfg->ReadInt("re_y2", real[2].y);

	delete cfg;
}

void FGInputDevice::save_values(void)
{
	Config* cfg = new Config("calibration_data");

	cfg->WriteInt("im_x0", imaginary[0].x);
	cfg->WriteInt("im_y0", imaginary[0].y);
	cfg->WriteInt("im_x1", imaginary[1].x);
	cfg->WriteInt("im_y1", imaginary[1].y);
	cfg->WriteInt("im_x2", imaginary[2].x);
	cfg->WriteInt("im_y2", imaginary[2].y);

	cfg->WriteInt("re_x0", real[0].x);
	cfg->WriteInt("re_y0", real[0].y);
	cfg->WriteInt("re_x1", real[1].x);
	cfg->WriteInt("re_y1", real[1].y);
	cfg->WriteInt("re_x2", real[2].x);
	cfg->WriteInt("re_y2", real[2].y);

	delete cfg;
}

void FGInputDevice::Scale(int& x, int& y)
{
	POINT point = { x, y }, point2;
	getDisplayPoint(&point2, &point, &matrix);
	x = point2.x;
	y = point2.y;
}

void FGInputDevice::Calibration(void)
{
	if (working)
	{
		init( GetXRes(), GetYRes() );

		for(int i = 0; i < 3; i++)
		{
			DoPoint(imaginary+i, real+i);
		}
		save_values();
		setCalibrationMatrix(imaginary, real, &matrix);
	}
}

void FGInputDevice::DoPoint(POINT * displayPtr, POINT * screenPtr)
{
	int rc, type, key, x, y, buttons;

	delay(1000);
	while ( (rc = GetInputEvent(type,key,x,y,buttons)) != 0)
		delay(1);

	fill_circle(displayPtr->x, displayPtr->y, 48, 0, _GSET);
	fill_circle(displayPtr->x, displayPtr->y, 24, CWHITED, _GSET);
	draw_circle(displayPtr->x, displayPtr->y, 4, CWHITE, _GSET);
	draw_circle(displayPtr->x, displayPtr->y, 2, 0, _GSET);
	Snd(1000, 100);

	while ( (rc = GetInputEvent(type, key, x, y, buttons)) == 0)
		delay(1);
		
	screenPtr->x = x;
	screenPtr->y = y;
}

#ifdef FG_NAMESPACE
}
#endif

