/*
    OpenGUI - Drawing & Windowing library

    Copyright (C) 1996,2004  Marian Krivos

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    nezmar@atlas.sk
    declarations for widgets

  $Id: widgets.h,v 1.12 2004/05/30 11:52:55 majo Exp $

  $Log: widgets.h,v $
  Revision 1.12  2004/05/30 11:52:55  majo
  dox

  Revision 1.11  2004/04/16 08:58:19  majo
  added FGRadioGroup::SetUserData(void *)

  Revision 1.10  2004/03/24 20:06:54  majo

  WIN32 compiling

  Revision 1.9  2004/03/24 18:44:43  majo
  update FGColorPicker dox

  Revision 1.8  2004/03/24 17:13:58  majo
  added FGXColor
  added FGColorPicker

  Revision 1.7  2004/03/17 21:19:12  majo
  added FGColor, FGImage

  Revision 1.6  2004/02/23 20:08:01  majo
  all classes are with prefix FG* on now
  polygon functions uses FGPointArray from now
  class GuiEvent is renamed to FGEvent
  some by parameters overloaded methods was removed (class FGWindow)
  many other small changes

  Revision 1.5  2004/02/12 17:52:11  majo
  *

  Revision 1.4  2004/02/10 16:18:59  majo
  fixed compile warnings

  Revision 1.3  2004/02/10 16:16:53  majo
  added method: bool FGFileDialog::TestForOvervrite(void)

  Revision 1.2  2004/01/28 18:08:58  majo
  header file cleanups - widgets.h no more include fastgl.h

  Revision 1.1.1.1  2004/01/22 16:25:08  majo
  initial release

*/

//
//	WIDGETS.CC		20.11.1998
//

#ifndef _WIDGETS_H_
#define _WIDGETS_H_

#include <sys/stat.h>
#include <string.h>

#ifdef __WATCOMC__
#ifdef __DOS__
#include <direct.h>
#else
#include <dirent.h>
#endif // dos
#elif !defined(_MSC_VER)
#include <dirent.h>
#endif // watcom

#ifdef __BORLANDC__
#include <dir.h>
#endif

#ifdef _MSC_VER
#include <direct.h>
#include "mydirent.h"
#endif

#ifdef __BORLANDC__
#pragma option push -b
#pragma option push -a4
#endif

#ifdef FG_NAMESPACE
namespace fgl {
#endif

const unsigned FDIALOG_OPEN	= 1;
const unsigned FDIALOG_SAVE	= 2;
const unsigned FDIALOG_MODAL = 4;
const unsigned FDIALOG_SAVEDIR = 8;
const unsigned FDIALOG_MULTISELECT = 16;

/**
	Internal helper structure for FGFileDialog
*/
struct mydirent
{
	struct dirent item;
	bool		type;	// 0 for dir, 1 for file
	bool        selected;
	mydirent() { memset(this, 0, sizeof(*this)); }
};

/**
 A nice widget intended for the file selection by user for various things.
 filename length max. 511 chars
 max. visible is ~30 chars
 @see FGFont FGFontManager

 @image html fdialog.png

*/
class FGFileDialog : public FGConnector
{
	public:
	    enum    { name_width=200, max_pathlen=512 };
	private:
		char	path[max_pathlen];
		char	tmppath[max_pathlen];
		char 	filename[max_pathlen];
		char	*filter;
		int 	mode;
		int		files,fls,dirs;
		int		maxfiles;
		FileDialogCallBack fileselect2;
		void	reload(void);
		FGWindow	*wnd;
		FGEditBox *nameEBox;
		struct 	mydirent *filebuffer, *unused;
		int		param;
		FGListBox *list;
		FGListBox *list2;
		FGPushButton *ok, *cancel, *mkdir, *home, *updir;
		int		mdir;
//		int     num_of_selected;

		int		Selected();
		void	Refresh(char *);
		bool    SelectFile(void);
		void 	_init(char *dir, char *flt, char *namewnd, int m, int ink, int paper);

		static 	void InputName(CallBack);
		static 	void Mkdir(CallBack);
		static 	void Ok(CallBack);
		static 	void Updir(CallBack);
		static 	void Home(CallBack);
		static  void myproc(FGEvent *p);
		static  void drawone(int flag, FGWindow *wnd,int x,int y,void *str,int index);
		static  void drawone2(int flag, FGWindow *wnd,int x,int y,void *str,int index);
		static  char delimiter[2];
	protected:
	public:
		/**
		Creates visual object for the file selection by user.
		@param filesel your callback, it is called when you enter or click a filename
		@param dir the start directory, default is current app's directory or last used (with flag FDIALOG_SAVEDIR)
		@param flt text string that contains file extension patterns separated by space, ".exe .com" by example
		@param namewnd the caption
		@param m mode of operation (SAVE or OPEN). See to SetMode()
		@param ink the FGWindow foreground color
		@param paper the FGWindow background color
		*/
		FGFileDialog(FileDialogCallBack filesel, char *dir=0, char *flt=0, char *namewnd="File Dialog", int m=FDIALOG_OPEN, int ink=0, int paper=PM);
		virtual ~FGFileDialog()
		{
			delete [] filebuffer;
			if (wnd) if (wnd->GetStatus() & WEXIST) delete wnd;
		}
		/**
		Returns the current browsed directory.
		*/
		char*   FGAPI GetDir(void) { return path; }
		//! this filename is passed to the callback directly.
		char*   FGAPI GetName(void) { return filename; }
		/**
		* Makes full path && name string into user defined buffer
		*/
		void FGAPI BuildFullPathname(char *buffer);
		/**
		* Makes relative path && name string into user defined buffer
		*/
		void FGAPI BuildRelativePathname(char *buffer);
		/**
		Change the current directory and reload it.
		*/
		void	FGAPI SetDir(char *d) { strcpy(path, d); reload(); }
		/**
		Change the current input filter mask (".cpp .h" by example) and reload it.
		*/
		void	FGAPI SetFilter(char *f) { filter=f; reload(); }
		/**
		Sets object's mode:
		-	FDIALOG_OPEN
		-	FDIALOG_SAVE
		-	FDIALOG_MODAL
		-	FDIALOG_SAVEDIR
		*/
		void	FGAPI SetMode(int m) { mode=m; }
		/**
		* Shows dialog to ask to overwrite the selected file.
		* Is it possible to call this function from user callback (on file select)
		* @return true if file can be replaced.
		*/
		bool TestForOverwrite(void);
};

/**
An andvanced widget to edit the plain text files or create the new one.
It has a traditional Open, Save & Close menu and simply search engine.

@image html tedit.png

*/
class FGTextEditor : public FGConnector
{
	public:
		enum 	{ FGMAX_LINE=32768, FGMAX_LINESIZE=512 };
	private:
		class FGLINE
		{
			public:
				unsigned size;         		/* amount of memory allocated */
				unsigned char *text;         /* the characters */
		};

		class FGBUFFER
		{
			public:
			char name[256];              /* name of the file */
			char buffer[FGMAX_LINESIZE];   /* name of the file */
			int  posx, posy;
			int  line;      		        /* cursor position within the line */
			int  lines;                  // all lines
			int  flags;                  /* flags about the buffer state */
			int  top;                    /* top line on the screen */
			int  ovr;
			int  hscroll;                /* offset when screen is scrolled sideways */
			int  sel_line1;               /* line that the selection began at */
			int  sel_pos1;                /* selection offset within the line */
			int  sel_line2;               /* offset in lines from sel_line to cursor */
			int  sel_pos2;                /* selection offset within the line */
			FGBUFFER()
			{
				memset(this,0,sizeof(*this));
			}
		};

	protected:
		FGWindow *Text_EditorPtr;
		FGWindow *OptionsPtr;
		FGWindow *String_SearchPtr;
		int lfonly, font_save;
		FGBUFFER buf;
		FGLINE text[FGMAX_LINE];
		char s[1028];
		int WX, WY;
		char srch_str[33];
		int last_found;
		int _font;
		int TABSIZE;
		int nodraw,inblock,ronly;
		int X, Y, W, H;
		FGTextEditor **self;
//
		void FGAPI file(void);
		int  FGAPI _memcmp(unsigned char *from, unsigned char *co, unsigned kolko, unsigned l);
		int  FGAPI Find(int from, int count, int size);
		void FGAPI Clear(void);
		void FGAPI _goto(void);
		static void SearchProc(FGEvent *p);
		void FGAPI search(void);
		static void SetFont(CallBack);
		static void OptionsProc(FGEvent *p);
		static void file_proc(CallBack);
		void FGAPI options(void);
		static char* FGAPI terminate(char *s);
		FGLINE* FGAPI create_line(char *s, int at);
		void FGAPI Init(void);
		void FGAPI SaveAsBuffer2(void);
		void FGAPI OpenBuffer2(void);
		static void OpenBuffer1(char *s, FGFileDialog *);
		static void SaveAsBuffer1(char *s, FGFileDialog *);
		void FGAPI ShowCursor(int trigger);
		void FGAPI ShowLine(int scr, char *s);
		int  FGAPI isempty(char *s);
		void FGAPI CKey(int k);
		void FGAPI Key(int k);
		void FGAPI SaveChanges(void);
		static void Text_EditorProc(FGEvent *p);
		void FGAPI Open(char *s);
		void FGAPI Destruct(void);
	public:
		FGTextEditor(FGTextEditor **e=0, char *arg="Untitled", int font=FONT0816, int ink=CGRAY3, int paper=CBLACK);
		virtual ~FGTextEditor();
		void FGAPI ShowBuffer(int top=0);
		void FGAPI NewBuffer(void);
		void FGAPI OpenBuffer(char *s);
		void FGAPI ReopenBuffer(void);
		void FGAPI SaveBuffer(void);
		void FGAPI SaveAsBuffer(char *s);
		void FGAPI Close(void);
		void FGAPI Goto(int);
		/**
		Don't allow any changes on the file. View mode only.
		*/
		void FGAPI ReadOnlyMode(void) { ronly = 1; Text_EditorPtr->ResetChange(); };
		/**
		Move the editor at the top of all opened windows.
		*/
		void FGAPI SetFocus(void) { Text_EditorPtr->WindowFocus(); }
		/**
		Append this line at the end of text buffer.
		*/
		int  FGAPI AppendLine(char *str);
};

/**
A nice widget to color selecting by user for various things.
@see FGColorPicker
@image html cdialog.png

*/
class FGColorDialog : public FGConnector
{
		FGWindow 	*wnd;
		ColorDialogCallBack fnc;
		int			curr;
		unsigned 	color;
		FGSlideBarV *sl1, *sl2, *sl3;
		int		cd_r, cd_g, cd_b;

		static unsigned int custom_colors[256];
		void FGAPI FocusTo(int a);
		static void CSliderFnc(CallBack cb);
		static void SetColorsProc(FGEvent *p);
		void UpdatePalette(void);
	protected:
	public:
		virtual ~FGColorDialog()
		{
			if (wnd) if (wnd->GetStatus() & WEXIST) delete wnd;
		}
		/**
		Creates a widget for easy and visual color selecting.
		@param capture a FGWindow caption
		@param f the callback function, it is called with selected color
		@param bgc the background color of widget
		*/
		FGColorDialog(char *capture, ColorDialogCallBack f, FGPixel bgc=PM);
		/**
		Saves the current palette to the file. Usable in 8bit color modes only.
		@param name a filename (with the path optionally)
		@return false if a file error is occured
		*/
		static int FGAPI Import(char *name);
		/**
		Loads the current palette from the file. Usable in 8bit color modes only.
		@param name a filename (with the path optionally)
		@return false if a file error is occured
		*/
		static int FGAPI Export(char *name);
		/**
		Returns the raw palette data.
		*/
		unsigned int* FGAPI GetColors(void)
		{
			return custom_colors;
		}
};

const unsigned FONT_DIALOG_FIXED = 1;
const unsigned FONT_DIALOG_VARIABLE = 2;
const unsigned FONT_DIALOG_ALL = (FONT_DIALOG_FIXED | FONT_DIALOG_VARIABLE);

/**
 A nice widget intended for the font selection by user for various things.

 @image html fntdlg.png

*/
class FGFontDialog : public FGConnector
{
		FGWindow 	*wnd;
		FontDialogCallBack userfnc;
		int		curr;
		FGListBox *lb;
		static void SetFontProc(FGEvent *p);
		void ShowFont(void);
	public:
		/**
		Creates a widget for easy and visual color selecting.
		@param f the callback function, it is called with selected color
		@param which you can show to select only fixed, or only proportional or both fonts.
		Predefined consts are: FONT_DIALOG_FIXED, FONT_DIALOG_VARIABLE and FONT_DIALOG_ALL
		@param capture a FGWindow caption
		@param bgc the background color of widget
		*/
		FGFontDialog(FontDialogCallBack f, int which=FONT_DIALOG_ALL, char *capture="Font selection dialog", FGPixel bgc=PM);
		virtual ~FGFontDialog()
		{
		}
};

/**
Use FGUpDown to add an up-down control to a FGWindow. Up-down controls consist
of a pair of arrow buttons, such as the arrows that appear in a spin box.
Up-down controls allow users to change the size of a numerical value by clicking
on arrow buttons.
*/
class FGUpDown
{
		FGPushButton *up, *down;
		FGEditBox *eb;
		float step;
		static void handler1(CallBack);
		static void handler2(CallBack);

	public:
		FGUpDown(FGEditBox *eb, float _step);
		//! Disable object
		void Disable(void)
		{
			up->Disable();
			down->Disable();
			eb->Disable();
		}
		//! Enable object
		void Enable(void)
		{
			up->Enable();
			down->Enable();
			eb->Enable();
		}
};

/**
	Provides a container for multiple Enable/Disable of the Control
	in a FGWindow.
	Use this if you want work with some Controls as with a group.
*/
class FGWidgetContainer
{
	protected:
		void *ptr;

	public:
		FGWidgetContainer();
		~FGWidgetContainer();
		//! Disables all FGControls
		void DisableContainer(void);
		//! Enables all FGControls
		void EnableContainer(void);
		//! Adds new FGControl
		void AddToContainer(FGControl *);
		//! Set user data pointer
		void SetUserData(void* val);

};

/**
    Base class for RadioGroup classes
*/
class FGRadioGroup : public FGWidgetContainer
{
	protected:
		FGButtonGroup*      grp;
		unsigned 		    id;
		int*			    value;
		int					current;
		int                 count;
		void 			    (*userfnc)(FGControl* ,int);

		static void callback(CallBack cb)
		{
			FGRadioGroup* _THIS = (FGRadioGroup *)cb->GetPrivateData();
			_THIS->current = cb->GetId() - _THIS->id;
			if (_THIS->value) *_THIS->value = _THIS->current;
			if (_THIS->userfnc)
				_THIS->userfnc(cb, _THIS->current);
		}
	public:
		FGRadioGroup(int x, int y, char *names[], int *init_value, void (*cb)(FGControl* ,int), FGWindow* w, int step=20);
		/**
		* Call destructor on TERMINATEEVENT explicitly.
		*/
		virtual ~FGRadioGroup()
		{
			if (grp) delete grp;
			grp = 0;
		}
		void ChangeItem(int* new_value);
};

/**
* Widget to easier using of vertical arrays of RadioButtons.
* @image html radio.png
*/
class FGRadioGroupVertical : public FGRadioGroup
{
		FGControl* DrawOne(int x, int y, int offset, char* name, FGWindow* w, ControlCall cb)
	    {
		    return w->AddPointButton(x, y+offset, name, 0, 0, cb);
		}
	public:
		/**
		* Creates the group of PointButtons at exact coords into the FGWindow. The number of created
		* objects is equivalent of the number of passed strings.
		* @param x start x coord
		* @param y start y coord
		* @param names an array of ASCIIZ strings terminated by 0 pointer, by example: char *names[]={"one","two","three",0};
		* @param init_value an initial index of active object on create time, by example: 1 for string "two"
		* @param cb standard OpenGUI CallBack routine
		* @param w the parent FGWindow object
		* @param step offset of y coordinates for each next line (default=20)
		*/
		FGRadioGroupVertical(int x, int y, char *names[], int *init_value, void (*cb)(FGControl* ,int), FGWindow* w, int step=20)
		 : FGRadioGroup(x,y,names,init_value,cb,w,step)
		 {
			 FGControl *c;
			 for(int i=0; i<count; i++)
			 {
				 c = DrawOne(x, y, i*step, names[i], w, callback);
				 if (i==0) id = c->GetId();
				 c->SetPrivateData((void *)this);
				 grp->AddToGroup(c, init_value ? i == *init_value : 0);
				 AddToContainer(c);
			 }
		 }
};

/**
* Widget to easier using of horizontal arrays of RadioButtons.
* @image html radio.png
*/
class FGRadioGroupHorizontal : public FGRadioGroup
{
		FGControl* DrawOne(int x, int y, int offset, char* name, FGWindow* w, ControlCall cb)
	    {
		    return w->AddPointButton(x+offset, y, name, 0, 0, cb);
		}
	public:
		/**
		* Creates the group of PointButtons at exact coords into the FGWindow. The number of created
		* objects is equivalent of the number of passed strings.
		* @param x start x coord
		* @param y start y coord
		* @param names an array of ASCIIZ strings terminated by 0 pointer, by example: char *names[]={"one","two","three",0};
		* @param init_value an initial index of active object on create time, by example: 1 for string "two"
		* @param cb standard OpenGUI CallBack routine
		* @param w the parent FGWindow object
		* @param step offset of x coordinates for each next button
		*/
		FGRadioGroupHorizontal(int x, int y, char *names[], int *init_value, void (*cb)(FGControl* ,int), FGWindow* w, int step)
		: FGRadioGroup(x,y,names,init_value,cb,w,step)
		{
			FGControl *c;
			for(int i=0; i<count; i++)
			{
				c = DrawOne(x, y, i*step, names[i], w, callback);
				if (i==0) id = c->GetId();
				c->SetPrivateData((void *)this);
				grp->AddToGroup(c, init_value ? i == *init_value : 0);
				AddToContainer(c);
			}
		}
};

/**
    A widget for better color pixking and mixing.

    It supports RGB and HSV/HLS color models.

    @image html picker.png

	@see FGColor, FG2Colors, FG3Colors, FG4Colors, FGRelatedColors

	<h3>Using the Color scheme</h3>

	This application generates color schemes of several types. Every scheme is based on one (base) color, which is supplemented with additional colors making together the best optical imperssion - using one of the authentic algorithms.
	Selecting the base color

	The base color can be set in several ways. The application works primarily with the color wheel, and with colors defined by it - hue (the angle determining position of the color on the wheel), saturation (100 % is the most saturated color, 0 is a shade of gray), and brightness (100 % is lightest shade, 0 is black). Warm colors (purple to yellow) and cold colors (yellow-green to violet) are marked here as well.

	We can choose one of the elemetary colors on the wheel by clicking on its circle. Primary colors (red, yellow, blue) correspond to angles 0°, 120°, and 240°, secondary colors (orange, green, violet) to angles 60°, 180°, and 300°, tertiary colors are between them. Smaller circles (multiples of 15°) are standing for transitional colors. Clicking on any of these colors, you'll set it as the base color.

	You can change parameters of the base color (hue, saturation, brightness) in the input fields. There is a scale 0-360° inside the color wheel, containing a small pointer corresponding to the angle of actual color. Clicking on this scale you can change the hue without changing brightness and/or saturation (you have to click, you cannot drag the pointer). The angle the cursor points at is shown in the browser's status line. Similarly, you can set saturation and/or brightness by clicking on the scales beside the color wheel.

	Important note: This tool doesn't use the standard HSV or HSB model (the same HSV/HSB values ie. in Photoshop describe different colors!). The color wheel used here differs from the RGB spectre used on computer screens, it's more in accordance with the classical color theory. This is also why some colors (especially shades of blue) make less bright shades than the basic colors of the RGB-model. In plus, the RGB-model uses red-green-blue as primary colors, but the red-yellow-blue combination is used here. This deformation also causes incompatibility in color conversions from RGB-values. Therefore, the RGB input (eg. the HTML hex values like #F854A9) is not possible.
	Color schemes

	The resulting color set is made by installing the base color into one of the color schemes. You can select the scheme from the menu on the right. Several schemes allow additional settings, which are displayed bellow the menu. You can choose one of following schemes:
	Monochromatic

	@image html prev_mono.gif

	Monochormatic scheme is based on only one color tint, and uses only variations made by changing its saturation and brightness. Black and white colors are always added. The result is comfortable for eyes, even when using aggressive color. However, it's harder to find accents and highlights.

	The application makes only several monochromatic variants of each color. You'll be able to make others - more or less saturated, lighter or darker. Monochromatic variations are made for each color in other schemes, too. They are displayed in the color palette bellow the color sample.
	Contrast

	Base color is supplemented with its complemet (color on the opposite side of the wheel). One warm and one cold color is always created - we have to consider, which one will be dominant, and if the result should look warm, or cold. Suitable monochromatic variations of this two colors may be added to the scheme.
	"Soft" contrast

	Base color is supplemented with two colors, placed identically on both sides of its complement. Unlike the "sharp" contrast, this scheme is often more comfortable for the eyes, it's softer, and has more space for balancing warm and cold colors.

	Additional input field (will be displayed bellow the scheme menu) sets the distance of these colors from the base color complement. The less the value is, the closer the colors are to the contrast color, and are more similar. The best value is between 15-30°. Higher values aren't too suitable - except the shift by 60°, which makes another color scheme, the triade:
	Triade

	@image html prev_triad.gif

	The triade is made by three colors evenly distributed on the thirds of the color wheel (by 120 degrees). The triade-schemes are vibrating, full of energy, and have large space to make contrasts, accents and to balance warm and cold colors. You can make the triade in the "soft contrast" scheme setting the distance to 60°.
	"Double-contrast"

	@image html prev_tetrad.gif

	This scheme is made by a pair of colors and their complements. It's based on the tetrade - the foursome of colors evenly distributed on the fourths of the color wheel (by 90 degreees). The tetrade is very aggressive color scheme, requiring very good planning and very sensitive approach to relations of these colors.

	@image html prev_compl.gif

	Less distance between two base colors causes less tension in the result. However, this scheme is always more "nervous" and "action" than other schemes. While working with it, we have to take care especially of relations between one color and the complement of its adjacent color - in case of the tetrade (90° distance), good feeling and very sensitive approach are necessary.
	Analogic colors

	@image html prev_analog.gif

	This scheme is made by base color and its adjacent colors - two colors identically on both sides. It always looks very elegantly and clear, the result has less tension and it's uniformly warm, or cold. If a color on the warm-cold border is chosen, the color with opposite "temperature" may be used for accenting the other two colors.

	Additional input field sets the distance of adjacent colors - values between 15-30° are optimal. You can also add the contrast color, the scheme is then supplemented with the complement of the base color. It must be treated only as a complement - it adds tension to the palette, and it's too aggressive when overused. However, used in details and as accent of main colors, it can be very effective and elegant.
	Contradictory colors

	There is no scheme named "contradictory" here - on the contrary, it's a color scheme, which can't fit into any rule described above. This scheme may not be useless - there may be situations, when we have to create shocking, really gaudy, jazzy work. Even in this case this tool may be usefull - if your scheme differs from all described models, you've got it: nervous, loud, aggressive set of colors. Of course, it cannot be suitable for applications requiring not so intensive emotions.
	Color palette

	To keep the impression of the color scheme, monochromatic variations shouldn't be overused - only black and white are often recommended. If you really need other variations, choose just several variants of the base color; possibly one or two variants of other (contrast) colors as well. Another possibility is to use the same variation of all main colors (eg. the same theme in lighter, paler, or darker shades). In general, one application (one WWW page, one web site) shouldn't use - besides black and white - more than six colors (including all variants). This tool hopefully can help you harmonize them.

	Color schemes selector
	Copyright (c) 2002 pixy
	Homepage: www.pixy.cz

*/
class FGColorPicker
{
        FGWindow*   wnd;
        FGColor*    base_color;
        FGColor     the_color;

        FGControl*  ctrl_sat;
        FGControl*  ctrl_lum;
        FGControl*  ctrl_hue;
        FGControl*  ctrl_r;
        FGControl*  ctrl_g;
        FGControl*  ctrl_b;

    	int         m_saturation;
	    int         m_luminance;
    	int         m_hue;

    	int         m_r;
	    int         m_g;
    	int         m_b;

        static void PickerProc(FGEvent* e);
        void DrawColorCircle(void);
        static void Ok(CallBack cb)
        {
            FGColorPicker *_THIS = (FGColorPicker *)cb->GetOwner()->GetUserData();
            * (_THIS->base_color) = _THIS->the_color;
            cb->Close(cb);
        }
        static void DrawColorHSV(CallBack cb)
        {
			FGColorPicker *_THIS = (FGColorPicker *)cb->GetOwner()->GetUserData();
			_THIS->the_color.SetHLS(_THIS->m_hue, _THIS->m_luminance/100., _THIS->m_saturation/100.);
			_THIS->DrawColor();
		}
		static void DrawColorRGB(CallBack cb)
		{
			FGColorPicker *_THIS = (FGColorPicker *)cb->GetOwner()->GetUserData();
			_THIS->the_color.SetRGB(_THIS->m_r, _THIS->m_g, _THIS->m_b);
			_THIS->DrawColor();
		}
		void SetState(FGColor& c)
		{
			the_color = c;
			m_saturation = int(the_color.GetSaturation()*100.);
			m_luminance = int(the_color.GetLuminance()*100.);
			m_hue = int(the_color.GetHue());
			m_r = the_color.GetRed();
            m_g = the_color.GetGreen();
            m_b = the_color.GetBlue();
            DrawColor();
        }
        void DrawColor(void);
        void Box(int x, int y, FGColor paper);
    public:
        /**
            Creates dialog box for smart color picking.
            The dialog also shows two/three colors and related
            colors. You can also click to any color within dialog window to
            pick new color data.
            @param _base_color the pointer to the color
            if user click OK, then the pointer will be use to store
            the selected color.
        */
        FGColorPicker(FGColor* _base_color);
        ~FGColorPicker()
        {
            delete wnd;
            wnd = 0;
        }
};

#ifdef __BORLANDC__
#pragma option pop /* pop -a switch */
#pragma option pop /* pop -b */
#endif

#ifdef FG_NAMESPACE
}
#endif

#endif
