/*
    OpenGUI - Drawing & Windowing library

	Copyright (C) 1996,2003  Marian Krivos

    This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
	version 2 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
	License along with this library; if not, write to the Free
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	nezmar@atlas.sk

    fastgl.cc - graphics engine

  $Id: fastgl.cpp,v 1.29 2004/09/27 09:45:22 majo Exp $

  $Log: fastgl.cpp,v $
  Revision 1.29  2004/09/27 09:45:22  majo
  Calibration part II.

  Revision 1.28  2004/09/27 06:59:57  majo
  SetParameter(void *)
  Calibration part I.

  Revision 1.27  2004/09/07 06:19:01  majo
  centered resize

  Revision 1.26  2004/09/02 10:18:17  majo
  fixed size WindowPanel()

  Revision 1.25  2004/08/26 05:47:22  majo
  added FGButtonGroup::GroupEnable(), Disable()
  fixed resize of the centred window

  Revision 1.24  2004/08/25 21:16:10  majo
  priebezne zmeny

  Revision 1.23  2004/07/13 17:45:41  majo
  fixed FGApp::Run() - no more corrupted event in message queue

  Revision 1.22  2004/07/07 16:51:32  majo
  FGSize
  FGWindowMenu::AddteText()
  FGMatrix::scale_with_center()

  Revision 1.21  2004/05/24 17:38:27  majo
  documentation fixes

  Revision 1.20  2004/04/27 17:39:19  majo
  *

  Revision 1.19  2004/04/19 20:05:14  majo
  added color highlight to FGEditBox when text string is ready to delete (when type mode is activated)

  Revision 1.18  2004/04/16 08:54:02  majo
  fixed bug with minimal size for FGWindow::WindowResize()
  added int FGWindow::GetNumberOfTabPages()
  added void FGWindow::DeleteTabPage(char *)
  reimplemented FGBitmap

  Revision 1.17  2004/03/24 17:13:58  majo
  added FGXColor
  added FGColorPicker

  Revision 1.16  2004/03/18 16:05:27  majo
  readded changes to branch

  Revision 1.15  2004/03/17 21:19:12  majo
  added FGColor, FGImage

  Revision 1.14  2004/03/09 20:23:58  majo
  added FGImage
  improved FGColor & and its docs

  Revision 1.13  2004/03/01 15:19:15  majo
  _GTRANSP -> _GCOLORKEY
  added _GTRANSPARENT

  Revision 1.12  2004/02/29 19:58:19  majo
  devel state

  Revision 1.11  2004/02/26 21:41:28  majo
  class FGInternalState
  char* -> const char*
  int ink = -1  -> unsigned ink = UNDEFINED_COLOR

  Revision 1.10  2004/02/23 20:08:01  majo
  all classes are with prefix FG* on now
  polygon functions uses FGPointArray from now
  class GuiEvent is renamed to FGEvent
  some by parameters overloaded methods was removed (class FGWindow)
  many other small changes

  Revision 1.9  2004/02/17 17:43:16  majo
  added FGColorScheme class constructor
  bigger text. buffer for editbox (from 128 to 252)

  Revision 1.8  2004/02/12 17:51:52  majo
  added class FGDialog
  changed/fixed CScheme members
    edit_back -> edit_background_active
    edit_bord1 -> edit_border
    edit fore -> edit_foreground
    edit_disable -> edit_disable_fore
  added:
    edit_background

  Revision 1.7  2004/02/10 16:18:59  majo
  fixed compile warnings

  Revision 1.6  2004/02/09 16:12:13  majo
  fixed FGWindow::ShowModal() - modal window no more closes on Controls
  without explicit value assignments & FGApp:;AppDone() causes direct
  shutdown of :wq
  the application.

  Revision 1.5  2004/02/06 16:33:47  majo
  fixed FGTabPages and Contols when TabPage is not active

  Revision 1.4  2004/02/04 17:55:01  majo
  fixed TabPages & FGControl::draw() visual errors

  Revision 1.3  2004/01/31 10:06:58  majo
  Added FGWindow::AddPushButton(int x, int y, int w, int h, char* name, int key, FGDrawBuffer* background, ControlCall fnc);

  Revision 1.2  2004/01/27 16:27:14  majo
  gcc 3.4.0 update

  Revision 1.1.1.1  2004/01/22 16:25:08  majo
  initial release

*/

#include <stdarg.h>
#include <stdlib.h>
#include <errno.h>

#ifdef __WATCOMC__
#ifdef __DOS__
#include <direct.h>
#else
#include <dirent.h>
#endif // dos
#elif __linux__
#include <dirent.h>
#include <sys/time.h>
#include <unistd.h>
#elif _MSC_VER
#include <direct.h>

#endif // watcom

#ifdef __BORLANDC__
#include <dir.h>
#endif

#include "config.h"
#include "fastgl.h"
#include "_fastgl.h"
#include "drivers.h"

#include <list>

#ifdef FG_TTF
#include "fgttf.h"
#endif
#include "listbox.h"

//
// Global definitions & declaration
//

#ifdef FG_NAMESPACE
namespace fgl {
#endif

FGApp *cApp;

FGInternalState fgstate;

WindowDatabase *entryPoint;
const char *FGDrawBuffer::empty_string="";
int FGWindow::nWnd = 0;			// numbers of all windows
FGWindow *FGWindow::current = (FGWindow *) 0;
FGWindow *FGWindow::first = (FGWindow *) 0;

char FGEditBox::clip[bufsize];
int  FGEditBox::is_clip;

void 			(*FGApp::DelayProc) (void) = 0;
void 			(*FGApp::OnEverySecond)(int) = 0;
FGEvent 		FGApp::queque[MAX_EVENT_QUEQUE];
MainHwnd 		FGApp::appHandler;
int 			FGApp::quequeIndex = 0;
char			FGApp::currdir[pathsize];
char 	*		FGApp::homedir = 0;
FGWindow	*		FGApp::Root = 0;
int				FGApp::flags;
int				FGApp::video;
int				FGApp::Argc;
int				FGApp::background;
char	**		FGApp::Argv;
char		   *FGApp::name;
unsigned long 	FGApp::ticks;
int 			FGApp::ttf_support = 0;

static FGBaseMenu *inmenu;
FGMenuWindow *FGMenuWindow::CurrentMenu=0;
GuiHwnd FGMenuWindow::Proc=0;

static FGWindow *_Note_Ptr, *close_adept;

static unsigned int CRCTab[256];
static int alty[26] = {
ALT_A,ALT_B,ALT_C,ALT_D,ALT_E,ALT_F,ALT_G,ALT_H,
ALT_I,ALT_J,ALT_K,ALT_L,ALT_M,ALT_N,ALT_O,ALT_P,
ALT_Q,ALT_R,ALT_S,ALT_T,ALT_U,ALT_V,ALT_W,ALT_X,
ALT_Y,ALT_Z };

static void InitCRC(void);

/**
	Internal implementation of tab pages
	@internal
*/
class FGTabPage
{
		FGControl		*self;
		const char	*idstring;
		int			width, ypos, flag;
	public:
		FGTabPage() { self = 0; idstring = 0; width = 0; ypos = 0; flag = 0;}
		FGTabPage(const char *s, FGControl *p, int y) : self(p), idstring(s?strdup(s):0), ypos(y)
		{
			width = FGFontManager::textwidth(FONTSYSLIGHT,(char *)s) + NOTEGAP*2;
			flag = 0;
		}
		FGTabPage(const FGTabPage& old)
		{
			self = old.self;
			idstring = strdup(old.idstring);
			width = old.width;
			ypos = old.ypos;
			flag = old.flag;
		}
		~FGTabPage() { if (idstring) free((void *)idstring); idstring=0; }
		const char	*GetString(void) { return idstring; }
		int			GetWidth(void) { return width; }
		int			GetY(void) { return ypos; }
		int			GetFlag(void) { return flag; }
		void		SetFlag(int f) { flag = f; }
		FGControl 	*GetControls(void) { return self; }
};

typedef std::list<FGTabPage> TabPageList;
typedef std::list<FGTabPage>::iterator TabPageListPtr;

FGColorScheme::FGColorScheme()
{
	window_back     = CWHITED;
	window_fore     = CBLACK;		// window colors
	active_title    = CBLUELIGHT;   // active title
	inactive_title  = CBROWN;		// inactive title;
	wnd_bord1       = CWHITE;		// wnd border 1
	wnd_bord2       = CGRAYED;	    // wnd border 2
	wnd_bord3       = CGRAY2;		// wnd border 3
	statusbar       = CGRAY2;		// statusbar
	menu_back       = CGRAY2;		// menu bg
	menu_fore       = CBLUE;		// menu fore
	menu_back_active = CGRAY2;		// menu bg active
	menu_fore_active = CWHITE;		// menu fore active
	button_fore     = CBLACK;  	    // pbutton fore
	button_back     = CGRAY1;		// pbutton back
	button_fore_pushed = CDARK;		// pbutton fore - pushed
	button_back_pushed = CGRAY3;	// pbutton back - pushed
	button_bord1    = CWHITE;		// control::frame1
	button_bord2    = CBLACK;		// control::frame2
	button_bord3    = CGRAYED;	    // control::frame3
	edit_background_active = CWHITE;// edit back
	edit_foreground = CBLACK;		// edit fore
	edit_border     = CDARK; 		// edit bord1
	edit_background = CGRAY3;		// edit bord2
	slider          = CGRAY1;		// slider
	menuwindow_back = CGRAY2;		// menuwindow back
	menuwindow_fore = CBLACK;		// menuwindow fore
	pdmenu_back_active = CBLUELIGHT;// pdmenu back
	pdmenu_fore_active = CWHITE;	// pdmenu fore
	pdmenu_gray     = CGRAY3;		// pdmenu gray
	edit_fore_disable = CGRAY1;		// editbox disable
	notebook_active = CBLUELIGHT;   // notebook active
	slider_disable  = CGRAY3;		// slider disable
};

static FGColorScheme _FGCS;
FGColorScheme * CScheme = &_FGCS;

/**
	Internal implementation of Icon object
	@internal
*/
class IconWindow : public FGWindow
{
		FGWindow *original;
		static void Proc(FGEvent *p)
		{
			IconWindow *pp = (IconWindow *)p->wnd;
			if (p->GetType()==CLICKRIGHTEVENT)
				pp->original->WindowIconize();
			if (p->GetType()==KEYEVENT && p->GetKey()==ALT_F02)
				pp->original->WindowIconize();
			if (p->GetType()==WINDOWMOVEEVENT)
			{
				if (entryPoint) entryPoint->DatabaseSetIcon(p->GetX(), p->GetY(), pp->original->GetName());
				pp->original->SetIconPosition(p->GetX(), p->GetY());
			}
		}
	public:
		IconWindow(FGWindow *parent, int xs, int ys, FGDrawBuffer *ico)
			: FGWindow(0, xs, ys, ico?ico->GetW()+4:48, ico?ico->GetH()+4+10:48, "Icon", Proc, 0, CWHITE, WNOPICTO|WCLICKABLE|WNODRAWFROMCONSTRUCTOR|WICONWINDOWTYPE)
		{
			icon = ico;
			if (icon==0)
				icon = new FGDrawBuffer(44, 44, BMP_MEM, CGRAYED);
			original = parent;
			original->WindowAttachIcon(icon);
			WindowBox(0,0,w,h,CBLUELIGHT);
			WindowBox(2,10,w-4,2,CWHITE);
			WindowSetWorkRect(2,2,w-4,h-4);
			set_ppop(_GSET);
			set_font(0);
			draw();
			WindowText(0,0,parent->GetName(), CWHITE, CBLUELIGHT);
		}
		virtual void draw(void)
		{
			if (icon)
			{
				WindowPutBitmap(0,10,0,0,w,h,icon);
			}
		}
};

/**
	Redraws a whole FGControl in a parents FGWindow.
*/
void FGControl::update_owner(void)
{
	// draw  only if this is your own TabPage
	if (IsYourTabPage())
	{
		SAVE_CLIPPING(owner)
	    owner->bitblit(x,y,0,0,w,h,this);
		RESTORE_CLIPPING(owner)
	}
}

bool FGControl::IsYourTabPage(void)
{
	return (tab_page_controls == owner->Buttony);
}

int	FGControl::GetXr(void)
{
	return owner->GetX() + x;
}

int	FGControl::GetYr(void)
{
	return owner->GetY() + y;
}

void FGWindow::SetActive(FGControl *c)
{
	call_handler = c;
	c->OnActivate();
}

void FGWindow::RunHandler(void)
{
	if (this)
	{
		if (call_handler)
		{
			FGControl *c = call_handler;
			call_handler = 0; // no more call it
			if (c->GetHandler())
			{
				c->GetHandler()(c);
			}
			fgstate.__hint_x_menu = fgstate.__hint_y_menu = -1;
		}
	}
}

void RootResize(int ww, int hh)
{
	FGWindow::WindowVirtualRoot(0,0,ww,hh);
}

//
// Internal
// kresli ikonky na titulok okna
void FGWindow::WindowPiktograms(int x, int y)
{
	int action= status&WMINIMIZE?1:0;
	action |= status&WNOPICTO?0:2;
	switch(action)
	{
		case 0:	// nothing
			break;
		case 3:	// all (minimize+close)
			RamToRam(0, 0, PICTOSIZE*2, PICTOSIZE, x, y, w,h, PICTOSIZE*2, PICTOSIZE, _internal_bitmap0, image, BITBLIT_WITH_CLIPPING, _GSET);
			break;
		case 1:	// (minimize)
			RamToRam(0, 0, PICTOSIZE*2, PICTOSIZE, x+PICTOSIZE, y, w,h, PICTOSIZE, PICTOSIZE, _internal_bitmap0, image, BITBLIT_WITH_CLIPPING, _GSET);
			break;
		case 2:	// (close)
			RamToRam(PICTOSIZE, 0, PICTOSIZE*2, PICTOSIZE, x+PICTOSIZE, y, w,h, PICTOSIZE, PICTOSIZE, _internal_bitmap0, image, BITBLIT_WITH_CLIPPING, _GSET);
			break;
	}
}

//! A traditional printf in window form
/**
Writes formatted output to the window at current position.
When the whole FGWindow is printed out, image automatically scroll-up.

The printf function:
	-	Accepts a series of arguments
	-	Applies to each argument a format specifier contained in the format string *format
	-	Outputs the formatted data (to the screen, a stream, stdout, or a string)

There must be enough arguments for the format. If there are not, the results will
be unpredictable and likely disastrous.  Excess arguments (more than required
by the format) are merely ignored.

The format string, present in each of the printf function calls, controls how each
function will convert, format, and print its arguments.
@note There must be enough arguments for the format; if not, the results will
be unpredictable and possibly disastrous. Excess arguments (more than required
by the format) are ignored.

The format string is a character string that contains two types of objects:
	-	Plain characters are copied verbatim to the output stream.
	-	Conversion specifications fetch arguments from the argument list and apply formatting to them.

Plain characters are simply copied verbatim to the output stream.
Conversion specifications fetch arguments from the argument list and apply formatting to them.

see 'man printf' for more details

@return On success, printf returns the number of bytes output. On error, printf returns EOF.
@param format formatting string as detailed above
*/
int FGWindow::printf(const char *format, ...)
{
	char s[256];
	va_list arglist;

	va_start(arglist, format);
#ifndef _MSC_VER
	vsnprintf(s, sizeof(s)-1, format, arglist);
#else
	vsprintf(s, format, arglist);
#endif
	va_end(arglist);
	puts(s);
	return strlen(s);
}

/**
	A traditional printf in window form + position
	As printf() but at the exact position.
*/
int FGWindow::printf(int xx, int yy, const char *format, ...)
{
	char s[256];
	move(xx, yy);
	va_list arglist;

	va_start(arglist, format);
#ifndef _MSC_VER
	vsnprintf(s, sizeof(s)-1, format, arglist);
#else
	vsprintf(s, format, arglist);
#endif
	va_end(arglist);
	int len = strlen(s);
	if (s[len-1] == '\n')
	{
		s[--len] = 0;
	}
	WindowText(xx,yy,s);
	return len;
}

//! Put char to window at current print position
void FGAPI FGWindow::wputc(int c)
{
	static FGMutex TextOutput(recursiveMutEx);

	TextOutput.Lock();

	if (c == -1)
	{
		WindowRepaintUser(0, ypos, xpos, GetFontH());
		TextOutput.Unlock();
		return;
	}

	if (xpos == -1)
	{
		WindowScrollUp(0, GetFontH(), wwrk, hwrk - GetFontH() - hwrk % GetFontH(), GetFontH());
		xpos = 1;
	}

	if (xpos < (wwrk - GetFontW()))
	{
		if (c == '\n')
		{
			if (ypos > (hwrk - GetFontH() * 2))
			{
				WindowRepaintUser(0, ypos, xpos, GetFontH());
				xpos = -1;
			}
			else
			{
				WindowRepaintUser(0, ypos, xpos, GetFontH());
				ypos += GetFontH();
				xpos = 1;
			}
		}
		else
		{
			text(xpos, ypos, (char *) &c);
			xpos += GetFontW();
		}
	}
	else
	{
		xpos -= GetFontW();
		wputc('\n');
		wputc(c);
	}
	TextOutput.Unlock();
}

/**
 Internal.
 * return 0 if in window area
 * else 1 for title
 * 2 for close
 * 3 for minimize
 * 4 for resize
 */
int FGAPI FGWindow::TitleFind(FGEvent * e)
{
	if (!IsVisible()) return 0;

	if ((x <= e->GetX()) && ((x + w) >= e->GetX()))
	{
		if (((x + w - 13) <= e->GetX()) &&
			((y + h - 13 <= e->GetY()) && ((y + h) >= e->GetY())) && status & WSIZEABLE)
				return 4;
		if (status & WTITLED)
		{
			if ((y <= e->GetY()) && ((y + 3+TITLEH) >= e->GetY()))
			{
				if (!(status & WNOPICTO)) // close or close&minimize
				{
					if (e->GetX() >= (x + w - 3 - PICTOSIZE))
						return 2;
					if (e->GetX() >= (x + w - 3 - PICTOSIZE*2))
						if (status&WMINIMIZE) return 3;
						else return 0;
				}
				else if (status & WMINIMIZE)
				{
					if (e->GetX() >= (x + w - 3 - PICTOSIZE))
						return 3;
				}
				return 1;
			}
		}
		else if (status & WUNMOVED) return 0;
		else if ((y <= e->GetY()) && ((y + h) >= e->GetY()) && status&WICONWINDOWTYPE)
			return 1;
		else if ((y <= e->GetY()) && ((y + TITLEH+3) >= e->GetY()))
			return 1;
	}
	return 0;
}

/**
	internal
*/
void FGAPI FGWindow::DestroyWindow(void)
{
	if (nWnd && fgstate.verbose)
		::printf("WARNING: %d window(s) not closed at exit\n", nWnd);
}

int	FGAPI FGWindow::_over(int xa1, int ya1, int xa2, int ya2, int xx, int yy, int ww, int hh)
{
	register int xb1 = xx, xb2 = xx + ww, yb1 = yy, yb2 = yy + hh;

	if ((xa1 <= xb1 && xa2 >= xb1) || (xa1 <= xb2 && xa2 >= xb2))	// x sa prekryva
	{
		if (ya1 <= yb1 && ya2 >= yb1)
		{
			return 1;
		}
		if (ya1 <= yb2 && ya2 >= yb2)
		{
			return 2;
		}
		if (yb1 <= ya1 && yb2 >= ya1)
		{
			return 3;
		}
		if (yb1 <= ya2 && yb2 >= ya2)
		{
			return 4;
		}
	}
	else if ((xb1 <= xa1 && xb2 >= xa1) || (xb1 <= xa2 && xb2 >= xa2))
	{
		if ((yb1 <= ya1 && yb2 >= ya1))
		{
			return 5;
		}
		if ((yb1 <= ya2 && yb2 >= ya2))
		{
			return 6;
		}
		if (ya1 <= yb1 && ya2 >= yb1)
		{
			return 7;
		}
		if (ya1 <= yb2 && ya2 >= yb2)
		{
			return 8;
		}
	}
	return 0;
}

//! redraw block unconditionally
void FGAPI FGWindow::WindowUpdateBlock(int x, int y, int w, int h)
{
	clip_shape(x,y,w,h);
	RepaintBlock(x,y,w,h);
}

//! params are real size & position of cleared window
void FGAPI FGWindow::intersect(int _x, int _y, int _w, int _h)
{
	FGWindow *p = first;
	int rc;

	if (!(cApp->flags&APP_ROOTWINDOW)) fill_box(_x,_y,_w,_h,cApp->background,_GSET);

	// now we trace all windows from down to *this*
	while (p)
	{
		if (!(p->status&WHIDEN) && p!=this)
		{
			rc=_over(p->x, p->y, p->x+p->w, p->y+p->h, _x, _y, _w, _h);
			if (rc)
			{
				p->WindowUpdateBlock(_x-p->GetX(), _y-p->GetY(), _w, _h);
			}
		}
		p =	(FGWindow *) ((FGBaseGui *)p)->GetNext();
	}
}

/**
	redraw window intersection with original background
	dx,dy are deltas from original
*/
void FGAPI FGWindow::WindowUpdate(int dx, int dy, int all)
{
	if (all || abs(dx) > w || abs(dy) > h)
	{
		intersect(x,y,w,h);
		return;
	}
	if (dx>0) // right side
	{
		intersect(x,y,dx,h);
	}
	if (dy>0) // up side
	{
		intersect(x,y,w,dy);
	}

	if (dx<0) // left side
	{
		intersect(x+w+dx,y,-dx,h);
	}
	if (dy<0) // down side
	{
		intersect(x,y+h+dy,w,-dy);
	}
}

//! Test overlaping window, called from OdkryteOkno only
FGBaseGui *FGBaseGui::OverSprite(FGBaseGui * od, int xx, int yy, int ww, int hh)
{
	register int xa1, xa2, ya1, ya2;
	FGBaseGui *v = 0;
	int xb1 = xx, xb2 = xx + ww, yb1 = yy, yb2 = yy + hh;

	while (od != this)
	{
		if (!(od->status & WHIDEN))
		{
			xa1 = od->x,
				xa2 = od->x + od->w,
				ya1 = od->y,
				ya2 = od->y + od->h;

			if ((xa1 <= xb1 && xa2 >= xb1) ||
				(xa1 <= xb2 && xa2 >= xb2))	// x sa prekryva
			{
				if ((ya1 <= yb1 && ya2 >= yb1) ||
					(ya1 <= yb2 && ya2 >= yb2))
				{
					v = od;
					break;
				}
				else if (yb1 <= ya1 && yb2 >= ya2)
				{
					v = od;
					break;
				}
			}
			else if ((xb1 <= xa1 && xb2 >= xa1) ||
				(xb1 <= xa2 && xb2 >= xa2))
			{
				if ((yb1 <= ya1 && yb2 >= ya1) ||
					(yb1 <= ya2 && yb2 >= ya2))
				{
					v = od;
					break;
				}
				else if (ya1 <= yb1 && ya2 >= yb2)
				{
					v = od;
					break;
				}
			}
		}
		od = od->last;
	}
	return v;
}

//! Called from WindowRepaint() for the window overlap testing
FGWindow * FGAPI FGWindow::OdkryteOkno(FGWindow * pokial, int x, int y, int w, int h)
{
	FGWindow *Colise=0;

	if (!IsVisible())
		return (FGWindow *)-1;	// je skryte

	if (!GetNext())
		return 0;				// je aktivne

	while (pokial != this)
	{
		if ((Colise = (FGWindow *) OverSprite(pokial,x, y, w, h)) != 0)
		{
			return Colise;
		}
		pokial = (FGWindow *) pokial->GetLast();
	}
	return 0;
}

#define area(a,b,c,d) *out++=a,*out++=b,*out++=c,*out++=d; //, debug==1?(Ladenie("a,b,c,d - %d,%d,%d,%d\n",a,b,c,d)):0;

int FGAPI FGWindow::Rozporcuj(int X1, int Y1, int X2, int Y2, int *in, int *out)
{
	int x1, y1, x2, y2, f;

	if (*in == -1) return 0;
	while (*in != -1)
	{
		x1 = *in++;
		y1 = *in++;
		x2 = *in++;
		y2 = *in++;
		f = 0;
		if (x1 >= X1 && y1 >= Y1 && x2 <= X2 && y2 <= Y2)
			continue;				// cely prekryty

		if (x2 < X1 || y2 < Y1 || x1 > X2 || y1 > Y2)	// cely mimo
		{
			area(x1, y1, x2, y2);
			continue;
		}

		if (x1 <= X1 && x2 >= X1)
		{
			area(x1, y1, X1, y2);
			x1 = X1;
		}

		if (y1 <= Y1 && y2 >= Y1)
		{
			area(x1, y1, x2, Y1);
			y1 = Y1;
		}

		if (x2 >= X2 && x1 <= X2)
		{
			area(X2, y1, x2, y2);
		}

		if (y2 >= Y2 && y1 <= Y2)
		{
			area(x1, Y2, x2, y2);
		}
	}
	*out = -1;
	return 1;
}

//! actualize window at relative coors in it
void FGAPI FGWindow::WindowRepaint(int xr, int yr, int w, int h)
{
	ListMutex.Lock();

	int zoznam1[4096], zoznam2[4096], *p1, *p2, *p3;
	FGBaseGui *posledne = this;
	int opt=optimize;
	FGWindow *Colise;

opt=0;

	optimize=0;	// cancel accel hint
	if (status & WLOCKED)
	{
		status |= WDIRTY;
		goto exi;
	}

	// clipping
	if (clip_shape(xr,yr,w,h) == 0) goto exi;

	for (; posledne->GetNext(); posledne = posledne->GetNext()); // at the end

	// test overlap
	Colise = OdkryteOkno((FGWindow *) posledne, x + xr, y + yr, w, h);

	switch (int(Colise))
	{
		case -1:
			status |= WDIRTY;
			goto exi;
		case 0:
			switch(opt)
			{
				case 1:
					draw_line(x+line_x1, y+line_y1, x+line_x2, y+line_y2, optimize_color, state._ppop);
					break;
				default:
				case 0:
					RepaintBlock(xr, yr, w, h);
					break;
				case 2:
					fill_box(x+xr, y+yr, w, h, optimize_color, state._ppop);
					break;
				case 3:
					draw_point(x+xr, y+yr, optimize_color, state._ppop);
					break;
			}
			goto exi;
	}
	// chyba je v okne Colise a mozno aj v oknach medzi nimy
	posledne = Colise;
	p1 = zoznam1;
	p1[0] = x + xr;
	p1[2] = p1[0] + w;
	p1[1] = y + yr;
	p1[3] = p1[1] + h;
	p1[4] = -1;
	p2 = zoznam2;		// inicializacia poli
	p2[0] = -1;

	status |= WDIRTY;
	for (;;)
	{
		if (posledne->IsVisible())
		{
			if (!Rozporcuj(posledne->GetX(), posledne->GetY(), posledne->GetX() + posledne->GetW(), posledne->GetY() + posledne->GetH(), p1, p2))
				goto exi; // ???
			p3 = p1;
			p1 = p2;
			p2 = p3;
		}
		posledne = posledne->GetLast();
		if (posledne == this) // finito?
			break;
		if (int(OdkryteOkno((FGWindow *) posledne, x + xr, y + yr, w, h)) == -1)
			break;
	}
	while (*p1 != -1)
	{
		RepaintBlock(p1[0] - x, p1[1] - y, p1[2] - p1[0], p1[3] - p1[1]);
		p1 += 4;
	}
exi:
	ListMutex.Unlock();
}

//! redraws rectangle area in a user space of the FGWindow
void FGAPI FGWindow::WindowRepaintUser(int xr, int yr, int wr, int hr)
{
	xr += xoff;
	yr += yoff;
	if (clip(xr,yr,wr,hr))
	{
		WindowRepaint(xr,yr,wr,hr);
	}
}

//! Hide the FGWindow from the screen.
//! You can still draw to the FGWindow.
void FGWindow::WindowHide(void)
{
	WindowUpdate(0,0,1); // zmaze ho cele podkladom
}

//! Shows the hidden FGWindow on the screen.
void FGWindow::WindowShow(void)
{
	if (!iconized)
	{
		RepaintBlock(0, 0, w, h);
		SetStatus(GetStatus() & ~(WDIRTY));
	}
}

//! Show/hide the FGWindow conditionally.
void FGWindow::WindowRecopy(void)
{
	if (!IsVisible())
		WindowHide();
	else
		WindowShow();
}

//! internal
long FGAPI FGWindow::WindowStatus(long stat)
{
	int tmp;
	tmp = status;

	switch (stat)
	{
		case WHIDE:
			if (!IsVisible()) break;
			status |= WHIDEN;
			WindowRecopy();
			break;

		case WVISIBLE:
			if (IsVisible()) break;
			status &= (~WHIDEN);
			WindowRecopy();
			break;

		case WDEACTIVE:
			WindowDrawTitle(CTITLE);
			break;

		case WACTIVE:
			WindowDrawTitle(CTITLEACTIVE);
			break;
	}
	return tmp;
}

//! switches the FGWindow to the front - it get focus.
void FGAPI FGWindow::WindowFocus(void)
{
	FGWindow *old = current, *novy = this;
	FGEvent a1(LOSTFOCUSEVENT), a2(GETFOCUSEVENT);

	if (iconized && iconized!=(FGWindow *)-1)
		novy = iconized;
	old->SendToWindow(&a1);

	// reload because change of current is possible
	old = current;

	if (nWnd <= 1) // no more window
		return;
	if (novy == old) // no more to switch
		return;

	old->WindowStatus(WDEACTIVE);	// deactive old

	ListMutex.Lock();

	if (novy == first)				// if switch to first in queque
	{
		first = (FGWindow *) novy->GetNext();
		first->SetLast(0);
	}
	else // switch to other than first
	{
		(novy->GetLast())->SetNext(novy->GetNext());
		(novy->GetNext())->SetLast(novy->GetLast());
	}
	current = novy;
	old->SetNext((FGBaseGui *) novy);	  // exchange
	novy->SetLast((FGBaseGui *) old);
	novy->SetNext(0);

	ListMutex.Unlock();

	novy->WindowStatus(WACTIVE); // active new
	oldidm = 0;
	novy->RepaintBlock(0,0,w,h); // refresh all
	SendToWindow(&a2);
}

static char *skipstr(char *s)
{
	while (*s)
	{
		if (*s > ' ')
			return s;
		s++;
	}
	return s;
}

FGControl::FGControl(int xs, int ys, int ws, int hs, const char *nm, ObjectType typ, int keycode, int flag, FGWindow * own, ControlCall f, FGPixel p, FGPixel i):
  FGBaseGui(xs += own->xoff, ys += own->yoff, ws, hs, nm, typ, i, p, flag)
{
	owner = own;
	grp=0;
	if (keycode >= 'a' && keycode <= 'z')
		keycode = toupper(keycode);
	key = keycode;
	fnc = f;
	local_id = owner->nControls++;

	ListMutex.Lock();
	SetLast(owner->GetLastButton() );

	if (!owner->Buttony)		// first control
	{
		owner->Buttony = this;
	}
	else
	{
		GetLast()->SetNext(this);
	}
	ListMutex.Unlock();
	tab_page_controls = owner->Buttony;
}

FGControl::~FGControl()
{
	FGBaseGui *l, *n;

// WHY??? - 9.3.2003
//	if (!owner) return;
	if (this==owner->call_handler) owner->call_handler = 0;

	// erase from visible !
	owner->WindowBox(x-owner->xoff, y-owner->yoff, w, h, owner->GetPaper());

	ListMutex.Lock();

	l = GetLast();
	n = GetNext();
	if (n == (FGControl *) 0 && l == (FGControl *) 0)	// no more windows ...
	{
		owner->Buttony = 0;
	}
	else if (n == (FGBaseGui *) 0)	// last in queqe
	{
//		owner->lastButton = l;
	}
	else if (l == (FGBaseGui *) 0)	// first in queqe
	{
		owner->Buttony = (FGControl *) n;
	}
	else
	{
	}
	owner->nControls--;
	owner = 0;
	ListMutex.Unlock();
}

void FGControl::Enable(void)
{
	SetStatus((GetStatus() & (~7)) | BNORMAL);
	draw();
}

void FGControl::Disable(void)
{
	SetStatus((GetStatus() & (~7)) | BDISABLED);
	draw();
}

void FGEditBox::Disable(void)
{
	FlushInput();
	SetStatus((GetStatus() & (~7)) | BDISABLED);
	draw();
}

void FGControl::ClickUp(int a)	// if TRUE, so call handler
{
	if ((status&7) == BDISABLED) return;
	status ^= WTRIGGER;
	draw();
	if (a)
	{
		owner->SetActive(this);
	}
}

void FGBaseMenu::ClickUp(int a)	// if TRUE, so call handler
{
	if ((status&7) == BDISABLED) return;
	draw();
	if (a)
	{
		fgstate.__hint_x_menu = x + owner->x;
		fgstate.__hint_y_menu = y + owner->y+20;
		owner->SetActive(this);
	}
}

void FGBaseMenuItem::ClickUp(int a)	// if TRUE, so call handler
{
	if ((status&7) == BDISABLED) return;
	draw();
	if (a)
	{
		owner->SetActive(this);
	}
}

FGControl *FGControl::ButtonFind(FGEvent * e, int n)
{
	FGWindow *curr = FGWindow::GetCurrent();
	FGControl *p = curr->GetCurrentControls();
	int c;

	// test for notebools and its bookmarks
	if ( e->GetButtons()&FG_BUTTON_LEFT )
		if (curr->ClickTabPage(e->GetX(), e->GetY()))
			return 0;

	if (!p || curr->IsIconized())
		return 0;				// niesu buttony alebo je to ikona

	if (e->GetKey() >= 'a' && e->GetKey() <= 'z')
		c = toupper(e->GetKey());
	else c = e->GetKey();

	if (e->GetType() == KEYEVENT)
	{
		for (;;)
		{
			if ((p->GetKey() == c) && (p->GetStatus() & n & 7) == n)
				return p;
			if ((p = (FGControl *) p->GetNext()) == 0)
				return 0;
		}
	}
	else
	{
		do
		{
			if ((p->owner->GetX() + p->owner->GetW() >= e->GetX()) && ((p->owner->GetY() + p->owner->GetH()) >= e->GetY()))
				if ((p->GetXr() <= e->GetX()) && ((p->GetXr() + p->GetW()) >= e->GetX()))
					if ((p->GetYr() <= e->GetY()) && ((p->GetYr() + p->GetH()) >= e->GetY()))
						if ((p->GetStatus() & n & 7) == n)
						{
							if (p->GetType() == BASEMENU
								|| p->GetType() == BASEMENUITEM)
								inmenu = (FGBaseMenu *) p;
							return p;
						}
						else if (p->GetStatus() == BDISABLED)
						{
							return 0;
						}
		}
		while ((p = (FGControl *) p->GetNext()) != 0);
	}
	return 0;
}

FGBaseGui *FGBaseGui::BaseGuiFind(FGEvent * e)
{
	FGBaseGui *x = 0;
	FGBaseGui *p=this;

	if (!p) return p;
	if (p->next)
			x = p->next->BaseGuiFind(e);
	if (x)
		return x;
	if ((p->x <= e->x) && ((p->x + p->w) >= e->x))
	{
		if ((p->y <= e->y) && ((p->y + p->h) >= e->y))
			if (!(p->status & (WHIDEN|WTEMPHIDEN)) && (p->type!=ROOTWINDOW))
				x = p;
	}
	return x;
}

//! Returns the FGWindow for that the events has been emited.
FGWindow* FGAPI FGWindow::WindowFind(FGEvent * e)
{
	return (FGWindow *) first->FGBaseGui::BaseGuiFind(e);
}

//! Finds the FGWindow by ID.
FGBaseGui *FGBaseGui::FindId(int idw)
{
	FGBaseGui *p= this;
	for (; p;)
	{
		if (p->GetId() == idw)
			break;
		if (p->next == 0)
			return 0;
		p = p->next;
	}
	return p;
}

//! Draws the frame of FGControl.
void FGControl::frame(void)
{
	vline(0, 0, h, CScheme->button_bord1);
	hline(0, 0, w, CScheme->button_bord1);
	vline(0 + w - 2, 0 + 1, h - 2, CScheme->button_bord3);
	hline(0 + 1, 0 + h - 2, w - 2, CScheme->button_bord3);
	vline(0 + w - 1, 0, h - 0, CScheme->button_bord2);
	hline(0, 0 + h - 1, w - 1, CScheme->button_bord2);
}

//! Draws a FGPushButton
void FGPushButton::draw(void)
{
	int dl= textwidth(name);	// maximalny zobrazitelny pocet znakov

	int xx = w / 2 - dl / 2,
		yy = h / 2 - GetFontH() / 2;

	frame();

	switch (status & 7)
	{
		case BNORMAL:
		    if (status & WBITMAP)
		    	bitblit(2, 2, 0, 0, icon->GetW(), icon->GetH(), icon);
			else
				box(1, 1, w - 3, h - 3, CScheme->button_back);

	    	if (name[0])
			{
				unsigned oldp = set_ppop(_GCOLORKEY);
    			text(xx, yy, name,CScheme->button_fore, CScheme->button_back);
				set_ppop(oldp);
	    		Underscore(xx , yy, CDARK);
		    }
		    frame();
			break;

		case BPUSHED:
			box(0, 0, w, h,CScheme->button_bord2);
			if (status & WBITMAP)
			{
				bitblit(0 + 2, 0 + 2, 0, 0, icon->GetW(), icon->GetH(), pushed ? pushed : icon);
			}
			else
			{
				box(2, 2, w - 4, h - 4,CScheme->button_back_pushed);
			}
			vline(0 + w - 2, 0 + 1, h - 2, CScheme->button_bord1);
			hline(0 + 1, 0 + h - 2, w - 2, CScheme->button_bord1);

			if (name[0])
			{
				unsigned oldp = set_ppop(_GCOLORKEY);
				text(xx, yy, name,CScheme->button_fore_pushed, CScheme->button_back_pushed);
				Underscore(xx , yy, CDARK);
				set_ppop(oldp);
			}
			break;

		case BDISABLED:
			box(0,0,w,h,CScheme->button_bord1);

			if (status & WBITMAP)
			{
				bitblit(0 + 2, 0 + 2, 0, 0, icon->GetW(), icon->GetH(), disabled ? disabled : icon);
			}
			else
				box(2, 2, w - 4, h - 4,CScheme->button_back);

			vline(w - 2, 1, h - 2, CScheme->button_fore_pushed);
			hline(1, h - 2, w - 2, CScheme->button_fore_pushed);
			rect(0, 0, w, h);

			if (name[0])
			{
				unsigned oldp = set_ppop(_GCOLORKEY);
				text(xx, yy, name,CScheme->button_bord3);
				set_ppop(oldp);
			}
			break;

		case BUNVISIBLE:
			clear(paper);
			break;
	}
	update_owner();
	owner->WindowRepaint(x, y, w, h);
}

void FGPointButton::MoveIn(void)
{
	rect(0, 0, w, h, CYELLOW);
	update_owner();
	owner->WindowRepaint(x, y, w, h);
}

void FGPointButton::MoveOut(void)
{
	rect(0, 0, w, h, paper);
	update_owner();
	owner->WindowRepaint(x, y, w, h);
}

void FGCheckButton::MoveIn(void)
{
	rect(0, 0, w, h, CYELLOW);
	update_owner();
	owner->WindowRepaint(x, y, w, h);
}

void FGCheckButton::MoveOut(void)
{
	rect(0, 0, w, h, paper);
	update_owner();
	owner->WindowRepaint(x, y, w, h);
}

void FGEditBox::MoveIn(void)
{
	rect(w1, 0, w2, GetH(), CYELLOW);
	update_owner();
	owner->WindowRepaint(x + w1, y, w2, GetH());
}

void FGEditBox::MoveOut(void)
{
	rect(w1, 0, w2, GetH(), CScheme->edit_border);
	update_owner();
	owner->WindowRepaint(x + w1, y, w2, GetH());
}

void FGPushButton::MoveIn(void)
{
	rect(0,0,w,h,CYELLOW);
	update_owner();
	owner->WindowRepaint(x,y,w,h);
}

void FGPushButton::MoveOut(void)
{
	frame();
	update_owner();
	owner->WindowRepaint(x,y,w,h);
}

void FGPushButton::Push(void)
{
	SetStatus((GetStatus() & (~7)) | BPUSHED | WTRIGGER);
	draw();
}

void FGPushButton::ClickDown(void)
{
	Push();
}

void FGPushButton::Release(void)
{
	SetStatus((GetStatus() & (~(7|WTRIGGER))) | BNORMAL);
	draw();
}

void FGPushButton::ClickUp(int a)	// if TRUE, so call handler
{
	if ((status&7) == BDISABLED) return;
	Release();
	if (a)
	{
		// if error, set cControl to 0
		owner->SetActive(this);
	}
}

//! Draws a FGCheckButton
void FGCheckButton::draw(void)
{
//	set_font(FGFONT_BUTTON);
	if (strlen(name))
	{
		int ww = textwidth(name)+28;
		int hh = 18;//GetFontH();
		FGDrawBuffer::Resize(ww-w,hh-h);
	}
	clear(paper);
	box(1,1,14,14,CWHITE);
	rect(1,1,14,14,CDARK);
//	rect(0,0,15,15,CDARK);
	switch (status & 7)
	{
		case BNORMAL:
			text(20, 2, name,ink,paper);
			Underscore(20, 2, ink);
			break;
		case BDISABLED:
			text(20, 2, name,CScheme->edit_fore_disable,paper);
			Underscore(20, 2, CScheme->edit_fore_disable);
			break;
	}

	if (status & WTRIGGER)
	{
		if (variable)
			*variable |= mask;
		box(4,4,7,7,CBLACK);
	}
	else
	{
		if (variable) *variable &= ~mask;
	}
	update_owner();
	owner->WindowRepaint(x, y, w, h);
}

//! Draws a FGPointButton
void FGPointButton::draw(void)
{
	int  col;
//	set_font(FGFONT_BUTTON);

	if (strlen(name))
	{
		int ww = textwidth(name)+28;
		int hh = 18;//GetFontH();
		FGDrawBuffer::Resize(ww-w,hh-h);
	}

	clear(paper);

	if ((status&7)==BDISABLED)
		col = CScheme->edit_fore_disable;
	else
		col = GetInk();

	text(20, 2, name, col, paper);
	Underscore(20, 2, col);
	int xx=7, yy=4;

	fcircle(xx,yy+4,6,CWHITE);
	circle(xx,yy+4,6,CDARK);

	if (status & WTRIGGER)
	{
		if (variable) *variable |= mask;
		fcircle(xx,yy+4,3,CBLACK);
	}
	else
	{
		if (variable) *variable &= ~mask;
		fcircle(xx,yy+4,3,CWHITE);
	}
	update_owner();
	owner->WindowRepaint(x,y,w,h);
}

/**
	Print a text string to the status bar.
	@param x x position
	@param s null terminated string
	@param c foreground color
*/
void FGAPI FGWindow::WindowStatusBar(int x, const char *s, int c)
{
	int y;
	x += xoff+1;
	y  = h-STATUSBARH+4-(status&WFRAMED?BASEBORDER:0);
	SAVE_CLIPPING(this)
// hack!
	wwrk-=4;
	WindowText(x, y, s, c, statuscolor);
	RESTORE_CLIPPING(this)
}

/**
	Draws a filled triangle to the FGWindow.
	@param x1 x position of first vertice
	@param y1 y position of first vertice
	@param x2 x position of second vertice
	@param y2 y position of second vertice
	@param x3 x position of third vertice
	@param y3 y position of third vertice
	@param col foreground color
	@return the rectangle of triangle
	@see set_ppop()
*/
FGRect FGAPI FGWindow::WindowFillTriangle(int x1, int y1, int x2, int y2, int x3, int y3, unsigned col)
{
	FGRect r;
//	if (col != UNDEFINED_COLOR) set_fcolor(col);
	r = ftriangle(x1,y1,x2,y2,x3,y3, col);
	if (r.w) WindowRepaintUser(r.x, r.y, r.w, r.h);
	return r;
}

/**
	Draws a filled convex polygon in the FGWindow.
	@param vertices FGPointArray object contained vertices of polygon
	@param col foreground color
	@return the rectangle of polygon
	@see set_ppop() FGPointArray
*/
FGRect FGAPI FGWindow::WindowFillPolygon(const FGPointArray& vertices, unsigned col)
{
	StateLock();
	FGRect r;
//	if (col!=-1) set_fcolor(col);
	r = fpolygon(vertices, col);
	if (r.w) WindowRepaintUser(r.x, r.y, r.w, r.h);
	StateUnlock();
	return r;
}

/**
	Draws a convex polygon in the FGWindow.
	@param vertices FGPointArray object contained vertices of polygon
	@param col foreground color
	@return the rectangle of polygon
	@see set_ppop() FGPointArray
*/
void FGAPI FGWindow::WindowDrawPolygon(const FGPointArray& vertices, unsigned col)
{
	StateLock();
//	if (col!=-1) set_fcolor(col);
	polygon(vertices, col);
	WindowRepaintUser(0,0,wwrk,hwrk);
	StateUnlock();
}

/**
Draws a bezier spline using the four control points specified in the points array.
The bezier curve is specified by the four x/y
control points in the points array: points[0] and points[1] contain
the coordinates of the first control point, points[2] and points[3]
are the second point, etc. FGControl points 0 and 3 are the ends of the spline,
and points 1 and 2 are guides. The curve probably won't pass through points 1 and 2,
but they affect the shape of the curve between points 0 and 3 (the lines p0-p1 and p2-p3
are tangents to the spline). The easiest way to think of it is that the curve starts at p0,
heading in the direction of p1, but curves round so that it arrives at p3
from the direction of p2. In addition to their role as graphics primitives,
spline curves can be useful for constructing smooth paths around a series of control points.
*/
void FGAPI FGWindow::WindowSpline(int points[8], unsigned color)
{
	#define NPTS   64

	int xpts[NPTS], ypts[NPTS];
	int i;

	calc_spline(points, NPTS, xpts, ypts);

	for (i=1; i<NPTS; i++)
	{
		WindowLine(xpts[i-1], ypts[i-1], xpts[i], ypts[i], color);
		if (state._ppop == _GXOR)
			WindowPixel(xpts[i], ypts[i], color);
	}
}

/**
	Draws the text to the FGWindow. The output will
	takes font and rop from the current FGWindow values.
	@param x x position
	@param y y position
	@param s standard C string taht will be printed-out
	@param Ink foreground color
	@param Paper background color
	@see set_ppop()
	@see set_font()
*/
void FGAPI FGWindow::WindowText(int x, int y, const char *s, unsigned Ink, unsigned Paper)
{
	StateLock();
	text(x, y, s, Ink, Paper);
	WindowRepaintUser(x, y, textwidth(s), GetFontH());
	StateUnlock();
}

/**
	Draws the filled box into the FGWindow. The output will
	takes rop from the current FGWindow values.
	@param x x position
	@param y y position
	@param a width
	@param b height
	@param color foreground color
	@see set_ppop()
*/
void FGAPI FGWindow::WindowBox(int x, int y, int a, int b, unsigned color)
{
	StateLock();
	if (box(x, y, a, b, color))
	{
		optimize = 2; // as box
		optimize_color = color;
		WindowRepaintUser(x, y, a, b);
	}
	StateUnlock();
}

/**
	Draws the rectangle into the FGWindow. The output will
	takes rop from the current FGWindow values.
	@param x x position
	@param y y position
	@param a width
	@param b height
	@param color foreground color
	@see set_ppop()
*/
void FGAPI FGWindow::WindowRect(int x, int y, int a, int b, unsigned color)
{
	WindowLine(x, y, x+a-1, y,color);
	WindowLine(x+a-1, y+1, x+a-1, y+b-1,color);
	WindowLine(x+a-2, y+b-1, x, y+b-1,color);
	WindowLine(x, y+b-2, x, y+1,color);
}

/**
	Draws the patterned rectangle into the FGWindow. The output will
	takes rop from the current FGWindow values.
	@param x x position
	@param y y position
	@param a width
	@param b height
	@param pat the Pattern object
	@see set_ppop()
	@see Pattern
*/
void FGAPI FGWindow::WindowPatternRect(int x, int y, int a, int b, FGPattern *pat)
{
	int tmp = pat->reset;
	pat->reset=1;
	WindowPatternLine(x, y, x+a-1, y, pat);
	pat->reset=0;
	WindowPatternLine(x+a-1, y+1, x+a-1, y+b-1, pat);
	WindowPatternLine(x+a-2, y+b-1, x, y+b-1, pat);
	WindowPatternLine(x, y+b-2, x, y+1, pat);
	pat->reset = tmp;
}

/**
	Draws the line into the FGWindow. The output will
	takes rop from the current FGWindow values.
	@param x start x position
	@param y start y position
	@param a end x position
	@param b end y position
	@param color foreground color
	@see set_ppop()
*/
void FGAPI FGWindow::WindowLine(int x, int y, int a, int b, unsigned color)
{
	int x1, x2, y1, y2;
	StateLock();

	if (x < a)
	{
		x1 = x;
		x2 = a;
	}
	else
	{
		x1 = a;
		x2 = x;
	}
	if (y < b)
	{
		y1 = y;
		y2 = b;
	}
	else
	{
		y1 = b;
		y2 = y;
	}

	x += xoff;
	y += yoff;
	a += xoff;
	b += yoff;

	int cl = ClipLine(x,y,a,b,xoff,yoff,xoff+wwrk,yoff+hwrk);

	x -= xoff;
	y -= yoff;
	a -= xoff;
	b -= yoff;

	if (cl)
	{
		line_x1=x;
		line_y1=y;
		line_x2=a;
		line_y2=b;
		__line(x,y,a,b, color);
		optimize = 1; // as not clipped line
		optimize_color = color;
		WindowRepaintUser(x1, y1, x2 - x1+1, y2 - y1+1);
	}
	StateUnlock();
}

/**
	Draws the pixel into the FGWindow. The output will
	takes rop from the current FGWindow values.
	@param x start x position
	@param y start y position
	@param color foreground color
	@see set_ppop()
*/
void FGAPI FGWindow::WindowPixel(int x, int y, unsigned color)
{
	if (y <	0 || x < 0 || y >= hwrk || x >= wwrk)
	{
		return;
	}
	modify_point(image + w * (y+yoff) + x+xoff, color);
	optimize = 3; // as point
	optimize_color = color;
	WindowRepaintUser(x, y, 1, 1);
}

/**
	Draws the patterned line into the FGWindow. The output will
	takes rop from the current FGWindow values.
	@param x x position
	@param y y position
	@param a width
	@param b height
	@param p the Pattern object
	@see set_ppop()
	@see Pattern
*/
void FGAPI FGWindow::WindowPatternLine(int x, int y, int a, int b, FGPattern *p)
{
	StateLock();
	int x1, x2, y1, y2;
	x1 = (x < a) ? x : a;
	x2 = (x < a) ? a : x;
	y1 = (y < b) ? y : b;
	y2 = (y < b) ? b : y;
	pline(x, y, a, b, p);
	WindowRepaintUser(x1, y1, x2 - x1+1, y2 - y1+1);
	StateUnlock();
}

/**
	Draws the circle into the FGWindow. The output will
	takes rop from the current FGWindow values.
	@param x x position
	@param y y position
	@param r radius
	@param color foreground color
	@see set_ppop()
*/
void FGAPI FGWindow::WindowDrawCircle(int x, int y, int r, unsigned color)
{
	StateLock();
	circle(x, y, r, color);
	WindowRepaintUser(x - r, y - r, r + r + 1, r + r + 1);
	StateUnlock();
}

/**
	Draws the filled circle into the FGWindow. The output will
	takes rop from the current FGWindow values.
	@param x x position
	@param y y position
	@param r radius
	@param color foreground color
	@see set_ppop()
*/
void FGAPI FGWindow::WindowFillCircle(int x, int y, int r, unsigned color)
{
	StateLock();
	fcircle(x, y, r, color);
	WindowRepaintUser(x - r, y - r, r + r + 1, r + r + 1);
	StateUnlock();
}

/**
	Draws the ellipse into the FGWindow. The output will
	takes rop from the current FGWindow values.
	@param x x position
	@param y y position
	@param rx X radius
	@param ry Y radius
	@param color foreground color
	@see set_ppop()
*/
void FGAPI FGWindow::WindowDrawEllipse(int x, int y, int rx, int ry, unsigned color)
{
	StateLock();
	ellipse(x, y, rx, ry, color);
	WindowRepaintUser(x - rx, y - ry, rx + rx + 1, ry + ry + 1);
	StateUnlock();
}

/**
	Draws the filled ellipse into the FGWindow. The output will
	takes rop from the current FGWindow values.
	@param x x position
	@param y y position
	@param rx X radius
	@param ry Y radius
	@param color foreground color
	@see set_ppop()
*/
void FGAPI FGWindow::WindowFillEllipse(int x, int y, int rx, int ry, unsigned color)
{
	StateLock();
	fellipse(x, y, rx, ry, color);
	WindowRepaintUser(x - rx, y - ry, rx + rx + 1, ry + ry + 1);
	StateUnlock();
}

/**
Draws a circular arc with centre x, y and radius r, in an anticlockwise direction
starting from the angle a and ending when it reaches b. These values are specified
in radian format, with M_PI*2 equal to a full circle.
Zero is to the right of the centre point, and larger values rotate anticlockwise from there.
@param x x coordinate of the arc centre
@param y y coordinate of the arc centre
@param ang1 starting angle
@param ang2 ending angle
@param r radius in pixels
@param color color
*/
void FGAPI FGWindow::WindowDrawArc(int x, int y, double ang1, double ang2, int r, unsigned color)
{
	StateLock();
	arc(x, y, ang1, ang2, r, color);
	WindowRepaintUser(x - r, y - r, r + r + 1, r + r + 1);
	StateUnlock();
}

/**
	Scrolls-down the FGWindow about n-pixels.
	@param xx start x position
	@param yy start y position
	@param ww the width of scrolled block
	@param hh the height of scrolled block
	@param about the size to scroll
*/
void FGAPI FGWindow::WindowScrollDown(int xx, int yy, int ww, int hh ,int about)
{
	StateLock();
	bitblit(xx,yy+about,xx,yy,ww,hh,this,BITBLIT_WITH_DECREMENT);
	box(xx, yy, ww, about, paper);
	WindowRepaintUser(xx, yy, ww, hh + about);
	StateUnlock();
}

/**
	Scrolls-up the FGWindow about n-pixels.
	@param xx start x position
	@param yy start y position
	@param ww the width of scrolled block
	@param hh the height of scrolled block
	@param about the size to scroll
*/
void FGAPI FGWindow::WindowScrollUp(int xx, int yy, int ww, int hh ,int about)
{
	StateLock();
	bitblit(xx,yy-about,xx,yy,ww,hh,this);
	box(xx, yy + hh - about, ww, about, paper);
	WindowRepaintUser(xx, yy - about, ww, hh + about);
	StateUnlock();
}

/**
	Scrolls-left the FGWindow about n-pixels.
	@param xx start x position
	@param yy start y position
	@param ww the width of scrolled block
	@param hh the height of scrolled block
	@param about the size to scroll
*/
void FGAPI FGWindow::WindowScrollLeft(int xx, int yy, int ww, int hh ,int about)
{
	StateLock();
	bitblit(xx - about, yy, xx, yy, ww, hh, this);
	box(xx + ww - about, yy, about, hh, paper);
	WindowRepaintUser(xx - about, yy, ww + about, hh);
	StateUnlock();
}

/**
	Scrolls-right the FGWindow about n-pixels.
	@param xx start x position
	@param yy start y position
	@param ww the width of scrolled block
	@param hh the height of scrolled block
	@param about the size to scroll
*/
void FGAPI FGWindow::WindowScrollRight(int xx, int yy, int ww, int hh ,int about)
{
	StateLock();
	bitblit(xx + about, yy, xx, yy, ww, hh, this, BITBLIT_WITH_DECREMENT);
	box(xx, yy, about, hh, paper);
	WindowRepaintUser(xx, yy, ww + about, hh);
	StateUnlock();
}

void FGAPI FGWindow::WindowUnLock(void)
{
	SetStatus(GetStatus() & (~WLOCKED));
	if (GetStatus() & WDIRTY && IsVisible())
	{
		SetStatus(GetStatus() & (~WDIRTY));
		WindowRepaint(0,0,w,h);
	}
}

bool FGApp::IsShift(void) { return FGDriver::is_shift; }
bool FGApp::IsCtrl(void) { return FGDriver::is_ctrl; }
bool FGApp::IsAlt(void) { return FGDriver::is_alt; }


/**
	Construct the instance of applicaton object. There is allowed only
	one instantion per process only. The following steps are executed at the constructor time:
-	if possible, the window database  is accessed. This contains last sizes and positions for registered windows
-	if possible, the application configuration is loaded from appname.rc file
-	command line parameters are parsed, and arguments from the list of system parameters are extracted (consult  the "reserved parameters" chapter for details)
-	test for the availability of MMX processor and type of graphics card is performed
-	the screen is switched to graphics mode
-	if 8-bit colors are compiled/used, the color palette management is initialized
-	screen is cleared
-	mouse is checked and, if present, it is passed to the system
-	call to SetColorFuzzy() is performed
-	the ROOTWINDOW is created if needed
-	reset the Timer and FGClock() routines
-	release the root UID on LINUX or Solaris
-	global variables cApp and cCfg are set.
-	load the default mouse cursor
-	init TrueType font system (if enabled in compile time)

	AppFlags values - should be OR-ed together:

-	APP_WINDOWDATABASE - enable saving the size and position to the window database
-	APP_CFG	- enable configuration file management
-	APP_MAGNIFIER  - enable mouse drags with rubber-band rectangle selection
-	APP_ROOTWINDOW - enable drawing to the root window - the whole screen
-	APP_ENABLEALTX - enable termination by <ALT+X> key press
-	APP_LOCALDIR - all Config files will be placed in current directory instead in the HOMEDIR

	@param m number of the graphics mode (valid mode number values are described bellow).
	Look for graph_set_mode function.
	@param argv parameter is standard "C" command line parameter
	@param argc parameter is standard "C" command line parameter
	@param bck defines the initial color of the screen background.
	@param appFlags set of a bit flags that defines the behaviour of your
	application. The the meaning of bit flags is described in the table below.

*/
FGApp::FGApp(int m, int &argc, char **&argv, int bck, int appFlags)
{
	volatile int curr = 0, par = 0;
	char *apn=argv[0], *apn2, **cmdl=argv+1;

	mx = my = mb = oldx = oldy = oldb = drag = action = 0;

	locked = false;
	mouseflag = -1;
	mousefirst = 1;
	video = m;
	resize = vyrez = 0;
	background = bck;
	flags = appFlags;
	hold_event = new FGEvent(BUTTONHOLDEVENT);
	user_input_drv = 0;
	FGApp::test_mmx();

	if (cApp) IError("Multiple FGApp class not allowed !", 1);
	cApp = this;
	homedir = 0;
	if ((flags&APP_LOCALDIR) == 0)
		homedir = getenv("HOME");
#ifndef _MSC_VER
	getcwd(currdir, 127);
#else
	_getcwd(currdir, 127);
#endif
	if (homedir==NULL) homedir = ".";
	apn2 = apn;
	apn = apn+strlen(apn)-1;
	if (strcmp(apn-3, ".exe")==0 || strcmp(apn-3, ".EXE")==0)
	{
		apn = apn-4;
		apn[1]=0;
	}
	while ((*apn!='/' && *apn!='\\' && *apn!=':') && apn2<=apn) apn--;
	apn++;
	name = strdup(apn);
#ifdef __linux__
	setenv("EXENAME", name, 1);
#endif
	cCfg = new Config(apn, flags&APP_CFG);
	cCfg->ReadInt("default_videomode", video);

	if (argc > 1)
	{
		while (curr<(argc-1))
		{
			curr++;
			if (argv[curr][0] == '-') switch (tolower(argv[curr][1]))
			{
				case 'p':
					if (!memcmp(&argv[curr][1], "ps2", 3))
					{
						fgstate.force_ps2 = 1;
						if (fgstate.verbose) printf("PS2 fallback mouse initialization!\n");
						par ++;
					}
					else *cmdl++ = argv[curr];
					break;
				case 'r':
					if (!memcmp(&argv[curr][1], "res", 3))
					{
						curr++;
						if (argc)
							video = argv[curr][0] - '0';
						par += 2;
					}
					else *cmdl++ = argv[curr];
					break;
				case 'd':
					if (!memcmp(&argv[curr][1], "dac", 3))
					{
						fgstate.force_dac8 = 1;
						par ++;
					}
					else
						*cmdl++ = argv[curr];
					break;
				case 's':
#ifdef __linux__
					if (!memcmp(&argv[curr][1], "svg", 3))
					{
						fgstate.nofb = 1;	// no detect FRAMEBUFFER
						fgstate.nodga = 1;
						par++;
					}
					else if (!memcmp(&argv[curr][1], "syn", 3))
					{
						curr++;
						if (argc)
							fgstate.synch = atoi(&argv[curr][0]);
						par += 2;
					}
			   else
#endif
					   if (!memcmp(&argv[curr][1], "saf", 3))
					{
						video = G640x480;
	                  	fgstate.verbose = true;
						flags &= ~APP_WINDOWDATABASE;
						if (fgstate.verbose) printf("Safe mode entered ..\a\n");
						par++;
					}
					else *cmdl++ = argv[curr];
					break;
#ifndef __QNX__
				case 'g':
					if (!memcmp(&argv[curr][1], "geo", 3))
					{
						curr++;
						if (argc>1)
						{
							fgstate.__force_x = atoi(&argv[curr++][0]);
							fgstate.__force_y = atoi(&argv[curr][0]);
							video 	  = GCUSTOM;
							fgstate.windowed_mode = true;
							par+=3;
						} else par++;
					}
					else *cmdl++ = argv[curr];
					break;
#endif
#ifdef _WIN32
				case 'w':
					if (!memcmp(&argv[curr][1], "win", 3))
					{
						fgstate.windowed_mode = true;
						par++;
					}
					else *cmdl++ = argv[curr];
					break;
#endif
				case 'n':
					if (!memcmp(&argv[curr][1], "nof", 3))
					{
						fulldrag = 0;
						if (fgstate.verbose) printf("Full window drag enabled\n");
						par++;
					}
					else if (!memcmp(&argv[curr][1], "noa", 3))
					{
						fgstate.noaccel = 1;
						if (fgstate.verbose) printf("HW acceleration disabled\n");
						par++;
					}
					else
#ifdef __linux__
						if (!memcmp(&argv[curr][1], "nox", 3))
					{
						fgstate.nofb = 1;	// no detect FRAMEBUFFER
						fgstate.nox11 = 1;
						par++;
					}
					else if (!memcmp(&argv[curr][1], "nod", 3))
					{
						fgstate.noaccel = 1;
						if (fgstate.verbose) printf("DGA disabled\n");
						par++;
					}
					else
#endif
						*cmdl++ = argv[curr];
					break;
				case 'h':
					printf("\nOpenGUI %s - (c) Copyright 1996,2003 Marian Krivos (nezmar@atlas.sk)\n\n", FG_VERSION);
#ifdef FG_THREADED
					printf(" threading model: posix threads\n");
#else
					printf(" threading model: single\n");
#endif
					printf(" FGAPI = '%s'\n", APISTR);
					printf(" supported image file formats: BMP GIF PCX TGA");
#ifdef FG_JPEG
					printf(" JPG");
#endif
#ifdef FG_PNG
					printf(" PNG");
#endif
#ifdef FG_TIFF
					printf(" TIFF");
#endif
#ifdef FG_TTF
					printf(" TRUETYPE");
#endif
					printf("\n");
#ifdef __linux__
					printf(" backends: FRAMEBUFFER SVGALIB");
#ifdef X11_DRIVER
					printf(" X11");
#endif
#ifdef DGA_DRIVER
					printf(" DGA2");
#endif
#endif
				    printf(" with %d BPP\n", FASTGL_BPP);
#ifdef _WIN32
					printf(" backend: M$_DirectX");
#endif
				    printf(" SIMD instructions: %s\n", fgstate.__fgl_mmx?"MMX":"");
#ifdef __GNUC__
				    printf(" Compiled with: GNU C++\n");
#endif
#ifdef __ICC
					printf(" Compiled with: Intel C++\n");
#endif
#ifdef __BORLANDC__
					printf(" Compiled with: Borland C++\n");
#endif
					printf("\n   -resolution N\n");
					printf("          :force scr. resolution 'N', where 'N' = \n");
					printf("           1=320x200, 2=640x480, 3=800x600, 4=1024x768, 5=1280x1024\n");
					printf("           6=1600x1200\n");
					printf(" -svga    :don't use FBDEV driver\n");
					printf(" -nodga   :don't use DGA2 driver\n");
					printf(" -nox11   :don't use X11 driver\n");
					printf(" -sync    :monitor refresh\n");
					printf(" -geometry:set custom resolutions x y\n");
					printf(" -noaccel :don`t use HW accelerations\n");
					printf(" -nofull  :use faster window moving\n");
					printf(" -verbose :verbose output\n");
					printf(" -safe    :safe mode\n");
					printf(" -dac8    :force 8bit DAC/palette\n");
					printf(" -ps2     :force PS/2 mouse driver (LINUX/QNX)\n");
					printf(" -window  :run in windowed mode (WIN32)\n");
					getchar();
					par++;
					break;
				case 'v':
					if (!memcmp(&argv[curr][1], "ver", 3))
					{
						fgstate.verbose = 1;
						printf("\nOpenGUI %s - (c) Copyright 1996,2003 Marian Krivos\n", FG_VERSION);
#ifdef _REENTRANT
						printf("		with multi-threading\n");
#endif
						par++;
					}
					else *cmdl++ = argv[curr];
					break;
				default:
					*cmdl++ = argv[curr];
					break;
			}
			else *cmdl++ = argv[curr];
		}
	}
	argc -= par;
	Argv = argv;
	Argc = argc;

	if (flags&APP_WINDOWDATABASE) entryPoint = new WindowDatabase(apn);
	if (video < 1 || video > GCUSTOM)
		video = 1;

	if (!graph_set_mode(video))
	{
		printf("Fatal error: can't set graphics mode %d\n\a", video);
		exit(-2);
	}
	XX = X_width / 2;
	YY = Y_width / 2;
	SetColorFuzzy(3);
	if (background!=255)
		clear_frame_buffer(background);
	if (m>=4)
		idc_normal = &IDC_NORMAL_LARGE;
	else
		idc_normal = &IDC_NORMAL_SMALL;
	CursorLoad(idc_normal);
	SetDelayProc(0);
#ifndef _WIN32
	SetTimerProc(0, 1000);
#endif
	if (flags&APP_ROOTWINDOW) Root = new FGWindow(&Root, background);
#if ( !defined( __MSDOS__ ) && !defined( __rtems__ ) ) && !defined(_WIN32)
	setuid(getuid());
	setgid(getgid());
	seteuid(getuid());
	setegid(getgid());
#endif
#if defined( __linux__ ) || defined( __rtems__ )
	FGTimer::TimerInit();
#endif
#ifdef FG_TTF
	if ( TTF_Init() < 0 )
		fprintf(stderr, "Couldn't initialize TTF\n");
	else
	{
		ttf_support = 1;
		atexit(TTF_Quit);
	}
#endif
}

static void ClrScr(int ako)
{
	switch(ako)
	{
		default:
		case 0:
			{
				for(int i=0; i<(Y_width-1); i++)
				{
					draw_line(i,i,(X_width - 1), i,0,_GSET);
					draw_line(i,i,i, (Y_width-1),0,_GSET);
					if (i%32 ==0) delay(1);
				}
			}
			break;
	}
}

/**
This is called when the application terminates. It performs the following steps:
-	resets the app timer
-	resets color palette
-	closes windows database
-	closes config file
-	resets the windowing system
-	call system's 'sync()' to flus all disk operations
*/
FGApp::~FGApp()
{
#if defined( __linux__ ) || defined( __rtems__ )
	FGTimer::TimerQuit();
#endif
	if (flags&APP_ENABLEEFECTS)
		ClrScr(0);
	else
		clear_frame_buffer(CBLACK);
	SetDelayProc(0);
#ifndef _WIN32
	SetTimerProc(0);
#endif
#ifdef INDEX_COLORS
	_set_default_palette();
#endif
	delete cCfg;
	cCfg = 0;
	if (Root) delete Root;
	Root = 0;
	FGWindow::DestroyWindow();
	if (flags&APP_WINDOWDATABASE) delete entryPoint;
	entryPoint = 0;
	free(name);
	cApp = 0;
#if !defined(__MSDOS__) && !defined(__rtems__) && !defined(_WIN32)
	sync();
#endif
	delete hold_event;
	hold_event = 0;
}

//! Sends the message to the all windows
//! @param e the event to send
void FGAPI FGWindow::BroadcastMessage(FGEvent *e)
{
	FGWindow *w = first;
	while(w)
	{
		w->SendToWindow(e);
		w = (FGWindow *) (w->GetNext());
	}
}

/**
	Peeks the user input, translate to EVENT if any and push this EVENT to the
	event list. You can use this procedure to allowing mouse move when system
	is busy in your code and don't respond to user. By example, when some
	progressbar is running.
*/
void FGAPI FGApp::FGYield(void)
{
	FGEvent ev(0);
	GetUserEvent(ev, 1);	// do while ALT+X no pressed
}

/**
	When you create the instance of the FGApp class and you sets callback for "DelayProc" ,
	it is a right time to call the main application loop. System starts processing
	the user input (keyboard & mouse) until you call AppDone().
	The mouse's movements are automatically convert to move its pointer
	at the screen and you don't care about it. When you press
	left button at the FGControl (by example FGPushButton), system detects it, and sends right event
	to the appropriate window. When you press button on an empty area of
	the window, this one will get event CLICKLEFTEVENT. This events are send to the application procedure,
	when one is defined.
	@param hwnd the pointer at this mysterious procedure.
*/
void FGAPI FGApp::Run(MainHwnd hwnd)
{
	FGEvent *e=queque, event, tmpevent;

	if (hwnd)
		appHandler = hwnd;

	for (;fgstate.shut_down == 0;)
	{
		GetUserEvent(event, -1);	// do while ALT+X no pressed
		TranslateUserEvent(event);
		for (; quequeIndex;)
		{
			quequeIndex--;
			tmpevent = queque[quequeIndex];
			e = &tmpevent;

			// call event handler
			if (appHandler)
				appHandler(e);

			// call event methods
			switch(e->GetType())
			{
				case KEYEVENT:
					OnKeyPress(e->key);
					break;
				case MOVEEVENT:
					OnMouseMove(e->x, e->y);
					break;
				case CLICKLEFTEVENT:
					OnClick(e->x, e->y);
					break;
				case CLICKRIGHTEVENT:
					OnContextPopup(e->x, e->y);
					break;
				case CURSOROUTEVENT:
					OnCursorOut(e->key);
					break;
				case STARTDRAGLEFTEVENT:
					OnStartDrag(1, e->x, e->y);
					break;
				case STARTDRAGRIGHTEVENT:
					OnStartDrag(0, e->x, e->y);
					break;
				case DRAGLEFTEVENT:
					OnEndDrag(1, e->x, e->y, e->w, e->h);
					break;
				case DRAGRIGHTEVENT:
					OnEndDrag(0, e->x, e->y, e->w, e->h);
					break;
				case TABSWITCHEVENT:
					OnTabSwitch((const char *)e->key);
					break;
				case QUITEVENT:
					fgstate.shut_down = 1;
				case CLOSEEVENT:
					return; // force ret if QUITEVENT
			}
			e->type = NOEVENT;
		}
	}
}

/**
	Run modal this FGWindow until closed or any FGControl.
	@param which your FGWindow with the 'OK' and 'Cancel'
	buttons by example.
	@return one of predefined values (mrClose, mrQuit or mrNone)
*/
MODAL_RETURN FGAPI FGApp::RunModal(FGWindow *which)
{
	FGEvent *e=queque, event;

	for ( ; fgstate.shut_down == 0 ; )
	{
		GetUserEvent(event, -1);	// do while ALT+X no pressed

		TranslateUserEvent(event);

		if (quequeIndex)
		{
			quequeIndex = 0;
			e = &event;

			// call event handler
			if (appHandler)
				appHandler(e);

			// call event methods
			switch(e->GetType())
			{
				case TERMINATEEVENT:
					if (e->wnd == which)
						return mrClose;
					break;
				case ACCELEVENT:
					if (which == e->wnd)
						if (e->accel->GetParameter() != (void *) mrNone)
                        {
							int rc = (int)e->accel->GetParameter();
						 	return (MODAL_RETURN) (rc);
                        }
					break;
				case QUITEVENT:
					fgstate.shut_down = 1;
				case CLOSEEVENT:
					return mrQuit; // force ret if QUITEVENT
			}
			e->type = NOEVENT;
		}
	}
	return mrNone;
}

void FGAPI FGWindow::nextwnd(void)
{
	FGBaseGui *p = 0;
	static int cnt = 0;
	int i;

	if (nWnd <= 1 || current->GetStatus() & WMODAL)
	{
		Snd(1000,80);
		return;
	}

	for (i=nWnd; i; i--)
	{
		if (GetCounter() > cnt) cnt++;
		else cnt = 0;
		p = first->FindId(cnt);
		if (!p)
			continue;
		if (current == p || p->GetType()==ROOTWINDOW || (p->GetStatus()&WHIDEN))
			continue;
		break;
	}
	if (!i) for (i=nWnd; i; i--)
	{
		if (GetCounter() > cnt) cnt++;
		else cnt = 0;
		p = first->FindId(cnt);
		if (!p)
			continue;
		if (current == p || p->GetType()==ROOTWINDOW || (p->GetStatus()&WHIDEN))
			continue;
		break;
	}
	if (!i) for (i=nWnd; i; i--)
	{
		if (GetCounter() > cnt) cnt++;
		else cnt = 0;
		p = first->FindId(cnt);
		if (!p)
			continue;
		if (current == p || p->GetType()==ROOTWINDOW || (p->GetStatus()&WHIDEN))
			continue;
		break;
	}
	if (!i) return; // inner loop

	((FGWindow *) p)->WindowFocus();
}

//! draw only input text with/without cursor
void FGEditBox::CapsMode(int m)
{
	caps = m;
	char *s=buf;
	if (caps==EBOX_LOW)
	{
		while(*s)
		{
			if(isupper(*s)) *s+=32;
			s++;
		}
	}
	else if (caps==EBOX_UPR)
	{
		while(*s)
		{
			if(islower(*s)) *s-=32;
			s++;
		}
	}
	draw();
}

//! redraw input string
void FGEditBox::draw(bool cursor, bool first_time)
{
	char str[130];
	set_font(FONT0816);

	box(2 + w1, 2, w2 - 4, GetH() - 4, CScheme->edit_background_active);
	strcpy(str, buf);
	if (passwd)
	{
		char *x=str;
		while (*x)
		{
			if (*x != ' ') *x = '*';
			x++;
		}
	}
	str[maxpos+offset] = 0;
	if (cursor)
	{
		text(w1 + pos * 8 + 2 + 8, 1 + (GetH() / 2 - GetFontH() / 2), str + offset + pos, CScheme->edit_foreground, CScheme->edit_background_active);
		box(w1 + pos * 8 + 8, 3, 2, GetH() - 6, CScheme->edit_foreground);
		str[pos+offset] = 0;
		text(w1 + 8, 1 + (GetH() / 2 - GetFontH() / 2), str+offset, CScheme->edit_foreground, first_time?CScheme->edit_background : CScheme->edit_background_active);
	}
	else
	{
		text(w1 + 8, 1 + (GetH() / 2 - GetFontH() / 2), str+offset, CScheme->edit_foreground, CScheme->edit_background_active);
	}
	update_owner();
	owner->WindowRepaint(x, y, w, h);
}

//! draw all
void FGEditBox::draw(void)
{
	char str[132], *spt=0;
	set_font(FGFONT_BUTTON);
	int cp;
	unsigned ndl=w1-textwidth(name)-8;

	offset = pos = 0;
	TestRange();

	cp = ((GetStatus() == BDISABLED) ? CScheme->edit_fore_disable : (FGPixel)GetInk());
	box(0, 0, w1, h, paper);
	text(ndl, 1 + (h / 2 - GetFontH() / 2), name, cp, paper);
	Underscore(ndl, 2 + h / 2 - GetFontH() / 2, paper);

	switch (status & 7)
	{
		case BNORMAL:
			Underscore(ndl, 2 + h / 2 - GetFontH() / 2, ink);
		case BDISABLED:
			set_font(FONT0816);
			box(w1, 0, w2, h, CScheme->edit_foreground);
			vline(w1 + w2 - 2, 0 + 1, h - 2, CScheme->edit_background_active);
			hline(1 + w1, 0 - 2 + h, w2 - 2, CScheme->edit_background_active);
			box(2 + w1, 0 + 2, w2 - 4, h - 4, CScheme->edit_background);
			rect(w1, 0, w2, h,CScheme->edit_border);

			switch (data_type)
			{
				case EDIT_INT:
					if (!hex) sprintf(str, "%d", *(int *) ptr);
					else sprintf(str, "0x%x", *(int *) ptr);
					spt = str;
					break;
				case EDIT_STRING:
					 strcpy(str,(char *)ptr);
					spt = str;
					break;
				case EDIT_DOUBLE:
					sprintf(str, "%f", *(double *) ptr);
					if (*str=='0') strcpy(str, str+1);
					spt = str;
					break;
			}
			spt[maxpos] = 0;	// print visible part only
			if (passwd)
			{
				char *x=spt;
				while (*x)
				{
					if (*x != ' ') *x = '*';
					x++;
				}
			}
			text(w1 + 8, 1 + (h / 2 - GetFontH() / 2), spt, (status == BDISABLED) ? cp : CScheme->edit_foreground, CScheme->edit_background);
			break;

		case BUNVISIBLE:
			clear(paper);
			break;
	}
	update_owner();
	owner->WindowRepaint(x,y,w,h);
	set_font(FONT0816);
}

//! Returns 1 at the emd
int FGEditBox::inputproc(unsigned c)
{
	char *s, *d;
	int i;

	// input filter
	if (c>= ' ' && c<127)
		switch(int(data_type))
	{
		case EDIT_INT:
			if (!isdigit(c&255) && c != '-')
				if (hex)
				{
					if (c<'A' || c>'f') return 0;
					else if (c>'F' && c<'a') return 0;
				}
				else return 0;
			break;
		case EDIT_DOUBLE:
			if (!isdigit(c&255) && c != '-')
				if (c!= '.' && c!='e' && c!='E') return 0;
			break;
	}

	if (caps==EBOX_UPR && islower(c&255)) c-=32;
	else if (caps==EBOX_LOW && isupper(c&255)) c+=32;

	draw(0); // cursor off
	if (first)
	{
		first = 0;
		if ((isprint(c&255) || c == DEL || c == BACKSP) && c < 256)
		{
			for (i = 0; i < size; buf[i++] = ' ');
			offset = pos = 0;
		}
	}

	switch (c)
	{
		case CTRL_INSERT:
			memcpy(clip,buf,sizeof(clip));
			is_clip = 1;
			break;
		case INSERT:
			memcpy(buf,clip,sizeof(clip));
			break;
		case ESC:
			isinput = 0;
			owner->SetInInput(0);
			SetStatus(BNORMAL);
			draw(); // draw final
			return -1;
		case CR:
			isinput = 0;
			owner->SetInInput(0);
			switch (data_type)
			{
				case EDIT_INT:
					int n;
					if (hex) n = strtol(buf, 0, 16); else n = atoi(buf);
					{
						*(int *) ptr = n;
						TestRange();
					}
					break;
				case EDIT_STRING:
					pos = size;
					while (pos)
					{
						if (buf[pos - 1] != ' ')
							break;
						pos--;
					}
					if (buf[pos-1]=='\"' && pos>0) pos--;
					buf[pos]=0;
					strcpy((char *) ptr, buf);
					break;
				case EDIT_DOUBLE:
					double d;
					d = atof(buf);
					{
						*(double *) ptr = d;
						TestRange();
					}
					break;
			}
			SetStatus(BNORMAL);
			draw(); // draw final
			owner->SetActive(this);
			return 1;
		case HOME:
			offset = pos = 0;
			break;
		case END:
			{
				pos = size;
				while (pos)
				{
					if (buf[pos - 1] != ' ')
						break;
					pos--;
				}
				offset = pos>maxpos?pos-maxpos:0;
				pos = pos-offset;
			}
			break;
		case BACKSP:
			if (pos+offset)
			{
				s = &buf[offset+pos--];
				d = s - 1;
				while (*s)
					*d++ = *s++;
				*d = ' ';
				if (offset) { offset--; pos++; }
			}
			break;
		case KLEFT:
			if (pos+offset)
			{
				if (pos) pos--;
				else if (offset) offset--;
			}
			break;
		case KRIGHT:
			if (pos+offset < size)
			{
				if (pos<maxpos) pos++;
				else { offset++; }
			}
			break;
		case DEL:
			if (pos+offset < size)
			{
				d = &buf[pos+offset];
				s = d + 1;
				while (*s)
					*d++ = *s++;
				buf[size - 1] = ' ';
			}
			break;

		default:
			if (pos+offset < size)
			{
				if ((c > 31) | (c < 128))
				{
					d = buf + sizeof(buf) -1;
					while(d>(buf+pos+offset))
					{
						*d = d[-1];
						d--;
					}
					*d = (char)c;
				}
				if (pos+offset < size)
				{
					if (pos<maxpos) pos++;
					else { offset++; }
				}
			}
			break;
	}
	buf[size] = 0;
	draw(1); // cursor on
	return 0;
}

//! Starts the input line
void FGEditBox::input(void)
{
	char str[130];
	set_font(FONT0816);

	isinput = 1;
	owner->SetInInput(this);

	iline = GetY() + 1 + (GetH() / 2 - GetFontH() / 2);
	icol = GetX() + w1 + 8;
	switch (data_type)
	{
		case EDIT_INT:
			sprintf(str, hex?"%x":"%d", *(int *) ptr);
			break;
		case EDIT_STRING:
			strcpy(str, (const char *) ptr);
			break;
		case EDIT_DOUBLE:
			sprintf(str, "%f", *(double *) ptr);
			if (*str=='0') strcpy(str, str+1);
			break;
	}
	// copy to buffer (and fill it) and find the end of line
	for (int j = 0, i = 0; i < (int)sizeof(buf);)
		if (str[j] == 0)
			buf[i++] = ' ';
		else
			buf[i++] = str[j++];
	buf[size] = 0;
	pos = size-1;
	while (pos >= 0)
	{
		if (buf[pos] > ' ')
			break;
		pos--;
	}
	pos++;
	offset = pos>maxpos?pos-maxpos:0;
	pos -= offset;
	first = 1;
}

FGEditBox::FGEditBox(int xs,	int ys,	int ws1, int ws2, const char *nm, int	key, FGWindow *w,	int *pt, int ink, int paper, ControlCall f, int mn, int mx, int check)
	: FGControl(xs, ys, ws1+ws2, 21, nm, EDITBOX,	key, BNORMAL, w, f,paper,ink)
{
	init();
	w1 = ws1;
	w2 = ws2;
	ptr= pt;
	check_range = check;
	min= mn;
	max=mx;
	data_type=EDIT_INT;
	size = maxpos=(ws2-16)/GetFontW();
	draw();
}

FGEditBox::FGEditBox(int sz, int xs, int ys, int ws1, int ws2, const char *nm, int key, FGWindow *w, char *pt, int ink, int paper, ControlCall f)
: FGControl(xs, ys, ws1+ws2, 21, nm, EDITBOX,	key, BNORMAL, w, f,paper,ink)
{
	init();
	w1 = ws1;
	w2 = ws2;
	ptr= pt;
	data_type=EDIT_STRING;
	maxpos=(ws2-16)/GetFontW();
	size = sz<=0 ? maxpos : ((sz>127) ? 127:sz);
	draw();
}

FGEditBox::FGEditBox(int xs,	int ys,	int ws1, int ws2, const char *nm, int	key, FGWindow *w,	double *pt, int	ink, int paper,	ControlCall f, double mn, double mx, int check)
: FGControl(xs, ys, ws1+ws2, 21, nm, EDITBOX,	key, BNORMAL, w, f,paper,ink)
{
	init();
	w1 = ws1;
	w2 = ws2;
	ptr= pt;
	check_range = check;
	mind= mn;
	maxd=mx;
	data_type=EDIT_DOUBLE;
	size = maxpos=(ws2-16)/GetFontW();
	draw();
}

void FGEditBox::init(void)
{
	check_range = offset = 0;
	hex = passwd = caps = isinput = 0;
}

FGEditBox::~FGEditBox()
{
	if (isinput)
		owner->SetInInput(0);
}

void FGEditBox::ClickUp(int a)	// if TRUE, so call handler
{
	if ((status&7) == BDISABLED) return;
	if (!a)	return;
	owner->WindowFlushInput();
	SetStatus(BPUSHED);
	input();
	draw(true, true);
}

static int is_alted(int c)
{
	for(int i=0;i<26;i++)
		if (alty[i]==c) return i;
	return -1;
}

void FGControl::Underscore(int xo, int yo, int c)
{
	char *s;
	int pos;
	int size=0;
	if (key==0) return;

	if ((pos=is_alted(key)) >= 0)
	{
		if ((s = strchr(name, pos+'a')) != 0)
		{
			size = (int)s - (int) name;
			box(textwidth(name, size) +xo-1, 13+yo, charwidth(pos+'a')+1,1,CREDLIGHT);
		}
		else if ((s = strchr(name, pos+'A')) != 0)
		{
			size = (int)s - (int) name;
			box(textwidth(name, size) +xo-1, 13+yo, charwidth(pos+'A')+1,1,CREDLIGHT);
		}
	}
	else if ((s = strchr(name, key)) != 0)
	{
		size = (int)s - (int) name;
		box(textwidth(name, size) +xo-1, 13+yo, charwidth(key)+1,1,c);
	}
	else if ((s = strchr(name, key ^ 32)) != 0)
	{
		size = (int)s - (int) name;
		box(textwidth(name, size) +xo-1, 13+yo, charwidth(key)+1,1,c);
	}
}

void FGBaseMenu::MoveOut(void)
{
	draw();
}

void FGBaseMenu::MoveIn(void)
{
//	set_font(FGFONT_MENU);

	box(0, 0, w, MENUH-2, CScheme->menu_back_active);
	text(4, 0, name,CScheme->menu_fore_active,CScheme->menu_back_active);
	Underscore(4,0,CWHITE);
	update_owner();
	owner->WindowRepaint(x, y, w, MENUH);
}

void FGBaseMenu::draw(void)
{
	switch (status & 7)
	{
		case BNORMAL:
			box(0, -1, GetW(), MENUH-2, paper);
			text(4, 0, GetName(),GetInk(),GetPaper());
			Underscore(4,0,GetInk());
			update_owner();
			owner->WindowRepaint(GetX()-4, GetY()-2, GetW()+8, MENUH);
			break;
		case BDISABLED:
			text(4, 0, GetName(), CScheme->pdmenu_gray, GetPaper());
			Underscore(4,0,GetPaper());
			Underscore(4,1,GetPaper());
			update_owner();
			owner->WindowRepaint(GetX()-4, GetY(), GetW()+8, GetH());
			break;
	}
}

void FGBaseMenuItem::MoveOut(void)
{
	draw();
}

void FGBaseMenuItem::MoveIn(void)
{
//	set_font(FGFONT_MENU);

	box(0, 0, w, h, CScheme->pdmenu_back_active);
	text(4, 1, name, CScheme->pdmenu_fore_active, CScheme->pdmenu_back_active);
	Underscore(4,1,CScheme->pdmenu_fore_active);
	update_owner();
	owner->WindowRepaint(x, y, w, h);
}

void FGBaseMenuItem::draw(void)
{
	switch (status & 7)
	{
		case BNORMAL:
			box(0, 0, w, h, paper);
			text(4, 1, GetName(),GetInk(),GetPaper());
			Underscore(4,1,GetInk());
			update_owner();
			owner->WindowRepaint(x, y, w, h);
			break;
		case BDISABLED:
			text(4, 0, GetName(), CScheme->pdmenu_gray, GetPaper());
			Underscore(4,0,GetPaper());
			Underscore(4,1,GetPaper());
			update_owner();
			owner->WindowRepaint(GetX()-4, GetY(), GetW()+8, GetH());
			break;
	}
}

int FGAPI FGWindow::GetXM(int a)
{
	if (!(GetStatus() & WMENU))
	{
		::printf("You must use WMENU option for this window\n");
	}
	c_menupos += a;
	return c_menupos - a;
}

int FGAPI FGWindow::GetYM()
{
	return -(status&WNOTEBOOK ? MENUH+NOTEBOOKH : MENUH);
}

/**
	Resizes the FGWindow and/or changes the FGWindow 's position.
	@param sx a new x position
	@param sy a new y position
	@param wx a new width
	@param wy a new height
*/
void FGAPI FGWindow::WindowShape(int sx, int sy, int wx, int wy)
{
	bool changepos = (sx!=x || sy!=y), changesize=(wx!=w || wy!=h);

	// resizing of centred window is moving too.
	if (status&WCENTRED)
	{
		sx += wx/2;
		sy += wy/2;
		changepos = true;
	}

	if (changepos && !changesize) // iba zmena polohy
	{
		WindowMove(sx-x, sy-y);
		return;
	}
	else if (!changepos && changesize) // iba zmena velkosti
	{
		WindowResize(wx-w, wy-h);
		return;
	}

	StateLock();
	c_menupos = 1;
	if (status&WALIGNED)
	{
		sx &= ~3;
		sy &= ~3;
		wx &= ~3;
		wy &= ~3;
	}

	if (sx>X_width-wx) sx = X_width-wx;
	if (sy>Y_width-wy) sy = Y_width-wy;
	if (sx<0) sx = 0;
	if (sy<0) sy = 0;
	if (wx > X_width) wx = X_width;
	if (wx < 32)	  wx = 32;
	if (wy > Y_width) wy = Y_width;
	if (wy < (BASEBORDER*2+TITLEH+MENUH+STATUSBARH+NOTEBOOKH))
		wy = BASEBORDER*2+TITLEH+MENUH+STATUSBARH+NOTEBOOKH;

	int nx=sx-x, ny=sy-y, dx=wx-w, dy=wy-h;

	// zrus obrysy povodneho okna
	WindowUpdate(nx, ny, 1); // redraw on screen

	if (status&WNOTEBOOK)
		RelocateTabPage(nx, ny);
	else
		Relocate(nx, ny);
	Resize(dx, dy);
	xpos = 2;
	ypos = 0;
	// nakresli nove okno
	draw();

	StateUnlock();

	if (CallWindowHook && changepos)  CallWindowHook(x, y, nx, ny, 1); // 1 as move
	if (CallWindowHook && changesize) CallWindowHook(w, h, dx, dy, 2); // 2 as size

	SendEvent(WINDOWRESIZEEVENT, x,y,w,h);
	SendEvent(WINDOWMOVEEVENT, x,y,w,h);
}

/**
	This function will remap the root window to virtually any size and position relative to the top left corner.
	The call:
	FGWindow::WindowVirtualRoot(-1000,-1000, 2000, 2000);
	Creates an virtual root window with position [1000,1000] in a top left corner of screen.
	@param sx x start position, 0 is top left position, you can use both, positive or negative values.
	@param sy y start position, 0 is top left position, you can use both, positive or negative values.
	@param wx the width, the values is from 1 up to some thousands, it is limited by size of VIDEORAM
	@param wy the height, the values is from 1 up to some thousands, it is limited by size of VIDEORAM
*/
void FGAPI FGWindow::WindowVirtualRoot(int sx, int sy, int wx, int wy)
{
	FGWindow *Root = FGApp::GetRootWindow();
	if (!Root) return;
	int nx=sx-Root->x, ny=sy-Root->y, dx=wx-Root->w, dy=wy-Root->h;

	Root->Relocate(nx, ny);
	Root->Resize(dx, dy);
	Root->clear(0);
	ShowAll();
}

/**
	Moves FGWindow at the new position.
	@param nx the delta of (new_x-old_x) positions
	@param ny the delta of (new_y-old_y) positions
*/
void FGAPI FGWindow::WindowMove(int nx, int ny)
{
	if (status&WALIGNED && !cApp->fulldrag)
	{
		nx &= ~3;
		ny &= ~3;
	}
	if (CallWindowHook) CallWindowHook(x, y, nx, ny, 1); // 1 as move
	XWaitRetrace();
	StateLock();
	WindowUpdate(nx,ny);
	if (status&WNOTEBOOK)
		RelocateTabPage(nx, ny);
	else
		Relocate(nx, ny);
	WindowRepaint(0,0,w,h);
	StateUnlock();
	SendEvent(WINDOWMOVEEVENT, x,y,w,h);
}

/**
	Resizes the FGWindow.
	@param dx the delta of (new_w-old_w) widths
	@param dy the delta of (new_h-old_h) heights
*/
void FGAPI FGWindow::WindowResize(int dx, int dy)
{
	int span = 0;
	// calculate minimal Window vertical size
	if (status & WFRAMED)
		span += BASEBORDER;
	if (status & WTITLED)
		span += (TITLEH+2);
	if (status & WMENU)
		span += (MENUH+1);
	if (status & WNOTEBOOK)
		span += (NOTEBOOKH+1);

	if (status&WALIGNED && !cApp->fulldrag)
	{
		dx &= ~3;
		dy &= ~3;
	}
	if (w + dx > X_width
		|| w + dx < 32
		|| h + dy > Y_width
		|| h + dy < span)
	{
		return;
	}
	if (CallWindowHook) CallWindowHook(w, h, dx, dy, 2); // 2 as resize

	XWaitRetrace();
	StateLock();
	c_menupos = 1;
	if (dx<0 || dy<0)
		WindowUpdate(dx,dy);
	Resize(dx, dy);
	xpos = 2;
	ypos = 0;
	draw();
	StateUnlock();
	SendEvent(WINDOWRESIZEEVENT, x,y,w,h);
}

/**
	Draws a rectangle in-memory bitmap to the FGWindow. Function clips all coordinates and sizes correctly.
	If the target area is bigger than source area, then image will be tilted.
	@note the ROP and the ALPHA of source bitmap is used.
	@param x target x position
	@param y target y position
	@param xs source x position
	@param ys source y position
	@param w the width of source area
	@param h the height of source area
	@param p the pointer to the source bitmap
	@see FGDrawBuffer::bitblit() set_ppop() SetAlpha()
*/
void FGAPI FGWindow::WindowPutBitmap(int x, int y, int xs, int ys, int w, int h, FGDrawBuffer * p)
{
	StateLock();
	bitblit(x,y,xs,ys,w,h,p);
	WindowRepaintUser(x, y, w, h);
	StateUnlock();
}

/**
	Gets a rectangle area from the FGWindow int an in-memory bitmap.
	@param x source x position
	@param y source y position
	@param xs target x position
	@param ys target y position
	@param w the width of source area
	@param h the height of source area
	@param p the pointer to the source bitmap
*/
void FGAPI FGWindow::WindowGetBitmap(int x, int y, int xs, int ys, int w, int h, FGDrawBuffer * p)
{
	if (p->GetType() == BMP_NONE)
		return;
	StateLock();
	x += xoff;
	y += yoff;
	RamToRam(x, y, GetW(), GetH(), xs, ys, p->GetW(), p->GetH(), w, h, image, p->GetArray(), BITBLIT_WITH_CLIPPING, _GSET);
	StateUnlock();
}

//! internal for print error message
static void IErrorText(char *s, int flag)
{
	Snd(1000, 100);
	if (flag)
	{
		printf("FATAL ERROR: %s\n", s);
		getchar();
		exit(flag);
	}
	else
	{
		printf("WARNING: %s\n", s);
		getchar();
	}
}

void IError(char *s, int flag)
{
	int w = (strlen(s) + 8) * 4;
	static FGWindow *errWindow=0;

	if (fgstate.verbose) ::printf("IError:: %s\n", s);
	if (errWindow) return;

	if (X_width)
	{
		w = w < 75 ? 75 : w;
		Snd(1000,80);
		errWindow = new FGWindow(&errWindow, 0, 0, w * 2, 78, "Error", GuiHwnd(0), CWHITE, CREDLIGHT, ((flag&1) ? WMODAL : 0) | WCENTRED | WCENTRED | WESCAPE);

		if (flag&1) errWindow->WindowText(w - 70, 8, "* * * ERROR * * *", CYELLOW);
		else errWindow->WindowText(w - 78, 8, "* * * WARNING * * *", CYELLOW);

		errWindow->WindowText(32, 28, s, CWHITE);
		errWindow->WindowRect(0,0,2*w,78,CBLACK);

		if (flag&1) errWindow->AddPushButton(w - 40, 48, 80, 21, "Abort", CR, FGControl::Quit);
		else errWindow->AddPushButton(w - 40, 48, 80, 21, "Ok", CR, FGControl::Close);
	}
	else IErrorText(s, flag);
//#ifndef FG_THREADED
	if (flag&2) errWindow->ShowModal();
//#endif
}

WindowDatabase::WindowDatabase(char *exename)
{
	int magic;
	FILE *f;

#ifndef __MSDOS__
   sprintf(name, "%s/.windowxy.wdb", FGApp::homedir);
#else
   strcpy(name, "windowxy.wdb");
#endif
	xyCounter = 0;
	sprintf(exe, "%s::", exename);
	f = fopen(name, "rb");
	if (f)
	{
		fread(&magic, sizeof(int), 1,f);
		fread(&xyCounter, sizeof(int), 1,f);
		if (xyCounter<1 || xyCounter>100 || magic!=0x4357)
		{
			remove(name);
			xyCounter = 0;
			endpointer = &first;
			return; // possibly damaged config
		}
		WindowItem **previous = &first;

		for (int i = 0; i < xyCounter; i++)
		{
			*previous = new WindowItem(f);
			previous = &((*previous)->next);
		}
		*previous = 0;
		endpointer = previous;
		fclose(f);
	}
	else
	{
		endpointer = &first;
	}
}

WindowDatabase::~WindowDatabase()
{
	int magic=0x4357;
	FILE *f;

	if (xyCounter <= 0)	return;
	f = fopen(name, "wb");
	if (f)
	{
		WindowItem *p = first, *pp;
		fwrite(&magic, sizeof(int), 1,f);
		fwrite(&xyCounter, sizeof(int), 1,f);

		for (int i = 0; i < xyCounter; i++)
		{
			pp = p->next;
			p->SaveToFile(f);
			delete p;
			p = pp;
		}
		fclose(f);
	}
}

WindowItem* FGAPI WindowDatabase::findxy(char *s)
{
	WindowItem *p = first;
	char ss[128];

	sprintf(ss, "%s%s", exe, s);
	for (int i = xyCounter; i; --i)
	{
		if (!strcmp(p->name, ss))
			return p;
		p = p->next;
	}
	return 0;
}

WindowItem::WindowItem(FILE *f)
{
	char s[80],*p=s;
	int	c;
	fread(&x, sizeof(int),1,f);
	fread(&y, sizeof(int),1,f);
	fread(&w, sizeof(int),1,f);
	fread(&h, sizeof(int),1,f);
	fread(&icon_x, sizeof(int),1,f);
	fread(&icon_y, sizeof(int),1,f);
	fread(&flag, sizeof(int),1,f);
	while ((c=fgetc(f)) != 0)	*p++=(char)c;
	*p=0;
	name = new char[strlen(s)+1];
	strcpy(name,s);
}

WindowItem::WindowItem(int xx, int yy, int ww, int hh, char	*s)
{
	x=xx;
	y=yy;
	w=ww;
	h=hh;
	icon_x = -1;
	icon_y = -1;
	flag = 0;
	name = new char[strlen(s)+1];
	strcpy(name,s);
}

void FGAPI WindowItem::SaveToFile(FILE *f)
{
	fwrite(&x,sizeof(int),1,f);
	fwrite(&y,sizeof(int),1,f);
	fwrite(&w,sizeof(int),1,f);
	fwrite(&h,sizeof(int),1,f);
	fwrite(&icon_x,sizeof(int),1,f);
	fwrite(&icon_y,sizeof(int),1,f);
	fwrite(&flag,sizeof(int),1,f);
	fputs(name,f);
	fputc(0,f);
}

WindowItem::~WindowItem()
{
	delete [] name;
}

void FGAPI WindowDatabase::Add(int &xx, int &yy, int &ww, int &hh, char *s, int fl)
{
	WindowItem *w;
	if (strlen(s) == 0)
		return;

	lock();
	if ((w = findxy(s))==0)
	{
		char ss[128];
		sprintf(ss, "%s%s",exe, s);
		*endpointer = new WindowItem(xx, yy, ww, hh, ss);
		endpointer = &((*endpointer)->next);
		*endpointer = 0;
		xyCounter++;
	}
	else if (fl&WUSELAST)
	{
		xx = w->x;
		yy = w->y;
		if (fl&WSIZEABLE)
		{
			ww = w->w;
			hh = w->h;
		}
	}
	unlock();
}

void FGAPI WindowDatabase::DatabaseResize(int xx, int yy, char *s)
{
	lock();
	WindowItem *w = findxy(s);
	if (w)
	{
		w->w = xx;
		w->h = yy;
	}
	unlock();
}

void FGAPI WindowDatabase::DatabaseSetIcon(int xx, int yy, char *s)
{
	lock();
	WindowItem *w = findxy(s);
	if (w)
	{
		w->icon_x = xx;
		w->icon_y = yy;
	}
	unlock();
}

void FGAPI WindowDatabase::DatabaseGetIcon(int &xx, int &yy, char *s)
{
	lock();
	WindowItem *w = findxy(s);
	if (w)
	{
		xx = w->icon_x;
		yy = w->icon_y;
	}
	unlock();
}

void FGAPI WindowDatabase::DatabaseRelocate(int xx, int yy, char *s)
{
	lock();
	WindowItem *w = findxy(s);
	if (w)
	{
		w->x = xx;
		w->y = yy;
	}
	unlock();
}

// vykresli titulok okna
void FGWindow::WindowDrawTitle(int color)
{
	if (status & WTITLED) // modry/hnedy titulok
	{
		StateLock();
		SAVE_CLIPPING(this)
		save_state();

		set_font(FGFONT_TITLE);
		set_ppop(_GSET);

		int fr = 0, picto=(status&WNOPICTO?0:(status&WMINIMIZE)?PICTOSIZE*2:PICTOSIZE);
		if (GetStatus() & WFRAMED) fr = BASEBORDER-1;
		box(fr, fr, w - (fr + fr)-picto, TITLEH, color);
		box(fr, fr+TITLEH-1, w - (fr + fr)-picto, 1, CWHITE);
		box(w-((status & WFRAMED) ?4:0), fr, 1, TITLEH, CWHITE);
		if (status&WMINIMIZE && !(status&WNOPICTO))
			box(w-PICTOSIZE*2-4, fr, 1, TITLEH, CWHITE);
		if (status&WMINIMIZE || !(status&WNOPICTO))
			box(w-PICTOSIZE-4, fr, 1, TITLEH, CWHITE);
		text(5 + fr, fr + 1, skipstr(name), CWHITE, color);
		WindowPiktograms(w - PICTOSIZE*2-3, fr);
		WindowRepaint(fr, 0, w-fr-fr, BASEBORDER+TITLEH);

		restore_state();
		RESTORE_CLIPPING(this)
		StateUnlock();
	}
}

// nakresli okno
void FGWindow::draw(void)
{
	int ff = GetStatus() & WFRAMED;
	int	tt = GetStatus() & WTITLED;
	int	pp = GetStatus() & WNOPICTO;
	int	mm = GetStatus() & WMENU;
	int	mw = GetStatus() & WMENUWINDOWTYPE;
	int	ss = GetStatus() & WSTATUSBAR;
	int	nn = GetStatus() & WNOTEBOOK;
	int vs = BASEBORDER, vrch;

	SAVE_CLIPPING(this)
	if (tt) vs+=(TITLEH+2);
	if (mm) vs+=MENUH+1;
	if (nn) vs+=NOTEBOOKH+1;
	vrch=vs;
	if (ss) vs+=STATUSBARH;
	if (!ff)  // ak nemas FRAME, potom urob podklad pod titulok
	{
		vs -= 2;
		vrch -= BASEBORDER;
		if (tt) FGmemset(image+w * TITLEH, CBLACK, w); // standard
		else FGmemset(image+w * TITLEH, paper, w); // standard
	}
	if (!ff && !pp && tt)
	{
		FGmemset(image, BORD3, w * TITLEH); // standard
	}
   	else if (!tt) FGmemset(image, paper, w * TITLEH);
	if (ff)
	{
		// www optimise it for better resize time
		FGmemset(image, BORD3, w*h); // standard
		//
		box(0, 0, w, 1, BORD1);
		box(0, 0, 1, h, BORD1);
		box(1, h-1, w, 1, BORD2);
		box(w-1, 1, 1, h-1, BORD2);
		if (!mw)
		{
			box(BASEBORDER, vrch, w-BASEBORDER*2, h-BASEBORDER-vs, paper);
			rect(BASEBORDER-1, vrch-1, w-BASEBORDER*2+2, h-BASEBORDER-vs+2, CWHITE);
			if (nn)
				box(BASEBORDER-1, vrch-2, w-BASEBORDER*2+2,2, CBLACK);
			else
				box(BASEBORDER-1, vrch-1, w-BASEBORDER*2+2,1, CBLACK);
			box(BASEBORDER-1, vrch-1, 1, h-BASEBORDER-vs+2, CBLACK);
		}
		else
		{
			box(BASEBORDER-2, vrch-2, w-BASEBORDER*2+4, h-BASEBORDER-vs+4, paper);
		}
	}
	else FGmemset(image+w*TITLEH +w, paper, w*(h-TITLEH-1)); // standard

	if (status & WSIZEABLE && ff)
	{
		box(w - 13, h - 3, 12, 2, CYELLOW);
		box(w - 3,  h - 13, 2, 12, CYELLOW);
	}
	if (mm)
	{
		if (ff)
		{
			box(xoff-1, yoff - MENUH-2, w - xoff * 2+2, MENUH+2, CBLACK);
			box(xoff, vrch-MENUH, 1, MENUH+2, CWHITE);
		}
		else
		{
			box(xoff, yoff - MENUH-1-(tt?1:0), w - xoff * 2, MENUH+1+(tt?1:0), CBLACK);
			if (tt)
				box(xoff,MENUH+1, w-xoff*2, MENUH, MENUBG);
			else
				box(xoff,0, w-xoff*2, MENUH, MENUBG);
		}
		box(xoff, yoff - MENUH-1, w - xoff * 2, MENUH, MENUBG);
		box(xoff, yoff - MENUH-1, 1, MENUH, CBLACK);
	}
	if (ss)
	{
		if (ff == 0)
		{
			box(0, h-STATUSBARH, w, STATUSBARH,statuscolor);
		}
	}

	RESTORE_CLIPPING(this)

	if (first_draw)
		SendEvent(REPAINTEVENT);

	SAVE_CLIPPING2(this)

	// redraw notebooks
	if (status&WNOTEBOOK)
		DrawTabPages();

	FGBaseGui *pt = Buttony;
	while (pt)
	{
		((FGControl *) pt)->draw();
		pt = pt->GetNext();
	}
	RESTORE_CLIPPING(this)
	first_draw = 1;
	if (current==this)
	{
		WindowStatus(WACTIVE);
		RepaintBlock(0, 0, w, h);
	}
	else
	{
		WindowStatus(WDEACTIVE);
		WindowRepaint(0, 0, w, h);
	}
}

void FGApp::Timer(void)
{
	static unsigned int last=0,now=0;
	now = FGClock();
	if (last>now)
	{
		last = now;
		return;
	}
	if (now >= (last+ticks) || last>now)
	{
		last = now;
		__fg_fps_total = __fg_fps;
		__fg_fps = 0; // zero FPS counter
		if (OnEverySecond)
		{
			OnEverySecond(now/1000);
		}
		else OnTimer(now/1000);
	}
}

static void InitCRC(void)
{
  int I, J;
  unsigned int C;

  for (I=0;I<256;I++)
  {
    for (C=I,J=0;J<8;J++)
	  C=(C & 1) ? (C>>1)^0xEDB88320L : (C>>1);
	CRCTab[I]=C;
  }
}

unsigned int CalculateCRC(unsigned StartCRC, void *Addr, unsigned Size)
{
  unsigned I;
  if (!StartCRC) InitCRC();
  for (I=0; I<Size; I++)
	StartCRC = CRCTab[(unsigned char)StartCRC ^ ((unsigned char*)Addr)[I]] ^ (StartCRC >> 8);
  return(StartCRC);
}

/**
sets the new client space area in the FGWindow, i.e. clipping for drawing.
@param x new top left corner coordinate in x
@param y new top left corner coordinate in y
@param w a new width
@param h a new height
*/
bool FGAPI FGWindow::WindowSetWorkRect(int x,int y,int w,int h)
{
	if (set_clip_rect(x,y,w,h)==0)
		return false; // error!
	xpos = 2;
	ypos = 0;
	return true;
}

FGBaseGui::FGBaseGui(int xs, int ys, int ws, int hs, const char *nm, ObjectType typ, FGPixel i, FGPixel p, long flag)
	: FGDrawBuffer(ws,hs,nm,typ,i,p)
{
	ink = i;
	paper = p;
	status = flag;
	next = (FGBaseGui	*)0;
	private_data = user_data = (void *)-1;

	if (typ==WINDOW	&& flag&WUSELAST && entryPoint) entryPoint->Add(xs, ys, ws, hs, name,flag);
	// position check  & correction
	if (typ	== WINDOW || typ == MENUWINDOW)
	{
		if (ws > X_width) ws = X_width;
		if (hs > Y_width) hs = Y_width;
		if (ws < 8)	ws = 8;
		if (hs < 8)	hs = 8;
		if (status&WALIGNED)
		{
			xs &= ~3;
			ys &= ~3;
			ws &= ~3;
			hs &= ~3;
		}
		if ((xs + ws) >= X_width) xs = X_width - ws;
		if (xs < 0)	xs = 0;
		if ((ys + hs - 1) >= Y_width) ys = Y_width - hs;
		if (ys < 0)	ys = 0;
	}
	if (typ==WINDOW	&& flag&WCENTRED)
	{
		xs = (X_width-ws)/2;
		ys = (Y_width-hs)/2;
	}
	x =	xs;
	y =	ys;
	Resize(ws-w, hs-h);
	if (status & WMODAL) status &= ~WMINIMIZE;
}

FGBaseGui::~FGBaseGui()
{
	if (type == ROOTWINDOW) return;

	ListMutex.Lock();

	if (!next && !last) // no more windows ...
	{
	}
	else if	(next==0)	// last in queque
	{
		last->next	= 0;
	}
	else if	(last==0)	// first in queqe
	{
		next->last	= 0;
	}
	else if	(next && last)
	{
		last->next	= next;
		next->last	= last;
	}
	ListMutex.Unlock();
}

void FGAPI FGWindow::SendToWindow(FGEvent *e)
{
	if (image==0)
	{
		IError("Internal Error 1 (SendToWindow)",	1);
		return;
	}

	assert(this);
	e->wnd = this;
	int cc = GetId();

	StateLock();
	switch(e->GetType())
	{
		case REPAINTEVENT:
			OnPaint();
			break;
		case GETFOCUSEVENT:
			OnFocus();
			break;
		case LOSTFOCUSEVENT:
			OnLostFocus();
			break;
//		case INITEVENT:
//			OnInit();
//			break;
		case KEYEVENT:
			OnKeyPress(e->GetKey());
			break;
		case MOVEEVENT:
			OnMouseMove(e->GetX(), e->GetY());
			break;
		case ACCELEVENT:
			OnAccelerator(e->accel);
			break;
		case WINDOWMOVEEVENT:
			OnMove(e->GetX(), e->GetY());
			break;
		case WINDOWRESIZEEVENT:
			OnResize(e->GetW(), e->GetH());
			break;
		case BUTTONHOLDEVENT:
			break;
		case STARTDRAGLEFTEVENT:
			OnStartDrag(1, e->GetX(), e->GetY());
			break;
		case STARTDRAGRIGHTEVENT:
			OnStartDrag(0, e->GetX(), e->GetY());
			break;
		case DRAGLEFTEVENT:
			OnEndDrag(1, e->GetX(), e->GetY(), e->GetW(), e->GetH());
			break;
		case DRAGRIGHTEVENT:
			OnEndDrag(0, e->GetX(), e->GetY(), e->GetW(), e->GetH());
			break;
		case CLICKLEFTEVENT:
			OnClick(e->GetX(), e->GetY());
			break;
		case CLICKMIDDLEEVENT:
			OnMiddleButton(e->GetX(), e->GetY());
			break;
		case CLICKRIGHTEVENT:
			OnContextPopup(e->GetX(), e->GetY());
			break;
		case MOUSEWHEELEVENT:
			OnWheel(e->GetX(), e->GetY(), e->GetButtons());
			break;
		case TABSWITCHEVENT:
			OnTabSwitch((const char *)e->GetKey());
			break;
//		case TERMINATEEVENT:
//			OnClose();
//			break;
		case ICONIZEEVENT:
			OnIconize();
			break;
		case NOTIFYEVENT: // not implemented
//			OnCloseQuery(e->GetKey());
			break;
//		default:
//			::printf("Unhandled message %d in WndHandler!", e->GetType());
	}

	FGWindow *ww = WindowFind(cc);

	if (ww && handler)
		handler(e);

	// call only if don't removed
	ww = WindowFind(cc);
	if (ww)
		StateUnlock();
}

void FGMenuWindow::MenuWindowHandler(FGEvent *p)
{
	if (Proc) Proc(p);
	if (CurrentMenu==0) return;
	switch(p->GetType())
	{
		case KEYEVENT:
			if (p->GetKey() !=	ESC) break;
			delete CurrentMenu;		// QWERT
			break;
		case ACCELEVENT:
			p->wnd->RunHandler();
			if (!(p->GetObjectType() == LABEL
				|| p->GetObjectType() == PUSHBUTTON
				|| p->GetObjectType() == BASEMENU
				|| p->GetObjectType() == BASEMENUITEM))  break;
		case LOSTFOCUSEVENT:
		case TERMINATEEVENT:
			if (CurrentMenu == p->wnd)
				delete CurrentMenu; // do not delete others QWERT
			break;
	}
}

FGMenuWindow::FGMenuWindow(int w, int h, GuiHwnd proc):
	FGWindow(0,
		fgstate.__hint_x_menu==-1 ? cApp->GetMouseX()+4 : fgstate.__hint_x_menu,
		fgstate.__hint_y_menu==-1 ? cApp->GetMouseY()+4 : fgstate.__hint_y_menu,
		w,
		h,
		"",
		MenuWindowHandler,
		CScheme->menuwindow_fore,
		CScheme->menuwindow_back,
		WFRAMED|WCLICKABLE|WNOPICTO|WMENUWINDOWTYPE)
{
	draw();
	_initmw(proc);
}

void FGAPI FGWindow::default_clipping(void)
{
	if (status & WFRAMED)
		yoff += BASEBORDER;
	if (status & WTITLED)
		yoff += (TITLEH+2);
	if (status & WMENU)
		yoff += (MENUH+1);
	if (status & WNOTEBOOK)
		yoff += (NOTEBOOKH+1);
	if (status & WFRAMED)
		xoff += BASEBORDER;

	// set clipping window
	wwrk = w - xoff*2;
	hwrk = h - yoff;
	if (status & WFRAMED)
		hwrk -= BASEBORDER;
	if (status & WSTATUSBAR)
		hwrk -= STATUSBARH;
}

/**
	@param wp pointer at the pointer at	the window itself. It is an address of store of pointer to this window.
	When window	is destroyed, to this pointer is assigned NULL. You can basically
	test this pointer to NULL to prevent manipulating with destroyed
	object.
	@param xs X position
	@param ys Y position
	@param ws width of complete window (client space is this size minus the window frame.)
	@param hs height of complete window (client space is this size minus the window frame.)
	@param nm standard C string, which can be with any length. This name will be displayed it the title of the window
	(if title is present).
	@param hnd pointer to the window callback procedure. It is a big switch
	command with many cases. You can set this to 0 by default.
	@param i ink color
	@param p paper color
	@param flag for the window properties (e.g. has frame, has title etc.)
*/
FGWindow::FGWindow(FGWindow **wp, int xs, int ys, int ws, int hs, const char *nm, GuiHwnd hnd, int i, int p, int flag):
  FGBaseGui(xs, ys, ws, hs, nm, flag & WMENUWINDOWTYPE ? MENUWINDOW : WINDOW, i, p, flag|WEXIST), StateMutex(recursiveMutEx)
{
	FGWindow *lst;

	int ww = 0;
	int hh = 0;

	// fix minimal size - 5.feb.2003
	if (w<8) ww = 8;
	if (h<24) hh = 24;

	if (ww && hh)
		FGDrawBuffer::Resize(ww-w,hh-h);
	else if (ww)
		FGDrawBuffer::Resize(ww-w, 0);
	else if (hh)
		FGDrawBuffer::Resize(0, hh-h);

	SetInk(i);
	SetPaper(p);

	init();
	if (entryPoint)
		entryPoint->DatabaseGetIcon(icon_x, icon_y, name);

	if (current)
		if (current->GetStatus()&WMENUWINDOWTYPE && GetCounter()>1)
	{
		FGEvent e(TERMINATEEVENT);
		current->SendToWindow(&e);
	}

	Resize(0, 0);

	ListMutex.Lock();
	if (nWnd)			// first window
	{
		lst = current;
	}
	else
	{
		lst = (FGWindow *) 0;		// get addr of previous
		first = this;
	}
	current = this;
	nWnd++;
	ListMutex.Unlock();

	handler = hnd;
	if (wp) *wp = this;			// save pointer to itself
	itself = wp;

	SetLast(lst);
	statuscolor = CScheme->statusbar;
	if (lst)
	{
		FGEvent e(LOSTFOCUSEVENT);
		lst->SendToWindow(&e);
		lst->WindowStatus(WDEACTIVE);
		lst->SetNext(this);
	}
	first_draw = 0;
	default_clipping();
    if ((status&WNODRAWFROMCONSTRUCTOR) == false)
   		draw();
	FGEvent e(INITEVENT);
//	set_colors(i,p);
	SendToWindow(&e);
	if ((lst && flag & WLASTFOCUS) || (lst && lst->GetStatus() & WMODAL) && nWnd)
		lst->WindowFocus();
//	set_colors(i,p);
	oldidm = 0;
}

// ROOT wnd
FGWindow::FGWindow(FGWindow **wp, FGPixel color):
  FGBaseGui(0, 0, X_width, Y_width, "Root", ROOTWINDOW, CWHITE, CBLACK, WEXIST|WUNMOVED), StateMutex(recursiveMutEx)
{
	init();
	assert(!current);		// must be first
	Resize(0, 0);
	clear(color);

	if (wp) *wp = this;			// save pointer to itself
	itself = wp;
	assert(!nWnd);
	first = this;
	current = this;
	nWnd++;
	SetLast(0);
	fgstate.DoRootWindow = 1;
	FGEvent e(INITEVENT);
	SendToWindow(&e);
}

void FGAPI FGWindow::RemoveControls(void)
{
	FGBaseGui *butony = Buttony, *butony2;
	while (butony)
	{
		butony2 = butony->GetNext();
		delete butony;
		butony = butony2;
		nControls--;
	}
	Buttony = 0;
	cApp->AutoRepeatEnd();
	c_menupos = 1;
	oldidm = 0;
}

void FGAPI FGWindow::EnableControls(void)
{
	FGControl *butony = (FGControl *)Buttony, *butony2;
	while (butony)
	{
		butony2 = (FGControl *)butony->GetNext();
		butony->Enable();
		butony = butony2;
	}
	oldidm = 0;
}

void FGAPI FGWindow::DisableControls(void)
{
	FGControl *butony = (FGControl *)Buttony, *butony2;
	while (butony)
	{
		butony2 = (FGControl *)butony->GetNext();
		butony->Disable();
		butony = butony2;
	}
	oldidm = 0;
}

FGWindow::~FGWindow()
{
	FGBaseGui *posledny, *novy, *l, *n;

	if (iconized && iconized!=(FGWindow *)-1)
		delete iconized;

	StateLock();

	if (type!= ROOTWINDOW)
	{
		WindowFlushInput();
		status &= ~WEXIST;
		FGEvent e(TERMINATEEVENT);
		SendToWindow(&e);

		if (itself) *itself = 0;

		if (status&WNOTEBOOK)
			DeleteTabPages();
		else
			RemoveControls();

		if (!fgstate.shut_down) WindowUpdate(0,0,1);		// delete all
	}
	status = 0;			// no anythings

	ListMutex.Lock();

	posledny = GetLast();
	if (nWnd<=1/* || type== ROOTWINDOW*/)			// last in system
	{
		first = current = 0;
		nWnd  = 0;
		ListMutex.Unlock();
		StateUnlock();
		return;
	}
	novy = this==current ? posledny:current;
	if (!novy)
	{
		::printf("Internal error - window list mismatch!\a\n");
		exit(-1);
	}
	novy->SetNext(0);
	nWnd--;
	l = GetLast();
	n = GetNext();
	if (n == (FGWindow *) 0)		// last in queqe
	{
		current = (FGWindow *) l;
	}
	else if (l == (FGWindow *) 0)	// first in queqe
	{
		first = (FGWindow *) n;
	}
	oldidm = 0;

	ListMutex.Unlock();

	if (!fgstate.shut_down) ((FGWindow *) novy)->WindowStatus(WACTIVE);
	StateUnlock();
}

void FGAPI FGButtonGroup::DisableGroup(void)
{
	for(int i=0; i<count; i++)
	{
		array[i]->Disable();
	}
}

void FGAPI FGButtonGroup::EnableGroup(void)
{
	for(int i=0; i<count; i++)
	{
		array[i]->Enable();
	}
}

void FGAPI FGButtonGroup::AddToGroup(FGControl *p, int activ)
{
	int i;
	if (count>=size)
	{
		IError("FGButtonGroup - out of range!", 0);
		return; // maximum numbers reached
	}
	for(i=0;i<count;i++)
	{
		assert(p != array[i]);
	}
	i = p->GetType();
	if (i!=FGBaseGui::POINTBUTTON && i!=FGBaseGui::CHECKBUTTON && i!=FGBaseGui::PUSHBUTTON) return; // bad type
	if (type==0)
	{
		if (i==FGBaseGui::CHECKBUTTON || i==FGBaseGui::POINTBUTTON) type=1;
		else type=2;
	}
	else if (type==1 && i==FGBaseGui::PUSHBUTTON)
		IError("FGButtonGroup mismatch",0);
	else if (type==2 && (i==FGBaseGui::POINTBUTTON || i==FGBaseGui::CHECKBUTTON))
		IError("FGButtonGroup mismatch",0);

	array[count++] = p;

	p->RegisterToGroup(this); // say to button for group members

	if (activ && !curr)
	{
		curr = p;
		if (type==1) p->SetTrigger(1);
		else ((FGPushButton *)p)->Push();
	}
}

void FGControl::RefreshGroup(void)
{
	if (grp)
		grp->RefreshGroup(this);
	owner->oldidm = 0;
}

void FGAPI FGButtonGroup::RefreshGroup(FGControl *c)
{
	FGPushButton *pb1=(FGPushButton *)c;
	switch(type)
	{
		case 1:
			if (curr) curr->SetTrigger(0);
			if (curr == c) // to same
			{
				curr->SetTrigger(1);
			}
			break;
		case 2:
			if (curr) ((FGPushButton *)curr)->Release();
			pb1->Push();
			break;
	}
	curr = c;
}

void FGSlider::MoveIn(void)
{
	rect(0,0,w,h,CYELLOW);
	update_owner();
	owner->WindowRepaint(x, y, w, h);
}

void FGSlider::MoveOut(void)
{
	frame(0,0,w,h);
	update_owner();
	owner->WindowRepaint(x, y, w, h);
}

void FGSlider::frame(int x, int y, int w, int h, int f)
{
	FGPixel col = (f?CWHITE:CBLACK);
	vline(x, y, h, col);
	hline(x, y, w, col);
	vline(x + w - 2, y + 1, h - 2, CGRAYED);
	hline(x + 1, y + h - 2, w - 2, CGRAYED);
	col = (f?CBLACK:CWHITE);
	vline(x + w - 1, y + 0, h - 0, col);
	hline(x + 0, y + h - 1, w - 1, col);
}

void FGSlider::ClickDown(void)
{
	if (smer==0) modify((cApp->GetMouseX()-owner->GetX())-GetX()-10);
	else modify((cApp->GetMouseY()-owner->GetY())-GetY()-10);
}

void FGSlider::modify(int value)
{
	if (value>= 0)
	{
		*val = maxv;
	}
	else if (value < (maxv-minv)/steps+8)
	{
		*val = minv;
	}
	if (value>= 0 && value < (maxv-minv)/steps)
	{
	    *val = minv+value*steps;
	}
	draw();
}

void FGSlideBarV::draw(void)
{
	// WWW
	char str[40];
	unsigned dl=0;
	set_font(FONT0808);

	switch (status & 7)
	{
		case BNORMAL:
			box(1, 1, w - 3, h - 3,CScheme->slider);
			frame(0, 0, w, dl=(maxv-minv)/steps+16+6);
			frame(0,dl, w, h-dl);
			frame(2,3+((*val-minv)/steps),w-4,w,1);
			str[0] = char((*val%1000)/100+'0');
			str[1] = 0;
			str[2] = char((*val%100)/10+'0');
			str[3] = 0;
			str[4] = char(*val%10+'0');
			str[5] = 0;
			str[6] = char((*val%10000)/1000+'0');
			str[7] = 0;
			text(4,dl+3,str+6, CBLACK, CScheme->slider);
			text(4,dl+11,str, CBLACK, CScheme->slider);
			text(4,dl+19,str+2, CBLACK, CScheme->slider);
			text(4,dl+27,str+4, CBLACK, CScheme->slider);
			break;

		case BDISABLED:
			box(1,   1, w - 3, h - 3, CGRAY2);
			frame(0, 0, w, dl=(maxv-minv)/steps+16+6);
			frame(0, dl, w, h-dl);
			frame(2, 3+((*val-minv)/steps),w-4,w,1);
			str[0] = char((*val%1000)/100+'0');
			str[1] = 0;
			str[2] = char((*val%100)/10+'0');
			str[3] = 0;
			str[4] = char(*val%10+'0');
			str[5] = 0;
			str[6] = char((*val%10000)/1000+'0');
			str[7] = 0;
			text(4,dl+3,str+6, CScheme->slider_disable, CScheme->slider);
			text(4,dl+11,str, CScheme->slider_disable, CScheme->slider);
			text(4,dl+19,str+2, CScheme->slider_disable, CScheme->slider);
			text(4,dl+27,str+4, CScheme->slider_disable, CScheme->slider);
			break;
		case BUNVISIBLE:
			clear(paper);;
			break;
	}
	update_owner();
	owner->WindowRepaint(x,y,w,h);
	if (fnc) fnc(this);
}

void FGSlideBarH::draw(void)
{
	// WWW
	char str[40];
	unsigned dl=0;
	set_font(FONT0808);

	switch (status & 7)
	{
		case BNORMAL:
			box(1, 1, w - 3, h - 3,CScheme->slider);
			frame(0,0,dl=(maxv-minv)/steps+16+6,GetH());
			frame(dl,0,w-dl,h);
			frame(3+((*val-minv)/steps),2,h,h-4,1);
			sprintf(str, "%3d", *val);
			text(dl+2, 4,str,CBLACK, CScheme->slider);
			break;

		case BDISABLED:
			box(1, 1, w - 3, h - 3,CGRAY2);
			frame(0,0,dl=(maxv-minv)/steps+16+6,GetH());
			frame(dl,0,w-dl,h);
			frame(3+((*val-minv)/steps),2,16,12,1);
			sprintf(str, "%3d", *val);
			text(dl+2, 4,str,CScheme->slider_disable, CGRAY2);
			break;

		case BUNVISIBLE:
			clear(paper);;
			break;
	}
	update_owner();
	owner->WindowRepaint(x,y,w,h);
	if (fnc) fnc(this);
}

FGText::FGText(int xs, int ys, const char *nm, FGWindow *wnd, int i, int p)
	: FGControl(xs, ys, 8, 8, nm, WINDOWTEXT, 0, BNORMAL, wnd,	0, p, i)
{
	if (image) free(image);
	image = 0;
	set_font(wnd->get_font());
	draw();
}

void FGText::draw(void)
{
	if (IsYourTabPage() && name)
    {
        int ff = owner->get_font();
        owner->set_font(state._font);

    	switch (status & 7)
	    {
    		case BNORMAL:
	        	owner->WindowText(x-owner->GetXW(), y-owner->GetYW(), name,ink,paper);
                break;
    		case BUNVISIBLE:
	        	owner->WindowText(x-owner->GetXW(), y-owner->GetYW(), name,paper,paper);
		    	break;
        }
        owner->set_font(ff);
    }
}

FGImage::FGImage(int xs, int ys, FGDrawBuffer* img, FGWindow *wnd)
	: FGControl(xs, ys, img->GetW(), img->GetH(), 0, WINDOWIMAGE, 0, 0, wnd,	0, 0, 0)
{
	bitmap = img;
}

void FGImage::draw(void)
{
	if (IsYourTabPage())
    {
    	switch (status & 7)
	    {
    		case BNORMAL:
        		owner->WindowPutBitmap(x-owner->GetXW(), y-owner->GetYW(), 0, 0, w, h, bitmap);
                break;
    		case BUNVISIBLE:
        		owner->WindowBox(x-owner->GetXW(), y-owner->GetYW(), w, h, paper);
		    	break;
        }
    }
}

FGPanel::FGPanel(int xs, int ys, int ws, int hs, const char *nm, FGWindow *wnd, int i1, int i2, int p)
	: FGControl(xs, ys, ws, hs, nm, WINDOWPANEL, 0, 0, wnd,	0, p, i1), ink1(i1), ink2(i2)
{
	if (image) free(image);
	image = 0;
	owner->clip(x,y,w,h);
	draw();
}

void FGPanel::draw(void)
{
	if (IsYourTabPage())
	{
		owner->StateLock();
		SAVE_CLIPPING(owner)
		owner->hline(x, y, w, ink1);
		owner->vline(x, y, h, ink1);
		owner->hline(x, y + h - 1, w, ink2);
		owner->vline(x	+ w - 1, y, h, ink2);
		if (name && *name) owner->printf(x+8, y-8, " %s ", name);
		RESTORE_CLIPPING(owner)
		owner->WindowRepaint(x,y,w,h);
		owner->StateUnlock();
	}
}

/**
	Adds a separator to the menu.
*/
void FGMenuWindow::Separator(void)
{
	int i;
	WindowBox(0, i=GetYM(8)+3, GetWW(), 1, CBLACK);
	WindowBox(0, i+1, GetWW(), 1, CWHITE);
}

/**
	Adds a text form of FGPushButton to the FGWindow.
	@param xs x position in a window
	@param ys y position in a window
	@param ws width of the button
	@param hs height of the button, there is recomended values from 21 and above
	@param nm name of the button
	@param key a keycode used as a hot-key for the button (it isn't case sensitive). By example: 'a', 0x20 or ALT_G
	@param f a callback function of type ControllCall
	@param flag initial state, default is BNORMAL
	@return a pointer to created object, you can save this one if you can call any methods.
	@note You don't need save the pointer if you don't want call other methods ( e.g. Enable() ) or groups with other buttons.
	The object will be release automatically when the FGWindow will be destroyed. Don't destruct the object explicitly!
	@see FGBaseGui FGControl FGPushButton FGButtonGroup
*/
FGPushButton *FGWindow::AddPushButton(int xs, int ys, int ws, int hs, const char *nm,	int	key, ControlCall f, int flag)
{
	return new FGPushButton(xs, ys, ws, hs, nm, key, flag,this,f);
}

/**
	Adds an icon form of FGPushButton to the FGWindow.
	@param xs x position in a window
	@param ys y position in a window
	@param key a keycode used as a hot-key for the button (it isn't case sensitive). By example: 'a', 0x20 or ALT_G
	@param bm pointer to image object - ICON as a FGDrawBuffer object
	@param f a callback function of type ControllCall
	@param flag initial state, default is BNORMAL
	@return a pointer to created object, you can save this one if you can call any methods.
	@note You don't need save the pointer if you don't want call other methods ( e.g. Enable() ) or groups with other buttons.
	The object will be release automatically when the FGWindow will be destroyed. Don't destruct the object explicitly!
	@see FGBaseGui FGControl FGPushButton FGButtonGroup
*/
FGPushButton *FGWindow::AddPushButton(int xs, int ys, int key, FGDrawBuffer *bm, ControlCall f, int flag)
{
	if (bm==0 || bm->GetType()==0)
	{
		// fallback if no valid image was passed
		new FGPushButton(xs, ys, 128, 21, "image not found!!!", 0, BNORMAL, this, ControlCall(0) );
	}
	return new FGPushButton(xs, ys, key, flag,this,bm, f);
}

/**
	Adds an icon form of FGPushButton to the FGWindow. This method can set size & name of button directly.
	@param xs x position in a window
	@param ys y position in a window
	@param ws width of the button
	@param hs height of the button, there is recomended values from 21 and above
	@param nm name of the button
	@param key a keycode used as a hot-key for the button (it isn't case sensitive). By example: 'a', 0x20 or ALT_G
	@param bm pointer to image object - ICON as a FGDrawBuffer object
	@param f a callback function of type ControllCall
	@param flag initial state, default is BNORMAL
	@return a pointer to created object, you can save this one if you can call any methods.
	@note You don't need save the pointer if you don't want call other methods ( e.g. Enable() ) or groups with other buttons.
	The object will be release automatically when the FGWindow will be destroyed. Don't destruct the object explicitly!
	@see FGBaseGui FGControl FGPushButton FGButtonGroup
*/
FGPushButton *FGWindow::AddPushButton(int xs, int ys, int ws, int hs, const char *nm, int key, FGDrawBuffer *bm, ControlCall f, int flag)
{
	if (bm==0 || bm->GetType()==0)
	{
		// fallback if no valid image was passed
		new FGPushButton(xs, ys, 128, 21, "image not found!!!", 0, BNORMAL, this, ControlCall(0) );
	}
	return new FGPushButton(xs, ys, ws, hs, nm, key, flag, this, bm, f);
}

/**
	Adds FGTwoStateButton to the FGWindow.
	@param xs x position in a window
	@param ys y position in a window
	@param w width of the button
	@param h height of the button, there is recomended values from 21 and above
	@param nm name of the button
	@param key a keycode used as a hot-key for the button (it isn't case sensitive). By example: 'a', 0x20 or ALT_G
	@param variable a pointer to int variable. The value of this variable controls the state of button (true/false)
	@param f a callback function of type ControllCall
	@return a pointer to created object, you can save this one if you can call any methods.
	@note You don't need save the pointer if you don't want call other methods ( e.g. Enable() ) or groups with other buttons.
	The object will be release automatically when the FGWindow will be destroyed. Don't destruct the object explicitly!
	@see FGBaseGui FGControl FGTwoStateButton FGButtonGroup
*/
FGTwoStateButton	*FGWindow::AddTwoStateButton(int	xs,	int	ys,	int w, int h, const char *nm, int key, int * variable, ControlCall f)
{
	return new FGTwoStateButton(xs, ys, w, h, nm, key, this, variable, f);
}

/**
	Adds a FGCheckButton to the FGWindow.
	@param xs x position in a window
	@param ys y position in a window
	@param nm name of the button
	@param key a keycode used as a hot-key for the button (it isn't case sensitive). By example: 'a', 0x20 or ALT_G
	@param variable a pointer to int variable. The value of this variable controls the state of button (true/false)
	@param f a callback function of type ControllCall
	@param flag initial state, default is BNORMAL
	@return a pointer to created object, you can save this one if you can call any methods.
	@note You don't need save the pointer if you don't want call other methods ( e.g. Enable() ) or groups with other buttons.
	The object will be release automatically when the FGWindow will be destroyed. Don't destruct the object explicitly!
	@see FGBaseGui FGControl FGCheckButton FGButtonGroup
*/
FGCheckButton	*FGWindow::AddCheckButton(int	xs,	int	ys,	const char *nm, int key, int * variable, ControlCall f, int flag)
{
	return new FGCheckButton(1,xs, ys, nm, key,	flag,this,GetInk(),GetPaper(),f,variable);
}

/**
	Adds a FGCheckButton to the FGWindow.
	@param xs x position in a window
	@param ys y position in a window
	@param nm name of the button
	@param key a keycode used as a hot-key for the button (it isn't case sensitive). By example: 'a', 0x20 or ALT_G
	@param variable a pointer to bool variable. The value of this variable controls the state of button (true/false)
	@param f a callback function of type ControllCall
	@param flag initial state, default is BNORMAL
	@return a pointer to created object, you can save this one if you can call any methods.
	@note You don't need save the pointer if you don't want call other methods ( e.g. Enable() ) or groups with other buttons.
	The object will be release automatically when the FGWindow will be destroyed. Don't destruct the object explicitly!
	@see FGBaseGui FGControl FGCheckButton FGButtonGroup
*/
FGCheckButton	*FGWindow::AddCheckButton(int	xs,	int	ys,	const char *nm, int key, bool * variable, ControlCall f, int flag)
{
	return new FGCheckButton(xs, ys, nm, key,	flag,this,GetInk(),GetPaper(),f,variable);
}

/**
	Adds a FGPointButton to the FGWindow.
	@param xs x position in a window
	@param ys y position in a window
	@param nm name of the button
	@param key a keycode used as a hot-key for the button (it isn't case sensitive). By example: 'a', 0x20 or ALT_G
	@param variable a pointer to int variable. The value of this variable controls the state of button (true/false)
	@param f a callback function of type ControllCall
	@param flag initial state, default is BNORMAL
	@return a pointer to created object, you can save this one if you can call any methods.
	@note You don't need save the pointer if you don't want call other methods ( e.g. Enable() ) or groups with other buttons.
	The object will be release automatically when the FGWindow will be destroyed. Don't destruct the object explicitly!
	@see FGBaseGui FGControl FGPointButton FGButtonGroup
*/
FGPointButton	*FGWindow::AddPointButton(int	xs,	int	ys,	const char *nm, int key, int * variable, ControlCall f, int flag)
{
	return new FGPointButton(1,xs, ys, nm, key,	flag,this,GetInk(),GetPaper(),f,variable);
}

/**
	Adds a masked version FGCheckButton to the FGWindow. This is like a normal check box
	but it is controlled by one of a 32 bit in a triggered variable (of type uint32).
	This is useful when you want change only one bit of variable.
	@param mask bit mask value, by example 1 for a bit 0 or (1<<20) for twenth bit.
	@param xs x position in a window
	@param ys y position in a window
	@param nm name of the button
	@param key a keycode used as a hot-key for the button (it isn't case sensitive). By example: 'a', 0x20 or ALT_G
	@param variable a pointer to int variable. The value of this variable controls the state of button (true/false)
	@param f a callback function of type ControllCall
	@param flag initial state, default is BNORMAL
	@return a pointer to created object, you can save this one if you can call any methods.
	@note You don't need save the pointer if you don't want call other methods ( e.g. Enable() ) or groups with other buttons.
	The object will be release automatically when the FGWindow will be destroyed. Don't destruct the object explicitly!
	@see FGBaseGui FGControl FGCheckButton FGButtonGroup
*/
FGCheckButton	*FGWindow::AddCheckButtonMask(int mask, int	xs,	int	ys,	const char *nm, int key, int * variable, ControlCall f, int flag)
{
	return new FGCheckButton(mask, xs, ys, nm, key,	flag,this,GetInk(),GetPaper(),f,variable);
}

/**
	Adds a masked version FGPointButton to the FGWindow. This is like a normal check box
	but it is controlled by one of a 32 bit in a triggered variable (of type uint32).
	This is useful when you want change only one bit of variable.
	@param mask bit mask value, by example 1 for a bit 0 or (1<<20) for twenth bit.
	@param xs x position in a window
	@param ys y position in a window
	@param nm name of the button
	@param key a keycode used as a hot-key for the button (it isn't case sensitive). By example: 'a', 0x20 or ALT_G
	@param variable a pointer to int variable. The value of this variable controls the state of button (true/false)
	@param f a callback function of type ControllCall
	@param flag initial state, default is BNORMAL
	@return a pointer to created object, you can save this one if you can call any methods.
	@note You don't need save the pointer if you don't want call other methods ( e.g. Enable() ) or groups with other buttons.
	The object will be release automatically when the FGWindow will be destroyed. Don't destruct the object explicitly!
	@see FGBaseGui FGControl FGPointButton FGButtonGroup
*/
FGPointButton	*FGWindow::AddPointButtonMask(int mask, int	xs,	int	ys,	const char *nm, int key, int * variable, ControlCall f, int flag)
{
	return new FGPointButton(mask, xs, ys, nm, key,	flag,this,GetInk(),GetPaper(),f,variable);
}

/**
	Adds an FGEditBox for an one-line-edit box for an integer number without the test to in-a-range-number.
	@param xs x position in a window
	@param ys y position in a window
	@param ws1 width of the button caption, the text is right alligned
	@param ws2 width of the button input part
	@param nm name of the button
	@param key a keycode used as a hot-key for the button (it isn't case sensitive). By example: 'a', 0x20 or ALT_G
	@param p a pointer to the edited int variable
	@param f a callback function of type ControllCall
	@return a pointer to created object, you can save this one if you can call any methods.
	@note You don't need save the pointer if you don't want call other methods ( e.g. Enable() ) or groups with other buttons.
	The object will be release automatically when the FGWindow will be destroyed. Don't destruct the object explicitly!
	@see FGBaseGui FGControl FGEditBox
*/
FGEditBox	*FGWindow::AddEditBox(int	xs,	int	ys,	int	ws1, int ws2, const char *nm,	int	key, int *p, ControlCall f)
{
	return new FGEditBox(xs, ys, ws1,	ws2, nm, key, this,	p, GetInk(),GetPaper(),	f, 0, 0, 0); // no checking
}

/**
	Adds an FGEditBox for an one-line-edit box for an integer number with the  test to in-a-range-number.
	@param xs x position in a window
	@param ys y position in a window
	@param ws1 width of the button caption, the text is right alligned
	@param ws2 width of the button input part
	@param nm name of the button
	@param key a keycode used as a hot-key for the button (it isn't case sensitive). By example: 'a', 0x20 or ALT_G
	@param p a pointer to the edited int variable
	@param f a callback function of type ControllCall
	@param min minimal correct value from -2147483648 to 2147483647 for a range testing
	@param max maximal correct value from -2147483648 to 2147483647 for a range testing
	@return a pointer to created object, you can save this one if you can call any methods.
	@note You don't need save the pointer if you don't want call other methods ( e.g. Enable() ) or groups with other buttons.
	The object will be release automatically when the FGWindow will be destroyed. Don't destruct the object explicitly!
	@see FGBaseGui FGControl FGEditBox
*/
FGEditBox	*FGWindow::AddEditBox(int	xs,	int	ys,	int	ws1, int ws2, const char *nm,	int	key, int *p, ControlCall f, int min, int max)
{
	return new FGEditBox(xs, ys, ws1,	ws2, nm, key, this,	p, GetInk(),GetPaper(),	f, min,	max, 1);
}

/**
	Adds an FGEditBox for an one-line-edit box for an string. The max length of string is get from the ws2 parameter/8, i.e. you don't type bigger string string that that value.
	@param xs x position in a window
	@param ys y position in a window
	@param ws1 width of the button caption, the text is right alligned
	@param ws2 width of the button input part
	@param nm name of the button
	@param key a keycode used as a hot-key for the button (it isn't case sensitive). By example: 'a', 0x20 or ALT_G
	@param p a pointer to the edited int variable
	@param f a callback function of type ControllCall
	@return a pointer to created object, you can save this one if you can call any methods.
	@note You don't need save the pointer if you don't want call other methods ( e.g. Enable() ) or groups with other buttons.
	The object will be release automatically when the FGWindow will be destroyed. Don't destruct the object explicitly!
	@see FGBaseGui FGControl FGEditBox
*/
FGEditBox	*FGWindow::AddEditBox(int	xs,	int	ys,	int	ws1, int ws2, const char *nm,	int	key, char *p, ControlCall f)
{
	return new FGEditBox(0, xs, ys, ws1,	ws2, nm, key, this,	p, GetInk(),GetPaper(),	f);
}

/**
	Adds an FGEditBox for an one-line-edit box for an string. The max length of string is get from the ws2 parameter/8, i.e. you don't type bigger string string that that value.
	@param sz is the maximal size of input string in range from 1 to 127 characters.
	@param xs x position in a window
	@param ys y position in a window
	@param ws1 width of the button caption, the text is right alligned
	@param ws2 width of the button input part
	@param nm name of the button
	@param key a keycode used as a hot-key for the button (it isn't case sensitive). By example: 'a', 0x20 or ALT_G
	@param p a pointer to the edited int variable
	@param f a callback function of type ControllCall
	@return a pointer to created object, you can save this one if you can call any methods.
	@note You don't need save the pointer if you don't want call other methods ( e.g. Enable() ) or groups with other buttons.
	The object will be release automatically when the FGWindow will be destroyed. Don't destruct the object explicitly!
	@see FGBaseGui FGControl FGEditBox
*/
FGEditBox	*FGWindow::AddEditBox(int sz, int	xs,	int	ys,	int	ws1, int ws2, const char *nm,	int	key, char *p, ControlCall f)
{
	return new FGEditBox(sz, xs, ys, ws1,	ws2, nm, key, this,	p, GetInk(),GetPaper(),	f);
}

/**
	Adds an FGEditBox for an one-line-edit box for an double number with the  test to in-a-range-number.
	@param xs x position in a window
	@param ys y position in a window
	@param ws1 width of the button caption, the text is right alligned
	@param ws2 width of the button input part
	@param nm name of the button
	@param key a keycode used as a hot-key for the button (it isn't case sensitive). By example: 'a', 0x20 or ALT_G
	@param p a pointer to the edited double variable
	@param f a callback function of type ControllCall
	@param min minimal correct value for a range testing
	@param max maximal correct value for a range testing
	@return a pointer to created object, you can save this one if you can call any methods.
	@note You don't need save the pointer if you don't want call other methods ( e.g. Enable() ) or groups with other buttons.
	The object will be release automatically when the FGWindow will be destroyed. Don't destruct the object explicitly!
	@see FGBaseGui FGControl FGEditBox
*/
FGEditBox	*FGWindow::AddEditBox(int	xs,	int	ys,	int	ws1, int ws2, const char *nm,	int	key, double *p, ControlCall f, double min, double max)
{
	return new FGEditBox(xs, ys, ws1,	ws2, nm, key, this,	p, GetInk(),GetPaper(),	f, min,	max, 1);
}

/**
	Adds an FGEditBox for an one-line-edit box for an double number without the test to in-a-range-number.
	@param xs x position in a window
	@param ys y position in a window
	@param ws1 width of the button caption, the text is right alligned
	@param ws2 width of the button input part
	@param nm name of the button
	@param key a keycode used as a hot-key for the button (it isn't case sensitive). By example: 'a', 0x20 or ALT_G
	@param p a pointer to the edited double variable
	@param f a callback function of type ControllCall
	@return a pointer to created object, you can save this one if you can call any methods.
	@note You don't need save the pointer if you don't want call other methods ( e.g. Enable() ) or groups with other buttons.
	The object will be release automatically when the FGWindow will be destroyed. Don't destruct the object explicitly!
	@see FGBaseGui FGControl FGEditBox
*/
FGEditBox	*FGWindow::AddEditBox(int	xs,	int	ys,	int	ws1, int ws2, const char *nm,	int	key, double *p, ControlCall f)
{
	return new FGEditBox(xs, ys, ws1,	ws2, nm, key, this,	p, GetInk(),GetPaper(),	f, 0, 0, 0); // no checking
}

/**
	Adds an line menu item to the FGWindow. This used for pop-up menus. The position and size is calculated automatically.
	@param nm name of the button
	@param key a keycode used as a hot-key for the button (it isn't case sensitive). By example: 'a', 0x20 or ALT_G
	@param f a callback function of type ControllCall
	@return a pointer to created object, you can save this one if you can call any methods.
	@note You don't need save the pointer if you don't want call other methods ( e.g. Enable() ) or groups with other buttons.
	The object will be release automatically when the FGWindow will be destroyed. Don't destruct the object explicitly!
	@see FGBaseGui FGControl FGBaseMenu
*/
FGBaseMenu *FGWindow::AddBaseMenu(const char *nm,	int	key, ControlCall f)
{
	return new FGBaseMenu(nm, key, this, f, FGFONT_MENU);
}

/**
	Adds a text Label.
	@param x x position in a window
	@param y y position in a window
	@param s name of the button
	@param key a keycode used as a hot-key for the button (it isn't case sensitive). By example: 'a', 0x20 or ALT_G
	@param f a callback function of type ControllCall
	@param i explicit foreground color, default is color of parents
	@param p explicit background color, default is color of parents
	@return a pointer to created object, you can save this one if you can call any methods.
	@note You don't need save the pointer if you don't want call other methods ( e.g. Enable() ) or groups with other buttons.
	The object will be release automatically when the FGWindow will be destroyed. Don't destruct the object explicitly!
	@see FGBaseGui FGControl Label
*/
FGLabel *FGWindow::AddLabel(int x, int y, const char *s, int key, ControlCall f, unsigned i, unsigned p)
{
	return new FGLabel(x,y,s,key,this,f,i==UNDEFINED_COLOR ?this->ink:i, p==UNDEFINED_COLOR ?this->paper:p);
}

/**
	Adds panel to the FGWindow.
	@param x x position in a window
	@param y y position in a window
	@param w width of the button caption, the text is right alligned
	@param h width of the button input part
	@param s name of the panel
	@param paper explicit background color, default is color of parents
	@param ink explicit foreground color1, default is CWHITED
	@param ink2 explicit foreground color1, default is CDARK
	@note This object is replacement for an older WindowPanel() methods that is absolete
	and will be removed in OpenGUI 4.4.0
	The main reason is compatibility with TabPages - WindowPanel() doesn't work properly with one.
*/
FGPanel* FGWindow::AddPanel(int x, int y, int w, int h, const char *s, unsigned paper, unsigned ink, unsigned ink2)
{
	return new FGPanel(x,y,w,h,s,this,ink,ink2,paper==UNDEFINED_COLOR?this->paper:paper);
}

/**
	Adds dynamic text string to the FGWindow.
	@param x x position in a window
	@param y y position in a window
	@param s name of the panel
	@param ink explicit foreground color1, default is CWHITED
	@param paper explicit background color, default is color of parents
*/
FGText* FGWindow::AddText(int x, int y, const char *s, unsigned ink, unsigned paper)
{
	return new FGText(x,y,s,this,ink==UNDEFINED_COLOR?this->ink:ink,paper==UNDEFINED_COLOR?this->paper:paper);
}

/**
	Adds dynamic image to the FGWindow.
	The attached image is nor released at the end of live.
	@param x x position in a window
	@param y y position in a window
	@param img Image
*/
FGImage* FGWindow::AddImage(int x, int y, FGDrawBuffer* img)
{
	return new FGImage(x, y, img, this);
}

/**
	Adds a horizontal slider to the FGWindow.
	@param xs x position in a window
	@param ys y position in a window
	@param min the start value of slider
	@param max the last value of slider
	@param step thes stepping (1 and above)
	@param val a pointer to the controled int variable
	@param f a callback function of type ControllCall
	@return a pointer to created object, you can save this one if you can call any methods.
	@note You don't need save the pointer if you don't want call other methods ( e.g. Enable() ) or groups with other buttons.
	The object will be release automatically when the FGWindow will be destroyed. Don't destruct the object explicitly!
	@see FGBaseGui FGControl FGSlider FGSlideBarH
*/
FGSlideBarH *FGWindow::AddSlideBarH(int xs, int ys, int min, int max, int step, int *val, ControlCall f)
{
	return new FGSlideBarH(xs, ys, min, max, step, val, this,f);
}

/**
	Adds a vertical slider to the FGWindow.
	@param xs x position in a window
	@param ys y position in a window
	@param min the start value of slider
	@param max the last value of slider
	@param step thes stepping (1 and above)
	@param val a pointer to the controled int variable
	@param f a callback function of type ControllCall
	@return a pointer to created object, you can save this one if you can call any methods.
	@note You don't need save the pointer if you don't want call other methods ( e.g. Enable() ) or groups with other buttons.
	The object will be release automatically when the FGWindow will be destroyed. Don't destruct the object explicitly!
	@see FGBaseGui FGControl FGSlider FGSlideBarV
*/
FGSlideBarV *FGWindow::AddSlideBarV(int xs, int ys, int min, int max, int step, int *val, ControlCall f)
{
	return new FGSlideBarV(xs, ys, min, max, step, val, this,f);
}

/**
	Adds a LISTBOX container to the FGWindow.
	@param xs x position in a window
	@param ys y position in a window
	@param w width of the one item
	@param h height of the one item
	@param dropdown number of lines that list box occupies. This is not a real number items, it is only visual size.
	@param drawone a callback function of type FGListBoxCallBack
	@param userdata user defined data pointer
	@return a pointer to created object, you can save this one if you can call any methods.
	@note You don't need save the pointer if you don't want call other methods ( e.g. Enable() ) or groups with other buttons.
	The object will be release automatically when the FGWindow will be destroyed. Don't destruct the object explicitly!
	@see FGListBox listbox0
*/
FGListBox * FGWindow::AddListBox(int xs, int ys, int w, int h, int dropdown, FGListBoxCallBack drawone, void* userdata)
{
	return new FGListBox(xs, ys, w, h, dropdown, this, drawone, userdata);
}

/*
FGListBox * FGWindow::AddListBox(int xs, int ys, int wpol, int hpol, int w, int h, int *cnt, int *pos, FGListBoxCallBack drawone, void *data)
{
	FGListBox *l = new FGListBox(xs, ys, w, h, hpol, this, drawone);
	l->SetSize(*cnt);
	l->SetIndex(*pos);
	return l;
}

FGListBox * FGWindow::AddListBox(int xs, int ys, int wpol, int hpol, int w, int h, int cnt, FGListBoxCallBack drawone, void *data)
{
	FGListBox *l = new FGListBox(xs, ys, w, h, hpol, this, drawone);
	l->SetSize(cnt);
	l->SetIndex(0);
	return l;
}
*/

/**
	Adds a FGProgressBar into the FGWindow.
	@param xs x position in a window
	@param ys y position in a window
	@param ws width of widget
	@param hs height of widget
	@param size the progressed value
	@return a pointer to created object, you can save this one if you can call any methods.
	@note You don't need save the pointer if you don't want call other methods ( e.g. Enable() ) or groups with other buttons.
	The object will be release automatically when the FGWindow will be destroyed. Don't destruct the object explicitly!
	@see FGBaseGui FGControl FGProgressBar
*/
FGProgressBar *FGWindow::AddProgressBar(int xs, int ys, int ws, int hs, int size)
{
	return new FGProgressBar(this, xs,ys,ws,hs,size);
}

/* --------------------------- WINDOW MENU ----------------------------------------- */

FGBaseMenu::FGBaseMenu(const char *nm, int key, FGWindow *w, ControlCall f, int font)
	: FGControl(w->GetXM(FGFontManager::textwidth(font,nm)+8),
		w->GetYM(),
		FGFontManager::textwidth(font,nm)+8,
		FGFontManager::GetH(font),
		nm,
		BASEMENU,
		key,
		BNORMAL,
		w,
		f,
		CScheme->menu_back,
		CScheme->menu_fore)
{
	set_font(font);
	draw();
}

FGBaseMenuItem::FGBaseMenuItem(const char *nm, int key, FGMenuWindow *w, ControlCall f, int font)
	: FGControl(w->GetXM()-4,
		w->GetYM(FGFontManager::GetH(font)+4),
		w->GetWW(),
		FGFontManager::GetH(font)+4,
		nm,
		BASEMENUITEM,
		key,
		BNORMAL,
		w,
		f,
		w->GetPaper(),
		w->GetInk())
{
	set_font(font);
	draw();
}

/**
	Adds an single menu item to the FGMenuWindow. This used for pop-up menus. The position and size is calculated automatically.
	@param nm name of the button
	@param key a keycode used as a hot-key for the button (it isn't case sensitive). By example: 'a', 0x20 or ALT_G
	@param f a callback function of type ControllCall
	@return a pointer to created object, you can save this one if you can call any methods.
	@note You don't need save the pointer if you don't want call other methods ( e.g. Enable() ) or groups with other buttons.
	The object will be release automatically when the FGWindow will be destroyed. Don't destruct the object explicitly!
	@see FGBaseGui FGControl FGBaseMenuItem
*/
FGBaseMenuItem *FGMenuWindow::AddMenu(char *nm,	int	key, ControlCall f)
{
	return new FGBaseMenuItem(nm, key, this, f, FGFONT_MENU);
}

/**
	Adds a Label for an one-line-edit box for an string. The max length of string is get from the ws2 parameter/8, i.e. you don't type bigger string string that that value.
	@param nm name of the button
	@param key a keycode used as a hot-key for the button (it isn't case sensitive). By example: 'a', 0x20 or ALT_G
	@param f a callback function of type ControllCall
	@param i explicit foreground color, default is color of parents
	@param p explicit background color, default is color of parents
	@return a pointer to created object, you can save this one if you can call any methods.
	@note You don't need save the pointer if you don't want call other methods ( e.g. Enable() ) or groups with other buttons.
	The object will be release automatically when the FGWindow will be destroyed. Don't destruct the object explicitly!
	@see FGBaseGui FGControl Label
*/
FGLabel *FGMenuWindow::AddLabel(char *nm, int key, ControlCall f, unsigned i, unsigned p)
{
	return new FGLabel(GetXM(), GetYM(18),nm,key,this,f,i==UNDEFINED_COLOR?this->ink:i, p==UNDEFINED_COLOR?this->paper:p);
}

/**
	Adds dynamic text string to the FGMenuWindow.
	@param s name of the panel
	@param ink explicit foreground color1, default is CWHITED
	@param paper explicit background color, default is color of parents
*/
FGText* FGMenuWindow::AddText(const char *s, unsigned ink, unsigned paper)
{
	return new FGText(GetXM(), GetYM(18), s,this,ink==UNDEFINED_COLOR?this->ink:ink,paper==UNDEFINED_COLOR?this->paper:paper);
}


/**
	Adds a FGCheckButton to the FGMenuWindow.
	@param nm name of the button
	@param key a keycode used as a hot-key for the button (it isn't case sensitive). By example: 'a', 0x20 or ALT_G
	@param variable a pointer to int variable. The value of this variable controls the state of button (true/false)
	@param f a callback function of type ControllCall
	@param flag initial state, default is BNORMAL
	@return a pointer to created object, you can save this one if you can call any methods.
	@note You don't need save the pointer if you don't want call other methods ( e.g. Enable() ) or groups with other buttons.
	The object will be release automatically when the FGWindow will be destroyed. Don't destruct the object explicitly!
	@see FGBaseGui FGControl FGCheckButton FGButtonGroup
*/
FGCheckButton	*FGMenuWindow::AddCheckButton(char *nm, int key, int * variable, ControlCall f, int flag)
{
	return new FGCheckButton(1, GetXM(),	GetYM(18), nm, key,	flag,this,GetInk(),GetPaper(),f,variable);
}

/**
	Adds a FGCheckButton to the FGMenuWindow.
	@param nm name of the button
	@param key a keycode used as a hot-key for the button (it isn't case sensitive). By example: 'a', 0x20 or ALT_G
	@param variable a pointer to bool variable. The value of this variable controls the state of button (true/false)
	@param f a callback function of type ControllCall
	@param flag initial state, default is BNORMAL
	@return a pointer to created object, you can save this one if you can call any methods.
	@note You don't need save the pointer if you don't want call other methods ( e.g. Enable() ) or groups with other buttons.
	The object will be release automatically when the FGWindow will be destroyed. Don't destruct the object explicitly!
	@see FGBaseGui FGControl FGCheckButton FGButtonGroup
*/
FGCheckButton	*FGMenuWindow::AddCheckButton(char *nm, int key, bool * variable, ControlCall f, int flag)
{
	return new FGCheckButton(GetXM(),	GetYM(18), nm, key,	flag,this,GetInk(),GetPaper(),f,variable);
}

/**
	Adds a FGPointButton to the FGMenuWindow.
	@param nm name of the button
	@param key a keycode used as a hot-key for the button (it isn't case sensitive). By example: 'a', 0x20 or ALT_G
	@param variable a pointer to int variable. The value of this variable controls the state of button (true/false)
	@param f a callback function of type ControllCall
	@param flag initial state, default is BNORMAL
	@return a pointer to created object, you can save this one if you can call any methods.
	@note You don't need save the pointer if you don't want call other methods ( e.g. Enable() ) or groups with other buttons.
	The object will be release automatically when the FGWindow will be destroyed. Don't destruct the object explicitly!
	@see FGBaseGui FGControl FGCheckButton FGButtonGroup
*/
FGPointButton	*FGMenuWindow::AddPointButton(char *nm, int key, int * variable, ControlCall f, int flag)
{
	return new FGPointButton(1, GetXM(),	GetYM(18), nm, key,	flag,this,GetInk(),GetPaper(),f,variable);
}

/**
	Adds a masked version FGCheckButton to the FGMenuWindow. This is like a normal check box
	but it is controlled by one of a 32 bit in a triggered variable (of type uint32).
	This is useful when you want change only one bit of variable.
	@param m bit mask value, by example 1 for a bit 0 or (1<<20) for twenth bit.
	@param nm name of the button
	@param key a keycode used as a hot-key for the button (it isn't case sensitive). By example: 'a', 0x20 or ALT_G
	@param variable a pointer to int variable. The value of this variable controls the state of button (true/false)
	@param f a callback function of type ControllCall
	@param flag initial state, default is BNORMAL
	@return a pointer to created object, you can save this one if you can call any methods.
	@note You don't need save the pointer if you don't want call other methods ( e.g. Enable() ) or groups with other buttons.
	The object will be release automatically when the FGWindow will be destroyed. Don't destruct the object explicitly!
	@see FGBaseGui FGControl FGCheckButton FGButtonGroup
*/
FGCheckButton	*FGMenuWindow::AddCheckButtonMask(int m, char *nm, int key, int * variable, ControlCall f, int flag)
{
	return new FGCheckButton(m, GetXM(),	GetYM(18), nm, key,	flag,this,GetInk(),GetPaper(),f,variable);
}

/**
	Adds a masked version FGPointButton to the FGMenuWindow. This is like a normal check box
	but it is controlled by one of a 32 bit in a triggered variable (of type uint32).
	This is useful when you want change only one bit of variable.
	@param m bit mask value, by example 1 for a bit 0 or (1<<20) for twenth bit.
	@param nm name of the button
	@param key a keycode used as a hot-key for the button (it isn't case sensitive). By example: 'a', 0x20 or ALT_G
	@param variable a pointer to int variable. The value of this variable controls the state of button (true/false)
	@param f a callback function of type ControllCall
	@param flag initial state, default is BNORMAL
	@return a pointer to created object, you can save this one if you can call any methods.
	@note You don't need save the pointer if you don't want call other methods ( e.g. Enable() ) or groups with other buttons.
	The object will be release automatically when the FGWindow will be destroyed. Don't destruct the object explicitly!
	@see FGBaseGui FGControl FGPointButton FGButtonGroup
*/
FGPointButton	*FGMenuWindow::AddPointButtonMask(int m, char *nm, int key, int * variable, ControlCall f, int flag)
{
	return new FGPointButton(m, GetXM(),	GetYM(18), nm, key,	flag,this,GetInk(),GetPaper(),f,variable);
}

/**
	Adds an FGEditBox for an one-line-edit box for an integer number with the  test to in-a-range-number.
	@param ws1 width of the button caption, the text is right alligned
	@param ws2 width of the button input part
	@param nm name of the button
	@param key a keycode used as a hot-key for the button (it isn't case sensitive). By example: 'a', 0x20 or ALT_G
	@param p a pointer to the edited int variable
	@param f a callback function of type ControllCall
	@param min minimal correct value from -2147483648 to 2147483647 for a range testing
	@param max maximal correct value from -2147483648 to 2147483647 for a range testing
	@return a pointer to created object, you can save this one if you can call any methods.
	@note You don't need save the pointer if you don't want call other methods ( e.g. Enable() ) or groups with other buttons.
	The object will be release automatically when the FGWindow will be destroyed. Don't destruct the object explicitly!
	@see FGBaseGui FGControl FGEditBox
*/
FGEditBox	*FGMenuWindow::AddEditBox(int	ws1, int ws2, char *nm,	int	key, int *p, ControlCall f, int min, int max)
{
	return new FGEditBox(GetXM(),	GetYM(24), ws1,	ws2, nm, key, this, p, GetInk(),GetPaper(), f, min, max, 1);
}

/**
	Adds an FGEditBox for an one-line-edit box for an integer number without the test to in-a-range-number.
	@param ws1 width of the button caption, the text is right alligned
	@param ws2 width of the button input part
	@param nm name of the button
	@param key a keycode used as a hot-key for the button (it isn't case sensitive). By example: 'a', 0x20 or ALT_G
	@param p a pointer to the edited int variable
	@param f a callback function of type ControllCall
	@return a pointer to created object, you can save this one if you can call any methods.
	@note You don't need save the pointer if you don't want call other methods ( e.g. Enable() ) or groups with other buttons.
	The object will be release automatically when the FGWindow will be destroyed. Don't destruct the object explicitly!
	@see FGBaseGui FGControl FGEditBox
*/
FGEditBox	*FGMenuWindow::AddEditBox(int	ws1, int ws2, char *nm,	int	key, int *p, ControlCall f)
{
	return new FGEditBox(GetXM(),	GetYM(24), ws1,	ws2, nm, key, this, p, GetInk(),GetPaper(),f, 0, 0, 0);
}

/**
	Adds an FGEditBox for an one-line-edit box for an string. The max length of string is get from the ws2 parameter/8, i.e. you don't type bigger string string that that value.
	@param ws1 width of the button caption, the text is right alligned
	@param ws2 width of the button input part
	@param nm name of the button
	@param key a keycode used as a hot-key for the button (it isn't case sensitive). By example: 'a', 0x20 or ALT_G
	@param p a pointer to the edited int variable
	@param f a callback function of type ControllCall
	@return a pointer to created object, you can save this one if you can call any methods.
	@note You don't need save the pointer if you don't want call other methods ( e.g. Enable() ) or groups with other buttons.
	The object will be release automatically when the FGWindow will be destroyed. Don't destruct the object explicitly!
	@see FGBaseGui FGControl FGEditBox
*/
FGEditBox	*FGMenuWindow::AddEditBox(int	ws1, int ws2, char *nm,	int	key, char *p, ControlCall f)
{
	return new FGEditBox(0, GetXM(),	GetYM(24), ws1,	ws2, nm, key, this,	p, GetInk(),GetPaper(),	f);
}

/**
	Adds an FGEditBox for an one-line-edit box for an string. The max length of string is get from the ws2 parameter/8, i.e. you don't type bigger string string that that value.
	@param sz is the maximal size of input string in range from 1 to 127 characters.
	@param ws1 width of the button caption, the text is right alligned
	@param ws2 width of the button input part
	@param nm name of the button
	@param key a keycode used as a hot-key for the button (it isn't case sensitive). By example: 'a', 0x20 or ALT_G
	@param p a pointer to the edited int variable
	@param f a callback function of type ControllCall
	@return a pointer to created object, you can save this one if you can call any methods.
	@note You don't need save the pointer if you don't want call other methods ( e.g. Enable() ) or groups with other buttons.
	The object will be release automatically when the FGWindow will be destroyed. Don't destruct the object explicitly!
	@see FGBaseGui FGControl FGEditBox
*/
FGEditBox	*FGMenuWindow::AddEditBox(int sz, int	ws1, int ws2, char *nm,	int	key, char *p, ControlCall f)
{
	return new FGEditBox(sz, GetXM(),	GetYM(24), ws1,	ws2, nm, key, this,	p, GetInk(),GetPaper(),	f);
}

/**
	Adds an FGEditBox for an one-line-edit box for an double number with the  test to in-a-range-number.
	@param ws1 width of the button caption, the text is right alligned
	@param ws2 width of the button input part
	@param nm name of the button
	@param key a keycode used as a hot-key for the button (it isn't case sensitive). By example: 'a', 0x20 or ALT_G
	@param p a pointer to the edited double variable
	@param f a callback function of type ControllCall
	@param min minimal correct value for a range testing
	@param max maximal correct value for a range testing
	@return a pointer to created object, you can save this one if you can call any methods.
	@note You don't need save the pointer if you don't want call other methods ( e.g. Enable() ) or groups with other buttons.
	The object will be release automatically when the FGWindow will be destroyed. Don't destruct the object explicitly!
	@see FGBaseGui FGControl FGEditBox
*/
FGEditBox	*FGMenuWindow::AddEditBox(int	ws1, int ws2, char *nm,	int	key, double	*p, ControlCall f, double min, double max)
{
	return new FGEditBox(GetXM(),	GetYM(23), ws1,	ws2, nm, key, this,	p, GetInk(), GetPaper(),	f, min,	max, 1);
}

/**
	Adds an FGEditBox for an one-line-edit box for an double number without the test to in-a-range-number.
	@param ws1 width of the button caption, the text is right alligned
	@param ws2 width of the button input part
	@param nm name of the button
	@param key a keycode used as a hot-key for the button (it isn't case sensitive). By example: 'a', 0x20 or ALT_G
	@param p a pointer to the edited double variable
	@param f a callback function of type ControllCall
	@return a pointer to created object, you can save this one if you can call any methods.
	@note You don't need save the pointer if you don't want call other methods ( e.g. Enable() ) or groups with other buttons.
	The object will be release automatically when the FGWindow will be destroyed. Don't destruct the object explicitly!
	@see FGBaseGui FGControl FGEditBox
*/
FGEditBox	*FGMenuWindow::AddEditBox(int	ws1, int ws2, char *nm,	int	key, double	*p,ControlCall f)
{
	return new FGEditBox(GetXM(),	GetYM(23), ws1,	ws2, nm, key, this,	p, GetInk(),GetPaper(),	f, 0, 0, 0);
}

void FGProgressBar::draw(void)
{
	int tmp;
	char s[16];
	set_font(2);

	tmp = (int)((w/(float)steps)*value);
	if (tmp!=sirka) // changed
	{
		rect(2,-2,w+4,h+4, CWHITE);
		rect(0,0,w+1,h+1, CWHITE);
		rect(-2,-2,w+3,h+3, CBLACK);
		rect(-1,-1,w+1,h+1, CBLACK);
		box(0,0,tmp,h,paper);
		box(tmp,0,w-tmp,h,ink);

		sprintf(s,"%d%%", (int)((100./steps)*value)+1);
		if (tmp>=textwidth(s))
		{
			text(tmp/2-textwidth(s)/2,h/2-8,s,ink,paper);
		}
		update_owner();
		owner->WindowRepaint(x-2,y-2,w+4,h+4);
	}
	sirka = tmp;
}

FGProgressBar::FGProgressBar(FGWindow *win, int xx, int yy, int ww, int hh, int s)
	: FGControl(xx, yy, ww, hh, 0, PROGRESSBAR, 0, BNORMAL, win, 0)
{
	if (s<=0)
	{
		 s=100;
		 IError("FGProgressBar: (step <= 0)", 0);
	}
	steps = s;
	value = 0;
	sirka = -1;
	SetInk(CWHITE);
	SetPaper(CBLUELIGHT);
	draw();
}

FGLabel::FGLabel(int xs, int ys, const char	*nm, int key, FGWindow *w, ControlCall f, unsigned i, unsigned p)
: FGControl(
	xs,
	ys,
	16,
	16,
	nm,
	LABEL,
	key,
	BNORMAL,
	w,
	f,
	p,
	i)
{
	set_font(w->get_font());
	is_transparent = high_light = 0;
	draw();
}

void FGLabel::ClickUp(int a)	// if TRUE, so call handler
{
	if ((status&7) == BDISABLED) return;
	MoveOut();
	if (a)
	{
		// if error, set cControl to 0
		owner->SetActive(this);
	}
}

void FGLabel::draw(void)
{
//	set_font(FGFONT_BUTTON);
	if (strlen(name))
	{
		int ww = textwidth(name)+28;
		int hh = 17;//GetFontH();
		FGDrawBuffer::Resize(ww-w,hh-h);
	}

	if (is_transparent)
	{
		set_ppop(_GCOLORKEY);
		state.colorkey = paper;
	}
	else
		set_ppop(_GSET);
	clear(paper);
	text(0,0,name, !high_light?ink:paper, high_light?ink:paper);
	update_owner();
	owner->WindowRepaint(x,y,w,h);
}

void FGAPI FGBaseGui::Resize(int dx, int dy)
{
	FGDrawBuffer::Resize(dx, dy);
	if (type == WINDOW && entryPoint)
		entryPoint->DatabaseResize(w,h,name);
}

void FGAPI FGWindow::Relocate(int xx, int yy)
{
	x += xx;
	y += yy;
	if (entryPoint)
		entryPoint->DatabaseRelocate(x,y,name);
}

static void Yes_No(CallBack t)
{
	FGEvent e(NOTIFYEVENT, t->GetLocalId());
	delete _Note_Ptr;
	close_adept->SendToWindow(&e);
	close_adept = 0;
}

/**
	Shows a dialog for save changes if needed (window has set WNOTIFY and has been
	marked as changed (by SetChange()). If the window has set WNOTIFY flag but isn't
	changed then function sends NOTIFYEVENT to the window immediatelly and returns.
	@return true if notifycation has been performed else returns false.
	@see TextEditor SetChange() ResetChange() WindowClose()
*/
bool FGAPI FGWindow::ShowNotify(void)
{
	if (status&WNOTIFY && !close_adept && changed)
	{
		close_adept = this;
		_Note_Ptr = new FGWindow(&_Note_Ptr, 320, 334, 308, 148, "Warning!", 0, CYELLOW, CREDLIGHT, 0xa7);
		_Note_Ptr->AddPushButton(168, 72, 64, 21, "No", 'n', Yes_No);
		_Note_Ptr->AddPushButton(56, 72, 64, 21, "Yes", 'y', Yes_No);
		_Note_Ptr->WindowText(32, 24, "The contents has been changed. Save it?");
		return true;
	}
	else if (status&WNOTIFY && !close_adept)
 // not changed -> close immediatelly
	{
		FGEvent e(NOTIFYEVENT, 0);
		SendToWindow(&e);
		return true;
	}
	return false;
}

/**
Closes the FGWindow and release all the controls associated with it.
This is a wrapper for window delete operator. You can still use operator delete
that destroy window immediatelly and unconditionally but you lose a notify mechanism (WNOTIFYEVENT).
WindowClose() test the window for 'changed' flag and WNOTIFY and shows a dialog
"SAVE CHANGES?". For more see at TextEditor implementation where it is used.
@return false if the window has been destroyed or true if still not.
@see ~FGWindow TextEditor
*/
bool FGWindow::WindowClose(void)
{
	if (current->ShowNotify())
		return true;
	if (current->status&WNOPICTO)
	{
		IError("Can`t close this window",0);
		return true;
	}
	delete this;
	return false;
}

static void ShowEfect(int x1, int y1, int w1, int h1, int x2, int y2, int w2, int h2)
{
	int s=25, xs=(x2-x1)/s, ys=(y2-y1)/s, ws=(w2-w1)/s, hs=(h2-h1)/s;
	for(;s;s--)
	{
		draw_box(x1,y1,w1,h1,0,_GNOT);
		draw_box(x1+1,y1+1,w1,h1,0,_GNOT);
		draw_box(x1+1,y1+1,w1,h1,0,_GNOT);
		draw_box(x1,y1,w1,h1,0,_GNOT);
		x1 += xs;
		y1 += ys;
		w1 += ws;
		h1 += hs;
	}
}

int FGWindow::iconize(int xs, int ys)
{
	int rc;
	if (!(status&WMINIMIZE)) return 0; // bad window

	StateLock();
	if (iconized==0)
	{
		if (cApp->fulldrag && icon) ShowEfect(x,y,w,h,xs,ys,icon->GetW(),icon->GetH());
		WindowStatus(WHIDE);
		if (status&WNOICON) iconized = (FGWindow *)-1;
		else
		{
			iconized = new IconWindow(this, xs, ys, icon);
		}
		icon_x = xs;
		icon_y = ys;
		rc = 1;
	}
	else if (iconized != (FGWindow *)-1)
	{
		icon_x = iconized->x;
		icon_y = iconized->y;
		if (cApp->fulldrag && icon) ShowEfect(icon_x, icon_y, icon->GetW(),icon->GetH(),x,y,w,h);
		delete iconized;
		iconized=0;
		WindowStatus(WVISIBLE);
		WindowFocus();
		rc = 2;
	}
	else // hide and no icon
	{
		if (cApp->fulldrag && icon) ShowEfect(icon_x, icon_y, icon->GetW(),icon->GetH(),x,y,w,h);
		iconized=0;
		WindowStatus(WVISIBLE);
		WindowFocus();
		rc = 2;
	}
	if (entryPoint) entryPoint->DatabaseSetIcon(icon_x, icon_y, name);
	StateUnlock();
	FGEvent e(ICONIZEEVENT, IsIconized());
	SendToWindow(&e);
	return rc;
}

int FGAPI FGWindow::WindowIconize(void)
{
	if (icon_x==-1 && icon_y==-1)
		return iconize(cApp->GetMouseX()-16, cApp->GetMouseY()-16);
	else
		return iconize(icon_x, icon_y);
}

int FGAPI FGWindow::WindowIconize(int xs, int ys)
{
	return iconize(xs, ys);
}

void FGAPI FGWindow::init(void)
{
	c_menupos = 1;
	itself = 0;
	handler = 0;
	Buttony = 0;
	tabpage_list = 0;
	nControls = 0;
	CallWindowHook = 0;
	changed = 0;
	iconized = 0;
	icon = 0;
	ypos = 0;
	xpos = 1;
	icon_x = -1;
	icon_y = -1;
	status &= ~(WHIDEN|WLOCKED);
	curr_ctrl = -1;
	current_tab_page = "";
	optimize = 0;
	ininput = 0;
	call_handler = 0;
	oldidm =	0;
}

/**
	Draws a panel. This is an ordinary rectangle with caption.
	@see FGWindow::AddPanel()
	@deprecated
*/
void FGAPI FGWindow::WindowPanel(int x, int y, int a, int b, const char *s, unsigned ink1, unsigned ink2)
{
	AddPanel(x,y,a,b,s,ink1,ink2);
}

void FGAPI FGWindow::RepaintBlock(int x, int y, int w, int h)
{
	RamToVideo(image,x, y, this->GetX()+x, this->GetY()+y,w, h, GetW(), GetH(),
		BITBLIT_COPY, _GSET);
}

void FGControl::Close(CallBack cb)
{
	delete cb->owner;	// QWERT
}

void FGControl::Quit(CallBack)
{
	FGApp::AppDone();
}

void FGAPI FGApp::SendToApp(FGEvent	*x)
{
	if (quequeIndex<MAX_EVENT_QUEQUE)
	{
		memcpy(queque+quequeIndex++,x,sizeof(FGEvent));
	}
	else IError("Message queque is full!",0);
}

FGEditBox* FGAPI FGWindow::FindNextEditBox(void)
{
	FGBaseGui *c = ininput;
	while(c->GetNext())
	{
		c = c->GetNext();
		if (c->GetType()==EDITBOX && (c->GetStatus()&7) != BDISABLED)
			return (FGEditBox *)c;
	}
	c = Buttony;
	while(c && c!=ininput)
	{
		if (c->GetType()==EDITBOX && (c->GetStatus()&7) != BDISABLED)
			return (FGEditBox *)c;
		c = c->GetNext();
	}
	return 0;
}

void FGWindow::HideAll(void)
{
	if (!(cApp->flags&APP_ROOTWINDOW))
		clear_frame_buffer(cApp->background);
	else
		FGApp::GetRootWindow()->WindowUpdateBlock(0,0,X_width,Y_width);
}

void FGWindow::ShowAll(void)
{
	FGBaseGui *p = (FGBaseGui *)first;
	FGWindow *pw;

	while (p)
	{
		pw = (FGWindow *)p;
		pw->WindowUpdateBlock(0,0, pw->GetW(), pw->GetH());
		p = p->GetNext();
	}
}

void FGTwoStateButton::draw(void)
{
//	set_font(FGFONT_BUTTON);
	int dl= textwidth(name);	// maximalny zobrazitelny pocet znakov
	int xx = w / 2 - dl / 2, yy = h / 2 - GetFontH() / 2;

	frame();
	if (status & WTRIGGER)
	{
		if (variable) *variable |= mask;
			box(0,0,w,h,CScheme->button_bord1);
			vline(w - 2, 1, h - 2,CScheme->button_fore_pushed);
			hline(1, h - 2, w - 2,CScheme->button_fore_pushed);
			rect(0, 0, w, h,CScheme->button_fore_pushed);
			box(1, 1, w - 3, h - 3,CGRAY3);
			text(xx, yy, name, CScheme->button_fore, CGRAY3);
			Underscore(xx , yy, CDARK);
	}
	else
	{
		if (variable) *variable &= ~mask;
				box(1, 1, w - 3, h - 3, CScheme->button_back);
				text(xx, yy, name,CScheme->button_fore,CScheme->button_back);
				Underscore(xx , yy, CDARK);
	}
	update_owner();
	owner->WindowRepaint(x, y, w, h);
}

// last in list
FGBaseGui* FGAPI FGWindow::GetLastButton(void)
{
	if (Buttony == 0) return 0;
	FGBaseGui *p = Buttony;
	while(p->GetNext())
	{
		p = p->GetNext();
	}
	return p;
}

//! Add current controls to the named tabpage
bool FGAPI FGWindow::AddTabPage(const char *name)
{
	if ((status & WNOTEBOOK) == 0)
	{
		IError("You must specify WNOTEBOOK window flag before AddTabPage()", 0);
		return false;
	}
	FGTabPage *tp = new FGTabPage(name, Buttony, yoff-NOTEBOOKH-2);

	if (tabpage_list == 0) // first use
	{
		tabpage_list = new TabPageList;	// create an empty list
	}
	else
	{
		TabPageList *l = (TabPageList *)tabpage_list;
		TabPageListPtr iter = l->begin();

		while(iter != l->end())
		{
			if (strcmp(iter->GetString(), name) == 0)
			{
				// already exists
				RemoveControls();
				Buttony = 0;
				delete tp;
				return false;
			}
			++iter;
		}
	}

	PrepareForAddTabPage();	// reset for next TabPage
	// insert TabPage to the list
	TabPageList * p = (TabPageList *)tabpage_list;
	p->push_back(*tp);
	delete tp;
	// redraw the whole window
	draw();
	return true;
}

//
// Set current named tabpage
//
FGTabPage* FGAPI FGWindow::SetTabPage(const char *name)
{
	FGTabPage *result=0;

	if ((status & WNOTEBOOK) == 0 || tabpage_list == 0)
	{
		IError("You must specify WNOTEBOOK window flag before SetTabPage()", 0);
		return 0;
	}
	TabPageList *l = (TabPageList *)tabpage_list;
	if (l == 0)
		return 0;

	TabPageListPtr iter = l->begin();

	while(iter != l->end())
	{
		iter->SetFlag(0);
		if (strcmp(iter->GetString(), name) == 0)
		{
			Buttony = iter->GetControls();
			// redraw the whole window
			iter->SetFlag(1);
			current_tab_page = iter->GetString();
			cApp->AutoRepeatEnd();
			c_menupos = 1;
			oldidm = 0;
			// fix - force end of input line on Tab Switching
			if (IsInInput())
				WindowInputChar(CR);
			result = &*iter;
		}
		++iter;
	}
	draw();
	SendEvent(TABSWITCHEVENT, (int)current_tab_page);
	return result;
}

void FGAPI FGWindow::DeleteTabPages(void)
{
	if ((status & WNOTEBOOK) == 0 || tabpage_list == 0)
	{
		return;
	}

	TabPageList *l = (TabPageList *)tabpage_list;
	if (l == 0)
		return;

	TabPageListPtr iter = l->begin();

	if (l->size())
		while(iter != l->end())
	{
		Buttony = iter->GetControls();
		RemoveControls();
		++iter;
	}
	delete l;
	Buttony = 0; // lost all buttons ???
	tabpage_list = 0;
}

void FGAPI FGWindow::DeleteTabPage(const char *name)
{
	if ((status & WNOTEBOOK) == 0 || tabpage_list == 0)
	{
		return;
	}

	TabPageList *l = (TabPageList *)tabpage_list;
	if (l == 0)	return;

	TabPageListPtr iter = l->begin();

	while(iter != l->end())
	{
		if (strcmp(iter->GetString(), name) == 0)
		{
			Buttony = iter->GetControls();
			RemoveControls();
			l->erase(iter);
			SetTabPage(l->begin()->GetString());
			return;
		}
		++iter;
	}
}

int	FGWindow::GetNumberOfTabPages(void)
{
	if (tabpage_list == 0)
		return 0;
	TabPageList *l = (TabPageList *)tabpage_list;
	if (l == 0)
		return 0;
	return l->size();
}

//! Redraw bookmarks only
void FGAPI FGWindow::DrawTabPages(void)
{
	if (tabpage_list == 0)
		return;
	TabPageList *l = (TabPageList *)tabpage_list;
	if (l == 0)
		return;
	TabPageListPtr iter = l->begin();
	if (l == 0) return;
	int _w = 0;
	int f = get_font();
	set_font(FONTSYSLIGHT);

	if (l->size()) while(iter != l->end())
	{
		int _y = iter->GetY();
		int _x = _w+4;
		int sirka = iter->GetWidth()-2;

		box(_x+1, _y+1, sirka-2, NOTEBOOKH-1, paper);
		rect(_x, _y+1, sirka, NOTEBOOKH, 0);
		rect(_x+1, _y, sirka-2, NOTEBOOKH+1, 0);
		box(_x, _y+NOTEBOOKH, sirka+2, 2, 0);

		if (iter->GetFlag() == 0)
		{
			rect(_x, _y+1, sirka, NOTEBOOKH+1,0);
			text(_x+NOTEGAP-2, _y+4,iter->GetString(), ink, paper);
		}
		else
		{
			box(_x+2, _y+2, sirka-4, NOTEBOOKH,paper);
			text(_x+NOTEGAP-2, _y+4,iter->GetString(),CScheme->notebook_active, paper);
		}
		_w += iter->GetWidth();
		++iter;
	}
	set_font(f);
}

//! Test bookmarks for click, and make one current if needed
int FGAPI FGWindow::ClickTabPage(int xx, int yy)
{
	int _w = 0;
	if ((status & WNOTEBOOK) == 0 || tabpage_list == 0)
	{
		return 0;
	}
	xx -= x;
	yy -= y;

	TabPageList *l = (TabPageList *)tabpage_list;
	if (l == 0) return 0;
	TabPageListPtr iter = l->begin();

	// here we go for markers, return TRUE if gets one
	if (l->size()) while(iter != l->end())
	{
		int _y = iter->GetY();
		int _x = _w+4;
		int sirka = iter->GetWidth();

		if (_x<=xx && _x+sirka>xx)
			if (_y<=yy && _y+NOTEBOOKH+2>yy)
				if (iter->GetFlag() == 0)
				{
					SetTabPage(iter->GetString());
					return 1;
				}
				else return 0;
		_w += sirka;
		++iter;
	}
	return 0;
}

void FGAPI FGWindow::SetNextTabPage(void)
{
	if ((status & WNOTEBOOK) == 0 || tabpage_list == 0)
	{
		return ;
	}
	TabPageList *l = (TabPageList *)tabpage_list;
	if (l == 0) return;
	TabPageListPtr iter = l->begin();
	// here we go for markers, return TRUE if gets one
	if (l->size()) while(iter != l->end())
	{
		if (iter->GetFlag() == 1)
		{
			++iter;
			if (iter != l->end())
				SetTabPage(iter->GetString());
			else
				SetTabPage(l->begin()->GetString());
			return;
		}
		++iter;
	}
}

// relokuj vsetky zalozky
void FGAPI FGWindow::RelocateTabPage(int xx, int yy)
{
	x += xx;
	y += yy;
	if (entryPoint)
		entryPoint->DatabaseRelocate(x,y,name);
	if ((status & WNOTEBOOK) == 0)
	{
		return ;
	}
}

//! Run until window close or mrXXX happen
MODAL_RETURN FGAPI FGWindow::ShowModal(void)
{
	int st = status;
	status |= WMODAL;
	MODAL_RETURN rc = cApp->RunModal(this);
	status = st;
	return rc;
}

FGFontProperty::FGFontProperty(char *n, int sz)
{
#ifdef FG_TTF
	_font = TTF_OpenFont(n, sz);
	if (Valid() == false)
		IError("TrueType font open error!",0);
	else
		Solid();
#else
	IError("Using TTF code, but not compiled-in support for one (did you see config.mak?)", 0);
#endif
}

FGFontProperty::~FGFontProperty()
{
#ifdef FG_TTF
	if (_font) TTF_CloseFont((TTF_Font *)_font);
#endif
}

void FGFontProperty::Bold(int value)
{
#ifdef FG_TTF
	int tmp = TTF_GetFontStyle((TTF_Font *)_font);
	if (value)
		TTF_SetFontStyle((TTF_Font *)_font, tmp | TTF_STYLE_BOLD);
	else
		TTF_SetFontStyle((TTF_Font *)_font, tmp & ~TTF_STYLE_BOLD);
#endif
}

void FGFontProperty::Italic(int value)
{
#ifdef FG_TTF
	int tmp = TTF_GetFontStyle((TTF_Font *)_font);
	if (value)
		TTF_SetFontStyle((TTF_Font *)_font, tmp | TTF_STYLE_ITALIC);
	else
		TTF_SetFontStyle((TTF_Font *)_font, tmp & ~TTF_STYLE_ITALIC);
#endif
}

void FGFontProperty::Underline(int value)
{
#ifdef FG_TTF
	int tmp = TTF_GetFontStyle((TTF_Font *)_font);
	if (value)
		TTF_SetFontStyle((TTF_Font *)_font, tmp | TTF_STYLE_UNDERLINE);
	else
		TTF_SetFontStyle((TTF_Font *)_font, tmp & ~TTF_STYLE_UNDERLINE);
#endif
}

/**
	Draws an ascii text with TTF font described by FontProperty.
	@param x x position
	@param y y position
	@param s an ascii text string by null
	@param f a font face object
	@param color foreground
	@param bk background
	@note this function depends on truetype fonts functionality provided
	by 'freetype2' library (www.freetype.org). You can enable TTF support
	in the OpenGUI by editing 'config.mak' file. See the installation PDF file.
	@see FGFontProperty
*/
void FGAPI FGWindow::WindowText(int x, int y, const char *s, FGFontProperty * f, unsigned color, unsigned bk)
{
#ifdef FG_TTF
	StateLock();
	FGDrawBuffer *text=0;
	switch(f->GetAlgorithm())
	{
		default:
		case FGFontProperty::ttfSolid:
			text = TTF_RenderText_Solid((TTF_Font *)f->GetFace(), s, color, bk);
			break;
#ifndef BPP8
		case FGFontProperty::ttfShaded:
			text = TTF_RenderText_Shaded((TTF_Font *)f->GetFace(), s, color, bk);
			break;
		case FGFontProperty::ttfBlended:
//			text = TTF_RenderText_Blended((TTF_Font *)f->GetFace(), s, color, bk);
			break;
#endif
	}
	if (text)
	{
		bitblit(x,y,0,0,text->GetW(), text->GetH(), text);
		WindowRepaintUser(x, y, text->GetW(), text->GetH());
		delete text;
	}
	StateUnlock();
#endif
}

/**
	Draws an ascii text with TTF font described by FontProperty.
	@param x x position
	@param y y position
	@param s an UTF8 text string by null
	@param f a font face object
	@param color foreground
	@param bk background
	@note this function depends on truetype fonts functionality provided
	by 'freetype2' library (www.freetype.org). You can enable TTF support
	in the OpenGUI by editing 'config.mak' file. See the installation PDF file.
	@see FGFontProperty
*/
void FGAPI FGWindow::WindowTextUTF8(int x, int y, const char *s, FGFontProperty * f, unsigned color, unsigned bk)
{
#ifdef FG_TTF
	StateLock();
	FGDrawBuffer *text=0;
	switch(f->GetAlgorithm())
	{
		default:
		case FGFontProperty::ttfSolid:
			text = TTF_RenderUTF8_Solid((TTF_Font *)f->GetFace(), s, color, bk);
			break;
#ifndef BPP8
		case FGFontProperty::ttfShaded:
			text = TTF_RenderUTF8_Shaded((TTF_Font *)f->GetFace(), s, color, bk);
			break;
		case FGFontProperty::ttfBlended:
//			text = TTF_RenderUTF8_Blended((TTF_Font *)f->GetFace(), s, color, bk);
			break;
#endif
	}
	if (text)
	{
		bitblit(x,y,0,0,text->GetW(), text->GetH(), text);
		WindowRepaintUser(x, y, text->GetW(), text->GetH());
		delete text;
	}
	StateUnlock();
#endif
}

/**
	Draws an ascii text with TTF font described by FontProperty.
	@param x x position
	@param y y position
	@param s 16bit text wide string terminated by null
	@param f a font face object
	@param color foreground
	@param bk background
	@note this function depends on truetype fonts functionality provided
	by 'freetype2' library (www.freetype.org). You can enable TTF support
	in the OpenGUI by editing 'config.mak' file. See the installation PDF file.
	@see FGFontProperty
*/
void FGAPI FGWindow::WindowTextUnicode(int x, int y, const unsigned short *s, FGFontProperty * f, unsigned color, unsigned bk)
{
#ifdef FG_TTF
	StateLock();
	FGDrawBuffer *text=0;
	switch(f->GetAlgorithm())
	{
		default:
		case FGFontProperty::ttfSolid:
			text = TTF_RenderUNICODE_Solid((TTF_Font *)f->GetFace(), s, color, bk);
			break;
#ifndef BPP8
		case FGFontProperty::ttfShaded:
			text = TTF_RenderUNICODE_Shaded((TTF_Font *)f->GetFace(), s, color, bk);
			break;
		case FGFontProperty::ttfBlended:
//			text = TTF_RenderUNICODE_Blended((TTF_Font *)f->GetFace(), s, color, bk);
			break;
#endif
	}
	if (text)
	{
		bitblit(x,y,0,0,text->GetW(), text->GetH(), text);
		WindowRepaintUser(x, y, text->GetW(), text->GetH());
		delete text;
	}
	StateUnlock();
#endif
}

char * FGFontProperty::PrepareFont(variable_record *var, int& size, int count, int offset, char *name)
{
#ifdef FG_TTF
	char *pixdata[256], *ptr, *retptr;
	int sizes[256];
	char str[64];
	TTF_Font * f;
	int mw=1, mh=1, sz=0, w, h;

	size = 0;
	memset(pixdata, 0, sizeof(pixdata));
	memset(sizes, 0, sizeof(sizes));

	f = (TTF_Font *)GetFace();
	sprintf(str,"%s-%s-%d", TTF_FontFaceFamilyName(f) , TTF_FontFaceStyleName(f), TTF_FontHeight(f) );

	for (int i=offset; i<(offset+count); i++)
	{
		w = h = 0;
		ptr = TTF_RenderChar(f, i, w, h, sz);

		if (ptr && sz)
		{
			if (mw<w) mw=w;
			if (mh<h) mh=h;
			pixdata[i] = (char *)calloc(sz,1);
			sizes[i] = sz;
			memcpy(pixdata[i], ptr, sz);
		}
		else
		{
			pixdata[i] = 0;
			sizes[i] = 0;
		}
		var->off[i] = size;
		size += sz;
		var->w[i] = w;
	}
	// alloc for entire bitmap
	retptr = ptr = (char *)malloc(size);

	for (int i=offset; i<(offset+count); i++)
	{
		if (pixdata[i])
		{
			memcpy(ptr, pixdata[i], sizes[i]);
			ptr += sizes[i];
			free(pixdata[i]);
		}
	}
	var->mw = mw;
	var->mh = mh;
	strncpy(name,str,23);
	name[23]=0;
	return retptr;
#else
     return 0;
#endif
}

//! try support of MMX for your CPU
int FGApp::test_mmx(void)
{
    // Kylix3 compiler bug!!!
    // (when val is not static)
	fgstate.__fgl_mmx = 0;
#ifdef __BORLANDC__
	static int val = 0;
		asm {
				push	eax
				push	ebx
				push	ecx
				push	edx
				pushfd
				pop		eax
				mov		edx,eax
				xor		eax,0x200000
				push	eax
				popfd
				pushfd
				pop		eax
				cmp		eax,edx
				mov		eax,0
				jz		mmx2
				mov		eax,1
				cpuid
				test	edx,0x800000
				mov		eax,0
				jz		mmx2
				inc		eax	// 1
				emms
				emms
mmx2:			mov		[val],eax
				pop		edx
				pop		ecx
				pop		ebx
				pop		eax
		};
        fgstate.__fgl_mmx = val;
#endif
#if (defined(__linux__) && defined(__GNUC__)) || defined(__ICC)
	__asm
	(
		// Check for CPUID instruction
		"\tpushl %%ebx\n"              	//; save EBX for fpic
		"\tpushf\n"              	//; save EFLAGS
		"\tpop %%eax\n"             	//; store EFLAGS in EAX
		"\tmov %%eax, %%edx\n"        	//; save in EBX for later testing
		"\txor $0x00200000, %%eax\n"  	//; toggle bit 21
		"\tpush %%eax\n"            	//; put to stack
		"\tpopf\n"               	//; save changed EAX to EFLAGS
		"\tpushf\n"              	//; push EFLAGS to TOS
		"\tpop %%eax\n"             	//; store EFLAGS in EAX
		"\tcmp %%edx, %%eax\n"        	//; see if bit 21 has changed
		"\tjz NO_ATHLON\n"        	//; if no change, no CPUID

		// Check for extended functions
		"\tmov $0x80000000, %%eax\n"  	//; query for extended functions
		"\tCPUID\n"               	//; get extended function limit
		"\tcmp $0x80000000, %%eax\n"  	//; is 8000_0001h supported?
		"\tjbe NO_ATHLON\n"       	//; if not, 3DNow! command set not supported

		"\tmov $0x80000001, %%eax\n"  	//; setup extended function 1
		"\tCPUID\n"               	//; call the function
		"\ttest $0x80000000, %%edx\n" 	//; test bit 31
		"\tjz NO_ATHLON\n"        	//; 3DNow! command set not supported
		"\tmov $1, %%eax\n"
		"\tjmp DONE\n"
"NO_ATHLON:\n"
		"\txor %%eax, %%eax\n"
"DONE:\n"
		"\tpopl %%ebx\t\n \tmov %%eax, %0\n"         //; store to bAthlon
		// optimizer hints
		: "=r" (fgstate.__fgl_mmx)          // let compiler choose output register
		:                           // no input selection for compiler
		: "eax", "ecx", "edx"	    // modifies
	);
#endif
	return fgstate.__fgl_mmx;
}

void FGApp::set_mmx(void)
{
	if (fgstate.__fgl_mmx & 1)
	{
		fgstate.mmx_state++;
#ifdef __BORLANDC__
		asm emms;
#endif
#if (defined(__linux__) && defined(__GNUC__)) || defined(__ICC)
		asm("emms");
#endif
	}
}

void FGApp::reset_mmx(void)
{
	if (fgstate.__fgl_mmx & 1)
	{
		fgstate.mmx_state--;
#ifdef __BORLANDC__
		asm emms;
#endif
#if (defined(__linux__) && defined(__GNUC__)) || defined(__ICC)
		asm("emms");
#endif
	}
}

/**
* Parameters are the same as std::printf()
*/
FGDialog::FGDialog(char* format, ...)
{
	char text[256];
	va_list arglist;
	va_start(arglist, format);
#ifndef _MSC_VER
	vsnprintf(text, 255, format, arglist);
#else
	vsprintf(text, format, arglist);
#endif
	va_end(arglist);

	width = FGFontManager::textwidth(FONTSYSLIGHT, text)+32;

	if (width < 280) width = 280;
	wnd = new FGWindow(&wnd, 0,0, width<GetXRes() ? width : GetXRes(), 120, "Error Dialog Box", 0, CBLACK, CREDLIGHT, WSTANDARD|WCENTRED|WNOPICTO);
	wnd->set_font(FONTSYSLIGHT);
	wnd->WindowText(24,16,text);
}

FGDialog::FGDialog(unsigned ink, unsigned paper, const char* title, char* format, ...)
{
	char text[256];
	va_list arglist;
	va_start(arglist, format);
#ifndef _MSC_VER
	vsnprintf(text, 255, format, arglist);
#else
	vsprintf(text, format, arglist);
#endif
	va_end(arglist);

	width = FGFontManager::textwidth(FONTSYSLIGHT, text)+32;

	if (width < 280) width = 280;
	wnd = new FGWindow(&wnd, 0,0, width<GetXRes() ? width : GetXRes(), 120, title, 0, ink, paper, WSTANDARD|WCENTRED|WNOPICTO);
	wnd->set_font(FONTSYSLIGHT);
	wnd->WindowText(24,16,text);
}

/**
* Returns mrOk on user click.
*/
MODAL_RETURN FGDialog::ShowOk(void)
{
	wnd->RemoveControls();
	ctrl = wnd->AddPushButton(width/2-40,48,80,23,"Ok",'o');
	ctrl->SetParameter((void *) mrOk);
	return wnd->ShowModal();
}

/**
* Returns mrYes or mrNo on user click.
*/
MODAL_RETURN FGDialog::ShowYesNo(void)
{
	wnd->RemoveControls();
	ctrl = wnd->AddPushButton(width/2-80,48,64,23,"Yes",'y');
	ctrl->SetParameter((void *) mrYes );
	ctrl = wnd->AddPushButton(width/2+24,48,64,23,"No",'n');
	ctrl->SetParameter((void *)  mrNo );

	return wnd->ShowModal();
}

/**
* Returns mrRetry, mrIgnore or mrCancel on user click.
*/
MODAL_RETURN FGDialog::ShowRetryIgnoreCancel(void)
{
	wnd->RemoveControls();
	ctrl = wnd->AddPushButton(width/4*1-32, 48, 64, 23, "Retry",'r');
	ctrl->SetParameter((void *) (mrRetry));
	ctrl = wnd->AddPushButton(width/4*2-32, 48, 64, 23, "Ignore",'i');
	ctrl->SetParameter((void *) (mrIgnore));
	ctrl = wnd->AddPushButton(width/4*3-32, 48, 64, 23, "Cancel",'c');
	ctrl->SetParameter((void *) (mrCancel));
	return wnd->ShowModal();
}

#ifdef FG_NAMESPACE
}
#endif


